/////////////////////////////////////////////////////////////////////
//
// \author Jeongnim Kim  jnkim@ncsa.uiuc.edu
//
// Measuring abstraction penalty using TinyVector and Vector based on PETE
// src/PETE      (copy of PETE library http://www.acl.lanl.gov)
// src/OhmmsPETE (Ohmms implementation of a fixed vector and an array)
//////////////////////////////////////////////////////////////////////
#include <iostream>
#include <vector>
#include <iomanip>
using namespace std;

#ifdef __GNUC__
#define restrict __restrict__
#endif
 
#if !defined(SP2)
#define addmult addmult_
#endif

#include "OhmmsPETE/TinyVector.h"
#include "OhmmsPETE/OhmmsVector.h"
#include "Utilities/RandomGenerator.h"
#include "Utilities/Clock.h"

template<class T, unsigned D, class RNG>
struct 
RandomVector<TinyVector<T,D>, RNG> {

  typedef TinyVector<T,D> Return_t;
  RandomVector(RNG& gen): RG(gen){ }
  inline Return_t operator()() {
    Return_t res(0);
    for(int i=0; i<D; i++) res[i] = 2*RG()-1;
    return res;
  }
  RNG& RG;
};


// specialized for TinyVector<T,D>
template<class RNG>
struct RandomVector<TinyVector<double,3>, RNG> {
  typedef TinyVector<double,3> Return_t;
  RandomVector(RNG& gen): RG(gen){ }

  inline Return_t operator()() {
    return Return_t(2*RG()-1, 2*RG()-1, 2*RG()-1);
  }
  RNG& RG;
};

extern "C" { // fortran call
  void addmult(double* a,const double* b,const double* c, const double *d,
	       int* ntot);
}

inline void 
caddmult(double* restrict a, const double* restrict b, 
	 const double* restrict c, const double* restrict d, int ntot) 
{
  for(int i=0; i<ntot; i++) {
     a[i] = b[i] + 1.3*c[i] + 1.7*d[i];
  } 
}

RandomGenerator_t Random;

int main(int argc, char **argv) {

  typedef TinyVector<double,OHMMS_DIM> Pos_t;
  typedef Vector<Pos_t> PosArray_t;

  bool makeseed = true;
  Random.init(makeseed); 
  RandomVector<Pos_t, RandomGenerator_t> rvgen(Random);

  int n=10;
  int i=0;
  while(i<argc) {
    if(!strcmp(argv[i],"-n")) {
      n = atoi(argv[++i]);
      break;
    }
    i++;
  }

  cout << endl << "Starting vector operations:" << n << " " << OHMMS_DIM 
       <<"-dimensional elements" << endl;

  Pooma::Clock myclock;
  vector<double> timingdata(10);

  myclock.start();
  PosArray_t A(n), B(n), C(n), D(n);
  myclock.stop();
  timingdata[0] = myclock.cpu_time();

  myclock.start();
  A = 0.0;
  for(int i=0; i<n; i++) {
    B[i] = rvgen();
  }
  for(int i=0; i<n; i++) {
    C[i] = rvgen();
  }
  D = 1.1*B;

  myclock.stop();
  timingdata[1] = myclock.cpu_time();

  myclock.start();
  A = B + 1.3*C + 1.7*D;
  myclock.stop();
  timingdata[2] = myclock.cpu_time();

  myclock.start();
  for(int i=0; i<n; i++)
    A[i] = B[i] + 1.3*C[i] + 1.7*D[i];
  myclock.stop();
  timingdata[3] = myclock.cpu_time();
 
  if(n<=100) {
    for(i=0; i<n; i++)   {
      cout << A[i] << " " << B[i] << " " << C[i] <<  " " << D[i] << endl;
    }
  }

  typedef Vector<double> ValArray_t;
  ValArray_t a(OHMMS_DIM*n), b(OHMMS_DIM*n), c(OHMMS_DIM*n), d(OHMMS_DIM*n);
  int ii=0;
  for(int i=0; i<n; i++) {
    for(int j=0; j<OHMMS_DIM; j++)
      b[ii++] = B[i][j];
  }

  ii=0;
  for(int i=0; i<n; i++) {
    for(int j=0; j<OHMMS_DIM; j++)
      c[ii++] = C[i][j];
  }

  ii=0;
  for(int i=0; i<n; i++) {
    for(int j=0; j<OHMMS_DIM; j++)
      d[ii++] = D[i][j];
  }

  myclock.start();
  a = b + 1.3*c + 1.7*d;
  myclock.stop();
  timingdata[4] = myclock.cpu_time();

  int ntot = OHMMS_DIM*n;
  myclock.start();
  for(int i=0; i<ntot; i++) a[i] = b[i] + 1.3*c[i] + 1.7*d[i];
  myclock.stop();
  timingdata[5] = myclock.cpu_time();

  myclock.start();
  caddmult(a.begin(), b.begin(), c.begin(), d.begin(), ntot);
  myclock.stop();
  timingdata[6] = myclock.cpu_time();

  myclock.start();
  addmult(a.begin(),b.begin(),c.begin(), d.begin(),&ntot);
  myclock.stop();
  timingdata[7] = myclock.cpu_time();

  cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::right, ios::adjustfield);

  cout << "Vector<TinyVector<3> > Allocation " << setw(10) << timingdata[0] << endl;
  cout << "Vector<TinyVector<3> > Assignment " << setw(10) << timingdata[1] << endl;

  cout << endl
       << "Operations C = x*A + B     "  
	<< setw(10) << "CPU secs" << setw(10) << "Penalty"
        << " C-penalty" << endl;
  cout << "AbstractionTest0           " 
       << setw(10) << timingdata[2] 
       << setw(10) << timingdata[2]/timingdata[5] 
       << setw(10) << timingdata[2]/timingdata[7]<< endl;

  cout << "AbstractionTest1           "
       << setw(10) << timingdata[3] 
       << setw(10) << timingdata[3]/timingdata[5] 
       << setw(10) << timingdata[3]/timingdata[7]<< endl;

  cout << "AbstractionTest2           " 
       << setw(10) << timingdata[4] 
       << setw(10) << timingdata[4]/timingdata[5] 
       << setw(10) << timingdata[4]/timingdata[7]<< endl;

  cout << "AbstractionTest3           " 
       << setw(10) << timingdata[5] 
       << setw(10) << timingdata[5]/timingdata[5] 
       << setw(10) << timingdata[5]/timingdata[7]<< endl;

  cout << "C inline function restrict " 
       << setw(10) << timingdata[6] 
       << setw(20) << timingdata[6]/timingdata[7]<< endl;
  cout << "fortran subroutine         " 
       << setw(10) << timingdata[7] 
       << setw(20) << timingdata[7]/timingdata[7]<< endl;

  cout << endl;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: vector.cpp 151 2004-11-02 23:51:14Z jnkim $
 ************************************************************************/
