//////////////////////////////////////////////////////////////////
// (c) Copyright 2003-  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_CONFIGURATION_H
#define OHMMS_CONFIGURATION_H

/**\mainpage ohmms main page
 *\section intro_sec Introduction
 *
 * The versing 0.2.4 is release on 2004-07-21 with revised xml schema and 
 * documentations for the developers and users.
 *
 * Main changes:
 * - The template parameter for ParticleBase is a trait class not the layout.
 * - Trait classes, PtclAttribTraits and PtclOnLatticeTraits are implemented.
 * - ParticleTags class to manage the data types of particle attributes and the object names of the attributes.
 * - OhmmsParticle is replaced by OHMMS::ParticleSet.
 * - NEBPotential is introduced to provide cleaner interfaces with model Hamiltonians.
 * - NNSpatialLayout is introduced for spatial decomposition of a large-scale problem.
 * - Remove singleton patterns to define simulation parameters.
 * - Major bug fixes related to k-point.
 * - xml tags are all in lower case.
 *
 * Todo:
 * - Make species local to ParticleSet.
 * - Implement PtclTraits for open system and overlapping systems for multi-scale simulations.
 * - Unit checking and conversions.
 * - ....
 *
 *\section conf_sec Configuration
 *
 *ohmms core libraries and applications can be configured by
 * - autoconf/automake: configure
 * - cmake : CMakeLists.txt
 * - perl  : configure.pl
 *
 *For typical ohmms applications without GUI supports on generic
 *unix/linux machines, configuring with autoconf/automake is the best
 *choice and is recommended.  On some High-performance computers (so
 *called, Super Computers), autoconf/automake often fails to work
 *properly. Things will improve over the time but JK found that cmake
 *or the perl script she wrote sometimes works better and easier to
 *change for the local environments. Especially, if OpenGL/Qt should
 *be used, using cmake is recommended and most likely that only cmake
 *will work.
 *
 *The perl script is kept for historical reasons and is available only
 *to those who develop ohmms applications.
 *
 *Once configuration is successful, build/src/ohmms-config.h (See \ref
 *ohmms-config.h) is generated. The compiler macros are set based on
 *the configuration options and the local environments.
 *
 *More on configuration can be found at http://www.mcc.uiuc.edu/ohmms/docs/dg/configtools.html
 *
 *\section io_sec I/O utilities
 *
 *The main instructions of ohmms applications are in xml files and the
 *documentation on the schema is available at
 *http://www.mcc.uiuc.edu/ohmms/docs/schema/index.html
 *. The schema
 *should be considered as a proposal, although it is fully functional
 *with ohmms version 0.2.x.  Since there is no schema/DTD available
 *for materials simulations, ohmms is using its own schema. The
 *situation can change as "some" standards are established within
 *materials research community.
 *
 *The actual parsing of xml documents is done by libxml2 library with
 *few template classes and functions implemented to reduce the amount
 *of coding required to use libxml2 directly (see libxmldefs.h).
 *
 *The main instructions are parsed by a DOM parser and the entire tree
 *is stored in memory. The assumption here is that the instructions
 *are small and large-scale data are stored in other formats,
 *typically in hdf5 formats, and the main xml file merely maps the
 *name of the object to the physical data in hdf5 files.
 *
 *Since DOM tree consists of text nodes, the conversion from the text
 *node to the actual objects is done by the ohmms classes. In general,
 *the classes that should handle a xml node are inherited from
 *OhmmsElementBase and implement the virtual function put(xmlNodePtr
 *cur). 
 *
 *The current design DOES NOT require that there is one-to-one mapping
 *between the xml node to a physical object. The node can be handled
 *by a simple object or can be handled by factory classes. An object
 *can overwrite, add or remove its child nodes and the node itself,
 *if the content is no longer used by other objects.
 *
 */

/**\file ohmms-config.h
 *\brief Configuration file for ohmms core libraries and applications.
 *
 *Configuration tools will generate build/src/ohmms-config.h and 
 *-DHAVE_CONFIG_H and be added to the compiler options.
 *The ohmms core libraries and applications should be built with the same
 *configuration file to maintain consistency of the data types, physical
 *dimension and the main xml parser libraries.
 */

/**\def ENABLE_OPENMP
 * Enable OpenMP parallelization. 
 *
 *openmp parallelization is enable when configured with "--enable-openmp" 
 *and the compiler supports openmp parallelization.
 *Current list includes intel compilers, SGI, IBM xlC, DEC cxx.
 */
#define ENABLE_OPENMP 1

/* Define to dummy `main' function (if any) required to link to the Fortran 77
   libraries. */
/* #undef F77_DUMMY_MAIN */

/**\def HAVE_ATEXIT
 * Define to 1 if you have the `atexit' function. 
 */
#define HAVE_ATEXIT 1

/**\def HAVE_BLAS
 * Define if you have a BLAS library. */
#define HAVE_BLAS 1

/* Define to 1 if you have the `clock_gettime' function. */
/* #undef HAVE_CLOCK_GETTIME */

/**\def HAVE_COMPLEX_MATH1
 * define if the compiler has complex math functions 
*/
#define HAVE_COMPLEX_MATH1 

/* define if the compiler has more complex math functions */
/* #undef HAVE_COMPLEX_MATH2 */

/**\def HAVE_DLFCN_H
 * Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 if you have the `floor' function. */
/* #undef HAVE_FLOOR */

/**\def HAVE_GETTIMEOFDAY
 * Define to 1 if you have the `gettimeofday' function. */
#define HAVE_GETTIMEOFDAY 1

/**\def HAVE_INTTYPES_H
 *Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/**\def HAVE_LAPACK
 * Define if you have LAPACK or equivalent library, e.g., MKL. */
#define HAVE_LAPACK 1

/**\def HAVE_LIBHDF5
 * Define to 1 if you have the `hdf5' library (-lhdf5). 
 *
 *If hdf5 libraries are in the path, it is automatically enabled.
 *Otherwise, configure --with-hdf5=DIR
 */
#define HAVE_LIBHDF5 1

/**\def HAVE_LIBSPRNG
 * Define to 1 if you have the `sprng' library (-lsprng). 
 *
 *If sprng library is in the path, it is automatically enabled.
 *Otherwise, configure --with-sprng=DIR
 */
#define HAVE_LIBSPRNG 1

/**\def HAVE_LIBXML2
 * Define to 1 if you have the libxml2 library.
 *
 *Configure "--with-xml2-prefix=DIR" if the header files and the library
 *are not available in the paths.
 */
#define HAVE_LIBXML2 1

/* Define to 1 if you have the <limits.h> header file. */
#define HAVE_LIMITS_H 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the `memset' function. */
#define HAVE_MEMSET 1

/* Define to 1 if you have MPI library */
/* #undef HAVE_MPI */

/* define if the compiler implements namespaces */
#define HAVE_NAMESPACES 

/* Define to 1 if you have OOMPI library */
/* #undef HAVE_OOMPI */

/* Define to 1 if you have hdf5 */
/* #undef HAVE_PKG_hdf5 */

/* Define to 1 if you have sprng */
/* #undef HAVE_PKG_sprng */

/* Define to 1 if you have the `pow' function. */
/* #undef HAVE_POW */

/* Define to 1 if you have the `sqrt' function. */
/* #undef HAVE_SQRT */

/* Define to 1 if you have the <stddef.h> header file. */
#define HAVE_STDDEF_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the `strcasecmp' function. */
#define HAVE_STRCASECMP 1

/* Define to 1 if you have the `strdup' function. */
#define HAVE_STRDUP 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the `strstr' function. */
#define HAVE_STRSTR 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/time.h> header file. */
#define HAVE_SYS_TIME_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the `uname' function. */
#define HAVE_UNAME 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/**\def OHMMS_DIM
 * Define the physical dimension of appliation. 
 *
 *Configure with "--enable-dimension=int" to overwrite the physical dimension.
 *Default is 3.
 */
#define OHMMS_DIM 3

/**\def OHMMS_PRECISION
 * Define application-wise precision to double/float/double double.
 *
 *Many of classes are templated by the scalar type and OHMMS_PRECISION
 *is used to set the precision consistently for the libraries and applications.
 */
#define OHMMS_PRECISION double

/**\def PACKAGE
 * Name of package 
 */
#define PACKAGE "ohmms"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT ""

/* Define to the full name of this package. */
#define PACKAGE_NAME "OHMMS"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "OHMMS 0.2.3"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "ohmms"

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.2.3"

/* Define as the return type of signal handlers (`int' or `void'). */
#define RETSIGTYPE void

/* The size of a `bool', as computed by sizeof. */
#define SIZEOF_BOOL 0

/* The size of a `int', as computed by sizeof. */
#define SIZEOF_INT 4

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/**\def TB_BLOCK
 * Define the block size of spare matrices for optimization. 
 *
 * When TB_BLOCK is 0, any block sparse matrix can be used.
 * Any positive integer enables the block optimizations.
 */
#define TB_BLOCK 0

/* Define to 1 if you can safely include both <sys/time.h> and <time.h>. */
#define TIME_WITH_SYS_TIME 1

/* Version number of package */
#define VERSION "0.2.3"

/* Define to empty if `const' does not conform to ANSI C. */
/* #undef const */

/**\def restrict
 *Define to equivalent of C99 restrict keyword, or to nothing if this is not supported. 
 *
 *Do not define if restrict is supported directly. 
 */
#define restrict __restrict__

/* Define to `unsigned' if <sys/types.h> does not define. */
/* #undef size_t */

/* Define to empty if the keyword `volatile' does not work. Warning: valid
   code using `volatile' can become incorrect without. Disable with care. */
/* #undef volatile */

#endif // OHMMS_CONFIGURATION_H
