//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "AppBase/GeometryBuilder.h"
#include "AppBase/DistributeParticle.h"
#include "Utilities/RandomGenerator.h"

namespace OHMMS {

  bool GeometryComponent::canExchange(int sid) {
    double r= Random();
    if(r<SwapProb[sid])  
      return true;
    else 
      return false;
  }

  bool GeometryComponent::put(xmlNodePtr cur) {

    Swap.resize(SpeciesRef.getTotalNum());
    SwapProb.resize(SpeciesRef.getTotalNum());
    for(int i=0; i<Swap.size(); i++) Swap[i]=i;
    for(int i=0; i<Swap.size(); i++) SwapProb[i]=0.0;

    if(cur == NULL) return true;

    xmlChar* removal=xmlGetProp(cur,(const xmlChar*)"remove");
    if(removal) {
      exclude=atoi((const char*)(removal));
    }

    xmlNodePtr cursave=cur;
    m_param.put(cur);
    cur=cursave->children;
    while(cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "swap") {
	xmlChar *a=xmlGetProp(cur, (const xmlChar *) "source");
	xmlChar *b=xmlGetProp(cur, (const xmlChar *) "target");
	if(a&&b) {
	  int i=SpeciesRef.getSpeciesID((const char*)a);
	  int j=SpeciesRef.getSpeciesID((const char*)b);
	  Swap[i]=j;
	  xmlChar *r=xmlGetProp(cur, (const xmlChar *) "mix");
	  if(r) {
	    SwapProb[i] = atof((const char*)r);
	  } else {
	    SwapProb[i] = 1.0;
	  }
	}
      }
      cur = cur->next;
    }
    return true;
  }

  DummyComponent::DummyComponent(SpeciesSet& species,xmlNodePtr cur):
    GeometryComponent(species)
  { 
    put(cur);
  }

  bool DummyComponent::exchange(int& sid, const SingleParticlePos_t& pos) {
    if(exclude) 
      sid=-1;
    else
      if(canExchange(sid)) sid = Swap[sid];
    return true;
  }

  Slab::Slab(SpeciesSet& species, xmlNodePtr cur):
    GeometryComponent(species), axis(2) {
    m_param.add(z_min,"lower","UC");
    m_param.add(z_max,"upper","UC");
    m_param.add(axis,"axis","int");
    put(cur);
  }

  bool 
  Slab::exchange(int& sid, const SingleParticlePos_t& pos) {
    if(pos[axis] >= z_min && pos[axis] < z_max) { 
      if(exclude) 
	sid=-1;
      else {
	if(canExchange(sid)) sid = Swap[sid];
      }
      return true;
    } else {
      return false;
    }	    
  }

  Tube::Tube(SpeciesSet& species,xmlNodePtr cur):
    GeometryComponent(species), axis(2),r_min(0.0) {
    m_param.add(axis,"axis","int");
    m_param.add(r_min,"inner","UC");
    m_param.add(r_max,"outer","UC");
    m_param.add(center,"center","UC");
    put(cur);
    if(axis == 0) {
      ind_x = 1; ind_y =2;
    } else if(axis == 1) {
      ind_x = 0; ind_y =2;
    } else {
      ind_x = 0; ind_y =1;
    }
    r_min = r_min*r_min;
    r_max = r_max*r_max;
  }

  bool 
  Tube::exchange(int& sid, const SingleParticlePos_t& pos) {
    double rho2 = pow(pos[ind_x]-center[ind_x],2)+pow(pos[ind_y]-center[ind_y],2);
    if(rho2<r_min || rho2 >= r_max) return false;
    if(exclude) 
      sid=-1;
    else {
      if(canExchange(sid)) sid = Swap[sid];
    }
    return true;
  }


  RegularPyramid::RegularPyramid(SpeciesSet& species, xmlNodePtr cur):
    GeometryComponent(species), axis(2),dz_max(-1.0),height(-1),theta(-1) {
    m_param.add(z_min,"lower","UC");
    m_param.add(dz_max,"upper","UC");
    m_param.add(axis,"axis","int");
    m_param.add(base,"base","UC");
    m_param.add(height,"height","UC");
    m_param.add(center,"center","UC");
    put(cur);
    if(dz_max<0) {
      dz_max=height;
    } else {
      dz_max -= z_min;
    }
    if(height>0) {
      ratio = base/height;
    } else {
      if(dz_max>0) {
	height=dz_max;
	ratio = base/height;
      }
    }
  }

  bool RegularPyramid::exchange(int& sid, const SingleParticlePos_t& pos) {
    double dz=pos[2]-z_min;
    if(dz<0 || dz> dz_max) return false;
    double bound=base-dz*ratio;
    if(fabs(pos[0]-center[0])>bound) return false;
    if(fabs(pos[1]-center[1])>bound) return false;
    if(exclude) 
      sid=-1;
    else {
      if(canExchange(sid)) sid = Swap[sid];
    }
    return true;
  }

  Sphere::Sphere(SpeciesSet& species, xmlNodePtr cur): 
    GeometryComponent(species){
      in_radius2=0.0;
    m_param.add(center,"center","UC");
    m_param.add(out_radius2,"radius","UC");
    m_param.add(in_radius2,"core_radius","UC");
    put(cur);
    out_radius2=out_radius2*out_radius2;
    in_radius2=in_radius2*in_radius2;
  }

  bool Sphere::exchange(int& sid, const SingleParticlePos_t& pos) {
    SingleParticlePos_t pp = pos - center;
    double sep=dot(pp,pp);
    if(sep< out_radius2 && sep>=in_radius2) {
      if(exclude) 
	sid=-1;
      else
	if(canExchange(sid)) sid = Swap[sid];
      return true;
    }
    else 
      return false;
  }

  Dome::Dome(SpeciesSet& species, xmlNodePtr cur):
    GeometryComponent(species), axis(2),z_min(-1.0),z_max(-1.0),height(1.e-10),radius(0.0) {
    m_param.add(axis,"axis","UC");
    m_param.add(z_min,"lower","UC");
    m_param.add(z_max,"upper","int");
    m_param.add(height,"height","UC");
    m_param.add(radius,"radius","UC");
    m_param.add(center,"center","UC");
    put(cur);

    //overwrite unphysical bounds
    if(z_min<0) z_min=center[axis];
    if(z_max<0) z_max=center[axis]+height;

    curvature = (height*height+radius*radius)/(2.0*height);
    center[axis]-=(curvature-height);
    rcsq = curvature*curvature;
  }

  bool 
  Dome::exchange(int& sid, const SingleParticlePos_t& pos) {
 
   if(pos[axis] < z_min || pos[axis] >= z_max) return false; 
   SingleParticlePos_t dr = pos-center;
   if(dot(dr,dr)> rcsq) return false; 
   if(exclude) 
     sid=-1;
   else {
     if(canExchange(sid)) sid = Swap[sid];
   }
   return true;
  }

  bool GeometryBuilder::put(xmlNodePtr cur) {

    typedef Particle_t::ParticleLayout_t      ParticleLayout_t;
    typedef Particle_t::Scalar_t              Scalar_t;
    typedef Particle_t::ParticlePos_t         ParticlePos_t;
    typedef Particle_t::ParticleIndex_t       ParticleIndex_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;

    SingleParticleIndex_t uc_grid(1);
    const int mpigrid  = ParticleLayout_t::MPI_GRID;
    const int ompgrid  = ParticleLayout_t::OMP_GRID;
    const int finegrid = ParticleLayout_t::SPATIAL_GRID;
    const int gl       = ParticleLayout_t::GridLevel;

    //grid info
    vector<SingleParticleIndex_t> grid(gl,SingleParticleIndex_t(1));
 
    //copy the in_.Lattice.Grid
    for(int igl=0; igl<gl; igl++)
      grid[igl]=in_.Lattice.Grid[igl];

    vector<GeometryComponent*> parts;
    bool grid_reset=false;
    xmlNodePtr curin=cur;
    xmlChar* remove_bulk=xmlGetProp(cur,(const xmlChar*)"remove");
    cur = cur->xmlChildrenNode;
    bool bulkisset=false;
    double r_max = -1;
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "parameter") {
        string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
	if(aname == "uc_grid" || aname == "unitcell_grid") {
	  putContent(uc_grid,cur);
	} else if(aname == "grid") {
          grid_reset=true;
          putContent(grid[finegrid],cur);
        } else if(aname == "ompgrid") {
          putContent(grid[ompgrid],cur);
        } else if(aname == "mppgrid") {
          putContent(grid[mpigrid],cur);
        } else if(aname == "rmax") {
          putContent(r_max,cur);
        }
      }  else if(cname == "geometry") {
        xmlChar* shape_ptr= xmlGetProp(cur, (const xmlChar *) "name");
        if(shape_ptr) {
          string shape =  (const char*)shape_ptr;
          if(shape == "sphere") {
            parts.push_back(new Sphere(in_.getSpeciesSet(),cur));
          } else if(shape == "slab") {
  	    parts.push_back(new Slab(in_.getSpeciesSet(),cur));
          } else if(shape == "pyramid") {
  	    parts.push_back(new RegularPyramid(in_.getSpeciesSet(),cur));
          } else if(shape == "bulk") {
	    bulkisset=true;
	    parts.push_back(new DummyComponent(in_.getSpeciesSet(),cur));
	  } else if(shape == "dome") {
  	    parts.push_back(new Dome(in_.getSpeciesSet(),cur));
	  } else if(shape == "tube") {
  	    parts.push_back(new Tube(in_.getSpeciesSet(),cur));
	  }
        }
      }
      cur=cur->next;
    }

    if(!bulkisset) {
      //Always set the default
      parts.push_back(new DummyComponent(in_.getSpeciesSet(),0));
      if(remove_bulk) {
	parts.back()->exclude=atoi((const char*)(remove_bulk));
      }
    }

    //first expand the unit cell
    out_.Lattice.set(in_.Lattice,&uc_grid[0]);
    size_type ntot = uc_grid[0]*uc_grid[1]*uc_grid[2]*in_.getLocalNum();

    in_.convert2Unit(in_.R);
    ParticleIndex_t *gid = out_.getIndexAttrib("groupid");

    out_.create(ntot);
    out_.R.setUnit(PosUnit::CartesianUnit);
    int newid;
    size_type nloc(0);
    SingleParticlePos_t del,pos;
    const ParticlePos_t& R = in_.R;

    for(int kc=0; kc<uc_grid[2]; kc++) {
      del[2] = static_cast<Scalar_t>(kc);
      for(int jc=0; jc<uc_grid[1]; jc++) {
	del[1] = static_cast<Scalar_t>(jc);
        for(int ic=0; ic<uc_grid[0]; ic++) {
          del[0] = static_cast<Scalar_t>(ic);
	  for(int nb=0; nb<in_.getLocalNum(); nb++) {
	    pos =R[nb]+del;//unit of in_.Lattice
	    newid = in_.GroupID[nb];	
	    bool exchanged=false;
	    int ipart=0;
	    while(!exchanged) {
	      exchanged = parts[ipart++]->exchange(newid,pos);
	    }
	    if(newid>=0) {      
	      out_.R[nloc]=in_.Lattice.toCart(pos);
	      out_.GroupID[nloc]=newid;
	      out_.ID[nloc]=nloc;
	      (*gid)[nloc] = ipart-1;
	      nloc++;
	    }
	  }
	}
      }
    }

    if(nloc != ntot) {
      out_.erase(nloc,ntot);
      cout << "Total number of particles " << out_.getLocalNum() << endl;
    }

    for(int ipart=0; ipart<parts.size();ipart++) delete parts[ipart];

    //overwrite the grid with the unit-cell grid
    if(!grid_reset) {
      for(int dim=0;dim<OHMMS_DIM; dim++) 
        grid[finegrid][dim]*=uc_grid[dim];
    }

   out_.Lattice.makeGrid(grid);

#ifdef HAVE_LIBMETIS
    DistributeParticle dist(out_,out_);
    dist.put(curin);
#endif

    return true;
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: GeometryBuilder.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
