//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifdef HAVE_CONFIG_H
#include "ohmms-config.h"
#endif
#include "AppBase/OhmmsAppBase.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/Communicate.h"
#include "AppBase/ParticleSets.h"
#include "AppBase/ProjectData.h"
#include "AppBase/RandomNumberControl.h"

namespace OHMMS {

  /*! \brief Constructor 
   * Register OhmmsElementBase objects corresponding to the major tags of input xml files.
   * OhmmsElementSet contains default objects \ref ohmmsdefaults and the constructor
   * adds OhmmsElementBase subclasses that depend on the default objects.
   *
   */
  OhmmsAppBase::OhmmsAppBase(int argc, char** argv): m_doc(NULL), m_context(NULL) {
    
    OhmmsInfo ohmms(argc,argv,Controller->mycontext());
    NodeList = new OhmmsElementSet;

  }

  OhmmsAppBase::~OhmmsAppBase() {
     delete NodeList;
  }

  /*!\fn bool OhmmsAppBase::parse(const char* fname) {
   *\param fname an input xml file to be parsed.
   *\return true if successful.
   *\page parsexml
   * The main routine to process an xml file.
   * Valid input files should contain
   * &lt;simulation&gt;
   * &lt;process&gt;
   *    ................
   * &lt;/process&gt;
   * &lt;/simulation&gt;
   *
   * Each &lt;process/&gt; is a self-contained application to perform a simulation.
   * The XML elements registered by OhmmsElementSet and OhmmsAppBase 
   * by OhmmsElementSet::addNode function are recognized and other unknown
   * elements are ignored.
   */
  bool OhmmsAppBase::parse(const char* fname) {
    
    LOGMSG("\tGoing to parse " << fname)
    xmlNsPtr ns;
    xmlNodePtr cur;
    
    // build an XML tree from a the file;
    m_doc = xmlParseFile(fname);
    if (m_doc == NULL) return false;
    
    // Check the document is of the right kind
    cur = xmlDocGetRootElement(m_doc);
    if (cur == NULL) {
      fprintf(stderr,"empty document\n");
      xmlFreeDoc(m_doc);
      return false;
    }

    ns = NULL;

    string cname((const char*)(cur->name));
    if(cname != "Main" && cname != "simulation") {
      fprintf(stderr,"document of the wrong type, root node != simulation");
      xmlFreeDoc(m_doc);
      return false;
    }
    
    // First level we expect just Jobs
    cur = cur->xmlChildrenNode;
    while ( cur && xmlIsBlankNode ( cur ) )
      {
	cur = cur -> next;
      }
    if ( cur == 0 )
      return false;

    //initialize Random number first with the first <random/>
    m_context = xmlXPathNewContext(m_doc);

    ///initialize the random number generators
    bool async = true;

    xmlXPathObjectPtr result
      = xmlXPathEvalExpression((const xmlChar*)"//application",m_context);
    if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
      xmlAttrPtr att = result->nodesetval->nodeTab[0]->properties;
      while(att != NULL) {
	std::string aname((const char*)(att->name));
	std::string vname((const char*)(att->children->content));
	if(aname == "condition") {
	  if(vname == "sync") async = false;
	}
	att = att->next;
      }
    }

    if(async) {
      LOGMSG("\tClear the particle sets upon completion of process")
    } else {
      LOGMSG("\tKeep the particle sets after completion of process")
      WARNMSG("\tMemory usage is not controlled...")
    }

    ParticleSets *pset= dynamic_cast<ParticleSets*>(NodeList->getNode(Key_particle));    
    RandomNumberControl *rc 
    = dynamic_cast<RandomNumberControl*>(NodeList->getNode(Key_random));    
    xmlNodePtr rptr = rc->initialize(m_context);
    if(rptr) xmlAddChild(cur,rptr);

    vector<xmlNodePtr> process_ptr;
    while(cur != NULL) {
      string cname((const char*)cur->name);
      if(cname == "comment") { cur = cur->next; continue;}
      if(cname == "process" || cname == "Jobs") {
	process_ptr.push_back(cur);
      } else {
        OhmmsElementBase* anode = NodeList->getNode(cname);
        if(anode) { 
          anode->put(cur);
        } 
      } 
      cur = cur->next;
    }

    ProjectData* project 
      =  dynamic_cast<ProjectData*>(NodeList->getNode(Key_project));

    vector<xmlNodePtr>::size_type np = process_ptr.size();
    if(np) {
      if(np>1) {
	LOGMSG("Launching a multiple jobs.")
        if(np != Controller->ncontexts()) {
	  WARNMSG("MPI processors != the number of processes. Squential simulations")
	  for(int ijob=0; ijob<np; ijob++) {
	    put(process_ptr[ijob]);
	    if(async) pset->clearAll();
	    project->advance();
	  }
	} else {
	  put(process_ptr[Controller->mycontext()]);
	}
      } else {
	put(process_ptr[0]);
      }
    } else {
      fprintf(stderr,"document of the wrong type, was '%s', process expected",
	      cur->name);
      fprintf(stderr,"xmlDocDump follows\n");
      xmlDocDump ( stderr, m_doc );
      fprintf(stderr,"xmlDocDump finished\n");
      xmlFreeDoc(m_doc);
      return false;
    }
    
    dump();
    LOGMSG("Completed the jobs. Bye!")
    return true;
  }

  bool OhmmsAppBase::put(xmlNodePtr cur) {

    if(xmlHasProp(cur, (const xmlChar *) "src")) {

      string fname((const char*)(xmlGetProp(cur, (const xmlChar*)"src")));
      // build an XML tree from a the file;
      xmlDocPtr adoc = xmlParseFile(fname.c_str());
      if (adoc == NULL) {
	ERRORMSG("\tA source " << fname << " does not exist")
	return false;
      }    

      //get the set of process in an external file.
      //External sources defining process CANNOT have external sources!
      xmlXPathContextPtr acontext = xmlXPathNewContext(adoc);
      xmlXPathObjectPtr result
	=xmlXPathEvalExpression((const xmlChar*)"//process",acontext);
      for(int i=0; i<result->nodesetval->nodeNr; i++) {
	NodeList->put(result->nodesetval->nodeTab[i]);
      }
      xmlXPathFreeObject(result);
      xmlXPathFreeContext(acontext);
      xmlFreeDoc(adoc);
    } else {
      NodeList->put(cur);
    }
    return true;
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 368 $   $Date: 2005-07-22 20:44:48 -0500 (Fri, 22 Jul 2005) $
 * $Id: OhmmsAppBase.cpp 368 2005-07-23 01:44:48Z jnkim $ 
 ***************************************************************************/
