//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Message/Communicate.h"
#include "Message/OpenMP.h"
#include "AppBase/TransformParticle.h"
#include "AppBase/ParticleSets.h"
#include "AppBase/GeometryBuilder.h"
#include "ParticleBase/ParticleFunctions.h"
#ifdef HAVE_LIBMETIS
#include "AppBase/DistributeParticle.h"
#endif

using namespace OHMMS;

  bool 
  ParticleSets::transformParticle(xmlNodePtr cur) {

    string whattodo("none");
    string pname("none");

    ///process attributes: type or format
    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "action" || aname== "role") {
	whattodo = vname;
      } else if(aname == "target") {
        pname = vname;
      }
      att = att->next;
    }

    if(pname == "all") {
      ERRORMSG("\tTransformation on all the particle sets is forbidden")
    } else {

      //prepare an input particle set to transform
      Particle_t* pin = get(pname);
      if(!pin) pin = ActivePtcl;

      //the pointer to the out particle set is NULL
      Particle_t* pout = NULL;

      if(pin) {
	int i=0;
	if(whattodo == "expand") {
	  pout = new Particle_t(pin->getSpeciesSet());
	  ExpandParticle expandit(*pin,*pout);
	  expandit.put(cur);
	} else if(whattodo == "shape") {
	  pout = new Particle_t(pin->getSpeciesSet());
	  GeometryBuilder shapeit(*pin,*pout);
	  shapeit.put(cur);
	} else if(whattodo == "translate") {
	  TranslateParticle translateit(*pin,*pin);
	  translateit.put(cur);
	} else if(whattodo == "scale") {
	  ScaleParticleLayout scaleit(*pin,*pin);
	  scaleit.put(cur);
	} 
#ifdef HAVE_LIBMETIS
	else if(whattodo == "distribute") {
	  DistributeParticle divideit(*pin,*pin);
	  divideit.put(cur);
	}
#endif
	if(pout) {
	  LOGMSG("\tTransformation \"" << whattodo << " \" removes the input particleset")
          NewPtclList.push_back(pout);
	  pout->setName(pin->getName());
	  remove(pin);
	  ActivePtcl = pout;
	} else {
	  ActivePtcl = pin;
	}
      } 
    }
    return true;
  }


  bool TranslateParticle::put(xmlNodePtr cur) {

    Particle_t::SingleParticlePos_t tvec,displacement;
    string selectedset; 
    bool inunit = false;    

    if(xmlHasProp(cur, (const xmlChar *) "name")) {
      selectedset = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
      if(!xmlStrcmp(cur->name, (const xmlChar *)"parameter")) {
        string aname 
	  = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
	if(aname == "displacement") {
	  putContent(tvec,cur);
	  if(xmlHasProp(cur, (const xmlChar*)"condition")) {
	    string u((const char*)(xmlGetProp(cur, (const xmlChar *) "condition")));
	    if(u == "lattice") {
	      inunit = true;
	    }
	  }
	} 
      }
      cur=cur->next;
    }

    displacement = tvec;
    if(out_.R.getUnit() == PosUnit::LatticeUnit) {
      if(!inunit) displacement = out_.Lattice.toUnit(tvec);
      LOGMSG("Translating positions by " << displacement <<" (Unit-Cell unit)")
    } else {
      if(inunit) displacement = out_.Lattice.toCart(tvec);
      LOGMSG("Translating positions by " << displacement <<" (Cartesian Unit in AA)")
    }

    out_.R += displacement;
    //out_.NeedPartition = false;
    return true;
  }


  bool ExpandParticle::put(xmlNodePtr cur) {

    typedef Particle_t::ParticleLayout_t      ParticleLayout_t;
    typedef Particle_t::ParticleIndex_t       ParticleIndex_t;
    typedef Particle_t::ParticlePos_t         ParticlePos_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::Scalar_t              Scalar_t;
    SingleParticleIndex_t uc_grid(1);

    const int finegrid = ParticleLayout_t::SPATIAL_GRID;
    const int ompgrid  = ParticleLayout_t::OMP_GRID;
    const int mpigrid  = ParticleLayout_t::MPI_GRID;
    const int gl       = ParticleLayout_t::GridLevel;
    bool grid_defined = false;
    bool distribute  = false;
    bool omp_distribute  = false;
    vector<SingleParticleIndex_t> grid(gl,SingleParticleIndex_t(1));
    string selectedset; 

    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "name") {
	selectedset = vname;
      } else if(aname == "distribute") {
	if(vname == "true" || vname == "1") distribute =true;
      }
      att = att->next;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
      if(!xmlStrcmp(cur->name, (const xmlChar *)"parameter")) {
        string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
	if(aname == "uc_grid" || aname == "unitcell_grid") {
	  putContent(uc_grid,cur);
	} else if(aname == "grid") {
	  putContent(grid[finegrid],cur);
	  grid_defined = true;
	} else if(aname == "ompgrid") {
          omp_distribute = true;
	  putContent(grid[ompgrid],cur);
	} else if(aname == "mppgrid") {
	  putContent(grid[mpigrid],cur);
	}
      }
      cur=cur->next;
    }

    ///\todogoing to use selectedset
    if(uc_grid[0] == 1 && uc_grid[1] == 1 && uc_grid[2] == 1) {
      WARNMSG("Expand function is requested with 1x1x1")
      return false;
    }

    WARNMSG("Expand function initialize R and ID only.")

    if(distribute) {
      //Prepare lattice
      out_.Lattice.set(in_.Lattice,uc_grid.begin());
      ParticleLayout_t& lattice = out_.Lattice;

      lattice.makeGrid(grid);

      typedef ParticleLayout_t::Grid_t Grid_t;
      //UniformCartesianGrid<Scalar_t,OHMMS_DIM> SuperGrid;
      Grid_t SuperGrid;
      SuperGrid.setGrid(grid[mpigrid]);

      //reference to the Spatial Grid
      Grid_t* sg(lattice.getGrid(finegrid));
      vector<int> PID(sg->getTotalNum());
      int ng = 0;
      for(int ig=0; ig<lattice.ngrid(finegrid,0); ig++) {
	for(int jg=0; jg<lattice.ngrid(finegrid,1); jg++) {
	  for(int kg=0; kg<lattice.ngrid(finegrid,2); kg++) {
	    PID[ng++] = SuperGrid.loc(sg->center(ig,jg,kg));
	  }
	}
      }

      in_.convert2Unit(in_.R);

      int mycontext = Controller->mycontext();
      SingleParticlePos_t del;
      const ParticlePos_t& R = in_.R;
      const ParticleIndex_t& ionid = in_.GroupID;
      vector<int> myID;
      vector<SingleParticlePos_t> myR;
      vector<int> nsub_m(SuperGrid.getTotalNum(),0);
      for(int ic=0; ic<uc_grid[0]; ic++) {
	del[0] = static_cast<Scalar_t>(ic);
	for(int jc=0; jc<uc_grid[1]; jc++) {
	  del[1] = static_cast<Scalar_t>(jc);
	  for(int kc=0; kc<uc_grid[2]; kc++) {
	    del[2] = static_cast<Scalar_t>(kc);
	    for(int nb=0; nb<in_.getLocalNum(); nb++) {
	      SingleParticlePos_t pos 
		= lattice.toUnit(in_.Lattice.toCart(R[nb]+del));
	      int pid = PID[sg->loc(pos)];
	      if(pid == mycontext) {
		myID.push_back(ionid[nb]);
		myR.push_back(pos);
	      }
	      nsub_m[pid]++;
	    }
	  }
	}
      }

      out_.clear();
      out_.create(myID.size());
      int nacc = 0;
      out_.R.setUnit(PosUnit::LatticeUnit);

      int offset = 0;
      for(int i=0; i<mycontext; i++) offset+=nsub_m[i];
      for(int iat=0; iat<myID.size(); iat++) {
        out_.GroupID[nacc] = myID[iat];
	out_.R[nacc] = myR[iat];
	out_.ID[nacc] = offset;
	++nacc; ++offset;
      }
      //out_.Distributed = true;
      out_.setMode(Particle_t::Distributed);
      return true;
    } 

    //Not very helpful: data locality but cache trashing may be the reason
//#ifdef ENABLE_OPENMP
//    ExpandSuperCellOMP(in_,out_,uc_grid,grid);
//#else
    ExpandSuperCell(in_,out_,uc_grid,grid);
//#endif
    return true;
  }


  bool ScaleParticleLayout::put(xmlNodePtr cur) {

    if(in_.getName() != out_.getName()) {
      out_.assign(in_); 
    }

    out_.convert2Unit(out_.R);

    //the lattice tensor to be scaled
    Particle_t::Tensor_t lattice = out_.Lattice.R;

    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
      string cname((const char*)cur->name);
      if(cname == "parameter") {
	xmlAttrPtr att = cur->properties;
	string pname("none"),ttype("scalar");
	while(att != NULL) {
	  string aname((const char*)(att->name));
	  if(aname == "name") {
	    pname = (const char*)(att->children->content);
	  } else if(aname == "datatype") {
	    ttype =(const char*)(att->children->content);
	  }
	  att = att->next;
	}
	if(pname == "scale") {
	  if(ttype == "scalar") {
	    double sc;
	    putContent(sc,cur);
	    lattice *= sc; 
	  } else if(ttype == "vector") {
	    Particle_t::SingleParticlePos_t scv;
	    Particle_t::Tensor_t t(0.0);
	    putContent(scv,cur);
	    for(int dim=0; dim<OHMMS_DIM; dim++) t(dim,dim) = scv[dim];
	    lattice = dot(lattice,t);
	  } else if(ttype == "tensor") {
	    Particle_t::Tensor_t sct;
	    putContent(sct,cur);
	    lattice = dot(sct,lattice);
	  }
	  //update the out lattice
	  out_.Lattice.set(lattice);
	}
      }
      cur = cur->next;
    }
    return true;
  }

// namespace OHMMS {

//   bool 
//   TransformParticleFactory(ParticleSets::ParticleSetList_t& pin,
// 			   ParticleSets::Particle_t& pout, 
//  			   xmlNodePtr cur) {

//     string whattodo("none");
//     ///process attributes: type or format
//     xmlAttrPtr att = cur->properties;
//     while(att != NULL) {
//       string aname((const char*)(att->name));
//       if(aname == "action" || aname== "role") {
//        whattodo = (const char*)(att->children->content);
//       }
//       att = att->next;
//     }

//     int i=0;
//     if(whattodo == "expand") {
//       ExpandParticle expandit(*(pin.front()),pout);
//       return expandit.put(cur);
//     } 
//     if(whattodo == "shape") {
//       ShapeParticle shapeit(*(pin.front()),pout);
//       return shapeit.put(cur);
//     } 
//     if(whattodo == "translate") {
//       if(pout.getTotalNum()== 0) {
//         pout.assign(*pin.front());
//       }
//       TranslateParticle translateit(pout,pout);
//       return translateit.put(cur);
//     }
//     ///no transformation is done
//     return false;
//   }
// }

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: TransformParticle.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
