//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
/**@file KPointBase.IO.cpp
 *@brief Define I/O functions of KPointBase class
 */
#include <fstream>
#include "CrystalSymmetry/KPointBase.h"
#include "ParticleBase/ParticleUtility.h"
#include "Utilities/SimpleParser.h"
#include "OhmmsData/FileUtility.h"

namespace OHMMS {

  void createKPointMesh(KPointBase* kengine, istream& kfile); 

  bool KPointBase::get(std::ostream& os) const {
    print(os,0);
    return true;
  }

  bool KPointBase::put(std::istream& is) {
    ///do not know what to do
    return true;
  }
  
  /** initialize k-point set
   *@param cur the current xml node
   *
   *A rather complex logic is applied to initialize a k-point set.
   *Preparation steps are:
   *-- it checks the attributes
   *-- is cur pointing to an external file?
   *   -- yes: use the file
   *       -- if it is a valid xml file with //kpoint, call this function recursively. Hopefully, it stops.
   *       -- if it is not a valid xml file, use the KPointBuilder. In future, this should go away.
   *   -- no: everything should be contained in the child nodes. Determine
   *       -- mode : kpt_mode = (auto(MP)| list|band)
   *       -- condition: reduced = true|false
   *       -- size : the number of input k-points. if kpt_mode == list, has to be positive integer
   *       -- role : none|distributed
   *
   *
   * Comments:
   * -- If reduced==true, symmetry to find a minium set.
   * -- With the combination of mode and condition flags, decide if the symmetry groups of the
   *complete particle system will be obtained. Due to numerical problems of the underlaying f77 codes, 
   *finidng the group operations can fail. Since MD normally uses a half-sphere, 
   *we make a fake one-atom system with the same supercell and use the group operations. 
   *
   */
  bool KPointBase::put(xmlNodePtr cur) {
    bool success = false;

    //local default values
    string kfile("none");
    bool reduced_in = false;
    bool distributed_in = false;
    int nkpt_in = 0;
    bool getsymops = true;

    //default kpoint mode is MP
    int kpt_mode = kpt_mp;

    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "src" || aname == "file") {
	kfile = vname;
      } else if(aname == "datatype") {
	if(vname == "auto") {
	  kpt_mode = kpt_mp;
	} else if(vname == "list") {
	  kpt_mode = kpt_list;
	} else if(vname == "band") {
	  kpt_mode = kpt_band;
	}
      } else if(aname == "condition") {
	if(vname == "reduced") {
	  reduced_in = true;
	}
      } else if(aname == "size") {
	nkpt_in = atoi(vname.c_str());
      }
      att = att->next;
    }

    //fun begins
    //case 1: external file is provided with src
    if(kfile != "none") {
      string kformat = getExtension(kfile);
      bool used_ext = false;
      if(kformat == "xml") {      
	// build an XML tree from a the file;
	xmlDocPtr adoc = xmlParseFile(kfile.c_str());
	if (adoc == NULL) {
	  ERRORMSG("\t"<<kfile << " is not valid.")
	  return false;
	}
	// Check the document is of the right kind
	cur = xmlDocGetRootElement(adoc);
	if (cur == NULL) {
	  ERRORMSG("\t"<<kfile << " is empty.")
	  xmlFreeDoc(adoc);
	  return false;
	}
	
	//initialize Random number first with the first <random/>
	xmlXPathContextPtr acontext = xmlXPathNewContext(adoc);
	xmlXPathObjectPtr result
	  =xmlXPathEvalExpression((const xmlChar*)"//kpoint",acontext);
	if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
	  used_ext = put(result->nodesetval->nodeTab[0]);
	}
	xmlXPathFreeObject(result);
	xmlXPathFreeContext(acontext);
	xmlFreeDoc(adoc);
      } else {
        std::ifstream kin(kfile.c_str());
	if(kin) createKPointMesh(this,kin);
	used_ext = true;
      }
      if(used_ext)
        LOGMSG("\tK-point is read from " << kfile)
      else
        ERRORMSG("\tK-point file is invalid " << kfile)
      return used_ext;
    }

    //Everything is in cur. 
    //size is not valid. Exit.
    if(kpt_mode == kpt_list && !nkpt_in) {
      return false;
    } 

    //symmetry properties
    if(reduced_in) {
      findSpaceGroup(*ref_);
    } else {
      Particle_t ptcl_new;
      ptcl_new.Lattice = ref_->Lattice;
      ptcl_new.create(1);
      ptcl_new.R[0] = 0.0;
      ptcl_new.GroupID[0] = 0;
      findSpaceGroup(ptcl_new);
    }

    Particle_t::SingleParticleIndex_t kgrid(1);
    Particle_t::SingleParticlePos_t korigin(0.0);

    //Reduced = reduced_in;
    if(kpt_mode == kpt_mp) {  
      cur = cur->children;
      while(cur != NULL) {
	string cname((const char*)cur->name);
	if(cname == "parameter") {
	  att = cur->properties;
	  while(att != NULL) {
	    string aname((const char*)(att->name));
	    string vname((const char*)(att->children->content));
	    if(aname == "name") {
	      if(vname == "grid") 
		putContent(kgrid,cur);
	      else if(vname == "origin") 
		putContent(korigin,cur);
	    }
	    att = att->next;
	  }
	} else if(cname == "grid") {
	  putContent(kgrid,cur);
	} else if(cname == "origin"){
	  putContent(korigin,cur);
	}
	cur = cur->next;
      }
      buildKpoint(kgrid,korigin);
      if(!reduced_in) extend2FBZ();
    } else if(kpt_mode == kpt_list) {
      putSimpleList(cur,nkpt_in);
      Reduced = reduced_in;
    } else if(kpt_mode == kpt_band) {
      Reduced = false;
      if(nkpt_in) 
	putSimpleList(cur,nkpt_in);
      else {
	vector<int> nbin;
	vector<SingleParticlePos_t> sk;
	int nsk=0;
	nkpt_in = 0;//reset to zero to count
	cur = cur->children;
	while(cur != NULL) {
	  string cname((const char*)cur->name);
	  if(cname == "sk") {
	    nbin.push_back(0);
	    sk.push_back(SingleParticlePos_t(0.0));
	    att = cur->properties;
	    while(att != NULL) {
	      string aname((const char*)(att->name));
	      if(aname == "size") { 
		nkpt_in += nbin[nsk] = atoi((const char*)(att->children->content));
	      }
	      att = att->next;
	    }
	    putContent(sk[nsk],cur);
	    nsk++;
	  }
	  cur=cur->next;
	}
	if(nsk == 0) return false;
	/////////////////////////////////
	nkpt_in -= nbin[0]; //remove the first size
	K.resize(nkpt_in+1);
	W.resize(nkpt_in+1,1.0/static_cast<Scalar_t>(nkpt_in+1));
	K[0] = sk[0]; 
	for(int i=1, k=1; i<nsk; i++) {
	  SingleParticlePos_t ki = sk[i-1]; // starting adn end point
	  SingleParticlePos_t kf = sk[i];
	  double dn = 1.0/static_cast<double>(nbin[i]);
	  SingleParticlePos_t dk =dn*(kf-ki);
	  for(int ik=0; ik<nbin[i]; ik++,k++) {
	    K[k] = ki += dk; 
	  }   
	}
      }
    }
    if(kpt_mode == kpt_band) 
      Kmode = kpt_band;
    else 
      Kmode = kpt_list;

    //Reduced = reduced_in;
    partition(1);
    //came this far
    return true;
  }

  void KPointBase::putSimpleList(xmlNodePtr cur, int nkpt_in){
    cur = cur->children;
    while(cur != NULL) {
      string cname((const char*)cur->name);
      if(cname == "attrib") {
	string ktype("k");
	xmlAttrPtr att = cur->properties;
	while(att != NULL) {
	  string aname((const char*)(att->name));
	  string vname((const char*)(att->children->content));
	  if(aname == "name") { ktype = vname;}
	  att = att->next;
	}
	if(ktype == "k") 
	  putContent(K,cur);
	else if(ktype == "weight") 
	  putContent(W,cur);
	else {
	  vector<TinyVector<Scalar_t,4> > kw;
	  putContent(kw,cur);
	  W.resize(nkpt_in);
	  K.resize(nkpt_in);
	  for(int k=0; k<nkpt_in;k++) {
	    K[k][0] = kw[k][0];
	    K[k][1] = kw[k][1];
	    K[k][2] = kw[k][2];
	    W[k] = kw[k][3];
	  }
	}
      }
      cur = cur->next;
    }
    if(W.empty()) W.resize(nkpt_in,1.0/static_cast<Scalar_t>(nkpt_in));
    NumKpt = nkpt_in;
  }

  void KPointBase::reset() {


  }

  void KPointBase::begin_node(std::ostream& os) const {
  }

  void KPointBase::end_node(std::ostream& os) const {
  }

  /*** simple print function
   *@param fout output stream
   *@param i output mode 
   *
   */
  void KPointBase::print(ostream& fout, int i) const {
    
    int nkpts = K.size();
    if(Kmode == kpt_band) {
      fout << "<kpoint datatype=\"band\" size=\"" 
	   << nkpts << "\" condition=\"band\">" << endl;
    } else {
      fout << "<kpoint datatype=\"list\" size=\"" << nkpts << "\" condition=\"";
      if(Reduced) 
	fout << "reduced\">" << endl; 
      else             
	fout << "half\">" << endl;
    }

    fout << "<attrib name=\"k\">" << endl;
    for(int ikpt=0; ikpt<nkpts; ikpt++) {
      fout << k_unit(ikpt) << endl;
    }
    fout << "</attrib>" << endl;
    if(Kmode != kpt_band) {
      fout << "<attrib name=\"weight\">" << endl;
      for(int ikpt=0; ikpt<nkpts; ikpt++) {
	fout << W[ikpt] << endl;
      }
      fout << "</attrib>" << endl;
    }
    fout << "</kpoint>" << endl;
    
    fout << "<symmetry pointgroup=\"" << hgroup[HolohedralGroup] 
	 << "\" size=\""<< NumSym << "\">" << endl;
    fout << "<rotationops>" << endl;
    for(int isym=0; isym<NumSym; isym++) {
      for(int i=0; i<D; i++) {
	for(int j=0; j<D; j++) {
	  fout << setw(3) << is[isym][j][i];
	}
      }
      fout << endl;
    }
    fout << "</rotationops>" << endl;
    fout << "<translationops>" << endl;
    for(int isym=0; isym<NumSym; isym++) {
      fout << setw(20) << setprecision(10) << tau[isym][0]
	   << setw(20) << setprecision(10) << tau[isym][1]
	   << setw(20) << setprecision(10) << tau[isym][2] << endl;
    }
    fout << "</translationops>" << endl;
    fout <<"</symmetry>" << endl;
    
    //   fout << " xred " << endl;
    //   for(int iatom=0; iatom<xredsym.size(); iatom++) {
    //     fout << setw(20) << setprecision(10) << xredsym[iatom][0]
    // 	 << setw(20) << setprecision(10) << xredsym[iatom][1]
    // 	 << setw(20) << setprecision(10) << xredsym[iatom][2] << endl;
    //   }
    
    //   Scalar_t kptnrm = 0.0;
    //   for(int ikpt=0; ikpt<NumKpt; ikpt++) {
    //     kptnrm += W[ikpt];
    //   }
    //   fout << " kptnrm " << kptnrm << endl;
    //   fout << " wkt " << endl;
    //   for(int ikpt=0; ikpt<NumKpt; ikpt++) {
    //     fout << setw(20) << setprecision(10) << W[ikpt];
    //     if(ikpt%3 == 2) fout << endl;
    //   }
    //   fout << endl;
    //   int mshdim  = 100000;
    //   vector<int> rvec(mshdim*3);
    //   vector<double> rsq(mshdim);
    //   SingleParticlePos_t a1 = LatRef.a(0);
    //   SingleParticlePos_t a2 = LatRef.a(1);
    //   SingleParticlePos_t a3 = LatRef.a(2);
    //   int irmag;
    //   double rmag, smag;
    
    //   spchek(&(a1[0]),&(a2[0]),&(a3[0]), mshdim, rsq.begin(),rvec.begin(),
    // 	 NumKpt,&(K[0][0]), W.begin(), NumSym, &(is[0][0][0]),
    // 	 irmag,rmag,smag);
    
    //   fout << " # phase sum first nonzero at shell " << irmag 
    //        << "  sum= " << smag
    //        << " |R|= " << rmag << endl;
    //   fout << " ==> grid ''efficiency'' (shell/nkpt) = " 
    //        << static_cast<double>(irmag)/static_cast<double>(NumKpt)+1e-6 << endl;
    
  }

void
createKPointMesh(KPointBase* kengine, istream& kfile) 
{

  typedef KPointBase::Particle_t          Particle_t;
  typedef KPointBase::Scalar_t            Scalar_t;
  typedef KPointBase::SingleParticlePos_t SingleParticlePos_t;

  if(kengine) {
    vector<string> slist;
    bool findkpt = true;
    do {
      int nw = getwords(slist,kfile);
      if(nw == 0) continue;
      if(slist[0] == "<KPOINT>") {
	findkpt = false;
      }
    } while(findkpt);
    getwords(slist,kfile);
    
    int nkpt =atoi(slist[0].c_str());
    
    bool getsymops = true;
    if(nkpt > 0) getsymops = false; // externally provided
    else {
      if(slist.size() > 1) { // MC grid but check if nosym option is given
	if(slist[1] == "nosym") 
	  getsymops = false;
	else if(slist[1] == "reduced") 
	  kengine->Reduced = true;
      }   
    }
    
    vector<SingleParticlePos_t> kptin;
    vector<Scalar_t>  wgtin;
    
    if(getsymops)  
      kengine->findSpaceGroup(*kengine->ref_);
    else {
      Particle_t ptcl_new;
      ptcl_new.Lattice = kengine->ref_->Lattice;
      ptcl_new.create(1);
      ptcl_new.R[0] = 0.0;
      ptcl_new.GroupID[0] = 0;
      kengine->findSpaceGroup(ptcl_new);
    }
    
    //if(nkpt <= 0) 
    //  kengine->findSpaceGroup(ptclin,cout);
    
    if(nkpt > 0) { // reading in nkpt points, do not get the space group 
      
      //\warning When kpoints and weights are given, assum that no need to extend k-points
      kengine->Reduced = false;
      
      kptin.resize(nkpt);
      wgtin.resize(nkpt);
      for(int i=0; i<nkpt; i++) {
	getwords(slist,kfile);
	kptin[i][0] = atof(slist[0].c_str());
	kptin[i][1] = atof(slist[1].c_str());
	kptin[i][2] = atof(slist[2].c_str());
	wgtin[i] = atof(slist[3].c_str());
      }
      kengine->buildKpoint(kptin,wgtin);
      
    } else if(nkpt == 0){ // creating MP special k-points
      
      KPointBase::SingleParticleIndex_t nk(1);
      KPointBase::SingleParticlePos_t origin;
      getwords(slist,kfile);
      nk[0] = atoi(slist[0].c_str());
      nk[1] = atoi(slist[1].c_str());
      nk[2] = atoi(slist[2].c_str());
      origin[0] = atoi(slist[3].c_str());
      origin[1] = atoi(slist[4].c_str());
      origin[2] = atoi(slist[5].c_str());
      kengine->buildKpoint(nk,origin);
      
    } else if(nkpt < 0) { // reading for band structure
      
      int nbranch = atoi(slist[1].c_str());
      
      vector<int> npts(nbranch);
      kptin.resize(nbranch);
      
      int nktot = 0;
      for(int i=0; i<nbranch; i++) {
	getwords(slist,kfile);
	kptin[i][0] = atof(slist[0].c_str());
	kptin[i][1] = atof(slist[1].c_str());
	kptin[i][2] = atof(slist[2].c_str());
	if(i) { // not for first k-point 
	  npts[i] = atoi(slist[3].c_str());
	  nktot += npts[i];
	}
      }
      
      kengine->K.reserve(nktot+1);//for efficiency
      kengine->W.reserve(nktot+1);//for efficiency
      
      Scalar_t wgts = 1.0/static_cast<double>(nktot+1); // not meaningful but...
      kengine->addKpt(kptin[0],wgts);// adding the first kpoint
      
      for(int i=1; i<nbranch; i++) {
	SingleParticlePos_t ki = kptin[i-1]; // starting adn end point
	SingleParticlePos_t kf = kptin[i];
	double dn = 1.0/static_cast<double>(npts[i]);
	SingleParticlePos_t dk =dn*(kf-ki);
	for(int ik=0; ik<npts[i]; ik++) {
	  ki += dk; 
	  kengine->addKpt(ki,wgts);// adding this kpoint
	}   
      }
      kengine->Reduced = false; // this kpoint set is not meant for total energy
      kengine->partition(1);
    }
  }
  //OhmmsInform kout(" ", "KPOINT.dat");//,false);
  //if(kout.open()) kengine->print(kout.getStream());
}

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: KPointBase.IO.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
