//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
// (c) Copyright 1998-2002 by Jeongnim Kim
//   All rights reserved.
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//                    CONTACT INFORMATION
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/*************************************************************************
 * template<class T>
 * class BoostLapackSymSolver
 *
 * Input Matrices = Lower Triangle Packed (in fortran)
 * using triangle_matrix<T,upper> 
 *
 * Expert driver to calculate M eigenstates, M < N 
 *************************************************************************/
#ifndef OHMMS_BOOST_LAPACK_SYMSOLVER_PACKED_H
#define OHMMS_BOOST_LAPACK_SYMSOLVER_PACKED_H

#ifndef EIGENSPACEDIMENSION_H
#include "EigSolver/EigenSpaceDimension.h"
#endif
#include <boost/numeric/ublas/config.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/traits.hpp>

#ifdef ADD_
#define dspevx dspevx_
#define zhpevx zhpevx_
#define dspgvx dspgvx_
#define zhpgvx zhpgvx_
#endif
extern "C" {

  void dspevx(const char& jobz, const char& range, const char& uplo, 
              const int& n, double* a, const double& vl, const double& vu,
              const int& il, const int& iu, const double& tol, const int& m, 
              double* w, double* z, const int& ldz, double* workd, 
              int* iwork, int* ifail, int& info); 

  void zhpevx(const char& jobz, const char& range, const char& uplo, 
              const int& n, complex<double>* a, 
              const double& vl, const double& vu,
              const int& il, const int& iu, const double& tol, const int& m, 
              double* w, complex<double>* z, const int& ldz, 
              complex<double>* workd, double* rworkd, int* iwork,
              int* ifail, int& info); 

  void dspgvx(const int& itype, const char& jobz, 
              const char& range, const char& uplo, 
              const int& n, double* a, double* b, 
              const double& vl, const double& vu,
              const int& il, const int& iu, const double& tol, const int& m, 
              double* w, double* z, const int& ldz, double* workd, 
              int* iwork, int* ifail, int& info); 

  void zhpgvx(const int& itype, const char& jobz, 
              const char& range, const char& uplo, 
              const int& n, complex<double>* a, complex<double>* b,
              const double& vl, const double& vu,
              const int& il, const int& iu, const double& tol, const int& m, 
              double* w, complex<double>* z, const int& ldz, 
              complex<double>* workd, double* rworkd, int* iwork,
              int* ifail, int& info); 
}

template<class T>
class BoostLapackSymSolver {

public:

  typedef T value_type;
  typedef boost::numeric::ublas::upper                               upper;
  typedef typename boost::numeric::ublas::type_traits<T>::real_type  real_type;
  typedef boost::numeric::ublas::triangular_matrix<value_type,upper> HamiltonianMatrix_t;
  typedef boost::numeric::ublas::matrix<value_type>                  EigenStates_t;
  
  typedef boost::numeric::ublas::unbounded_array<value_type>         WorkArray_t;
  typedef boost::numeric::ublas::unbounded_array<real_type>          RealArray_t;
  typedef boost::numeric::ublas::unbounded_array<int>                IndexArray_t;
  typedef typename WorkArray_t::size_type                            size_type;

  EigenSpaceDimension& States;       
  HamiltonianMatrix_t  H, S;
  
  BoostLapackSymSolver(EigenSpaceDimension& dim, bool nonortho=false): 
  States(dim), Itype(1), Range('I'), Uplo('L'), Jobz('N'),MyTol(0){ 
    H.resize(States.N, States.N);
    if(nonortho)     
      S.resize(States.N, States.N);
  }

  ~BoostLapackSymSolver() { }

  //bool getEigvectors(HamiltonianMatrix_t& H, EigenStates_t& eigvec) {
  template<class InHamiltonian_t>
  bool getEigvectors(InHamiltonian_t& h, EigenStates_t& Psi) {
    H.clear();  // clean the Hmailtonian
    copy(h,H);  // H = h*phase
    Jobz = 'V';
    return solve(H.data().begin(), Psi.data().begin()); 
  }

  template<class InHamiltonian_t>
  bool getEigvalues(InHamiltonian_t& h) {
    H.clear();         // clean the Hmailtonian
    copy(h,H);  // H = h*phase
    Jobz = 'N';
    return solve(H.data().begin(), H.data().begin()); 
  }


  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvectors(InHamiltonian_t& h, IterType phi, EigenStates_t& Psi) {
    H.clear();         // clean the Hmailtonian
    copy(h,phi,H);  // H = h*phase
    Jobz = 'V';
    return solve(H.data().begin(), Psi.data().begin()); 
  }

  template<class InHamiltonian_t, class IterType>
  bool getEigvalues(InHamiltonian_t& h, IterType phi) {
    H.clear();         // clean the Hmailtonian
    copy(h,phi,H);  // H = h*phase
    Jobz = 'N';
    return solve(H.data().begin(), H.data().begin()); 
  }


  template<class InHamiltonian_t>
  bool 
  getEigvectors(InHamiltonian_t& h, InHamiltonian_t& s, EigenStates_t& Psi) {
    H.clear();         // clean the Hmailtonian
    copy(h,H);  // H = h*phase
    S.clear();         // clean the Hmailtonian
    copy(s,S);  // H = h*phase

    Jobz = 'V';
    return solve(H.data().begin(), S.data().begin(), Psi.data().begin());
  }

  template<class InHamiltonian_t>
  bool getEigvalues(InHamiltonian_t& h, InHamiltonian_t& s) {
    H.clear();         // clean the Hmailtonian
    copy(h,H);  // H = h*phase
    S.clear();         // clean the Hmailtonian
    copy(s,S);  // H = h*phase

    Jobz = 'N';
    return solve(H.data().begin(), S.data().begin(), H.data().begin());
  }


  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvectors(const InHamiltonian_t& h, const InHamiltonian_t& s, 
		IterType phi, EigenStates_t& Psi) {
    H.clear();      // clean the Hmailtonian
    S.clear();      // clean the Hmailtonian

    OHMMS::copy(h,phi,H);  // H = h*phase
    OHMMS::copy(s,phi,S);  // H = h*phase

    Jobz = 'V';
    return solve(H.data().begin(), S.data().begin(), Psi.data().begin()); 
  }

  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvalues(InHamiltonian_t& h, InHamiltonian_t& s, IterType phi) {
    H.clear();         // clean the Hmailtonian
    S.clear();         // clean the Hmailtonian
    OHMMS::copy(h,phi,H);  // H = h*phase
    OHMMS::copy(s,phi,S);  // H = h*phase
    Jobz = 'N';
    return solve(H.data().begin(), S.data().begin(), H.data().begin()); 
  }

//    template<class OtherMat_t>
//    bool copyEigvectors(OtherMat_t& psi0) {
//      copy(Psi,psi0);
//    }

  inline real_type eigval(size_type i) const { return eigVal[i];}

private:
  int Itype;
  int Info;
  char Range, Jobz, Uplo;
  real_type MyTol;

  // \todo make these arrays static workspace
  IndexArray_t Ifail, Iwork;
  WorkArray_t workd; 
  RealArray_t rworkd;
  RealArray_t eigVal;

  inline bool solve(double* a, double *z) {

    check_workspace(States.N, 8, 0);
    dspevx(Jobz, Range, Uplo, States.N, a, 
           States.MinE, States.MaxE, States.First, States.Last, 
           MyTol, States.M, eigVal.begin(), z, States.N, 
           workd.begin(), Iwork.begin(), Ifail.begin(), Info);
    return true;
  }

  inline bool solve(double* ap, double *bp, double* z) {

    check_workspace(States.N, 8, 0);
    dspgvx(Itype, Jobz, Range, Uplo, States.N, ap, bp, 
           States.MinE, States.MaxE, States.First, States.Last, 
           MyTol, States.M, eigVal.begin(), z, States.N, 
           workd.begin(), Iwork.begin(), Ifail.begin(), Info);
    return true;
  }

  inline bool solve(complex<double>* a, complex<double> *z) {

    check_workspace(States.N, 2, 7);
    zhpevx(Jobz, Range, Uplo, States.N, a, 
           States.MinE, States.MaxE, States.First, States.Last, 
           MyTol, States.M, eigVal.begin(), z, States.N, 
           workd.begin(), rworkd.begin(), Iwork.begin(), Ifail.begin(), Info);
    return true;
  }

  inline bool solve(complex<double>* ap, complex<double>* bp, complex<double>* z) {

    check_workspace(States.N, 2, 7);
    zhpgvx(Itype, Jobz, Range, Uplo, States.N, ap, bp, 
           States.MinE, States.MaxE, States.First, States.Last, 
           MyTol, States.M, eigVal.begin(), z, States.N, 
           workd.begin(), rworkd.begin(), Iwork.begin(), Ifail.begin(), Info);
    return true;
  }

  void check_workspace(int n, int workd_size, int rworkd_size) {

    if(Ifail.size() < n)  Ifail.resize(n);
    if(Iwork.size() < 5*n) Iwork.resize(5*n);
    if(eigVal.size() < n) eigVal.resize(n);
    if(workd.size() < workd_size*n) workd.resize(workd_size*n);
    if(rworkd.size() < rworkd_size*n) rworkd.resize(rworkd_size*n);
  }
};

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: BoostLapackSymSolver.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

