//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
// (c) Copyright 1998-2002 by Jeongnim Kim
//   All rights reserved.
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//                    CONTACT INFORMATION
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/*************************************************************************
 * template<class T>
 * class BoostLapackSymSolver
 *
 * Input Matrices = Lower Triangle Packed (in fortran)
 * using triangle_matrix<T,upper> 
 *
 * Expert driver to calculate M eigenstates, M < N 
 *************************************************************************/
#ifndef OHMMS_NAG_SYMSOLVER_H
#define OHMMS_NAG_SYMSOLVER_H

#ifndef EIGENSPACEDIMENSION_H
#include "EigSolver/EigenSpaceDimension.h"
#endif
#ifdef OHMMS_MTL
#include "mtl/mtl.h"
#else
#include "OhmmsPETE/OhmmsVector.h"
#endif

// probably NAG both but just in case
#ifdef ADD_
#define f02fcf f02fcf_
#define f02fdf f02fdf_
#define f02hcf f02hcf_
#define f02hdf f02hdf_
#endif /*ADD_*/

extern "C" {
  // eigen solver Ax = lambda x A = symmetric
  void f02fcf(const char& jobz, const char& range, const char& uplo, 
	      const int& n, double* a, const int& lda,
	      const double& vl, const double& vu, 
	      const int& il, const int& iu, 
	      const int& mest, int& mconv,
	      double* w, double* z, const int& ldz, double* workd, 
	      const int& lwork, int* iwork, int& info); 
  
  // eigen solver Ax = lambda Bx A/B symmetic
  void f02fdf(const int& itype, const char& jobz, const char& uplo, 
	      const int& n, double* a, const int& lda,
	      double* b, const int& ldb, double* w, 
	      double* workd,  const int& lwork, int& ifail); 

  // eigen solver Ax = lambda x A hermitian
  void f02hcf(const char& jobz, const char& range, const char& uplo, 
	      const int& n, complex<double>* a, const int& lda,
	      const double& vl, const double& vu, 
	      const int& il, const int& iu, 
	      const int& mest, int& mconv,
	      double* w, complex<double>* z, const int& ldz, 
	      complex<double>* workd,  const int& lwork, 
	      double* rwork, int* iwork, int& info); 

  // eigen solver Ax = lambda Bx A/B symmetic
  void f02hdf(const int& itype, const char& jobz, const char& uplo, 
	      const int& n, complex<double>* a, const int& lda,
	      complex<double>* b, const int& ldb, double* w, 
	      double* rwork,
	      complex<double>* workd,  const int& lwork, int& ifail); 
}

template<class T, class Ham_t, class EigVec_t>
class NagSymSolver {

  EigenSpaceDimension& States;       

public:

  typedef typename OHMMS::type_traits<T>::value_type value_type;
  typedef typename OHMMS::type_traits<T>::real_type  real_type;
  typedef Ham_t    HamiltonianMatrix_t;
  typedef EigVec_t EigenStates_t;
#ifdef OHMMS_MTL
  typedef mtl::dense1D<value_type> WorkArray_t;
  typedef mtl::dense1D<real_type>  RealArray_t;
  typedef mtl::dense1D<int>        IndexArray_t;
#else
  typedef Vector<value_type> WorkArray_t;
  typedef Vector<real_type>  RealArray_t;
  typedef Vector<int>        IndexArray_t;
#endif
  typedef typename WorkArray_t::size_type                           size_type;

  HamiltonianMatrix_t H, S;

  NagSymSolver(EigenSpaceDimension& dim, bool nonortho=false): 
    States(dim), Itype(1), Range('I'), Uplo('L'), Jobz('N'){ 
#ifdef OHMMS_MTL
    // resize for packed storage is not working properly
    H = HamiltonianMatrix_t(States.N,States.N);
    if(nonortho) S = HamiltonianMatrix_t(States.N, States.N);
#else
    H.resize(States.N, States.N);
    if(nonortho) S.resize(States.N, States.N);
#endif
  }


  ~NagSymSolver() { }


  //bool getEigvectors(HamiltonianMatrix_t& H, EigenStates_t& eigvec) {
  template<class InHamiltonian_t>
  bool getEigvectors(InHamiltonian_t& h, EigenStates_t& Psi) {
    copy(h,H);  // H = h*phase
    Jobz = 'V';
    return solve(H.data(), Psi.data()); 
  }

  template<class InHamiltonian_t>
  bool getEigvalues(InHamiltonian_t& h) {
    copy(h,H);  // H = h*phase
    Jobz = 'N';
    return solve(H.data(), H.data()); 
  }


  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvectors(InHamiltonian_t& h, IterType phi, EigenStates_t& Psi) {
    copy(h,phi,H);  // H = h*phase
    Jobz = 'V';
    return solve(H.data(), Psi.data()); 
  }

  template<class InHamiltonian_t, class IterType>
  bool getEigvalues(InHamiltonian_t& h, IterType phi) {
    copy(h,phi,H);  // H = h*phase
    Jobz = 'N';
    return solve(H.data(), H.data()); 
  }


  template<class InHamiltonian_t>
  bool 
  getEigvectors(InHamiltonian_t& h, InHamiltonian_t& s, EigenStates_t& Psi) {
    copy(h,H);  // H = h*phase
    copy(s,S);  // H = h*phase

    Jobz = 'V';
    return solve(H.data(), S.data(), Psi.data());
  }

  template<class InHamiltonian_t>
  bool getEigvalues(InHamiltonian_t& h, InHamiltonian_t& s) {
    copy(h,H);  // H = h*phase
    copy(s,S);  // H = h*phase

    Jobz = 'N';
    return solve(H.data(), S.data(), H.data());
  }


  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvectors(const InHamiltonian_t& h, const InHamiltonian_t& s, 
		IterType phi, EigenStates_t& Psi) {
    OHMMS::copy(h,phi,H);  // H = h*phase
    OHMMS::copy(s,phi,S);  // H = h*phase

    Jobz = 'V';
    return solve(H.data(), S.data(), Psi.data()); 
  }

  template<class InHamiltonian_t, class IterType>
  bool 
  getEigvalues(InHamiltonian_t& h, InHamiltonian_t& s, IterType phi) {
    OHMMS::copy(h,phi,H);  // H = h*phase
    OHMMS::copy(s,phi,S);  // H = h*phase
    Jobz = 'N';
    return solve(H.data(), S.data(), H.data()); 
  }


  inline real_type eigval(size_type i) const { return eigVal[i];}

private:

  int Itype;
  int Info;
  int Lwork, Mconv, Ifail;
  char Range, Jobz, Uplo;

  // \todo make these arrays static workspace
  IndexArray_t Iwork;
  WorkArray_t workd; 
  RealArray_t rworkd;
  RealArray_t eigVal;

  inline bool solve(double* a, double *z) {

    Mconv = 0;
    check_workspace(States.N, 64, 0);
    f02fcf(Jobz, Range, Uplo, 
	   States.N, a, States.N,
	   States.MinE, States.MaxE, States.First, States.Last, 
	   States.M, Mconv, eigVal.data(), z, States.N,
	   workd.data(), Lwork, Iwork.data(), Info);

//      if(Jobz == 'V') {
//        OhmmsBlas::copy(States.N*States.M,a,z);
//      }

    return true;
  }

  inline bool solve(double* ap, double *bp, double* z) {

    check_workspace(States.N, 64, 0);
    f02fdf(Itype, Jobz, Uplo,
	   States.N, ap, States.N, bp, States.N, 
	   eigVal.data(), workd.data(), Lwork, Ifail);

    if(Jobz == 'V') {
      OhmmsBlas::copy(States.N*States.M,ap,z);
    }

    return true;
  }

  inline bool solve(complex<double>* a, complex<double> *z) {

    Mconv = 0;
    check_workspace(States.N, 64, 7);
    f02hcf(Jobz, Range, Uplo, 
	   States.N, a, States.N,
	   States.MinE, States.MaxE, States.First, States.Last, 
	   States.M, Mconv, eigVal.data(), z, States.N,
	   workd.data(), Lwork, rworkd.data(), Iwork.data(), Info);
    return true;
  }

  inline bool solve(complex<double>* ap, 
		    complex<double>* bp, complex<double>* z) {

    check_workspace(States.N, 64, 3);
    f02hdf(Itype, Jobz, Uplo,
	   States.N, ap, States.N, bp, States.N, 
	   eigVal.data(), rworkd.data(), workd.data(), Lwork, Ifail);

    //need to copy eigen states
    if(Jobz == 'V') {
      OhmmsBlas::copy(States.N*States.M,ap,z);
    }
    return true;
  }

  void check_workspace(int n, int workd_size, int rworkd_size) {
    Lwork = workd_size*n;
    if(workd.size() < Lwork) { workd.resize(Lwork);}
    if(Iwork.size() < 5*n) Iwork.resize(5*n);
    if(eigVal.size() < n) eigVal.resize(n);
    if(rworkd.size() < rworkd_size*n) rworkd.resize(rworkd_size*n);
  }
};

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: NagSymSolver.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

