      subroutine cpointg(b1,b2,b3,ntype,natom,nkatom,atom,
     &           nsym,rot,tnons)

c Tests a putative space group to see if atomic coordinates agree.

c Input:
c  b1(3),b2(3),b3(3)=reciprocal lattice primitive translations
c  ntype=number of types of atoms
c  natom=number of atoms
c  nkatom(ntype)=number of atoms of each type
c  atom(3,natom)=dimensional cartesian coordinates of the atoms
c  nsym=number of space group symmetries
c  rot(3,3,nsym)=cartesian rotation matrices for symmetries
c  tnons(3,nsym)=nonsymmorphic translations (cartesian)
c Output:
c  Stops with error message if symmetries are not consistent.

      implicit none

      integer ntype,natom,nsym
      integer nkatom(ntype)
      double precision b1(3),b2(3),b3(3),atom(3,natom),rot(3,3,nsym),
     &  tnons(3,nsym)

      
      integer iatom,itype,inum,is,ir,latm,il,iat
      double precision rtol,twopi,rx,ry,rz,arg,phase1,phase2,phase3
      double precision rr(3)
      data rtol/1.0d-10/
 
      twopi = 8.d0 * atan(1.d0)

      iatom = 0
      do 110 itype = 1, ntype
      do 110 inum = 1, nkatom(itype)

      iatom = iatom + 1
c      Skip atom at origin
       if ( abs(atom(1,iatom)).lt.1.0e-6 .and.
     &      abs(atom(2,iatom)).lt.1.0e-6 .and.
     &      abs(atom(3,iatom)).lt.1.0e-6) go to 110
 
      do is = 2, nsym
 
       do ir = 1, 3
        rr(ir) = 0.0d0
       enddo
 
c      Hit atom with rotation
       do ir = 1, 3
        rr(1) = rr(1) + rot(ir,1,is) * atom(ir,iatom)
        rr(2) = rr(2) + rot(ir,2,is) * atom(ir,iatom)
        rr(3) = rr(3) + rot(ir,3,is) * atom(ir,iatom)
       enddo
c      ... and nonsymmorphic translation
       rr(1) = rr(1) + tnons(1,is)
       rr(2) = rr(2) + tnons(2,is)
       rr(3) = rr(3) + tnons(3,is)
 
c      check whether this vector coincident with other atom coord.
c      this is tested by asking whether cos(2*Pi*G.diff)=1
c      for G=each reciprocal lattice translation, diff=coord difference
c      i.e. diff can be a direct lattice vector
 
       latm = 0
       do il = 1, itype - 1
        latm = latm + nkatom(il)
       enddo
 
       do iat = 1, nkatom(itype)
        latm = latm + 1
        rx = rr(1) - atom(1,latm) 
        ry = rr(2) - atom(2,latm) 
        rz = rr(3) - atom(3,latm) 
c       Note that b1, b2, b3 do not carry the 2*Pi
        arg = twopi * (rx * b1(1) + ry * b1(2) + rz * b1(3))
        phase1 = cos(arg)
        arg = twopi * (rx * b2(1) + ry * b2(2) + rz * b2(3))
        phase2 = cos(arg)
        arg = twopi * (rx * b3(1) + ry * b3(2) + rz * b3(3))
        phase3 = cos(arg)

c       Break out of iat loop when a match is found
        if (abs(phase1-1.0d0).lt.rtol .and.
     &      abs(phase2-1.0d0).lt.rtol .and.
     &      abs(phase3-1.0d0).lt.rtol) goto 101

       enddo

c      If you get here, no match was found

       write(*,1000) iatom,is
 1000  format(/,'cpointg error: atom ',i5,' doesn''t fit symmetry ',
     &   i5)
       write(*,1010) rr,(atom(ir,latm),ir=1,3),rx,ry,rz,
     &  phase1,phase2,phase3
 1010  format(' rotated coord',3f16.12,/,
     &        ' other atom   ',3f16.12,/,
     &        ' difference   ',3f16.12,/,
     &        ' phases       ',3f16.12)
       write(*,1020) ((rot(ir,il,is),ir=1,3),il=1,3)
 1020  format(' rotation matrix:',9f6.2)
       write(*,1030) (tnons(ir,is),ir=1,3)
 1030  format(' nonsymmorphic transl:',3f16.12)
       stop

  101  continue

      enddo
 
  110 continue
 
      return
      end
