      program kptgen

c This program, kptgen.f and related subroutines, takes input unit cell
c and atomic coordinate data and some control parameters and deduces
c the space group and Bravais lattice and computes a special k point
c Monkhorst-Pack type Brillouin zone quadrature grid.

c It needs a single file name as input, the name of the input file.
c The rest of the input comes from the input file, described below.
c Output is sent to two hard-coded files: symmetry.summary and 
c abinit.data.  The first is detailed analysis of crystal symmetry,
c the second is a section of a file which would be readable by ABINIT.

c Most recent revisions by Douglas C. Allan, Corning Incorporated, Mar 1995.
c Updated for ABINIT, XG, February 1999.
c This program was in public domain and presumably remains so.
c Not sure of its history, see comments in subroutines for some indication.

c I left some vestiges of old routines in place, even though I did not 
c modify or upgrade them, assuming they might possibly be useful to others
c who further upgrade this package.  dca

c=====================================================================
c Description of input file:
c title 
c acell(3)
c rprim(3,3)
c natom, ntype
c (type(iatom),iatom=1,natom)
c ((xred(mu,iatom),mu=1,3),iatom=1,natom)
c ibrav, inpgd
c nkpt
c  mp1 mp2 mp3  k1 k2 k3  [depending on nkpt, see below]  -OR-
c  kp11 kp21 kp31
c  kp21 kp22 kp32 ... etc. one k point per line
c  wt1 wt2 wt3 ... etc.   [depending on nkpt, see below]
c=====================================================================
c More detail on input lines:
c title -> up to 80 characters, single line
c acell -> length scale for each primitive translation, bohr
c rprim -> primitive translation vectors (first 3 for first vector, etc.)
c natom -> number of atoms
c ntype -> number of types of atoms (number of different species)
c type  -> integer for each type of atom, usually 1 1 1 ... 2 2 ... 
c xred  -> reduced coordinates of atoms (in basis of (a1 a2 a3))
c ibrav -> Bravais lattice specification: 0 to generate Bravais lattice
c          from input coordinates, else see pointg.f for definitions
c          NOTE that ibrav > 0 is probably not working (3/95 dca).
c inpgd -> point group index: 1 is c1, else see pointg.f
c           (I have only run this with inpgd=0, dca)
c nkpt  -> control parameter: 
c  (nkpt<=0) => generate a Monkhorst-Pack special point grid based on
c               the next line of input, which gives mp1, mp2, mp3 as the
c               three Monkhorst-Pack indices and k1, k2, k3 as the origin
c               of the grid (in basis of recip translations)
c               Usually want k1, k2, k3 = 0, 0, 0.
c  (nkpt >0) => read k points in basis of recip translations, then read
c               associated weights, and use these to test the form factors

c Note regarding choice of k1, k2, k3: should see A. H. MacDonald, 
c Phys. Rev. B 18, 5897 (1978) especially for cubic systems and large
c numbers of k points.

c Checking the accuracy of the k point grid, using spchek, generally
c takes longer and uses more memory than the generation of the
c grid originally using speck.

c For special k point quadrature grids, see
c  Hendrik J. Monkhorst and James D. Pack,
c  ``special points for brillouin-zone integrations'',
c  Phys. Rev. B 13, 5188-5192 (1976);
c  James D. Pack and Hendrik J. Monkhorst,
c  ``"Special points for brillouin-zone integrations"--a reply'',
c  Phys. Rev. B  16, 1748-1749 (1977);
c  A. H. MacDonald,
c  ``comment on special points for Brillouin-zone integrations'',
c  Phys. Rev. B 18, 5897-5899 (1978).

c=====================================================================
c abinit.data can be grepped for # to get a quick summary of grid.
c Also # indicates a comment line for an ABINIT input file, so
c these lines can be included in an input file to ABINIT to describe
c the k grid.  The rest of the output, symrel and kpt, etc., is designed
c for cutting and pasting into an ABINIT input file.
c=====================================================================
 
      implicit none

c     mkpt sets maximum number of k points in irreducible BZ
c     matom sets maximum number of atoms
c     mtype sets maximum number of kinds of atoms
      integer mkpt, matom, mtype
      parameter (mkpt = 4096, matom = 250 ,mtype = 4) 

c     mshdim sets dimension for array used in spchek.
      integer mshdim
      parameter (mshdim=100000)
 
      double precision ck(4,mkpt), gk0(3)
      integer nk(3)
      double precision a1(3),a2(3),a3(3),b1(3),b2(3),b3(3)
      double precision r(49,3,3),tau(3,48),rot(3,3,48)
      integer nkatom(mtype)
      double precision atom(3,matom)
      integer is(3,3,48), ty(matom), f0(48,matom)
      double precision  tnons(3,48)

      double precision acell(3),rprim(3,3)
      double precision xred(3,matom),xredsym(3,matom)

      double precision kpt(3,mkpt),wtk(mkpt)
      integer rvec(3,mshdim)
      double precision rsq(mshdim)

      character*80 title
      character*100 infile 

c     unit00 is where the summary of crystal symmetry information is sent
c     This is called symmetry.summary and is overwritten on each run.

c     unit01 is where a data file suitable for ABINIT is sent
c     This is called abinit.data and is also overwritten on each run.

c     unit02 is the input file

      integer unit00,unit01,unit02
      parameter (unit00=10, unit01=11, unit02=12)

c     parameter (bohr=0.529177249d0,bohri=1.8897259885789232d0)

      integer ihg,ihc,ibrav,inpgd,isymp,irmag,iatom,i,j,n,isym,
     &  symorp,nkpt,nsym,iinv,natom,ikind,ntype,jtype,ikpt
      double precision ucvol,rmag,smag,r1,r2,r3
      integer trmlen
      external trmlen

      integer natom0,nat0,iinv0,k,nbrav,ihc0
      integer symlat(3,3,48),ty0(1),ind(48)
      double precision atom0(3,1),tau0(3,48)
      
 
c     ------------------------------------------------------------------

      write(*,1010)
 1010 format('please input a file name -->',$)
      read (*,1020) infile
 1020 format(a)

      write(*,1030)
 1030 format(/,'Output will be sent to symmetry.summary and ',
     &  'abinit.data',/)
 
      open (unit=unit02,file=infile,form='formatted',
     &  status='old')
      open (unit=unit00,file='symmetry.summary',form='formatted',
     &  status='unknown')
      open (unit=unit01,file='abinit.data',form='formatted',
     &  status='unknown')

c     Place name of input file at top of the output files
      write(unit00,1035) infile(1:trmlen(infile))
      write(unit01,1035) infile(1:trmlen(infile))
 1035 format('#Input file was ',a)

c     First line of input file is title line
      read (unit02,1020) title
      write(unit00,1040) title(1:trmlen(title))
      write(unit01,1040) title(1:trmlen(title))
 1040 format('#',a)
 
c     Read 3 lattice constants in bohr units
      read (unit02,*) acell

c     Read real space primitive translation vectors (dimensionless)
c     as columns of the 3x3 array rprim
      read (unit02,*) rprim

c     Define dimensional real space primitive translations
      do i=1,3
       a1(i)=rprim(i,1)*acell(1)
       a2(i)=rprim(i,2)*acell(2)
       a3(i)=rprim(i,3)*acell(3)
      enddo

c     Read number of atoms in cell and number of species
      read (unit02,*) natom,ntype

      if (natom.gt.matom) then
       write(*,1050) natom,matom
 1050  format(/,' Error: input natom=',i10,' exceeds dimension',
     &  ' matom=',i10,/,' raise matom in kptgen.f and recompile.',/)
       stop
      endif
      if (ntype.gt.mtype) then
       write(*,1060) ntype,mtype
 1060  format(/,' Error: input ntype=',i10,' exceeds dimension',
     &  ' mtype=',i10,/,' raise mtype in kptgen.f and recompile.',/)
       stop
      endif

c     Read integers which specify which type of atom (1 1 1 ... 2 ...)
      read (unit02,*) (ty(i),i=1,natom)

c     Determine how many atoms of each kind there are
      do ikind=1,mtype
       nkatom(ikind)=0
      enddo
      jtype=ty(1)
      do iatom=1,natom
       if (ty(iatom).gt.mtype) then
        write(*,1070) iatom,ty(iatom),mtype
 1070   format(/,'Error: for input atom number ',i6,' input type was',
     &  i6,/,' which exceeds allowed number of types mtype=',i6,/,
     &  ' Raise dimension mtype and recompile kptgen.',/)
        stop
       endif
       nkatom(ty(iatom))=nkatom(ty(iatom))+1
       jtype=max(jtype,ty(iatom))
      enddo
      if (jtype.ne.ntype) then
       write(*,1080) jtype,ntype
 1080  format(/,' Warning: highest input atom type ',i5,' != input ',
     & 'ntype ',i5,' ==> ntype reset.',/)
       ntype=jtype
      endif

c     Compute reciprocal lattice vectors and unit cell volume
      call recip(a1,a2,a3,b1,b2,b3,ucvol)

c     Read fractional coordinates of atoms (in basis of a1, a2, a3)
      do iatom = 1, natom
       read(unit02,*) r1, r2, r3
       xred(1,iatom)=r1
       xred(2,iatom)=r2
       xred(3,iatom)=r3
c      convert to dimensional cartesian coordinates for atoms
       atom(1,iatom) = r1 * a1(1) + r2 * a2(1) + r3 * a3(1)
       atom(2,iatom) = r1 * a1(2) + r2 * a2(2) + r3 * a3(2)
       atom(3,iatom) = r1 * a1(3) + r2 * a2(3) + r3 * a3(3)
      enddo

c     Write some data
      write(unit00,1090) a1,a2,a3
      write(unit00,1100) b1,b2,b3
      write(unit00,1110) ucvol,ucvol/dble(natom)
 1090 format(' a1 =',3f15.10/' a2 =',3f15.10/' a3 =',3f15.10)
 1100 format(' b1 =',3f15.10/' b2 =',3f15.10/' b3 =',3f15.10)
 1110 format(' ucvol(bohr^3)=',f14.6,'  per atom=',f14.6)

c     Read specification of Bravais lattice and point group
      read(unit02,*) ibrav,inpgd
 
c *****************************************************************
c     Warning: I have not checked carefully the working of the code
c     when ibrav is input > 0.  It is probably buggy (dca 3/95).
c *****************************************************************

      if (ibrav .gt. 0) then

       write(*,1120)
 1120  format(/,' Warning in kptgen: ibrav>0 needs more work.',/)
 
       if(inpgd .lt. 0) then
        read (unit02,*) nsym, isymp
        do n = 1, nsym
          read(unit02,*) ((is(i,j,n),i=1,3),j=1,3)
        enddo
        if(isymp .eq. 0) then
         do n = 1, nsym
          read (unit02,*) r1,r2,r3
          tnons(1,n) = r1 * a1(1) + r2 * a2(1) + r3 * a3(1)
          tnons(2,n) = r1 * a1(2) + r2 * a2(2) + r3 * a3(2)
          tnons(3,n) = r1 * a1(3) + r2 * a2(3) + r3 * a3(3)
          tau(1,n) = tnons(1,n)
          tau(2,n) = tnons(2,n)
          tau(3,n) = tnons(3,n)
         enddo
        endif
       else if(inpgd .eq. 0) then
        nsym = 1
       else
        call pointg(unit00,ibrav,inpgd,is,nsym,iinv)
c       iinv=1 means there is inversion symmetry in point group
        do n = 1, nsym
        do j = 1, 3
        do i = 1, 3
         rot(i,j,n) = dble(is(i,j,n))
        enddo
        enddo
        enddo
       endif
c
      else
 
c      Generate the space group from the input coordinates

       call group1 (unit00,a1,a2,a3,natom,ty,matom,atom,b1,b2,b3,
     &  ihg,ihc,symorp,iinv,nsym,ind,tau,f0,r)

c      symorp is 1 for symmorphic, 0 for nonsymmorphic

       write(unit00,1130) nsym
 1130  format(1x,'nsym from point1',i5)
c      Recall: the nonsymmmorphic translations tau are in terms of (a1,a2,a3)
       do isym=1,nsym
        write(unit00,1150)         (r(ind(isym),1,j),j=1,3),
     &     tau(1,isym)
        write(unit00,1140)ind(isym),(r(ind(isym),2,j),j=1,3),isym,
     &     tau(2,isym)
        write(unit00,1150)         (r(ind(isym),3,j),j=1,3),
     &     tau(3,isym)
c       Get nonsymmorphic translations in dimensional cartesian coordinates
        tnons(1,isym) = tau(1,isym)*a1(1)+tau(2,isym)*a2(1)+
     &                  tau(3,isym)*a3(1)
        tnons(2,isym) = tau(1,isym)*a1(2)+tau(2,isym)*a2(2)+
     &                  tau(3,isym)*a3(2)
        tnons(3,isym) = tau(1,isym)*a1(3)+tau(2,isym)*a2(3)+
     &                  tau(3,isym)*a3(3)
        do j = 1, 3
         do i = 1, 3
          rot(i,j,isym) = r(ind(isym),j,i)
         enddo
        enddo
       enddo
 1140  format('  r(',i2,') = ',3f6.2,8x,'   tau(',i2,') = ',f11.8)
 1150  format('          ',3f6.2,8x,'             ',f11.8)

       call symgen(a1,a2,a3,b1,b2,b3,nsym,rot,is)

       write(unit00,1160)
 1160  format(' Reciprocal space rotation matrices and translations')
       write(unit00,1170) nsym, symorp
       do isym = 1, nsym
        write(unit00,1170) ((is(i,j,isym),i=1,3),j=1,3),
     &   (tau(i,isym),i=1,3)
       enddo
 1170  format(9i4,2x,3f12.5)

c      Provide symmetrized fractional atomic coordinates, including
c      possible "cleaning up" of nonsymmorphic translations

       call symatm(natom,xred,f0,nsym,is,tau,xredsym,tau)
 
c      Also provide the transposes of rotation arrays, which
c      express the rotations in the basis of real space translations,
c      expressed in a format readable by ABINIT.
c      write(unit01,1180)
c1180  format(/,'Real space rotation matrices and translations symrel',
c    &  ' and tnons follow;',/,'also symmetrized fractional',
c    &  ' coordinates xred:')
       write(unit01,1190) nsym
 1190  format(' nsym ',i5)
       write(unit01,'(1x,a6,1x,(t9,3(3i3,1x),4x,3(3i3,1x)))') 'symrel',
     $   (((is(j,i,isym),i=1,3),j=1,3),isym=1,nsym)
c      write(unit01,'(1x,a6,1x,(t9,3f11.8,3x,3f11.8))')
c    $    'tnons',((tau(i,isym),i=1,3),isym=1,nsym)
c      write(unit01,1200) 'xred',((xredsym(i,iatom),i=1,3),iatom=1,natom)
c1200  format(1x,a6,1x,(t9,3f20.12))
c      Print more digits to avoid precision problems with symmetry
       write(unit01,'(1x,a6)') 'tnons'
       write(unit01,1200) ((tau(i,isym),i=1,3),isym=1,nsym)
       write(unit01,'(1x,a6)') 'xred'
       write(unit01,1200) ((xredsym(i,iatom),i=1,3),iatom=1,natom)
 1200  format(t9,3f22.15)
      endif
c
      write(unit00,1210) ibrav,inpgd,nsym
 1210 format(/,'bravais lattice index --> ',i3,
     &      /'point group index --> ',i3,
     &      /'no. of point group elements in scf--> ',i3)
 
c     check whether the point group symmetry is correct or not
 
      call cpointg(b1,b2,b3,ntype,natom,nkatom,atom,
     &   nsym,rot,tnons)            

c     Determine the lattice point group (no atoms) needed for k point stuff
c     Note that this ignores any actual atoms, it only considers
c     the point group symmetries of the lattice by itself
c     Place a single atom at the origin as a place holder
      natom0 = 1
      nat0 = 1
      atom0(1,1) = 0.0d0
      atom0(2,1) = 0.0d0
      atom0(3,1) = 0.0d0
      ty0(1) = 1
      call group1 (unit00,a1,a2,a3,natom0,ty0,nat0,atom0,b1,b2,b3,
     &  ihg,ihc0,symorp,iinv0,nbrav,ind,tau0,f0,r)

c     nbrav is returned as the number of elements in the point group
c     iinv0=0 means no inversion; >0 means inversion is present
c     r(3,3,ind(i)),i=1,nbrav are the point group operations

      if (iinv0 .eq. 0) then
       write(*,1215)
 1215  format(/,' kptgen error: every Bravais lattice must have',
     &  ' inversion symmetry.',/)
       stop
      endif


c     Obtain the lattice symmetries in the reciprocal translation basis
      do i=1,nbrav
       do j=1,3
        do k=1,3
         rot(j,k,i)=r(ind(i),j,k)
        enddo
       enddo
      enddo
      call symgen(a1,a2,a3,b1,b2,b3,nbrav,rot,symlat)
 
c     Done finding symmetry, now input k point information

      read(unit02,*) nkpt

      if (nkpt.gt.mkpt) then
       write(6,1220) nkpt,mkpt
 1220  format(/,'Error: input nkpt ',i10,' exceeds dimensioned mkpt ',
     &  i10,/,'Raise mkpt in kptgen.f and recompile.',/)
       stop
      endif
 
      if (nkpt .gt. 0) then

c      Read k points in terms of recip translations, and weights
       do ikpt=1,nkpt
        read (unit02,*) (ck(i,ikpt),i=1,3)
       enddo
       read (unit02,*) (ck(4,ikpt),ikpt=1,nkpt)
 
      else
 
c      Read the Monhorst-Pack indices and origin for this grid
       read(unit02,*) nk,gk0

c      Compute special point grid
       call speck(unit00,unit01,nk,gk0,b1,b2,b3,nsym,
     &           iinv,ck,mkpt,nkpt,is,ihc0,nbrav,symlat)
 
      endif

      write(*,1230) nkpt
 1230 format('number of special k points in irr bz nkpt is ',i9)

c     check the accuracy of special points
 
      do i=1,nkpt
       kpt(1,i)=ck(1,i)
       kpt(2,i)=ck(2,i)
       kpt(3,i)=ck(3,i)
       wtk(i)  =ck(4,i)
      enddo

      call spchek(a1,a2,a3,mshdim,rsq,rvec,nkpt,kpt,wtk,
     &  nsym,is,irmag,rmag,smag)

      write(unit00,1240) irmag,smag,rmag
      write(unit01,1240) irmag,smag,rmag
      write(*,1240)      irmag,smag,rmag
 1240 format('# phase sum first nonzero at shell',i5,
     &  ', sum=',f10.6,', |R|=',f13.5,' bohr'/)

      write(*,1250) dble(irmag)/dble(nkpt)+1.d-6
 1250 format('==> grid ''efficiency'' (shell/nkpt) = ',f7.1)

      write(*,1500)
 1500 format(/,' kptgen ends normally')

      end
