//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file CrystalLattice.cpp
 *@brief Member function definitions of CrystalLattice<T,D>, included by CrystalLattice.h
 */
#include "Lattice/MakeCrystalLattice.h"

/*! \fn CrystalLattice::CrystalLattice()
 *  Default constructor. Initialized to a \p 1x1x1 cubic supercell.
 */
template<class T, std::size_t D,bool ORTHO>
CrystalLattice<T,D,ORTHO>::CrystalLattice() {
  R.diagonal(1e10);  
  G = R;
  M = R;
  Volume = 1;
  for(std::size_t i=0; i<D; i++) BoxBConds[i] = 1;
  reset();
}

template<class T, std::size_t D,bool ORTHO>
CrystalLattice<T,D,ORTHO>::CrystalLattice(const CrystalLattice<T,D>& rhs) {

  R = rhs.R;
  BConds = rhs.BConds;
  for(std::size_t idir=0; idir<D; idir++) {
    BoxBConds[idir] = rhs.BoxBConds[idir];//!< Bonundary Condition
  }
  reset();
}

template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::set(int argc, char **argv) {
  std::vector<string> opt;
  for(int i=0; i<argc; i++) opt.push_back(argv[i]);
  set(opt);
}

template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::set(vector<string>& argv) {
  makelattice<CrystalLattice<T,D,ORTHO> >::apply(*this, argv);
}


template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::set(const Tensor<T,D>& lat) {
   R = lat;
   reset();
}

template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::set(T sc, T* restrict lat) {
  if(lat) {
    std::copy(lat,lat+D*D,R.data());
    R *= sc;
  } else {
    R = 0.0e0;
    R.diagonal(sc);
  }
  reset();
}

template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::reset() {

  G = inverse(R); //G = transpose(Inverse(R));
  Volume = fabs(det(R));
  //M = dot(transpose(R),R);
  M = dot(R,transpose(R));
  T t = pow(TWOPI,2.0);
  Mg = t*dot(transpose(G),G);
  for(std::size_t i=0; i<D; i++)
    for(std::size_t j=0; j<D; j++)
      Rv[i][j] = R(i,j);
  for(std::size_t i=0; i<D; i++)
    for(std::size_t j=0; j<D; j++)
      Gv[i][j] = G(j,i);

  const T rad_to_deg = 180.0/M_PI;
  T a0=sqrt(dot(Rv[0],Rv[0]));
  T a1=sqrt(dot(Rv[1],Rv[1]));
  T a2=sqrt(dot(Rv[2],Rv[2]));
  ABC[0] = rad_to_deg*acos(dot(Rv[0],Rv[1])*a0*a1);
  ABC[1] = rad_to_deg*acos(dot(Rv[1],Rv[2])*a1*a2);
  ABC[2] = rad_to_deg*acos(dot(Rv[2],Rv[0])*a2*a0);
}

/*! \fn  CrystalLattice<T,D>::operator=(const CrystalLattice<T,D>& rhs)
 *  \param rhs a CrystalLattice to be copied
 *  \brief Copy all the properties of the lattice, 
 *   including boundary conditions and grid paritions.
 */
template<class T, std::size_t D,bool ORTHO>
CrystalLattice<T,D,ORTHO>& 
CrystalLattice<T,D,ORTHO>::operator=(const CrystalLattice<T,D,ORTHO>& rhs) {

  R = rhs.R;
  BConds = rhs.BConds;
  for(std::size_t idir=0; idir<D; idir++) {
   BoxBConds[idir] = rhs.BoxBConds[idir];//!< Bonundary Condition
  }
  reset();
  return *this;
}

/*! \fn  CrystalLattice<T,D>::operator=(const Tensor<T,D>& rhs)
 *  \param rhs a Tensor to be copied
 */
template<class T, std::size_t D,bool ORTHO>
CrystalLattice<T,D,ORTHO>& 
CrystalLattice<T,D,ORTHO>::operator=(const Tensor<T,D>& rhs) {
  R = rhs;
  reset();
  return *this;
}

/*! \fn  CrystalLattice<T,D>::operator*=(T sc)
 *  \param sc A scaling factor.
 *  \brief Rescale this supercell by a scalar.
 */
template<class T, std::size_t D,bool ORTHO>
CrystalLattice<T,D,ORTHO>& 
CrystalLattice<T,D,ORTHO>::operator*=(T sc) {
  R *= sc;
  reset();
  return *this;
}

template<class T, std::size_t D,bool ORTHO>
void CrystalLattice<T,D,ORTHO>::print(ostream& os, int level) const {

  /*\note level == 0: print only the lattice vectors
   *      level == 1: lattice vectors, boundary conditions, grid 
   *      level == 2: + all the internal values
   */

  os << "<lattice>" << endl; 
  for(std::size_t i=0; i<D; i++) {
    os << Rv[i] << endl;
  }
  os << "</lattice>" << endl; 

  if(level > 0) {
    os << "<bconds> ";
    for(std::size_t i=0; i<D; i++) {
      if(BoxBConds[i]) os << " p "; 
      else             os << " n ";
    }
    os << " </bconds>" << endl;
  }

  if(level > 1) {
    os << "Volume (A^3) = " << Volume << endl;
    os << "Reciprocal vectors without 2*pi.\n";
    for(std::size_t i=0; i<D; i++) {
      os << "g_"<< i+1<< " = " << Gv[i] << endl;
    }
    os << "Metric tensor in real-space.\n";
    for(std::size_t i=0; i<D; i++) {
      os << "h_"<< i+1<< " = ";
      for(std::size_t j=0; j< D; j++) {
        os << M(i,j) << " ";
      }
      os << endl;
    }
    os << "Metric tensor in g-space.\n";
    for(std::size_t i=0; i<D; i++) {
      os << "h_"<< i+1<< " = ";
      for(std::size_t j=0; j< D; j++) {
        os << Mg(i,j) << " ";
      }
      os << endl;
    }
  }
}


template<class T, std::size_t D,bool ORTHO>
inline bool operator==(const CrystalLattice<T,D,ORTHO>& lhs, 
                       const CrystalLattice<T,D,ORTHO>& rhs) {
  for(std::size_t i=0; i<D*D; i++) 
    if(abs(lhs.R[i]-rhs.R[i]) > numeric_limits<T>::epsilon()) return false;
  return true;
}

template<class T, std::size_t D,bool ORTHO>
inline bool operator!=(const CrystalLattice<T,D,ORTHO>& lhs, 
                       const CrystalLattice<T,D,ORTHO>& rhs) {
  return !(lhs == rhs);
}


// free function to check if a CrystalLattice is orthorhombic
template<class T, std::size_t D,bool ORTHO>
inline bool orthorombic(const CrystalLattice<T,D,ORTHO>& a) {
  return ORTHO;
}

//
////!< Returns true if the off-diagonal elements of Rm are zero.
//template<class T, std::size_t D, bool ORTHO>
//bool orthorombic(const CrystalLattice<T,3>& a, T) {
//  const T eps = 1e-6;
//  return ((a.R(0,1) < eps) && (a.R(0,2) < eps) &&
//          (a.R(1,0) < eps) && (a.R(1,2) < eps) &&
//          (a.R(2,0) < eps) && (a.R(2,1) < eps) );
//}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: CrystalLattice.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

