//////////////////////////////////////////////////////////////////
// (c) Copyright 2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_UNIFORMGRIDLAYOUT_3D_NONTEMP_H
#define OHMMS_UNIFORMGRIDLAYOUT_3D_NONTEMP_H

#ifdef HAVE_CONFIG_H
#include "ohmms-config.h"
#endif
#include <vector>
#include <iostream>
#include "Lattice/CrystalLattice.h"
#include "Lattice/UniformCartesianGrid.h"

/**@file Uniform3DGridLayout.h
 *@brief Concrete uniform grid layout for 3D
 */

/** A concrete, specialized class equivalent to UniformGridLayout<T,3>
 *
 *This class represents a supercell and is one of the layout classes 
 *to define a trait class PT for ParticleBase<PT> (as a matter of fact,
 *the only layout class implemented in ohmms).
 *
 *It is inherited from CrystalLattice<OHMMS_PRECISION,3> and
 *UniformCartesianGrid<OHMMS_PRECISION,3>: CrystalLattice defines a general
 *supercell and UniformCartesianGrid defines a cubic-cell view of a
 *general supercell. This layout class is intended for typical
 *condensed matter or molecular systems, where a uniform grid partition
 *works fairly well.
 *
 *The additional interfaces of UniformGridLayout class are primarily
 *for NNEngine classes, which use the parition information to evaluate
 *the neighbor lists efficiently. 
 *
 */
//class Uniform3DGridLayout: public CrystalLattice<OHMMS_PRECISION,3,OHMMS_ORTHO>,
//			   public UniformCartesianGrid<OHMMS_PRECISION,3>
class Uniform3DGridLayout: public CrystalLattice<OHMMS_PRECISION,3,OHMMS_ORTHO>
{ 

public:

  enum {IsOrthogonal = OHMMS_ORTHO};

  /**enumeration for grid levels*/
  enum {SPATIAL_GRID=0, /*!< spatial level */
	OMP_GRID,     /*!< open mp level */
        MPI_GRID, /*!< mpi level */
	GridLevel = 3 /*!< counter used to tell the number of levels */
  };

  typedef OHMMS_PRECISION                          value_type;
  typedef UniformCartesianGrid<value_type,int,3>   Grid_t;
  typedef CrystalLattice<value_type,3,OHMMS_ORTHO> Base_t;
  typedef Uniform3DGridLayout                      This_t;

  //using Base_t::

  ///typedef for DistributedIndex_t
  typedef Grid_t::DistributedIndex_t    DistributedIndex_t;
  ///type of index
  typedef Grid_t::Index_t               Index_t;
  ///iterator of the distributed index
  typedef Grid_t::iterator              iterator;
  ///const_iterator of the distributed index
  typedef Grid_t::const_iterator        const_iterator;

  typedef Base_t::SingleParticlePos_t   SingleParticlePos_t;
  typedef Base_t::SingleParticleIndex_t SingleParticleIndex_t;

  /**default constructor
   *
   *Create 1x1x1 cubic view for a supercell assuming that no
   *parallelization is used.
   */
  inline Uniform3DGridLayout():MaxConnections(0), NCMax(0), Status(0){
    for(int ig=0; ig<GridLevel; ig++)
      SuperGrid.push_back(new Grid_t);
  } 

  ~Uniform3DGridLayout() {
    for(int ig=0; ig<GridLevel; ig++) {
      delete SuperGrid[ig];
    }
  }

  /**copy constructor
   */
  inline Uniform3DGridLayout(const Uniform3DGridLayout& pl){
    copy(pl);
  } 

  /** assignmenet operator
   */
  inline Uniform3DGridLayout& operator=(const Uniform3DGridLayout& pl){
    copy(pl);
    return *this;
  } 


  /**copy function
   *
   * Only copy the CrystalLattice and saptial layout. Dataset are not copied.
   */
  inline void copy(const Uniform3DGridLayout& pl){
    MaxConnections=0;
    Base_t::makeCopy(pl);
    *(SuperGrid[SPATIAL_GRID]) = *(pl.SuperGrid[SPATIAL_GRID]);
  } 

  /**copy grid only
   * @param cL Uniform3DGridLayout
   */
  void copyGrid(const This_t& cL) {
    for(int gl=0; gl<GridLevel; gl++) {
      Grid[gl]=cL.Grid[gl];
      *(SuperGrid[gl])=*(cL.SuperGrid[gl]);
    }
  }

  ///return the first cell connected to the ig cell
  inline int first_connected(int ig) const { return c_offset[ig];}
    
  ///return the last cell connected by the interaction radius to the ig cell
  inline int last_connected(int ig) const { return c_max[ig];}

  ///return the last cell connected by the connectivity to the ig cell
  inline int max_connected(int ig) const { return c_offset[ig+1];}
    
  /// return the cell index
  inline int id_connected(int j) const { return c_id[j];}
    
  /// return the correction vector according to the boundary conditions
  inline SingleParticlePos_t bc(int j) const { return c_bc[j];}

  /// return the total number of paritions
  inline std::size_t ncontexts() const { return c_offset.size()-1;}
    
  /// return the maximum number of connected cells
  int connectGrid(value_type int_rad, value_type con_rad=-1);

  void makeShell(std::vector<SingleParticleIndex_t>& RS,
		 std::vector<bool>& FS,
		 const SingleParticleIndex_t& nc,
		 value_type rr);
    
  template<class GIM>
  inline void makeGrid(const GIM& mgrid) {
    int n=GridLevel;
    n = (mgrid.size()<n)? mgrid.size():n;
    for(int ig=0; ig<n; ig++) {
      for(int idir=0; idir<3; idir++) 
        Grid[ig][idir]=mgrid[ig][idir];
      SuperGrid[ig]->setGrid(mgrid[ig][0], mgrid[ig][1],mgrid[ig][2]);
    }
  }

  inline Grid_t* 
  addGrid(int glevel, const SingleParticleIndex_t& agrid) {
    SuperGrid[glevel]->setGrid(agrid);
    return SuperGrid[glevel];
  }

  inline Grid_t* getGrid(int glevel) { return SuperGrid[glevel];}

  inline const Grid_t* getGrid(int glevel) const {return SuperGrid[glevel]; }

  void initGrid(Grid_t& agrid, const Grid_t& subgrid);

  void checkGrid(value_type int_rad);

  inline int ngrid(int ig) const { 
    return Grid[SPATIAL_GRID][ig];
  }

  inline int ngrid(int glevel, int ig) const { 
    return Grid[glevel][ig];
  }

  void print(std::ostream& os) const;

  inline void update() {
    for(int i=0; i<u_bc.size(); i++)  c_bc[i]=toCart(u_bc[i]);
  }

  /** set the lattice vector with a tensor
   *@param lat a tensor representing a supercell
   *
   *In addition to setting the internal parameters, 
   *it updates the Cartesian displacement vectors.
   */
  inline void update(const Tensor<value_type,3>& lat) {
    Base_t::set(lat);
    connectGrid(InteractionRadius,ConnectionRadius);
  }

  //inline void createData(int gl=SPATIAL_GRID) { SuperGrid[gl].createData();}
  //inline void clearData(int gl=SPATIAL_GRID) { SuperGrid[gl].clearData();}
  //inline void addData(int ig, int iat, int gl=SPATIAL_GRID) {
  //  SuperGrid[gl].addData(ig,iat);
  //}

  ///Uniform grid partitions at MPI_GRID, OPENMP_GRID, and SPATIAL_GRID levels.
  SingleParticleIndex_t Grid[GridLevel];

  ///integer to define a status
  unsigned int Status;

  ///The maximum number of connections per cell
  int MaxConnections;

  ///The interaction radius
  value_type InteractionRadius;
 
  ///The radius to define a connected cell
  value_type ConnectionRadius;

  SingleParticleIndex_t NCMax;

  ///UniformCartesianGrid for multi levels.
  std::vector<Grid_t*> SuperGrid;

  ///offsets to determine cell conection
  std::vector<int> c_offset, c_max;

  ///cell index connected to each cell
  std::vector<int> c_id;

  ///displacement vectors due to the boundary conditions
  std::vector<SingleParticlePos_t> c_bc;

  ///displacement vectors due to the boundary conditions in the unit-cell unit
  std::vector<SingleParticlePos_t> u_bc;
};
#endif  
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: Uniform3DGridLayout.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
