//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_UNIFORMCARTESIANGRID_H
#define OHMMS_UNIFORMCARTESIANGRID_H

/**@file UniformCartesianGrid.h
 *@brief Class declarations for UniformCartesianGrid
 */
#include "Utilities/DistributedIndex.h"
#include "OhmmsPETE/TinyVector.h"
#include <iostream>

/* \class UniformCartesianGrid
 * \brief Class to manage a uniform grid.
 *
 * Does nothing and needs to be specialized.
 */
template<class T, class IDType, std::size_t D>
struct UniformCartesianGrid {};

/** Specialization of UniformCartesianGrid<T,D> to 3-Dim domains.
 *
 *Regardless of the shape of the domain, the view of UniformCartesianGrid
 *is a cubic cell. The 3-D position vector is valid only if \f$v = ([0,1),[0,1),[0,1))\f$.
 *
 *The main function of this class is to partition a domain defined by
 *3 orthogonal axes, \f${\hat x}, {\hat y}\f$ and \f${\hat z}\f$.
 *A region or domain is specified by three indices (i,j,k)
 *and they satisfy the conditions \f[ i \in [0,NP[0])\f] \f[ j \in
 *[0,NP[1])\f] \f[ k \in [0,NP[2])\f] The subdomains are stored in an
 *one-dimensional array and key is used to map (i,j,k) to the
 *corresponding domain.  Member function key is provided in
 *anticipation of the need to use a sparse storage of the subdomains
 *to handle inhomogeneous systems.
 */
template<class T, class IDT>
class UniformCartesianGrid<T,IDT,3> {

public:

  typedef IDT                                         Index_t;
  typedef DistributedIndex<list<IDT> >                DistributedIndex_t;
  typedef typename DistributedIndex_t::iterator       iterator;
  typedef typename DistributedIndex_t::const_iterator const_iterator;
  typedef UniformCartesianGrid<T,IDT,3>               This_t;

  ///default constructor
  inline UniformCartesianGrid() { 
    NumGrids = 1;
    NP[0] = 1; NP[1] = 1; NP[2] = 1; 
    Delta[0] = 1; Delta[1] = 1; Delta[2] =1;
    InvDelta[0] = 1; InvDelta[1] = 1; InvDelta[2] =1;
  }

  ///copy constructor
  inline UniformCartesianGrid(const This_t& gr) { 
    setGrid(gr.NP[0],gr.NP[1],gr.NP[2]);
  }

  /**constructor 
   *@param ng a 3-Dim index vector that sets the partition
   */
  inline UniformCartesianGrid(TinyVector<IDT,3>& ng) {
    setGrid(ng[0],ng[1],ng[2]);
  }

  ///destructor
  virtual inline ~UniformCartesianGrid() { }

  ///copy operator
  This_t& operator=(const  This_t& gr) {
    setGrid(gr.NP[0],gr.NP[1],gr.NP[2]);
    return *this;
  }

  ///copy function
  inline void copy(const  This_t& gr) {
    setGrid(gr.NP);
  }

  ///return the total number of sub domains/grids
  inline int getTotalNum() const { return NumGrids;}

  ///return the number of grid in the i-th direction
  inline unsigned int size(int i) const { return NP[i];}

  ///return true if the distributed data set is empty
  inline bool empty() const { return I.empty(); }

  /**set the parameters to partition a 3-Dim domain
   *@param ng a 3-Dim index vector that sets the partition
   */
  template<class IV>
  inline void setGrid(const IV& ng) {
    setGrid(ng[0],ng[1],ng[2]);
  }

  inline void setGrid(Index_t ig, Index_t jg, Index_t kg) {
    NumGrids = ig*jg*kg;
    NP[0] = ig;    NP[1] = jg;    NP[2] = kg;
    InvDelta[0] =static_cast<T>(ig);
    InvDelta[1] =static_cast<T>(jg);
    InvDelta[2] =static_cast<T>(kg);
    Delta[0] =1./InvDelta[0];
    Delta[1] =1./InvDelta[1];
    Delta[2] =1./InvDelta[2];
  }

  /**get a unique key for a domain(i,j,k)
     @param i the index of the first dimension
     @param j the index of the second dimension
     @param k the index of the third dimension
     @return the key for a domain(i,j,k)
     *
     *@warning broken. Should fail.
  */
  inline int key(int i, int j, int k) const {
    return CellKey(k+NP[2]*(j+NP[1]*i));
  }

  /**get a index of a domain(i,j,k)
     @param i the index of the first dimension
     @param j the index of the second dimension
     @param k the index of the third dimension
     @return the storage index of a domain(i,j,k)
  */
  inline Index_t loc(Index_t i, Index_t j, Index_t k) const {
    if(i<0) i+=NP[0];
    if(i>=NP[0]) i-= NP[0];
    if(j<0) j+=NP[1];
    if(j>=NP[1]) j-= NP[1];
    if(k<0) i+=NP[2];
    if(k>=NP[2]) k-= NP[2];
    return k+NP[2]*(j+NP[1]*i);
  }

  /**get a domain index of a 3-D vector v(x,y,z)
   @param x the position in the first dimension
   @param y the position in the second dimension
   @param z the position in the third dimension
   @return the storage index of a domain whose position is p(x,y,z)
   */
  inline Index_t loc(T x, T y, T z) const {
    return loc(static_cast<Index_t>(x*InvDelta[0]),
               static_cast<Index_t>(y*InvDelta[1]), 
               static_cast<Index_t>(z*InvDelta[2]));
  }

  /**get a domain index of a 3-D vector p
   @param p a 3-D vector with operator[]
   @return the storage index of a domain whose position is p
   */
  template<class POSTYPE>
  inline Index_t loc(const POSTYPE& p) const {
    return  static_cast<Index_t>(p[2]*InvDelta[2])
      +NP[2]*(static_cast<Index_t>(p[1]*InvDelta[1])+NP[1]*static_cast<Index_t>(p[0]*InvDelta[0]));
  }

  /**get a 3-D index vector for a position r
   *@param r the position
   *@return a index vector containing the cell indices in the three directions
   */
  inline TinyVector<Index_t,3> index(const TinyVector<T,3>& p) {
    return TinyVector<Index_t,3>(static_cast<Index_t>(p[0]*InvDelta[0]),
			     static_cast<Index_t>(p[1]*InvDelta[1]),
			     static_cast<Index_t>(p[2]*InvDelta[2]));
  }

  /**get a center position of a domain(i,j,k)
     @param i the index of the first dimension
     @param j the index of the second dimension
     @param k the index of the third dimension
     @return the position of the domain(i,j,k)
   */
  inline TinyVector<T,3> center(int i, int j, int k) const {
    return TinyVector<T,3>((static_cast<T>(i)+0.5)*Delta[0],
			   (static_cast<T>(j)+0.5)*Delta[1],
			   (static_cast<T>(k)+0.5)*Delta[2]);
  }

  /**get a center position of a domain(i,j,k)
     @param i the index of the first dimension
     @param j the index of the second dimension
     @param k the index of the third dimension
     @return the position of the domain(i,j,k)
   */
  inline TinyVector<T,3> vertex(int i, int j, int k) const {
    return TinyVector<T,3>((static_cast<T>(i))*Delta[0],
			   (static_cast<T>(j))*Delta[1],
			   (static_cast<T>(k))*Delta[2]);
  }

  /**distribute the domains over the processors
   *@param ntot the total number of processors
   *
   *The processor can be a MPI node or an OpenMP thread.
   *The domains are distributed over the processors as evenly as possible.
   *However, the computataionl nodes are not factored in for load balancing.
   */
  inline void distribute(int ntot) {
    I.distribute(ntot,NumGrids);
  }

  /**distribute the domains over the processors [first, last)
   *@param first iterator for the starting node
   *@param last iterator for the ending node
   *
   *A similar function to the void distribute(int ntot).
   */
  template <class _InputIterator>
  void distribute(_InputIterator first, _InputIterator last) { 
    I.distribute(first,last);
  }

  /**print the domain partition information
   *@param os ostream to write to
   *
   *This is for debug/test only.
   */
  void printGrid(std::ostream& os) const { 
    const int nw = 15;
    os << "number of sub regions " << NumGrids << endl;
    os << "Attribute distribution  = " << endl;
    os << "grid spacing = " << setw(nw) << Delta[0] << setw(nw) << Delta[1]
       << setw(nw) << Delta[2] << endl;

    TinyVector<T,3> origin;
    for(int ig=0; ig<NP[0]; ig++) {
      origin[0] = static_cast<T>(ig)*Delta[0];
      for(int jg=0; jg<NP[1]; jg++) {
	origin[1] = static_cast<T>(jg)*Delta[1];
	for(int kg=0; kg<NP[2]; kg++) {
	  origin[2] = static_cast<T>(kg)*Delta[2];
	  os << setw(nw) << origin[0] << " | " << setw(nw) << origin[0]+Delta[0]
	     << setw(nw) << origin[1] << " | " << setw(nw) << origin[1]+Delta[1]
	     << setw(nw) << origin[2] << " | " << setw(nw) << origin[2]+Delta[2]
	     << std::endl;
	}
      }
    }
  }

  /** resize the grouping of domains
   *@param ng the number of groups
   */
  inline void resizeGroup(int ng) { PID.resize(ng);}

  ///return the processor ID of the group ig 
  inline Index_t group(int ig) const { return PID[ig]; }
  ///assign the processor ID of the group ig 
  inline Index_t& group(int ig) { return PID[ig];}

  inline void createData() { I.create(NumGrids);}
  inline void clearData() { I.clear(); }
  inline void addData(int ig, Index_t iat) {I.add(ig,iat);}

  inline DistributedIndex_t& getDataSets() { return I;}
  inline const DistributedIndex_t& getDataSets() const { return I;}

  inline int getNumData() const {return I.getNumData();}
  inline int getMaxDataPerGrid() const {return I.capacity();}
  inline int getNumDataSets() const { return I.getNumDataSets();}
  inline int getNumData(int i) const { return I.size(i);}
  inline int firstData(int i) const { return I.M[i];}
  inline int lastData(int i) const { return I.M[i+1];}
  inline iterator beginData(int i) { return I.begin(i);}
  inline iterator endData(int i) { return I.end(i);}
  inline const_iterator beginData(int i) const { return I.begin(i);}
  inline const_iterator endData(int i) const  { return I.end(i);}

  void printData(std::ostream& os) const {
    I.print(os);
  }

  int NumGrids;
  Index_t NP[3];
  T Delta[3],InvDelta[3];
  DistributedIndex_t I;
  std::vector<Index_t> PID;
  std::vector<Index_t> CellKey;
};

#endif
  
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: UniformCartesianGrid.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
