//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "NEB/NEBPropagator.h"
#include "ParticleBase/ParticleUtility.h"
#include "Propagators/ConstraintBase.h"
#include "Message/Communicate.h"
#include "Message/OpenMP.h"
using namespace OHMMS;

NEBPropagator::NEBPropagator(): 
Rid(0),T_ref(0.0), Dt(1.0), Friction(0.05)
{ 
  /*! add NEBPropagator specific <parameter/>  */
  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
}


void 
NEBPropagator::setImages(vector<Particle_t*>& images) {
  if(Images.size()) Images.clear();
  for(int ic=0; ic<images.size(); ic++)
    Images.push_back(images[ic]);
}

void 
NEBPropagator::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  myPot = potin;

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  Scalar_t tmp = Friction*Dt*0.5e0;
  c0 = 2.0e0/(1.0e0+tmp);
  c1 = 1.0e0 - c0;
  c2 = Dt*Dt/(1.0e0+tmp);
  oneover2dt = 0.5e0/Dt;

  if(!Initialized) { 
    myPtcl=myPot->getParticleSet();
    myConstraint->set(myPtcl);
    if(TwoOverDOF < 0.0e0) {
      DOF = myConstraint->dof();
      TwoOverDOF = 2.0/DOF;
      LOGMSG("\tThe number of particles " << myPtcl->getTotalNum() 
          << " Degrees of Freedom " << DOF)
    }
    resetBase();
    for(int ic=0; ic<Images.size(); ic++){
      myPtcl=Images[ic];
      LocIndex0.resize(omp_get_max_threads()+1);
      FairDivide(myPtcl->getLocalNum(),omp_get_max_threads(),LocIndex0);

      //add an attribute for previous position and save the index to access the attribute
      Rid =myPtcl->addAttribute(postype_tag, "r1");
      ParticlePos_t& R1 = *myPtcl->getVectorAttrib(Rid);

      R1.setUnit(PosUnit::CartesianUnit);
    }
    Initialized = true;
  }

  for(int ic=0; ic<Images.size(); ic++){
    ParticlePos_t& R1 = *(Images[ic]->getVectorAttrib(Rid));
    R1 = Images[ic]->R;
  }
  ScaledV = 1.0;
  RescaleTemperature = false;
}

void 
NEBPropagator::setTemperature(bool rescale, Scalar_t tref) {
  //do nothing
} 

bool 
NEBPropagator::propagate(Interaction_t* pot) {

  size_type first=0;
  size_type last=myPtcl->getLocalNum();
  int ic=1;
  int nc=Images.size()-1;
  int movingchains=0;
  myPtcl=pot->getParticleSet();
  const ParticleIndex_t& gid = myPtcl->GroupID;

  while(ic<nc) {
    ParticlePos_t& R0(Images[ic]->R);
    ParticlePos_t& F(Images[ic]->F);
    ParticlePos_t& R1(*(Images[ic]->getVectorAttrib(Rid)));
    Scalar_t ke=0.0;
    for(size_type iat=0; iat<last; iat++) {
      Index_t itype(gid[iat]);
      SingleParticlePos_t r2(R1[iat]);
      R1[iat]=R0[iat];
      R0[iat]=c0*R1[iat]+c1*r2+c2*minv[itype]*F[iat];
      SingleParticlePos_t v(oneover2dt*(R0[iat]-r2));
      ke += mass[itype]*dot(v,v);
    }

    //evalKineticEnergy();
    E[KINETIC_ENERGY]= 0.5*myPtcl->collect(ke);
    E[TEMPERATURE] = E[KINETIC_ENERGY]*TwoOverDOF*eV_to_Kelvin;
    movingchains += (E[TEMPERATURE]>KE_min_K);
    ic++;
  }

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  if(movingchains) {
    applyBConds(false);
    return true;
  } else {
    return false;
  }
}

void NEBPropagator::applyBConds(bool wrapping) {
#ifdef HAVE_MPI
  //Exchanging end configurations
  //using Blocking MPI functions directly
  int nitem = OHMMS_DIM*Images[0]->getTotalNum();
  int last_active = Images.size()-2;
  int last_dormant = Images.size()-1;
  MPI_Status status;

  if(I[right_tag] >= 0) 
    MPI_Send(&(Images[last_active]->R[0][0]), nitem, MPI_DOUBLE, 
        I[right_tag], I[send_to_right],MPI_COMM_WORLD);

  if(I[left_tag] >= 0)
    MPI_Recv(&(Images[0]->R[0][0]), nitem, MPI_DOUBLE, 
        I[left_tag], I[send_to_right],MPI_COMM_WORLD, &status);

  if(I[left_tag] >= 0)
    MPI_Send(&(Images[1]->R[0][0]), nitem, MPI_DOUBLE, 
        I[left_tag], I[send_to_left],MPI_COMM_WORLD);

  if(I[right_tag] >= 0)
    MPI_Recv(&(Images[last_dormant]->R[0][0]), nitem, MPI_DOUBLE, 
        I[right_tag], I[send_to_left],MPI_COMM_WORLD, &status);
#endif
  Counter = 0;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 381 $   $Date: 2005-09-05 15:24:10 -0500 (Mon, 05 Sep 2005) $
 * $Id: NEBPropagator.cpp 381 2005-09-05 20:24:10Z jnkim $ 
 ***************************************************************************/
