//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "OhmmsApp/DimerRun.h"
#include "OhmmsApp/RecordPropertyList.h"
#include "OhmmsApp/ProjectData.h"
#include "OhmmsApp/ParticleSets.h"
#include "OhmmsApp/IonicPotential.h"
#include "OhmmsApp/WorkingPropagator.h"
#include "Potentials/PotentialBase.h"
#include "Propagators/DimerPropagator.h"
#include "Utilities/Clock.h"
#include "Utilities/OhmmsInfo.h"
namespace OHMMS {

  DimerRun::DimerRun(OhmmsElementSet& nodelist, const char* aname): 
    OhmmsElementBase(aname),NodeList(nodelist)
  { 
    blocks = 1;
    dimer_steps = 1000;
    relax_steps = 1000;
    dimer_stride = 1;
    relax_stride =10;
    max_cpu = -1.0; // maximum CPU time
  }

  bool DimerRun::get(ostream& os) const {
    return true;
  }

  bool DimerRun::put(istream& is){
    return true;
  }


#if defined(HAVE_LIBXML2)

  bool DimerRun:: put(xmlNodePtr cur){

    DimerPropagator mover;

    RecordPropagator dimer_record(mover);
    RecordPropertyListManager recordlist(NodeList);

    cur = cur->xmlChildrenNode;

    ///for positive temperature, velocity will be assigned with the input temperature
    double t_temperature = -1;

    int irecord = NodeList.size();
    map<string,int> checklist;
    checklist["particlesets"]   = Key_particle;
    checklist["particlelayout"] = Key_playout;
    checklist["ensemble"]       = Key_ensemble;
    checklist["potential"]      = Key_potential;
    checklist["random"]         = Key_random;
    checklist["record"]         = irecord;
    checklist["cpulimit"]       = irecord+1;
    checklist["search"]         = irecord+2;

    while (cur != NULL) {
      ///The first character of the tag is capitalized.
      string cname((const char*)(cur->name));
      //cname[0]=toupper(cname[0]);
      map<string,int>::iterator it = checklist.find(cname);
      if(it != checklist.end()) {
	int inode = (*it).second;
	if(inode < irecord) {
	  NodeList.getNode(inode)->put(cur);
	} else {
	  inode-=irecord;
	  switch(inode) {
	  case(0):
	    recordlist.put(cur);
	    break;
	  case(1):
	    putContent(max_cpu,cur);
	    break;
	  case(2):
	    initSearch(mover,cur);
	    break;
	  }
	}
      }
      cur = cur->next;
    }

    if(max_cpu<0) max_cpu = 43200.; // default is 12 hours

    ParticleSets::Particle_t* workingptcl
      = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle)))->get();

    if(!workingptcl) {
      ERRORMSG("No particle set exists to propagate. Do nothing.")
      return false;
    }

    ///Do not use this anymore
    /// workingptcl->put(cur);

    ProjectData* project 
      = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));

    //setting the object name of particles
    workingptcl->setRoot(project->CurrentRoot());

    if(workingptcl->getTotalNum()) {

      ///init/reset random number generator
      NodeList.getNode(Key_random)->reset();

      PotentialBase* workingpot  
	= dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential))->get(); 

      PropagatorBase* optimizer
	= dynamic_cast<WorkingPropagator*>(NodeList.getNode(Key_ensemble))->get();
      RecordPropagator relax_record(*optimizer);

      ///resetting the recording strides with input data
      dimer_record.stride = dimer_stride;
      relax_record.stride = relax_stride;

      ///connecting particles to a potential
      workingpot->set(workingptcl);

      ///connecting particles to DimerPropagator
      mover.initialize(workingptcl,false);

      ///connecting particles to optimization
      optimizer->initialize(workingptcl,false);

      if(t_temperature > 0) mover.setTemperature(true,t_temperature); 

      ///do the simulation
      int iblock=0;
      bool moving = true;
      int totalsteps = 0;
      Pooma::Clock d_timer;
      d_timer.start();

      do {
	int istep = 0;

        ///prepare all the record and save engines for dimer move
        recordlist.setstride(dimer_stride);
        recordlist.reset();
        dimer_record.reset(project->CurrentRoot());
 	do{
 	  moving = mover.propagate(workingpot);        

	  recordlist.report(totalsteps);
          dimer_record.report(totalsteps);

 	  istep++;totalsteps++;
 	} while(istep < dimer_steps && moving);

	///finalize this
	recordlist.report(totalsteps);
        dimer_record.report(totalsteps);

	///relaxation from the saddle point
	istep = 0;
	project->advance();
	recordlist.setstride(relax_stride);
        recordlist.reset();
	relax_record.reset(project->CurrentRoot());
 	do{
 	  moving = optimizer->propagate(workingpot);        
	  recordlist.report(istep);
          relax_record.report(istep);
 	  istep++;
 	} while(istep < relax_steps && moving);

	///finalize this
	recordlist.report(totalsteps);
        relax_record.report(totalsteps);
	project->advance();


	mover.adjust(false,true); ///start a new dimer search
 	iblock++;
	if(d_timer.cpu_time() > max_cpu) iblock = blocks;
      } while(iblock < blocks);

      return true;

    } else
      return false;
  }
#endif

  void DimerRun::reset() {  }

  bool DimerRun:: initSearch(DimerPropagator& mover,
			     xmlNodePtr cur){
    if(xmlHasProp(cur, (const xmlChar *) "blocks")) {
      blocks = atoi((const char*)(xmlGetProp(cur, (const xmlChar *) "blocks")));
    }

    xmlNodePtr subcur = cur->xmlChildrenNode;
    while(subcur != NULL) {
      string cname2((const char*)(subcur->name));
      if(cname2 == "upmove") {
	xmlNodePtr subcur_save = subcur;
	mover.put(subcur);       
	xmlNodePtr tcur = subcur_save->xmlChildrenNode;
	while(tcur != NULL) {
	  if ((!xmlStrcmp(tcur->name, (const xmlChar *)"parameter"))) {
	    string aname 
	      = (const char*)(xmlGetProp(tcur, (const xmlChar *) "name"));
	    if(aname == "steps") {
	      putContent(dimer_steps,tcur);
	    } else if(aname== "stride") {
	      putContent(dimer_stride,tcur);
	    }
	  }
	  tcur = tcur->next;
	}
      }
      else if(cname2 == "downmove") {
	xmlNodePtr subcur_save = subcur;
	NodeList.getNode(Key_ensemble)->put(subcur);
	xmlNodePtr tcur = subcur_save->xmlChildrenNode;
	while(tcur != NULL) {
	  if ((!xmlStrcmp(tcur->name, (const xmlChar *)"parameter"))) {
	    string aname 
	      = (const char*)(xmlGetProp(tcur, (const xmlChar *) "name"));
	    if(aname == "steps") {
	      putContent(relax_steps,tcur);
	    } else if(aname== "stride") {
	      putContent(relax_stride,tcur);
	    }
	  }
	  tcur = tcur->next;
	}
      }
      subcur = subcur->next;
    }
    return true;
  }


}



/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: DimerRun.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

