//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "OhmmsApp/MDRun.h"
#include "AppBase/ProjectData.h"
#include "OhmmsApp/RecordPropertyList.h"
#include "Utilities/OhmmsInfo.h"
#include "Utilities/Clock.h"
using namespace std;
namespace OHMMS {

  MDRun::MDRun(OhmmsElementSet& nodelist, const char* aname): 
    OhmmsElementBase(aname),NodeList(nodelist)
  {  }

  bool MDRun::get(ostream& os) const {
    return true;
  }

#if defined(HAVE_LIBXML2)

  bool MDRun:: put(xmlNodePtr cur){

    //RecordList is cleared if new <Record/> are found
    bool clearRecordList = true;

    //LINUX+MPI Dies with this
    //NodeList.put(cur);
    cur = cur->xmlChildrenNode;

    int irecord = NodeList.size();
    int iloop = irecord+1;

    map<string,int> nodes;
    nodes["particlesets"]   = Key_particle;
    nodes["particlelayout"] = Key_playout;
    nodes["ensemble"]       = Key_ensemble;
    nodes["potential"]      = Key_potential;
    nodes["record"]         = irecord;
    nodes["loop"]           = iloop;

    vector<xmlNodePtr> Loops;
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      map<string,int>::iterator it = nodes.find(cname);
      if(it != nodes.end()) {
	int inode = (*it).second;
	if(inode < irecord) {
	  NodeList.getNode(inode)->put(cur);
	} else {
	  if(inode == irecord) {
	    RecordList.push(cur);
	  } else if(inode == iloop) {
	    //================================================
            ///Found <loop/> Add the node to Loops
	    //================================================
	    Loops.push_back(cur);
	  }
	}
      }
      cur = cur->next;
    }

    Pooma::Clock timer;
    timer.start();
    bool success = performMDLoops(Loops);
    timer.stop();
    LOGMSG("MD run " << timer.cpu_time() << " secs")
    return success;
  }

  /** erform Molecular Dynamics Simulations by processing loops
   *@param Loops a series of Loops to be processed
   *@return true if successful
   */
  bool MDRun::performMDLoops(vector<xmlNodePtr>& Loops) { 

    //================================================
    //get information about the project.
    //================================================
    ProjectData* project 
      = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));
    bool firstrun = (project->m_series == 0);
    
    //================================================
    //get the working particle
    //================================================
    ParticleSets::Particle_t* workingptcl
      = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle)))->get();
    
    if(!workingptcl) {
      ERRORMSG("No particle set exists to propagate. Do nothing.")
      return false;
    }

    if(!workingptcl->getTotalNum()) {
      ERRORMSG("No particles are added to the particle set. Do nothing.")
      return false;
    }
    //setting the object name of particles
    workingptcl->setRoot(project->CurrentRoot());

    //================================================
    //get the interactions for the particle
    //================================================
    PotentialBase* workingpot  
      =(dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential)))->get();

    if(!workingpot) {
      ERRORMSG("Potential requested is not created.")
      return false;
    }
    workingpot->setParticleSet(workingptcl);

    //================================================
    //get the propagator and initialize it
    //================================================
    PropagatorBase* mover
      =(dynamic_cast<PropagatorBuilder*>(NodeList.getNode(Key_ensemble)))->get();
    mover->initialize(workingptcl,workingpot,false);

    //================================================
    //Create the record list : saved queue<xmlNodePtr> FIFO
    //================================================
    RecordPropertyListManager records(NodeList);
    while(RecordList.size()) {
      records.put(RecordList.front()); RecordList.pop();
    }
    records.reset();

    //================================================
    //set loop parameters and prepare a loop
    //================================================
    int totalsteps = 0;
    bool moving = true;
    bool applybc = true;
    bool resetrecord = false;
    typedef PropagatorBase::Scalar_t Scalar_t;

    int LastLoop = Loops.size()-1;
    for(int iloop=0; iloop<Loops.size(); iloop++) {
      bool newrecord=false;
      int blocks = 1;
      int steps =1;
      bool toadvance = false;
      xmlNodePtr tcur = Loops[iloop];
      xmlChar* b_ptr = xmlGetProp(tcur,(const xmlChar*)"blocks");
      xmlChar* s_ptr = xmlGetProp(tcur,(const xmlChar*)"steps");
      xmlChar* a_ptr = xmlGetProp(tcur,(const xmlChar*)"advance");
      if(b_ptr) blocks = atoi((const char*)b_ptr);
      if(s_ptr) steps = atoi((const char*)s_ptr);
      if(a_ptr) toadvance = atoi((const char*)a_ptr);

      //For the last loop, do not make advance 
      if(iloop == LastLoop) toadvance = false;

      //pass the node so that mover can grep the parameters
      bool changed = mover->setParameters(Loops[iloop]);
      //bool changed = false;
      tcur = tcur->xmlChildrenNode;
      int nparams=0; 
      while(tcur != NULL) {
	string cname2((const char*)(tcur->name));
	//================================================
	// //mdsim/loop/record  Clear the existing list.
	//================================================
	if(cname2 == "record" || cname2 == "Record") {
	  if(!newrecord) {
	    records.push_back();
	    newrecord=true;
	  } 
	  records.put(tcur);
	} 
	tcur = tcur->next;
      }

      if(newrecord) records.reset_top();
      if(changed) mover->initialize(workingptcl,workingpot,false);
      ///initialize or rescale the temperature
      //if(t_temperature>0) mover->setTemperature(firstrun,t_temperature); 
      
      ///always false, after the first call
      firstrun = false;

      //================================================
      //go over the blocks
      //================================================
      int iblock = 0;
      do {

	//================================================
	//go over the steps
	//================================================
        int istep = 0;
	do{
	  moving = mover->propagate(workingpot);        
	  records.report(totalsteps);
	  istep++;totalsteps++;
	} while(istep < steps && moving);
	//mover->applyBConds(applybc);
	iblock++;
	mover->applyBConds(applybc);
      } while(iblock < blocks && moving);
      
      //records.report(totalsteps);
      records.report(-1);
      if(newrecord) records.pop_back();

      if(toadvance) {
         records.finalize();
         project->advance();
         records.reset();
         workingptcl->setRoot(project->CurrentRoot());
      }
    }
    records.finalize();
    return true;
  }
#else
  bool MDRun:: put(xmlNodePtr cur){
    return true;
  }
  bool MDRun:: put(vector<xmlNodePtr>& Loops){
    return true;
  }
#endif

  void MDRun::reset() {  }

  ///useless for now but keeping it for future reference
#if defined(ENABLE_GUI)

  bool MDRun::put(istream& is) {

    wxTextCtrl* temp;
    wxString t;

    wxChoice *choice = (wxChoice*)(wxWindow::FindWindowById(ID_IONIC_ENSEMBLE));
    mdmode = choice->GetSelection();
    mdtype = choice->GetString(mdmode);

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_MDBLOCK));
    blocks = atoi(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_MDSTEP));
    steps = atoi(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_MDLOGSTRIDE));
    stride = atoi(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_MDLOGLEVEL));
    outlevel = atoi(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_IONIC_TEMPERATURE));
    temperature = atof(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_IONIC_TIMESTEP));
    dt = atof(temp->GetValue().c_str());

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_IONIC_PRESSURE));
    pressure = atof(temp->GetValue().c_str());  

    temp = (wxTextCtrl*)(wxWindow::FindWindowById(ID_IONIC_FRICTION));
    friction = atof(temp->GetValue().c_str());  

// #if defined(ENABLE_OHMMS)
//     wxCheckBox* evst = (wxCheckBox*)(wxWindow::FindWindowById(ID_RUNTIMEPLOT_EVST));
//     //first overwrite the stride to be at least equal to the number of steps
//     if(evst->GetValue() && stride <= 0) stride = steps;
//     if(stride > 0) {
//       OHMMS::AddDynamicsAnalysis(analizers,OHMMS::report_energies,stride,
// 				 evst->GetValue());
//     }
//     evst = (wxCheckBox*)(wxWindow::FindWindowById(ID_RUNTIMEPLOT_GOFR));
//     if(evst->GetValue() && stride <= 0) stride = steps;
//     if(stride > 0) {
//       OHMMS::AddDynamicsAnalysis(analizers,OHMMS::report_gofr,stride,
// 				 evst->GetValue());
//     }
// #endif  // ENABLE_OHMMS
  }
#else //!defined(ENABLE_GUI)
  bool MDRun::put(istream& is){
    return true;
  }
#endif //ENABLE_GUI


}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 317 $   $Date: 2005-03-01 17:13:41 -0600 (Tue, 01 Mar 2005) $
 * $Id: MDRun.cpp 317 2005-03-01 23:13:41Z jnkim $ 
 ***************************************************************************/

