//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "OhmmsApp/MinimaHopping.h"
#include "Propagators/CGPropagator.h"
#include "Propagators/EVNVerlet.h"
#include "AppBase/ProjectData.h"
#include "OhmmsApp/IonicPotential.h"
#include "OhmmsApp/PropagatorBuilder.h"
#include "OhmmsApp/RecordPropertyList.h"
#include "Utilities/OhmmsInfo.h"
#include "OhmmsData/ParameterSet.h"
#include "ParticleIO/RecordParticle.h"
#include "ParticleIO/XMLParticleIO.h"
using namespace std;
namespace OHMMS {

  /** constructor
   *
   *Set the control parameters 
   */
  MinimaHopping::MinimaHopping(OhmmsElementSet& nodelist, const char* aname): 
    OhmmsElementBase(aname),NodeList(nodelist),print_debug(false),
    md_engine(0),sd_minimizer(0),cg_minimizer(0),
    recorders(0),myPtcl(0), myPot(0),
    MDSteps(1000),MinSteps(500),SDMinSteps(50),
    KEtarget(300),Etol(1e-3),LowTemperature(5),pe_ptr(0)
  {
    beta1=1.05;
    beta2=beta1;
    beta3=1.0/beta1;
    alpha2=1.02;
    alpha1=1.0/alpha2;
  }

  MinimaHopping::~MinimaHopping() {
    if(cg_minimizer) delete cg_minimizer;
    if(sd_minimizer) delete sd_minimizer;
  }

  bool MinimaHopping::get(ostream& os) const {
    return true;
  }

  bool MinimaHopping::put(std::istream& is) {
    return true;
  }

  void MinimaHopping::reset() {
  }

  bool MinimaHopping::put(xmlNodePtr cur) {

    //read the input parameters
    processXmlNode(cur);

    //================================================
    //get information about the project.
    //================================================
    ProjectData* project 
      = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));
    bool firstrun = (project->m_series == 0);
    
    //================================================
    //get the working particle
    //================================================
    myPtcl = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle)))->get();
    
    if(!myPtcl) {
      ERRORMSG("No particle set exists to propagate. Do nothing.")
      return false;
    }

    if(!myPtcl->getTotalNum()) {
      ERRORMSG("No particles are added to the particle set. Do nothing.")
      return false;
    }
    //setting the object name of particles
    myPtcl->setRoot(project->CurrentRoot());

    //================================================
    //get the interactions for the particle
    //================================================
    myPot=(dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential)))->get();

    if(!myPot) {
      ERRORMSG("Potential requested is not created.")
      return false;
    }
    myPot->setParticleSet(myPtcl);

    //================================================
    //get the propagator and initialize it
    //================================================
    md_engine=(dynamic_cast<PropagatorBuilder*>(NodeList.getNode(Key_ensemble)))->get();
    md_engine->setParameter("temperature",KEtarget);
    md_engine->initialize(myPtcl,myPot,false);

    if(!cg_minimizer) {
      sd_minimizer = new EVNVerlet();
      sd_minimizer->setName("sdmin");
      sd_minimizer->setParameter("timestep",1.0);
      sd_minimizer->setParameter("temperature",0.0);
      sd_minimizer->setParameter("friction",0.5);
      sd_minimizer->setParameter("update",20);
      sd_minimizer->initialize(myPtcl,myPot,false);

      cg_minimizer = new CGPropagator();
      cg_minimizer->initialize(myPtcl,myPot,false);
    }

    myPtcl->convert2Cart(myPtcl->R);
    double pe = getMinimumConfig(myPot);

    //Save current minimum configuration.
    Rmin.resize(myPtcl->R.size());
    Rmin.setUnit(myPtcl->R.getUnit());
    Rcur.resize(myPtcl->R.size());
    Rcur.setUnit(myPtcl->R.getUnit());
    Vcur.resize(myPtcl->R.size());
    Vcur.setUnit(PosUnit::CartesianUnit);
    Rmin = myPtcl->R;
    
    double pe_min = pe;
    PE.push_back(pe);
    Visited.push_back(0);

    int curconf = PE.size()-1;
    saveConfiguration(project->CurrentRoot(),curconf);

    //apply random velocities
    //md_engine->setTemperature(true,KEtarget);

    RecordParticle xyzrecord(*myPtcl);
    xyzrecord.setAnimationRecord(1);
    xyzrecord.reset(project->CurrentRoot());

    char logname[128];
    sprintf(logname,"%s.mh.report",project->CurrentRoot());
    ofstream fout(logname);
    fout.setf(ios::scientific, ios::floatfield);
    fout << setw(5) << curconf 
         << setw(15) << setprecision(5) << PE[curconf] 
         << setw(5) << 1 << endl;

    int trial = 0, totalstep=0, accepted=0,rejected = 0;
    bool apply_bconds=true;
    do {

      //always use rescaling
      if(print_debug) LOGMSG("trial,temperature = " << trial << " " <<KEtarget)
      md_engine->setTemperature(false,KEtarget);

      trial++;

      for(int mdstep=0; mdstep<MDSteps; mdstep++,totalstep++) {
	md_engine->propagate(myPot);        
	if(totalstep%100 == 99) 
	  LOGMSG(totalstep << " " << md_engine->getProperty(PropagatorBase::TEMPERATURE) << " " << md_engine->getProperty(PropagatorBase::FREE_ENERGY))
      }

      //save the current position
      myPtcl->convert(myPtcl->R,Rcur);

      pe = getMinimumConfig(myPot);

      int visited = matchingConfig(pe,curconf);
      if(visited>-1) {

        Visited[visited]++;
	if(visited == curconf) 
	  KEtarget *= beta1;
	else 
	  KEtarget *= beta2;

	//starting with the configuration where minimization has started
	myPtcl->convert(Rcur,myPtcl->R);
	continue;

      } else {
	KEtarget *= beta3;
      }

      if(print_debug)
	LOGMSG("Did not find a match. Energy " << pe << " pe_min = " << pe_min)

      //adjust the downward preference
      if(fabs(pe-PE[curconf]) < Ediff) {
	curconf++;//include history
	accepted++;//only current run
        PE.push_back(pe); 
        Visited.push_back(0);
        Ediff*= alpha1;

	//write to a new configuration
	saveConfiguration(project->CurrentRoot(),curconf);

	pe_min = std::min(pe,pe_min);
        xyzrecord.report(trial);

	//save the minimum configuration, don't know what to do yet
	//myPtcl->Lattice.convert(myPtcl->R,Rmin);
        fout << setw(5) << curconf 
	     << setw(15) << setprecision(5) << PE[curconf] 
	     << setw(5) << 1 << endl;

      } else {//rejected
        Ediff*= alpha2;
	rejected++;
      }

    }while(accepted < NumTargets);

    fout << "##################################################" << endl;
    fout << "#         Summary                                #" << endl;
    fout << "# Accepted configurations = " << accepted << endl;
    fout << "# Rejected configurations = " << rejected << endl;
    fout << "##################################################" << endl;
    for(int i=0; i<PE.size(); i++)
      fout << setw(5)  << i
	   << setw(15) << setprecision(5) << PE[i] 
	   << setw(5)  << Visited[i] << endl;

    xmlNodeSetContent(pe_ptr,(const xmlChar*)" ");
    getContent(PE,pe_ptr);

    return true;
  }


  void 
  MinimaHopping::saveConfiguration(const char* fileroot, int i) {
    XMLSaveParticle xmlfile(*myPtcl);
    char newconfig[128];
    sprintf(newconfig,"%s.f%03d",fileroot,i);
    xmlfile.reset(newconfig);
    xmlfile.stride = 1;
    xmlfile.report(1);
  }



  /** Check if the new configuration is one of the minimum configurations
   *@param pe the potential energy of a new minimum
   *@param curconf the current minimum configuration
   *@return -1 if the new minimum is found. Otherwise, the matching configuration.
   *
   *Using energies with a meV tolerance. Can be improved with better methods.
   */
  int MinimaHopping::matchingConfig(double pe, int curconf) {

    while(curconf>-1) {
      if(fabs(pe-PE[curconf])<Etol) return curconf;
      curconf--;
    }
    return curconf;
  }

  /** Relax the current configuration to a local minimum
   *@param workingpot interactions of the particles
   *@return the potential energy of the local minimum
   *
   *In order to prevent instabilities of CG minimization starting from a configuration
   *far from a local minimum, perform the steepest-decent minimization until
   *the temperature is within LowTemperature.
   *Use block simulations for the steepest-descent minimization before
   *checking the temperature. This could be done after time step.
   *Once the temperature is less than LowTemperature, CG minimization starts
   *until the energy or force convergent criterion are met.
   */
  double MinimaHopping::getMinimumConfig(Interaction_t* workingpot) {

    bool moving = true;
    double CurT = 1000.0;
    int sd_steps=0;

    //set the temperature to zero
    sd_minimizer->setTemperature(false,0.0);
    do {
      for(int iter=0; iter<SDMinSteps; iter++,sd_steps++) {
	moving = sd_minimizer->propagate(workingpot);        
      }
      sd_minimizer->applyBConds(true);
      CurT = sd_minimizer->getProperty(PropagatorBase::TEMPERATURE);
    }while(CurT>LowTemperature);

    double sdpe = sd_minimizer->getProperty(PropagatorBase::POT_ENERGY);
    int istep=0;
    do{
      moving = cg_minimizer->propagate(workingpot);        
      istep++;
    } while(istep < MinSteps && moving);

    double pe = cg_minimizer->getProperty(PropagatorBase::POT_ENERGY);

    if(print_debug)
      LOGMSG("Relaxation " << sd_steps << " " << sdpe << " " << istep << " " << pe)
    return pe;
  }

  /** process xml elements to initialize MinimaHopping
   *@param cur the current xml node
   *
   *Accept 
    - <parameter name="name">value</parameter>.
    - <minima> list of the energies of the known minima </minima>.
   *Other valid sub nodes:
    - <particlesets/>
    - <particlelayout/>
    - <potential/>
    - <ensemble/>
    *Ensemble is for a MD engine, typically tvn, LangevinVVerlet is used.
    */
  void MinimaHopping:: processXmlNode(xmlNodePtr cur){

    int iminima = NodeList.size();
    map<string,int> nodes;
    nodes["particlesets"]   = Key_particle;
    nodes["particlelayout"] = Key_playout;
    nodes["potential"]      = Key_potential;
    nodes["ensemble"]       = Key_ensemble;
    nodes["minima"]         = iminima;
    //nodes["record"]         = irecord;

    ParameterSet m_param;
    m_param.add(print_debug, "debug","bool");
    m_param.add(Ediff,     "ediff","eV");
    m_param.add(KEtarget,  "temperature","K");
    m_param.add(Etol,      "etol","eV");
    m_param.add(beta1,     "beta1","none");
    m_param.add(beta2,     "beta2","none");
    m_param.add(beta3,     "beta3","none");
    m_param.add(alpha1,    "alpha1","none");
    m_param.add(alpha2,    "alpha2","none");
    m_param.add(NumTargets,"target","none");
    m_param.add(MDSteps,   "md-steps","none");
    m_param.add(SDMinSteps,"sd-steps","none");
    m_param.add(MinSteps,   "cg-steps","none");
    m_param.add(LowTemperature,"cg-temperature","none");

    xmlNodePtr cur_save = cur;
    m_param.put(cur);
    m_param.get(cout);

    //if(!recorders)
    //  recorders = new RecordPropertyListManager(NodeList);

    cur = cur_save->children;    
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      map<string,int>::iterator it = nodes.find(cname);
      if(it != nodes.end()) {
	int inode = (*it).second;
	if(inode < iminima) {
	  NodeList.getNode(inode)->put(cur);
	} else if(inode == iminima) {
	  pe_ptr = cur;
	  putContent(PE,cur);
	}
      }
      cur = cur->next;
    }

    //minima is not there, add one for a continuation
    if(!pe_ptr) {
      pe_ptr = xmlNewNode(NULL,(const xmlChar*)"minima");
      xmlAddChild(cur_save,pe_ptr);
    }
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 374 $   $Date: 2005-07-30 16:53:15 -0500 (Sat, 30 Jul 2005) $
 * $Id: MinimaHopping.cpp 374 2005-07-30 21:53:15Z jnkim $ 
 ***************************************************************************/

