//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <string>
#include <vector>
#include <utility>
#include <iostream>
#include <fstream>
using namespace std;
#ifdef HAVE_CONFIG_H
#include "ohmms-config.h"
#include "OhmmsApp/AppConfig.h"
#endif
#include "OhmmsApp/PropagatorBuilder.h"
#include "Propagators/NoPropagator.h"
#include "Propagators/CGPropagator.h"
#include "Propagators/EVNVerlet.h"
#include "Propagators/EVNVelocityVerlet.h"
#include "Propagators/TVNLangevin.h"
#include "Propagators/TPNLangevin.h"
#include "Propagators/LatticeRelaxation.h"
#if defined(ENABLE_FORCECONSTANT)
#include "Propagators/ForceConstant.h"
#endif
#include "Utilities/OhmmsInfo.h"

namespace OHMMS {

  enum {fixedion = 0, sdmin, cgmin, evn, tvn, epn, tpn, phonon, 
	forceconstant, elph, xyz2relaxed, translateall};
  
  int getPropagatorID(const string& mdname) {

    if(mdname == "no") return fixedion;
    else if(mdname == "static") return fixedion;
    else if(mdname == "sdmin") return sdmin;
    else if(mdname == "cgmin") return cgmin;
    else if(mdname == "evn") return evn;
    else if(mdname == "epn") return epn;
    else if(mdname == "tvn") return tvn;
    else if(mdname == "tpn") return tpn;
    else if(mdname == "phonon") return phonon;    
    return -1;
  }

  PropagatorBuilder::PropagatorBuilder(const char* aname):
    OhmmsElementBase(aname), 
    m_mover(NULL), mover_type(-1)
  { }


  PropagatorBase* PropagatorBuilder::get() {
    if(!m_mover) {
      LOGMSG("\tCreating NoPropagator (default) - Static Calculation.\n")
      mover_type = fixedion;
      m_mover = new NoPropagator; 
    }
    return m_mover;
  }

  PropagatorBase* PropagatorBuilder::getCopy() {

    string pengine = getPropagatorName(mover_type,"default");
    PropagatorBase *new_mover = createPropagator(pengine);
    if(new_mover) new_mover->put(m_cur);
    return new_mover;
  }

  void PropagatorBuilder::reset() {
    if(m_mover) m_mover->reset();
  }  

  /// generic output
  bool PropagatorBuilder::get(ostream& os) const {
    return  m_mover->get(os);
  }

  /// generic input
  bool PropagatorBuilder::put(istream& is) {
    return true;
  }

  bool PropagatorBuilder:: put(xmlNodePtr cur){

    ///save the pointer for later use
    m_cur = cur;

    string mdtype = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
    string integrator("default");

    const xmlChar* amethod = xmlGetProp(cur, (const xmlChar *) "method");
    if(amethod) {
      integrator =(const char*)amethod;
    }

    PropagatorBase* old_mover = NULL;
    if(m_mover) {//a Propagator already exists; either re-use or make a new
      int new_type = getPropagatorID(mdtype);
      if(mover_type == new_type) {
	return m_mover->put(cur);
      } else {
	WARNMSG("\tGetting ready to create a new Propagator.")
        mover_type = new_type;
	old_mover = m_mover;
	m_mover = 0;
      }
    } else {
      mover_type = getPropagatorID(mdtype);
    }

    string pengine("invalid");
    const xmlChar* pclass=xmlGetProp(cur,(const xmlChar*)"engine");
    if(pclass) {
      pengine = (const char*)pclass;
    } else {
      pengine = getPropagatorName(mover_type,integrator);
    }

    m_mover = createPropagator(pengine);
    if(m_mover == 0) {
      ERRORMSG("\tFailed to create a Propagator. " << pengine << " is not availble")
      return false;
    }

    LOGMSG("\tPropagator: " << mdtype << " class= " << pengine)
    m_mover->put(cur);

    if(old_mover) {
      WARNMSG("\tDeleting the old Propagator")
      m_mover->copy(old_mover);
      old_mover->getPtcl()->clearAttributes();
      delete old_mover;
    }
    return true;
  }


  std::string 
  PropagatorBuilder::getPropagatorName(int mtype, const string& integrator) {

    std::string pengine("invalid");

    switch (mtype) {
    case (evn):
      if(integrator == "verlet") {
	pengine = "EVNVerlet";
      } else {
	pengine = "EVNVelocityVerlet";
      }
      break;
    case (tvn):
      if(integrator == "nose-hoover") {
	pengine = "TVNNoseHoover";
      } else {
	pengine = "TVNLangevin";
      }
      break;
    case (sdmin):
      if(integrator == "lattice") {
	pengine="TPNLangevin";
      }else {
	pengine="EVNVerlet";
      }  
      break;
    case (cgmin):
      if(integrator == "lattice") {
	pengine="LatticeRelaxation";
      } else {
	pengine="CGPropagator";
      }
      break;
    case (epn):
      if(integrator == "hydro-verlet") {
	pengine = "EPNHydroVerlet";
      } else {
	pengine = "TPNLangevin"; 
      }
      break;
    case (tpn):
      if(integrator == "hydro-langevin") {
	pengine = "TPNHydroLangevin"; 
      } else {
	pengine = "TPNLangevin"; 
      }
      break;
    case (phonon):
      pengine = "ForceConstant";
      break;
    default:
      pengine = "NoPropagator";
      break;
    }
    //return the class name
    return pengine;

  }

  PropagatorBase* 
  PropagatorBuilder::createPropagator(const string& pengine) {

    if(pengine == "NoPropagator") 
      return new NoPropagator;

    if(pengine == "CGPropagator") 
      return new CGPropagator;

    if(pengine == "EVNVelocityVerlet") 
      return new EVNVelocityVerlet;

    if(pengine == "EVNVerlet") 
      return new EVNVerlet;

    if(pengine == "TVNLangevin") 
      return new TVNLangevin;

    if(pengine == "TPNLangevin") 
      return new TPNLangevin;

    if(pengine=="LatticeRelaxation")
      return new LatticeRelaxation;

#if defined(ENABLE_FORCECONSTANT)
    if(pengine=="ForceConstant") {
      return new ForceConstant;
    }
#endif

    return 0;
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: PropagatorBuilder.cpp 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/

