//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "OhmmsApp/ScanUnitCell.h"
#include "AppBase/ProjectData.h"
#include "OhmmsApp/IonicPotential.h"
#include "OhmmsApp/PropagatorBuilder.h"
#include "OhmmsApp/RecordPropertyList.h"
#include "Propagators/PropagatorBase.h"
#include "Utilities/OhmmsInfo.h"
#include "OhmmsData/ParameterSet.h"
#include "ParticleIO/RecordParticle.h"
#include "ParticleIO/XMLParticleIO.h"
using namespace std;
namespace OHMMS {

  /** constructor
   *
   *Set the control parameters 
   */
  ScanUnitCell::ScanUnitCell(OhmmsElementSet& nodelist, const char* aname): 
    OhmmsElementBase(aname),NodeList(nodelist),//print_debug(false),
    MinSteps(500),NumConfig(10),LowerScale(-1.0),UpperScale(-1.1),
    cg_minimizer(0),myPtcl(0), myPot(0)
  {
    FirstLattice.diagonal(1.0);
    LastLattice.diagonal(1.0);
  }

  ScanUnitCell::~ScanUnitCell() {
  }

  bool ScanUnitCell::get(ostream& os) const {
    return true;
  }

  bool ScanUnitCell::put(std::istream& is) {
    return true;
  }

  void ScanUnitCell::reset() {
  }

  bool ScanUnitCell::put(xmlNodePtr cur) {

    //read the input parameters
    processXmlNode(cur);

    //================================================
    //get information about the project.
    //================================================
    ProjectData* project 
      = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));
    bool firstrun = (project->m_series == 0);
    
    //================================================
    //get the working particle
    //================================================
    myPtcl = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle)))->get();
    
    if(!myPtcl) {
      ERRORMSG("No particle set exists to propagate. Do nothing.")
      return false;
    }

    if(!myPtcl->getTotalNum()) {
      ERRORMSG("No particles are added to the particle set. Do nothing.")
      return false;
    }
    //setting the object name of particles
    myPtcl->setRoot(project->CurrentRoot());

    //================================================
    //get the interactions for the particle
    //================================================
    myPot=(dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential)))->get();

    if(!myPot) {
      ERRORMSG("Potential requested is not created.")
      return false;
    }

    //This can be anything
    cg_minimizer =(dynamic_cast<PropagatorBuilder*>(NodeList.getNode(Key_ensemble)))->get();
    cg_minimizer->initialize(myPtcl,myPot,false);

    if(LowerScale>0 && UpperScale>LowerScale) {
      FirstLattice = LowerScale*myPtcl->Lattice.R;     
      LastLattice = UpperScale*myPtcl->Lattice.R;     
    }
      
    Scalar_t del = 1.0/static_cast<Scalar_t>(NumConfig);
    DeltaLattice = del*(LastLattice-FirstLattice);
    Tensor_t curLattice = FirstLattice;
    
    myPtcl->updateLayout(curLattice);

    cout.setf(ios::scientific, ios::floatfield);

    int what2do = 0;
    if(what2do) {

      double pe = getMinimumConfig(myPot);
      cout << myPtcl->Lattice.Volume << " " << pe << endl;
      for(int i=1; i<=NumConfig; i++) {
	curLattice += DeltaLattice;
	myPtcl->updateLayout(curLattice);
	pe = getMinimumConfig(myPot);
	cout << myPtcl->Lattice.Volume << " " << pe << endl;
      }
    } else {

      ParticleSets::Particle_t::ParticlePos_t F(myPtcl->getTotalNum());
      ParticleSets::Particle_t::ParticleScalar_t Eloc(myPtcl->getTotalNum());
      Tensor_t stress = 0.0;
      Scalar_t dedv;
      F=0.0;
      Eloc=0.0;
      cg_minimizer->propagate(myPot);
      double pe = myPot->evalForceAndPressure(F,dedv);            
      double pressure = trace(myPot->getStress())/OHMMS_DIM/myPtcl->Lattice.Volume;

      cout << "#Volume         PotEnergy      Pressure        dE/dVol      Pressure-dE/dVol" << endl;
      cout << "# AA^3            eV            eV/AA^3        eV/AA^3        eV/AA^3     " << endl;
	cout << myPtcl->Lattice.Volume << setw(15) << pe
	     << setw(15) << pressure 
	     << setw(15) << dedv 
	     << setw(15) << pressure-dedv << endl;
      for(int i=1; i<=NumConfig; i++) {
	curLattice += DeltaLattice;
	myPtcl->updateLayout(curLattice);
        myPot->setCurrentPos(myPtcl->R,true);
	stress=0.0;F=0.0;Eloc=0.0;
	pe = myPot->evalForceAndPressure(F,dedv);
	pressure = trace(myPot->getStress())/OHMMS_DIM/myPtcl->Lattice.Volume;
	cout << myPtcl->Lattice.Volume << setw(15) << pe
	     << setw(15) << pressure 
	     << setw(15) << dedv 
	     << setw(15) << pressure-dedv << endl;
      }

    }
    return true;
  }

  /** Relax the current configuration to a local minimum
   *@param workingpot interactions of the particles
   *@return the potential energy of the local minimum
   *
   *In order to prevent instabilities of CG minimization starting from a configuration
   *far from a local minimum, perform the steepest-decent minimization until
   *the temperature is within LowTemperature.
   *Use block simulations for the steepest-descent minimization before
   *checking the temperature. This could be done after time step.
   *Once the temperature is less than LowTemperature, CG minimization starts
   *until the energy or force convergent criterion are met.
   */
  double ScanUnitCell::getMinimumConfig(PotentialCollection* workingpot) {

    bool moving = true;
    int istep=0;
    do{
      moving = cg_minimizer->propagate(workingpot);        
      istep++;
    } while(istep < MinSteps && moving);

    double pe = cg_minimizer->getProperty(PropagatorBase::POT_ENERGY);
    return pe;
  }

  /** process xml elements to initialize ScanUnitCell
   *@param cur the current xml node
   *
   *Accept 
    - <parameter name="name">value</parameter>.
   *Other valid sub nodes:
    - <particlesets/>
    - <particlelayout/>
    - <potential/>
    - <ensemble/>
    *Ensemble is for a MD engine, typically tvn, LangevinVVerlet is used.
    */
  void ScanUnitCell:: processXmlNode(xmlNodePtr cur){

    map<string,int> nodes;
    nodes["particlesets"]   = Key_particle;
    nodes["particlelayout"] = Key_playout;
    nodes["potential"]      = Key_potential;
    nodes["ensemble"]       = Key_ensemble;

    ParameterSet m_param;
    m_param.add(NumConfig, "numconfig","none");
    m_param.add(LowerScale, "lower-scale","none");
    m_param.add(UpperScale, "upper-scale","none");
    m_param.add(FirstLattice, "lower-lattice","AA");
    m_param.add(LastLattice, "upper-lattice","AA");
    m_param.add(MinSteps, "cg-steps","none");

    xmlNodePtr cur_save = cur;
    m_param.put(cur);
    //m_param.get(cout);
    cur = cur_save->children;    
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      map<string,int>::iterator it = nodes.find(cname);
      if(it != nodes.end()) {
	int inode = (*it).second;
	NodeList.getNode(inode)->put(cur);
      } 
      cur = cur->next;
    }
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 317 $   $Date: 2005-03-01 17:13:41 -0600 (Tue, 01 Mar 2005) $
 * $Id: ScanUnitCell.cpp 317 2005-03-01 23:13:41Z jnkim $ 
 ***************************************************************************/

