//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_RECORDPROPERTYBASE_H__
#define OHMMS_RECORDPROPERTYBASE_H__

#include "OhmmsData/OhmmsElementBase.h"
#include <algorithm>
#include <vector>
#include <string>
#include <fstream>
#include <iomanip>

struct RecordProperty {
  
  int stride;
  std::string FileName;
  RecordProperty(): stride(-1),FileName("default"){ }
  
  virtual ~RecordProperty() { }
  virtual void reset(const char* fileroot) = 0;
  virtual void report(int i) = 0;
  virtual void finalize() = 0;
  virtual bool put(xmlNodePtr cur) =0;
};


class RecordPropertyList {

 public:

  typedef std::vector<RecordProperty*> RecordList_t;
  typedef RecordList_t::size_type size_type;
  
  RecordPropertyList(){}
  virtual ~RecordPropertyList(){ clear();}

  static void registerRecord(const std::string& oname,
                             RecordProperty* a) {
    std::map<std::string,RecordProperty*>::iterator it = RegisteredRecords.find(oname);
    if(it == RegisteredRecords.end()) {
       RegisteredRecords[oname] = a;
    } 
  }
  
  inline RecordProperty* addRecord(const std::string& oname) {
    std::map<std::string,RecordProperty*>::iterator it = RegisteredRecords.find(oname);
    if(it != RegisteredRecords.end()) {
      Properties.push_back((*it).second);
      Removable.push_back(false);
      return (*it).second;
    } else {
      return 0;
    }
  }
  inline void addRecord(RecordProperty* a) {
    Properties.push_back(a);
    Removable.push_back(true);
  }
  
  inline void setstride(int n) {
    RecordList_t::iterator it(Properties.begin());
    RecordList_t::iterator it_end(Properties.end());
    while(it != it_end) {
      (*it)->stride = n; ++it;
    }
  }

  inline void resetRecords(const char* aroot) {
    RecordList_t::iterator it(Properties.begin());
    RecordList_t::iterator it_end(Properties.end());
    while(it != it_end) {
      (*it)->reset(aroot);++it;
    } 
  }  

  inline void report(int iter) {
    RecordList_t::iterator it(Properties.begin());
    RecordList_t::iterator it_end(Properties.end());
    while(it != it_end) {
      (*it)->report(iter); ++it;
    }
  }
  
  inline void finalize() {
    RecordList_t::iterator it= Properties.begin();
    RecordList_t::iterator it_end = Properties.end();
    while(it != it_end) {
      (*it)->finalize(); ++it;
    }
  }
  inline void clear() {
    for(int i=0; i<Properties.size(); i++) {
      if(Removable[i]) delete Properties[i];
    }
    Properties.clear();
  }
  
 protected:
  std::vector<RecordProperty*> Properties;  
  std::vector<bool> Removable;
  static std::map<std::string, RecordProperty*> RegisteredRecords;
};


template<class T>
struct RecordNamedProperty: public RecordProperty {

  typedef typename std::vector<T>::size_type size_type;
  std::ostream *OutStream;
  std::vector<T> Values;
  std::vector<std::string> Name;

  RecordNamedProperty(): OutStream(NULL) {
    Values.reserve(20); 
    Name.reserve(20);
  }

  explicit RecordNamedProperty(size_type n) { 
    Values.resize(n,T());
    Name.resize(n);
  }

  ~RecordNamedProperty() { 
    if(OutStream) delete OutStream;
  }

  inline T operator[](size_type i) const { return Values[i];}
  inline T& operator[](size_type i) { return Values[i];}

  inline size_type add(const char* aname) {
    size_type i=0;
    while(i<Name.size()) {
      if(Name[i]== aname) return i;
      i++;
    }
    Name.push_back(aname);
    Values.push_back(T());
    return i;
  }

  inline size_type size() const { return Name.size();}

  inline void resize(size_type n) {
    std::vector<T> a = Values;
    std::vector<std::string> b = Name;
    Values.resize(n,T()); 
    for(size_type i=0; i<a.size(); i++) Values[i] = a[i];
    //std::copy_n(a.begin(), a.size(), Values.begin());
    Name.resize(n);  
    for(size_type i=0; i<a.size(); i++) Name[i] = b[i];
    //std::copy_n(b.begin(), b.size(), Name.begin());
  }


  ///implement virtual functions
  inline void reset(const char* fileroot) { 
    if(OutStream) delete OutStream;
    OutStream = new std::ofstream(fileroot);
    OutStream->setf(std::ios::scientific, std::ios::floatfield);
    OutStream->setf(std::ios::left,std::ios::adjustfield);
    *OutStream << "#   ";
    for(size_type i=0; i<Name.size(); i++) 
      (*OutStream) << std::setw(15) << Name[i].c_str();
    (*OutStream) << std::endl;
    OutStream->setf(std::ios::right,std::ios::adjustfield);
  }

  inline void report(int iter){ 
    if(stride && iter%stride == 0) {
      for(size_type i=0; i<Values.size(); i++) 
	(*OutStream) << std::setw(15) << Values[i];
      (*OutStream) << std::endl;
    }
  }

  void finalize() {}
  bool put(xmlNodePtr cur);
};

#if defined(HAVE_LIBXML2)
template<class T>
bool RecordNamedProperty<T>::put(xmlNodePtr cur){

  xmlAttrPtr att = cur->properties;
  while(att != NULL) {
    std::string aname((const char*)(att->name));
    if(aname == "stride") {
      stride = atoi((const char*)(att->children->content));
    }
    att = att->next;
  }
  return true;
}
#endif
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: RecordProperty.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
