//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/*! \class template<class T, class C> class Vector
 *  \brief A one-dimensional vector class based on PETE 
 *  Closely related to PETE STL vector example. 
 *
 *  Equivalent to blitz::Array<T,1>, pooma::Array<1,T>. 
 *  
 *  class C is a container class. Default is vector<T>
 *
 *  \todo Implement openMP compatible container class or evaluate function.
 *  \todo Implement get/put member functions for MPI-like parallelism 
 */
#ifndef OHMMS_PETE_VECTOR_H
#define OHMMS_PETE_VECTOR_H

#include "PETE/PETE.h"

#include <vector>
#include <iostream>


template<class T, class _Allocator=std::allocator<T> >
class Vector: public std::vector< T,_Allocator> {
public:

  typedef T                    Type_t;
  typedef Vector<T,_Allocator> This_t;
  typedef typename std::vector<T,_Allocator>::size_type     size_type;
  typedef typename std::vector<T,_Allocator>::pointer       pointer;
  typedef typename std::vector<T,_Allocator>::const_pointer const_pointer;

  Vector(){ } // Default Constructor initializes to zero.

  Vector(size_type n){ 
    resize(n);
  }

  // Copy Constructor 
  Vector(const Vector<T,_Allocator> &rhs){
    this->insert(this->begin(),rhs.begin(),rhs.end());
  }

  // Destructor 
  ~Vector() { }

  void create(size_type n){ // add n elements
    this->insert(this->end(), n, T());
  }

  // Assignment Operators
  This_t& operator=(const Vector<T,_Allocator> &rhs) {
    return assign(*this,rhs);
  }

  const This_t &operator=(const Vector<T, _Allocator> &rhs) const {
    return assign(*this, rhs);
  }

  template<class RHS>
  This_t& operator=(const RHS& rhs) {
    return assign(*this,rhs);
  }

  inline pointer data() {return &((*this)[0]);}// return &(X[0]);}
  inline const_pointer data() const { return &((*this)[0]);} //&(X[0]);}

  inline Type_t& operator()(size_type i) { 
    return (*this)[i];
  }

  inline Type_t operator()(size_type i) const { 
    return (*this)[i];
  }

  //----------------------------------------------------------------------
  // parallel communication

  //Message& putMessage(Message& m) const {
  //  m.setCopy(true);
  //  ::putMessage(m, X, X + D);
  //    return m;
  //}

  //Message& getMessage(Message& m) {
  //  ::getMessage(m, X, X + D);
  //  return m;
  //}

//private:
//  Container_t X;
};

//template<class T, class C>
//void Vector<T,C>::resize(unsigned n){
//  X = C(n,T());
//}
//template<class T, class C>
//void Vector<T,C>::create(unsigned n) {
//  X.insert(X.end(), n, T());
//}
 
#include "OhmmsPETE/OhmmsVectorOperators.h"
//-----------------------------------------------------------------------------
// We need to specialize CreateLeaf<T> for our class, so that operators
// know what to stick in the leaves of the expression tree.
//-----------------------------------------------------------------------------
template<class T, class C>
struct CreateLeaf<Vector<T, C> >
{
  typedef Reference<Vector<T, C> > Leaf_t;
  inline static
  Leaf_t make(const Vector<T, C> &a) { return Leaf_t(a); }
};

//-----------------------------------------------------------------------------
// We need to write a functor that is capable of comparing the size of
// the vector with a stored value. Then, we supply LeafFunctor specializations
// for Scalar<T> and STL vector leaves.
//-----------------------------------------------------------------------------
class SizeLeaf
{
public:

  SizeLeaf(int s) : size_m(s) { }
  SizeLeaf(const SizeLeaf &model) : size_m(model.size_m) { }
  bool operator()(int s) const { return size_m == s; }
  
private:
  
  int size_m;
  
};

template<class T>
struct LeafFunctor<Scalar<T>, SizeLeaf>
{
  typedef bool Type_t;
  inline static
  bool apply(const Scalar<T> &, const SizeLeaf &) 
  {
    // Scalars always conform.
    return true;
  }
};

template<class T, class C>
struct LeafFunctor<Vector<T, C>, SizeLeaf>
{
  typedef bool Type_t;
  inline static
  bool apply(const Vector<T, C> &v, const SizeLeaf &s) 
  {
    return s(v.size());
  }
};

//-----------------------------------------------------------------------------
// EvalLeaf1 is used to evaluate expression with vectors.
// (It's already defined for Scalar values.)
//-----------------------------------------------------------------------------

template<class T, class C>
struct LeafFunctor<Vector<T, C>,EvalLeaf1>
{
  typedef T Type_t;
  inline static
  Type_t apply(const Vector<T, C>& vec,const EvalLeaf1 &f)
  {
    return vec[f.val1()];
  }
};

//////////////////////////////////////////////////////////////////////////////////
// LOOP is done by evaluate function
//////////////////////////////////////////////////////////////////////////////////
template<class T, class C, class Op, class RHS>
inline void evaluate(Vector<T, C> &lhs, const Op &op, const Expression<RHS> &rhs)
{
  if (forEach(rhs, SizeLeaf(lhs.size()), AndCombine()))
    {
      // We get here if the vectors on the RHS are the same size as those on
      // the LHS.
      for (int i = 0; i < lhs.size(); ++i)
        {
          // The actual assignment operation is performed here.
          // PETE operator tags all define operator() to perform the operation.
          // (In this case op performs an assignment.) forEach is used
          // to compute the rhs value.  EvalLeaf1 gets the
          // values at each node using random access, and the tag
          // OpCombine tells forEach to use the operator tags in the expression
          // to combine values together.

          op(lhs[i], forEach(rhs, EvalLeaf1(i), OpCombine()));
        }
    }
  else
    {
      cerr << "Error: LHS and RHS don't conform." << endl;
      exit(1);
    }
}
// I/O
template<class T, class C>
std::ostream& operator<<(std::ostream& out, const Vector<T,C>& rhs)
{

  for(int i=0; i<rhs.size(); i++)
    out << rhs[i] << endl;
  return out;
}

#endif // VEKTOR_H

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: OhmmsVector.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
