//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <string>
#include <vector>
#include <deque>
#include <utility>
#include <iostream>
#include <fstream>
using namespace std;

#include "ParticleSets.h"
#include "TransformParticle.h"
#include "Particle/NNEngine.h"
#include "ParticleIO/ParticleLayoutIO.h"
#include "ParticleIO/ParticleParser.h"
#include "ParticleIO/RecordParticle.h"

namespace OHMMS {

  ParticleSets::ParticleSets(const char* aname): 
    OhmmsElementBase(aname), 
    PtclCollection(NULL),
    m_cur(NULL),
    RestartConfig("xml")
  { 
  }

  void ParticleSets::reset() {

    if(m_cur) {
      xmlFreeNodeList(m_cur->xmlChildrenNode);
      m_cur->xmlChildrenNode = NULL;
      xmlNodePtr 
	anode =xmlNewChild(m_cur,m_cur->ns,(const xmlChar*)"particleset",NULL);
      string newfile(get()->getRoot());
      newfile.append(".restart.");
      newfile.append(RestartConfig);
      xmlNewProp(anode,(const xmlChar*)"file",(const xmlChar*)(newfile.c_str()));
    }
  }  

  Particle_t* ParticleSets::get() {
    
    if(!PtclCollection) {
      if(PtclSet.size() == 1) {
	PtclCollection = PtclSet[0];
      }
    }
    return PtclCollection;
  }

  void ParticleSets::addParticleSet(Particle_t* apt) {
    PtclSet.push_back(apt);
  }

#if defined(HAVE_LIBXML2)

  extern bool TransformParticleFactory(vector<Particle_t*>& pin,
				       Particle_t& pout, 
				       xmlNodePtr cur);

  bool ParticleSets::put(xmlNodePtr cur) {

    m_cur = cur;
    cur = cur->xmlChildrenNode;
    ParticleLayout_t lat;
    bool gotalattice = false;

    vector<xmlNodePtr> builder;

    ///get the xmlNodePtr for ParticleLayout
    xmlNodePtr cur_layout = NULL;

    while (cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "particleset") {
	Particle_t* aptcl = new Particle_t;
	ParticleParser aHandle(*aptcl);
	aHandle.put(cur);
	addParticleSet(aptcl);
      } else if (cname == "UnitCell" || cname == "unitcell") {
	LatticeParser lat(Lattice);
	lat.put(cur);
	gotalattice = true;
      } else if(cname == "transform" || cname == "TransformParticle") {
	builder.push_back(cur);
      } else if(cname == "particlelayout") {
	cur_layout = cur;
      }
      cur = cur->next;
    }
    
    if(gotalattice) {
      ///overwrite the unit cell with the input
      WARNMSG("Overwriting unitcells of sub particles")
      ParticleSet_t::iterator it=PtclSet.begin();
      while(it != PtclSet.end()) {
	(*it)->Lattice = Lattice;it++;
      }
    } else {
      ///use the unitcell of the first particle set
      if(!PtclSet.empty()) {
	Lattice = PtclSet.front()->Lattice;
      }
    }

    ///transformation is required
    vector<xmlNodePtr>::iterator it = builder.begin();
    while(it != builder.end()) {
      bool needtoremove = false;
      if(!PtclCollection) {
	PtclCollection =  new Particle_t;
	needtoremove = true;
      }

      if(!TransformParticleFactory(PtclSet,*PtclCollection,(*it))) {
        ///transform is invalid
        if(needtoremove) {delete PtclCollection; PtclCollection=NULL;}
      }
      it++;
    }

    ///<ParticleLayout/> is found, create the layout
    if(cur_layout) {
      NNEngine* neweng = NNEngineBuilder::create(cur_layout);
      get()->setNNEngine(neweng);
    }

    return true;
  }

  RecordProperty* ParticleSets::createRecordProperty(xmlNodePtr cur){
    Particle_t* thisptcl = get();
    bool dumpit = false;
    string aform;

    ///process attributes: type or format
    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "type") {
	if(vname == "restart") dumpit = true;
      } else if(aname == "format") {
	aform =vname;
      }
      att = att->next;
    }

    RecordProperty *arecord = NULL;
    if(dumpit) {
      if(aform.size()) {
	///@warning hdf5 is called within xml file now.
	///@todo hdf5 should handle everything by itself
	if(aform == "hdf5") aform = "xml";
	RestartConfig = aform;
      }
      arecord = new SaveParticle(*thisptcl);
    } else {
      arecord = new RecordParticle(*thisptcl);
    }
    arecord->put(cur);
    return arecord;
  }

#endif
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ParticleSets.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
