//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////

/***************************************************************************
 * OptimizationBase is a base class to perform optimization.
 * Subroutines brent and mnbrak of NRC are slightly modified for C++.
 * Currently only works for minimization.
 * Any function to optimize should be inherited from OptimizationBase.
 * Inherited class has to provide Func(Element_t dl).
 * Only interface you need to call is
 *  bool lineoptimization()
 * Example:
 * struct Polynomial: public OptimizationBase {
 *   // Func is a virtual function to return a value for
     //    old variables + dl * perturbations (e.g., conjugate gradients)
 *   Element_t Func(Element_t dl) { return poly(x0 + dl*h);}
 *   Element_t x0, h;
 *   Element_t poly(Element_t x) { return pow(x+1,3) + pow(x-2,4);}  
 *   Element_t dpoly(Element_t x) { return -3*pow(x+1,2)-4.0e0*pow(x-2,3);}
 *
 *   Polynomial() { x0 = 0;}
 *   ~Polynomial() { }
 * };

 ***************************************************************************/
#ifndef OHMMS_CG_OPTIMIZATIONBASE_H
#define OHMMS_CG_OPTIMIZATIONBASE_H

/** trait class for CGOptimization */
template<class AT>
struct TypeTraits { };

template<>
struct TypeTraits<std::vector<double> > { 
  typedef double value_type;
  typedef double real_type;
};

template<>
struct TypeTraits<Tensor<double,3> > {
  typedef double value_type;
  typedef double real_type;
};


template<>
struct TypeTraits<ParticleAttrib<TinyVector<double,3> > > {
  typedef double value_type;
  typedef double real_type;
};


template<class T>
struct sign2 { };

template<>
struct sign2<double> {
  inline static double apply(double a, double b) {
    return (b > 0.0)? fabs(a): -fabs(a); 
  }
};

template<class AT>
struct CGOptimization {

  typedef typename TypeTraits<AT>::real_type Return_t;

  bool RestartCG;
  int CGmode;
  int ITMAX;
  int NumLineMin;
  int MinCGSteps;
  Return_t ZEPS, CGOLD, TOL, GLIMIT, TINY, GOLD;
  Return_t Lambda, LambdaMax;

  Return_t gdotg, gdotg0, gdoth, gamma;
  Return_t d_pe_save, deltaEmax, gdotgmax, fmax;

  inline CGOptimization():RestartCG(true),CGmode(0),ITMAX(100),NumLineMin(0),MinCGSteps(1) {
    ZEPS = 1.0e-10;
    CGOLD = 0.3819660e0;
    GOLD = 1.618034e0;
    TOL = 2.0e-4;
    GLIMIT = 100.0;
    TINY = 1.0e-20;
    LambdaMax = 0.2;
  }

  virtual ~CGOptimization() {}

  inline bool converged(Return_t fx) {
    if(fx<fmax) return true;
    if(gdotg<gdotgmax) return true;
    //if(fabs(gamma) < deltaEmax) return true;
    return false;
  }

  bool lineoptimization();


  inline void shift(Return_t& a, Return_t& b, Return_t& c, Return_t d) {
    a = b; b= c; c = d;
  }

  Return_t brentNRC(Return_t ax, Return_t bx, Return_t cx, Return_t & xmin);

  void mnbrakNRC(Return_t& ax,Return_t& bx,Return_t& cx,
		 Return_t& fa,Return_t& fb,Return_t& fc );

  //Return_t lineoptimization(const vector<Return_t>& e, bool min=true);
  virtual Return_t Func(Return_t dl) = 0;

};

template<class AT>
inline bool
CGOptimization<AT>::lineoptimization() { 
  Return_t ax = 0;
  Return_t bx, fa, fx, fb;
  Return_t xx = LambdaMax;
    
  mnbrakNRC(ax,xx,bx,fa,fx,fb);
    
  Lambda = 0.0e0;
  Return_t ep = brentNRC(ax,xx,bx,Lambda);
    
  if(Lambda <= 1.0e-5) return false;
  else {
    Return_t e0 = Func(Lambda);
    return true;
  }
}


template<class AT>
inline typename CGOptimization<AT>::Return_t
CGOptimization<AT>::brentNRC(Return_t ax, Return_t bx, Return_t cx, Return_t & xmin) {

  Return_t a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  Return_t e=0.0e0;

  a=((ax < cx) ? ax : cx);
  b=((ax > cx) ? ax : cx);

  x=w=v=bx;

  fw=fv=fx=Func(x);

  for(int iter=1;iter<=ITMAX;iter++) {
    xm=0.5*(a+b);
    tol2=2.0*(tol1=TOL*fabs(x)+ZEPS);
    if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
      xmin=x;
      return fx;
    }
    if (fabs(e) > tol1) {
      r=(x-w)*(fx-fv);
      q=(x-v)*(fx-fw);
      p=(x-v)*q-(x-w)*r;
      q=2.0*(q-r);
      if (q > 0.0) p = -p;
      q=fabs(q);
      etemp=e;
      e=d;
      if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
	d=CGOLD*(e=(x >= xm ? a-x : b-x));
      else {
	d=p/q;
	u=x+d;
	if (u-a < tol2 || b-u < tol2) d=sign2<Return_t>::apply(tol1,xm-x);
      }
    } else {
      d=CGOLD*(e=(x >= xm ? a-x : b-x));
    }
    u=(fabs(d) >= tol1 ? x+d : x+sign2<Return_t>::apply(tol1,d));
    fu = Func(u);// fu=(*f)(u);
    if (fu <= fx) {
      if (u >= x) a=x; else b=x;
      shift(v,w,x,u);
      shift(fv,fw,fx,fu);
    } else {
      if (u < x) a=u; else b=u;
      if (fu <= fw || abs(w-x)<TINY) { //(fu <= fw || w == x) {
	v=w;
	w=u;
	fv=fw;
	fw=fu;
      } else if (fu <= fv || fabs(v-x)<TINY || fabs(v-w)<TINY) { //(fu <= fv || v == x || v == w) {
	v=u;
	fv=fu;
      }
    }
  }

  xmin=x;
  return fx;
}

template<class AT>
inline void 
CGOptimization<AT>::mnbrakNRC(Return_t& ax,Return_t& bx,Return_t& cx,
			      Return_t& fa,Return_t& fb,Return_t& fc ) {
    
  Return_t ulim,u,r,q,fu,dum = 0.0e0;
    
  fa = Func(ax); // *fa=(*func)(*ax);
  fb = Func(bx); // *fb=(*func)(*bx);
    
  if (fb > fa) {
    shift(dum,ax,bx,dum);
    shift(dum,fb,fa,dum);
  }
    
  cx=bx+GOLD*(bx-ax);
  fc = Func(cx); // *fc=(*func)(*cx);
  while (fb > fc) {
    r=(bx-ax)*(fb-fc);
    q=(bx-cx)*(fb-fa);
    u=(bx)-((bx-cx)*q-(bx-ax)*r)/(2.0*sign2<Return_t>::apply(max(fabs(q-r),TINY),q-r));
    ulim=(bx)+GLIMIT*(cx-bx);
    if ((bx-u)*(u-cx) > 0.0) {
      fu = Func(u); // fu=(*func)(u);
      if (fu < fc) {
	ax=bx;
	bx=u;
	fa=fb;
	fb=fu;
	return;
      } else if (fu > fb) {
	cx=u;
	fc=fu;
	return;
      }
      u=cx+GOLD*(cx-bx);
      fu = Func(u);//fu=(*func)(u);
    } else if ((cx-u)*(u-ulim) > 0.0) {
      fu = Func(u);//fu=(*func)(u);
      if (fu < fc) {
	  
	shift(bx,cx,u,cx + GOLD*(cx-bx));
	shift(fb,fc,fu,Func(u)) ;
      }
    } else if ((u-ulim)*(ulim-cx) >= 0.0) {
      u=ulim;
      fu = Func(u);//fu=(*func)(u);
    } else {
      u=cx+GOLD*(cx-bx);
      fu = Func(u);//fu=(*func)(u);
    }
    shift(ax,bx,cx,u);
    shift(fa,fb,fc,fu);
  }
}


#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 258 $   $Date: 2005-01-14 11:44:55 -0600 (Fri, 14 Jan 2005) $
 * $Id: CGOptimization.h 258 2005-01-14 17:44:55Z jnkim $ 
 ***************************************************************************/
