//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
/***************************************************************************
 * OptimizationBase is a base class to perform optimization.
 * Subroutines brent and mnbrak of NRC are slightly modified for C++.
 * Currently only works for minimization.
 * Any function to optimize should be inherited from OptimizationBase.
 * Inherited class has to provide Func(Element_t dl).
 * Only interface you need to call is
 *  bool lineoptimization()
 * Example:
 * struct Polynomial: public OptimizationBase {
 *   // Func is a virtual function to return a value for
     //    old variables + dl * perturbations (e.g., conjugate gradients)
 *   Element_t Func(Element_t dl) { return poly(x0 + dl*h);}
 *   Element_t x0, h;
 *   Element_t poly(Element_t x) { return pow(x+1,3) + pow(x-2,4);}  
 *   Element_t dpoly(Element_t x) { return -3*pow(x+1,2)-4.0e0*pow(x-2,3);}
 *
 *   Polynomial() { x0 = 0;}
 *   ~Polynomial() { }
 * };

 ***************************************************************************/
#ifndef OHMMS_OPTIMIZATIONBASE_H
#define OHMMS_OPTIMIZATIONBASE_H

#include <vector>
using std::vector;

struct OptimizationBase {

  typedef double Return_t;

  // If successful, Lambda > 1.0e-5. Be sure to update the variables
  // after lineoptimization is successfully performed.
  bool lineoptimization(); 

  Return_t lineoptimization(const vector<Return_t>& e, bool min=true);
  virtual Return_t Func(Return_t dl) = 0;

  Return_t brentNRC(Return_t ax, Return_t bx, Return_t cx, Return_t & xmin);
  void mnbrakNRC(Return_t& ax,Return_t& bx,Return_t& cx,Return_t& fa,Return_t& fb,Return_t& fc );

  OptimizationBase();
  ~OptimizationBase() {}

  int ITMAX;
  Return_t ZEPS, CGOLD, TOL, GLIMIT, TINY, GOLD;
  Return_t Lambda, LambdaMax;
};
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 260 $   $Date: 2005-01-14 13:14:41 -0600 (Fri, 14 Jan 2005) $
 * $Id: OptimizationBase.h 260 2005-01-14 19:14:41Z jnkim $ 
 ***************************************************************************/
