//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Particle/BulkSystemBuilder.h"
#include "ParticleBase/ParticleFunctions.h"

namespace OHMMS {

  void CreateBulkSystem::printUsage(ostream& os) {
    os << endl
       << "Usage: --unitcell x1 y1 z1 x2 y2 z2 x3 y3 z3 | \\" << endl
       << "       --lattice cubic|fcc|bcc|hcp a [b] \\" << endl
       << "      [--crystal cubicdiamond|fccdiamond] " << endl << endl;
    os << "Without --crystal, a monoatomic crystal is created." << endl;
  } 

  /**@fn Particle_t* CreateBulkSystem::apply(int argc, char **argv)
   *@param argc the number of arguments
   * @param argv the argumens
   * @return a new OhmmsParticle
   * @brief
   * A simple function to create a bulk system using argumets.
   * Usage is 
   * --unitcell x1 y1 z1 x2 y2 z2 x3 y3 z3 | --lattice cubic|fcc|bcc|hcp a [b]
   * [--crystal cubicdiamond|fccdiamond|cubiczb]
   *
   * Only the last --unitcell or --lattice  is valid.
   * The option --crystal is optional. If no valid structure is provided or the
   * name is mispelled, it assumes that the crystal has one atom at the origin.
   */
  void CreateBulkSystem::apply(ParticleSet* sys,
			       int& argc, char**& argv) {

    typedef ParticleSet::ParticleLayout_t         ParticleLayout_t;
    typedef ParticleLayout_t::SingleParticlePos_t SingleParticlePos_t;

    //! Enumeration for the crystalline structure
    enum {MONOATOM=0, 
	  CUBICDIAMOND,  
	  FCCDIAMOND,
          CUBICZB,
          FCCZB,
          };    

    vector<string> options; 
    map<string,int> crystals;
    crystals["cubicdiamond"] = CUBICDIAMOND;
    crystals["fccdiamond"] = FCCDIAMOND;
    crystals["cubiczb"] = CUBICZB;
    crystals["fcczb"] = FCCZB;
    bool expand = false;

    string thiscrystal;
    TinyVector<int,OHMMS_DIM> uc_grid(1);

    int retargc = 1;
    char **retargv = new char *[argc];
    retargv[0] = argv[0];
    
    int i=1;
    while(i<argc) {
      string aword(argv[i]);
      if(aword == "--unitcell") {
	ParticleLayout_t::Tensor_t uc;
	for(int idim=0; idim<OHMMS_DIM; idim++) 
	  for(int jdim=0; jdim<OHMMS_DIM; jdim++) 
	    uc(idim,jdim) = atof(argv[++i]);
	sys->Lattice.set(uc);
      } else if(aword == "--lattice") {
	vector<string> v;
	v.push_back(argv[++i]);
	v.push_back(argv[++i]);
        if(i<argc-1) {
  	  string t(argv[++i]);
	  if(isdigit(t[0])) 
	    v.push_back(t);
	  else 
	    --i; //move back
        }
	sys->Lattice.set(v);
      } else if(aword == "--uc_grid") {
	expand = true;
	for(int idim=0; idim<OHMMS_DIM; idim++) uc_grid[idim] = atoi(argv[++i]);
      } else if(aword == "--crystal") {
	thiscrystal = argv[++i];
      } else {
	retargv[retargc++]  = argv[i];
      }
      ++i;
    }

    argc = retargc;
    for (i = 0; i < retargc; ++i) {argv[i] = retargv[i];}
    delete [] retargv;

    ///default is with one base atom per unit cell
    int crystaltype = MONOATOM;

    map<string,int>::iterator it = crystals.find(thiscrystal);
    if(it != crystals.end()) {
      crystaltype = (*it).second;
      sys->setName(thiscrystal);
    }
    sys->R.setUnit(PosUnit::LatticeUnit);

    switch(crystaltype) {
    case(MONOATOM):
      sys->create(1);
      sys->R[0] = SingleParticlePos_t(0.0,0.0,0.0);
      sys->ID[0] =0;
      sys->GroupID = 0;
      break;
    case(CUBICDIAMOND):
      sys->create(8);
      sys->GroupID = 0;
      for(int i=0; i<8; i++) sys->ID[i] = i;
      sys->R[0] = SingleParticlePos_t(0.0,0.0,0.0);
      sys->R[1] = SingleParticlePos_t(0.25,0.25,0.25);
      sys->R[2] = SingleParticlePos_t(0.5,0.5,0.0);
      sys->R[3] = SingleParticlePos_t(0.75,0.75,0.25);
      sys->R[4] = SingleParticlePos_t(0.5,0.0,0.5);
      sys->R[5] = SingleParticlePos_t(0.75,0.25,0.75);
      sys->R[6] = SingleParticlePos_t(0.0,0.5,0.5);
      sys->R[7] = SingleParticlePos_t(0.25,0.75,0.75);
      break;
    case(FCCDIAMOND):
      sys->create(2);
      sys->GroupID = 0;
      sys->R[0] = SingleParticlePos_t(0.0,0.0,0.0);
      sys->R[1] = SingleParticlePos_t(0.25,0.25,0.25);
      for(int i=0; i<2; i++) sys->ID[i] = i;
      break;
    case(CUBICZB):
      sys->create(8);
      for(int i=0; i<8; i++) sys->ID[i] = i;
      for(int i=0; i<8; i+=2) {sys->GroupID[i] = 0; sys->GroupID[i+1]=1;}
      sys->R[0] = SingleParticlePos_t(0.0,0.0,0.0);
      sys->R[1] = SingleParticlePos_t(0.25,0.25,0.25);
      sys->R[2] = SingleParticlePos_t(0.5,0.5,0.0);
      sys->R[3] = SingleParticlePos_t(0.75,0.75,0.25);
      sys->R[4] = SingleParticlePos_t(0.5,0.0,0.5);
      sys->R[5] = SingleParticlePos_t(0.75,0.25,0.75);
      sys->R[6] = SingleParticlePos_t(0.0,0.5,0.5);
      sys->R[7] = SingleParticlePos_t(0.25,0.75,0.75);
      break;
    case(FCCZB):
      sys->create(2);
      sys->R[0] = SingleParticlePos_t(0.0,0.0,0.0);
      sys->R[1] = SingleParticlePos_t(0.25,0.25,0.25);
      for(int i=0; i<2; i++) sys->ID[i] = i;
      sys->GroupID[0] = 0; sys->GroupID[1] = 1;
      break;
    }

    if(expand) {
      ExpandSuperCell(*sys,uc_grid);
    } else {
      TinyVector<int,3> sg(1);
      sys->Lattice.setGrid(sg);
    }
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: BulkSystemBuilder.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
