//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <fstream>
#include "Message/Communicate.h"
#include "Message/CommOperators.h"
#include "Message/OpenMP.h"
#include "Utilities/UtilityFunctions.h"
#include "Utilities/OhmmsInfo.h"
#include "Particle/NNEngine.h"
#include "ParticleBase/ParticleUtility.h"

using namespace OHMMS;

NNEngine::NNEngine():
  NeedPartition(true),  
  UseVisitingForce(false),
  NewConfiguration(true),
  UseLayout(false),
  Reordered(false),IndexOnly(false),Symmetric(false),Unique(true), 
  MethodID(0),Stride(1),Counter(-1),TotalNum(0), NumGhosts(0), 
  MaxPairsPerPtcl(0), TotalNumPairs(0),
  MaxPtclsPerCell(0), MaxConnectedCells(0), ActivePtcl(0),
  Rmax(3.0), Rscale(1.01),RmaxSq(1.0),RminSq(1e-6), PresetMaxNN(0), 
  OmpAttrib(0),Cell(0)
{ 
  DEBUGMSG("NNEngine::NNEngine")

  NumThreads=omp_get_max_threads();
  PtclDist.reserve(NumThreads+1);

  //assign default value for the grids and set flags to false
  Grid.resize(GridLevel);
  for(Index_t ig=0; ig<GridLevel; ig++) Grid[ig] = SingleParticleIndex_t(1);
  for(Index_t ig=0; ig<GridLevel; ig++) ModifyGrid[ig] = false;

  //create PairAttrib with null points 
  for(Index_t ip=0; ip<NumThreads; ip++) PairAttrib.push_back(0);
}

NNEngine::~NNEngine(){ 
  for(Index_t i=0; i<PairAttrib.size(); i++) if(PairAttrib[i]) delete PairAttrib[i];
  if(Cell) delete Cell;
}

void NNEngine::addPairList(Index_t nat, Index_t maxnn){
  if(PtclDist.empty()) PtclDist.push_back(0);
  Index_t myID = PairAttrib.size();
  PairAttrib.push_back(new PtclPairList_t());
  PairAttrib[myID]->resizePairs(nat,maxnn);
  PtclDist.push_back(PtclDist[myID]+nat);
}

/** evaluate the total number of pairs
 *
 *When OpenMP is enabled, it sets up an index array to query the thread id
 *where each particle resides if the interaction uses the visiting force model.
 */
void NNEngine::sync() {
  TotalNumPairs = 0; 
  PtclDist[0]=0;
  for(Index_t i=0; i<PairAttrib.size(); i++) {
    TotalNumPairs += PairAttrib[i]->getTotNadj();
    PtclDist[i+1]=PtclDist[i]+PairAttrib[i]->getLocalNum();
    PairAttrib[i]->set(i,PtclDist[i],PtclDist[i+1]);
  }
}

bool NNEngine::addConnection(const string& cname, vector<MetisConnection*>* deps) {
  map<string,vector<MetisConnection*>*>::iterator it(Connections.find(cname));
  if(it == Connections.end()) {
    Connections[cname]=deps;
  }
  return false;
}

/** a default partition mode
 * @param p Particletype
 *
 * It can handle either OpenMP or MPI parallelization.
 * This partition is for the replicated data algorithm without data distribution. 
 * Each processor has complete data but does work on different parts of the data.
 * Derived class should overwrite this function to handle more complex parallelization.
 * Use a global communicator OHMMS::Controller in Message/Communicate.h
 */
void NNEngine::makeFairPartition(Particle_t& p) {

  ///mpi overwrites openmp: single mpi partition is assumed.
  //\todo include using groups of mpi nodes
  if(Distributed[enable_mpi]) {
    p.setMode(Particle_t::Mixed);
    //=======================================================================
    //Distribute n particles over the mpi processors. Spatial locality is not
    //considered in distributing the particles. Simply divide n particles
    //over m processors.
    //=======================================================================
    vector<Index_t> mPtclDist;
    FairDivide(p.getTotalNum(),Controller->ncontexts(),mPtclDist);

    //=======================================================================
    //Distribute the particles in a node over the threads. Spatial locality is
    //ignored. 
    //=======================================================================
    Index_t nodeid = Controller->mycontext();
    Index_t offset = mPtclDist[nodeid];
    FairDivide(mPtclDist[nodeid+1]-offset, omp_get_max_threads(),PtclDist);
    for(Index_t i=0; i<PtclDist.size(); i++) PtclDist[i] += offset;
  } else {
    p.setMode(Particle_t::Serial);
    FairDivide(p.getTotalNum(),omp_get_max_threads(),PtclDist);
  }

}

void NNEngine::copy(const NNEngine& inda) {

  Index_t np=inda.size();

  //resize the PairAttib first
  PtclPairList_t::difference_type dn(np-size());
  if(dn>0) {
    while(dn) {
      PairAttrib.push_back(new PtclPairList_t());--dn;
    }
  } else if(dn<0){//need to remove extra
    for(Index_t i=np; i<size(); i++) delete PairAttrib[i];
    PairAttrib.erase(PairAttrib.begin()+np, PairAttrib.end());
  }

#pragma omp parallel for 
  for(Index_t i=0; i<np; i++) {
    PairAttrib[i]->copy(inda(i));
  }
}

bool NNEngine::get(ostream& os) const{
  os << "Method ID = " << getID() << endl;
  for(Index_t myID=0; myID< size(); myID++) {
    Index_t iL = 0;
    Index_t nnloc = 0;
    const PtclPairList_t& nnlist = *PairAttrib[myID];
    os << "My node = " << myID << " starting with atom " << PtclDist[myID] << endl;
    for(Index_t iG=PtclDist[myID]; iG<PtclDist[myID+1]; iL++, iG++) {
      os << iG << " has " << nnlist.nadj(iL) << endl;
      PtclPairList_t::const_iterator nit(nnlist.begin(iL));
      PtclPairList_t::const_iterator nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cpair(*nit);
        os << setw(5) << cpair.Next << setw(3) << cpair.Group
            << setw(5) << cpair.Where << " " << setw(10) << cpair.sep << " " << cpair.displ << endl;
        ++nit;
      }
      //for(int nni=0; nni<nnlist.nadj(iL); nni++, nnloc++) {
      //  const PairData_t& cpair = nnlist[nnloc];
      //  os << setw(5) << cpair.J << setw(5) << cpair.Loc << " " << setw(10) << cpair.sep << " " << cpair.displ << endl;
      //}//nni
    } // iG,iL

  }//myID

  return true;
}

bool NNEngine::put(istream& is) {
  return true;
}

void NNEngine::checkGrid(ParticleLayout_t& lattice) {
  for(Index_t glevel=0; glevel< GridLevel; ++glevel){
    if(ModifyGrid[glevel]) lattice.Grid[glevel] = Grid[glevel];
  }
}

bool NNEngine::put(xmlNodePtr cur){
#if defined(HAVE_LIBXML2)
  //overwrite Rscale
  const xmlChar* aptr=xmlGetProp(cur,(const xmlChar*)"scale");
  if(aptr) {
    Rscale=atof((const char*)aptr);
  }
  cur = cur->xmlChildrenNode;
  while (cur != NULL) {
    string cname((const char*)cur->name);
    if(cname == "parameter") {
      string aname((const char*)(xmlGetProp(cur, (const xmlChar *) "name")));
      if(aname == "grid") {
	putContent(Grid[SPATIAL_GRID],cur);
	ModifyGrid[SPATIAL_GRID] = true;
      } else if(aname == "ompgrid") {
	putContent(Grid[OMP_GRID],cur);
	ModifyGrid[OMP_GRID] = true;
      } else if(aname == "mppgrid" || aname == "mpigrid") {
	putContent(Grid[MPI_GRID],cur);
	ModifyGrid[MPI_GRID] = true;
      } else if(aname == "rcut") {
	putContent(Rmax,cur);
      } else if(aname == "maxnn") {
	putContent(PresetMaxNN,cur);
      }
    }
    cur = cur->next;
  }
  putExtra(cur);
  LOGMSG("\tNNEngine::Stride is set to " << Stride)
  if(Stride<0) NeedPartition=false;
  //pass it to the derived class to check its input parameters
  return true;
#else
  return false;
#endif
}

#if defined(HAVE_MPI)
/** 
 *\param fv ParticlePos_t to be collected.
 *\brief Initiate collecting paritial values obtained on each processor.
 * For instance, used to collect visiting forces from the neighboring processors.
 * fv is not a distributed object and each processor contains the complete data.
 * The matching action endCollection is to perform reduction operation
 * to put all the pieces together.
 */
void NNEngine::startCollect(ParticlePos_t& fv) { }

/** Matching endCollect for startCollect
 *\param gv The array whose values to be collected.
 *\brief When the particle is not distributed over the processor, do nothing.
 * When the t particle indices are distributed although
 * the data are not distributed, each processor evaluates the partial
 * value of fv and this function add the partial values.
 */
void NNEngine::endCollect(ParticlePos_t& p) {
  if(Distributed[enable_mpi]) {
    Index_t nat =p.size();
    ParticlePos_t pg(nat);
    MPI_Allreduce(&(p[0][0]), &(pg[0][0]), nat*OHMMS_DIM,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    p = pg;
  }
}

void NNEngine::collect(ParticlePos_t& p) {
  if(Distributed[enable_mpi]) {
    Index_t nat =p.size();
    ParticlePos_t pg(nat);
    MPI_Allreduce(&(p[0][0]), &(pg[0][0]), nat*OHMMS_DIM, 
		  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    p = pg;
  } 
}

void NNEngine::collect(ParticleScalar_t& q) {
  if(Distributed[enable_mpi]) {
    Index_t nat =q.size();
    ParticleScalar_t qg(nat);
    MPI_Allreduce(&(q[0]), &(qg[0]), nat, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    q = qg;
  }
}

NNEngine::Scalar_t NNEngine::collect(Scalar_t x) {
  if(Distributed[enable_mpi]) {
    Scalar_t xL(x);
    MPI_Allreduce(&(xL), &(x), 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  }
  return x;
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: NNEngine.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
