//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_NEARESTNEIGHBOR_ENGINE_H
#define OHMMS_NEARESTNEIGHBOR_ENGINE_H

/**\defgroup parallel The parallelization layer.
 */
#include "Particle/ParticleSet.h"
#include "Particle/MetisConnection.h"
#include "Message/OpenMP.h"

namespace OHMMS {

  /** Engine to evaluate the nearest-neighbo lists.
   */
  class NNEngine: 
    public ParticleTags,
    public OhmmsElementBase
  {

    friend class NNEngineBuilder;

  public:

    typedef ParticleSet                       Particle_t;

    typedef Particle_t::size_type             size_type;
    typedef Particle_t::Index_t               Index_t;
    typedef Particle_t::Scalar_t              Scalar_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;
    typedef Particle_t::Tensor_t              Tensor_t;
    typedef Particle_t::ParticlePos_t         ParticlePos_t;
    typedef Particle_t::ParticleScalar_t      ParticleScalar_t;
    typedef Particle_t::ParticleIndex_t       ParticleIndex_t;
    typedef Particle_t::ParticleLayout_t      ParticleLayout_t;
    typedef ParticleLayout_t::Grid_t          Grid_t;

    typedef Particle_t::PtclPairList_t        PtclPairList_t;
    typedef PtclPairList_t::PairData_t        PairData_t;

    enum {enable_omp= 0, enable_mpi};

    enum {SPATIAL_GRID = ParticleLayout_t::SPATIAL_GRID,
          OMP_GRID, 
          MPI_GRID,
          GridLevel
    };

    ///The engine uses layout
    bool UseLayout;

    NNEngine();
    virtual ~NNEngine();

    //@{virtual member functions of OhmmsElementBase
    bool get(ostream& os) const;
    bool put(istream& is);
    bool put(xmlNodePtr cur);
    void reset() { Counter = -1; }
    //@}

    //@ set/get functions
    inline Index_t getID() const { return MethodID;}
    inline void setID(Index_t id) { MethodID = id;}
    inline Index_t current() const { return Counter;}

    inline bool distributed(Index_t i = enable_mpi) const { 
      return Distributed[i];
    }

    inline void distribute(Index_t i, bool doit) {
      Distributed[i] = doit;
    }

    inline PtclPairList_t& getPairAttrib(size_type i) {
      return *PairAttrib[i];
    }

    inline const PtclPairList_t& getPairAttrib(size_type i) const {
      return *PairAttrib[i];
    }

    inline PtclPairList_t& operator()(size_type i) {
      return *PairAttrib[i];
    }

    inline const PtclPairList_t& operator()(size_type i) const {
      return *PairAttrib[i];
    }
    //@}

    inline Index_t getTotNum() const { return TotalNum;}
    inline Index_t getTotNadj() const { return TotalNumPairs;}
    inline Index_t getMaxNadjPerPtcl() const { return MaxPairsPerPtcl;}

    inline Index_t first_id(size_type id) const { return PtclDist[id];}
    inline Index_t last_id(size_type id) const { return PtclDist[id+1];}
    inline Index_t getLocalNum(size_type id) const { return PtclDist[id+1]-PtclDist[id];}
    inline omp_int_t size() const { return NumThreads;}
    
    inline bool needCleanUp(Index_t iter) { return iter%Stride == 0; }

    bool addConnection(const string& cname, vector<MetisConnection*>* deps);

#ifdef HAVE_MPI
    virtual void startCollect(ParticlePos_t& fv);
    virtual void endCollect(ParticlePos_t& fv);
    virtual void collect(ParticlePos_t& p);
    virtual void collect(ParticleScalar_t& p);
    virtual Scalar_t collect(Scalar_t x);
#else
    inline void startCollect(ParticlePos_t& p) {}
    inline void endCollect(ParticlePos_t& p){}
    inline void collect(ParticlePos_t& p){}
    inline void collect(ParticleScalar_t& p){}
    inline Scalar_t collect(Scalar_t x){ return x;}
#endif

    /** check any overwriting grid information by NNEngine::put 
     */
    void checkGrid(ParticleLayout_t& lattice); 

    /** initialize the engine
     * @param p ParticleSet whose NN list will be constructed.
     * @param needvisitingforce if true, prepare for the ghost cells.
     */
    virtual bool initialize(Particle_t& p, bool needvisitingforce) = 0;

    /** update the data associated with the spatial grid
     * @param pos current position used to build a list
     * @param pu position in LatticeUnit
     * @param lattice supercel
     */
    virtual Index_t updateGridData(const ParticlePos_t& pos, 
        ParticlePos_t& pu, ParticleLayout_t& lattice) =0;

    /** collect visiting forces
     * @param f forces
     * @param ip thread id
     */
    virtual void collectForce(ParticlePos_t& f, omp_int_t ip) ;

    /** Evaluation the pair lists 
     * @param lattice the Layout of the particleset
     * @param pos the position used to evaluate the pair relations
     *
     * The information of the previous step is cleared and new pair list 
     * is evaluated. Loops over the threads and calls evalPairsPerThread.
     */
    virtual void makeNewPairs(Particle_t& ptcl, const ParticlePos_t& pos);

    /** Update the pair lists
     * @param del displacement in Cartesian Coordiates
     *
     * Add the contributions from displacement to the previous position. The list,
     * e.g., the number of pairs, the id's of the connected particles, is
     * unchanged.
     */
    void updatePairs(Particle_t& ptcl, const ParticlePos_t& del);

    virtual void sync();
    void rescale(Scalar_t ascale);
    void update(const Particle_t& p, Index_t iat);

    //@{functions to perform particle-by-particle moves, e.g. MC
    void addPairList(Index_t nat, Index_t maxnn);
    void makeMove(Index_t i, const SingleParticlePos_t& d);
    void restoreMove(Index_t i);
    //@}
 
    void copy(const NNEngine& nn);
    inline void setStride(Index_t astride) { Stride = astride;}

    ////////////////////////////////////////////////////
    //Data members
    ////////////////////////////////////////////////////
    
    ///flag to tell repartition
    bool NeedPartition;

    ///flag to tell if aux storage is required to keep the visiting forces
    bool UseVisitingForce;

    ///flag for what????
    bool NewConfiguration;

    bool Reordered;
    bool IndexOnly;
    bool Symmetric;
    bool Unique; 
    bool Distributed[2]; 
    bool ModifyGrid[GridLevel];

    omp_int_t NumThreads; 
    Index_t MethodID;
    Index_t Counter;
    Index_t Stride;
    Index_t TotalNum;
    Index_t NumGhosts;
    Index_t PresetMaxNN;
    Index_t MaxPairsPerPtcl;
    Index_t TotalNumPairs;
    Index_t MaxPtclsPerCell;
    Index_t MaxConnectedCells;
    Index_t ActivePtcl;
 
    Scalar_t Rmax, Rscale, RmaxSq, RminSq;

    //temporary data for grid paritition: to overwrite ParticleLayout::Grid
    vector<SingleParticleIndex_t> Grid;

    /** For multi-threaded applications using OpenMP, each process operates on its PtclPairList_t
     */
    vector<PtclPairList_t*> PairAttrib;

    /** OmpAttrib->operator[](iat) is the thread number for the iat-th particle
     */
    ParticleIndex_t* OmpAttrib;

    /** Cell->operator[](iat) is the spatial-grid number for the iat-th particle
     */
    ParticleIndex_t* Cell;

    /** Auxiliary data used for striding. The size of PtclDist is PairAttrib.size()+1.
     *
     * PtclDist[i] is the starting index of the particles of a thread on a processor.  
     * For the i-th thread, the OpenMP parallel regions handle 
     * [PtclDist[i+1],PtclDist[i]) particles.  
     */
    vector<Index_t> PtclDist;

    /** List of connection data that depend on the pairlist */
    map<string,vector<MetisConnection*>* > Connections;

    /** Divide a particleset without considering the locality 
     *
     * The number of particles per thread differs by 1 (maximum).
     */
    void makeFairPartition(Particle_t& p);

    /** Allocate PairAttrib
     *
     *The PairAttrib is the most expensive object. Each thread will
     *allocate the data within openmp parallel for.
     */
    void allocateAttributes(Particle_t& p, bool needvisitingforce);

    virtual bool putExtra(xmlNodePtr cur) = 0;

    /** thread-safe function to evalaute a local PtclPairList_t
     */
    virtual void evalPairsPerThread(const ParticleLayout_t& lattice, 
                                    const ParticlePos_t& pos, 
                                    const ParticleIndex_t& grID,
                                    PtclPairList_t& myppl) =0;

    virtual void updatePairsPerThread(const ParticlePos_t& del, PtclPairList_t& myppl);

  };

  struct NNEngineBuilder {
    static int Counter;
    static NNEngine* create(xmlNodePtr cur=NULL);
    static NNEngine* create(int& argc, char**& argv);
  };

  //data introduced to handled particle-by-particle update for MC
  //  struct PairBandElement {
  //    Index_t loc;
  //    Scalar_t sep;
  //    SingleParticlePos_t dcos;
  //    SingleParticlePos_t dr;
  //    inline PairBandElement(): loc(0),sep(0.0),dcos(0.0) { }
  //    inline PairBandElement(Index_t l, Scalar_t r_in, const SingleParticlePos_t& dcos_in):
  //      loc(l),sep(r_in),dcos(dcos_in),dr(r_in*dcos_in) {
  //    }
  //  };

}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 361 $   $Date: 2005-07-13 13:31:31 -0500 (Wed, 13 Jul 2005) $
 * $Id: NNEngine.h 361 2005-07-13 18:31:31Z jnkim $ 
 ***************************************************************************/
