//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <set>
#include <list>
#include <numeric>
#include <algorithm>
#include <iterator>
#include <fstream>
#include "Message/OpenMP.h"
#include "Message/Communicate.h"
#include "Utilities/UtilityFunctions.h"
#include "OhmmsPETE/OhmmsMatrix.h"
#include "Particle/NNSpatialEngine.h"
#include "ParticleBase/ParticleFunctions.h"
#ifdef HAVE_OOMPI
#include "oompi.h"
#endif

using namespace OHMMS;

/** collect visiting forces
 *
 * @param f force container
 * @param ip thread number, ignored
 */
void NNSpatialEngine::collectForce(ParticlePos_t& f, int ip) {


  //alias to Fv
  ParticlePos_t& fv=PairAttrib[ip]->Fv;

  vector<OOMPI_Packed*> recvBuff;
  OOMPI_Request_array recvRequests(NumSendMsg);
  OOMPI_Status_array recvStatus(NumSendMsg);

  //start non-blocking recv using SendList
  //Returing data: (id,pos)* 
  int node=0;
  for(int from=0; from<SendList.size(); from++){
    int nitem = SendList[from]->size();
    if(nitem) {
      int nchar = nitem*(sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
      OOMPI_Packed *apacket = new OOMPI_Packed(nchar,OOMPI_COMM_WORLD);
      recvBuff.push_back(apacket);
      recvRequests[node++] = OOMPI_COMM_WORLD[from].Irecv(*apacket,MPI_ANY_TAG);
    }
  }

  //send ghost forces non-blocking send using RecvList
  for(int to=0; to<RecvCounter.size(); to++){
    int nitem=RecvCounter[to];
    if(nitem) {
      int nchar = nitem*(sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
      for(int ig=GhostOffset[to]; ig<GhostOffset[to+1]; ig++) {    
	apacket << GhostList[ig].ID;
	fv[ig+NumLocalPtcls].putMessage(apacket);
      }
      OOMPI_COMM_WORLD[to].Send(apacket);
    }
  }

  //"assign" the local forces from the visting forces
  for(int iat=0; iat<NumLocalPtcls; iat++) f[iat]=fv[iat];

  //check if all the irecvs are completed
  recvRequests.Waitall(recvStatus);


  //unpack del from ghost cells to GhostList::R (reuse)
  node=0;
  int iat,nmsg; 
  SingleParticlePos_t ftmp;
  while(node<NumSendMsg) {
    OOMPI_Packed &apacket = *recvBuff[node];
    apacket>>nmsg;
    while(nmsg) {
      apacket>>iat;
      ftmp.getMessage(apacket);
      f[iat] += ftmp;
      --nmsg;
    }
    node++;
  }

}


void 
NNSpatialEngine::updatePairsPerThread(const ParticlePos_t& del, PtclPairList_t& myppl){
  
  vector<OOMPI_Packed*> recvBuff;
  OOMPI_Request_array recvRequests(NumRecvMsg);
  OOMPI_Status_array recvStatus(NumRecvMsg);

  ///start non-blocking recv of displacement vectors
  int node=0;
  for(int from=0; from<RecvCounter.size(); from++){
    int nitem(RecvCounter[from]);
    if(nitem) {
      int nchar(nitem*OHMMS_DIM*sizeof(double)/sizeof(char));
      OOMPI_Packed *apacket = new OOMPI_Packed(nchar,OOMPI_COMM_WORLD);
      recvBuff.push_back(apacket);
      recvRequests[node++] 
        = OOMPI_COMM_WORLD[from].Irecv(*apacket,MPI_ANY_TAG);
    }
  }

  //start blocking send
  //message=nitem + nitem*(del)
  for(int to=0; to<SendList.size(); to++) {
    int nitem(SendList[to]->size());
    if(nitem) {
      int nchar(nitem*OHMMS_DIM*sizeof(double)/sizeof(char));
      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
      apacket.Start();
      IndexList_t::const_iterator sit(SendList[to]->begin());
      IndexList_t::const_iterator sit_end(SendList[to]->end());
      while(sit!=sit_end) {
	del[*sit++].putMessage(apacket);
      }
      apacket.End();
      OOMPI_COMM_WORLD[to].Send(apacket);
    }
  }

  //check if all the irecvs are completed
  recvRequests.Waitall(recvStatus);

  //unpack del from ghost cells to GhostList::R (reuse)
  node=0;
  for(int from=0; from<RecvCounter.size(); from++){
    int nitem = RecvCounter[from];
    if(nitem){
      OOMPI_Packed &apacket = *recvBuff[node];
      int ng(GhostOffset[from]);
      for(int item=0; item<nitem; item++, ng++) {
	GhostList[ng].R.getMessage(apacket);
      }
      ///remove the buffer
      delete recvBuff[node]; 
      ++node;
    }
  }

  for(int iL=0; iL<myppl.getLocalNum(); iL++) {
    PtclPairList_t::iterator nit(myppl.begin(iL)),nit_end(myppl.end(iL));
    const SingleParticlePos_t& d(del[myppl.I[iL]]);
    while(nit != nit_end) {
      int jat((*nit).Where);
      if(jat<NumLocalPtcls) {
        (*nit).addDelta(del[jat]-d);
      } else {
        (*nit).addDelta(GhostList[jat-NumLocalPtcls].R-d);
      }
      ++nit;
    }
  }
}

/** exchange ghost poarticles
 * 
 * Using Irecv (non-blocking recv) and Send (blocking send)
 * Each node posts Irecv's from the sources and posts sends to the destinations.
 * Packed messages:
 * For the iat-th particle to be sent to other processors, 
 * - the local index (for reverse communication by endCollection)
 * - the species index
 * - the cell index of the current position
 * - the current position
 * Note that the index iat is [0,p.getLocalNum())
 */
void
NNSpatialEngine::sendrecvGhostParticles(const ParticlePos_t& pos) {  

  ParticleIndex_t& grID(myPtcl->GroupID);

  vector<OOMPI_Packed*> recvBuff;
  OOMPI_Request_array recvRequests(NumRecvMsg);
  OOMPI_Status_array recvStatus(NumRecvMsg);

  ///start non-blocking recv
  int node=0;
  for(int from=0; from<RecvCounter.size(); from++){
    int nitem = RecvCounter[from];
    if(nitem) {
      int nchar=sizeof(int) 
	+ nitem*(3*sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
      OOMPI_Packed *apacket = new OOMPI_Packed(nchar,OOMPI_COMM_WORLD);
      recvBuff.push_back(apacket);
      recvRequests[node++] = OOMPI_COMM_WORLD[from].Irecv(*apacket,MPI_ANY_TAG);
    }
  }

  //start blocking send
  //message=nitem + nitem*(LocalID+GroupID+CellID+pos)
  for(int to=0; to<SendList.size(); to++) {
    int nitem(SendList[to]->size());
    if(nitem) {
      int nchar(sizeof(int)
	+nitem*(3*sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char));

      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
      apacket.Start();
      apacket.Pack(nitem);
      IndexList_t::const_iterator sit(SendList[to]->begin());
      IndexList_t::const_iterator sit_end(SendList[to]->end());
      while(sit !=sit_end) {
	int iat(*sit++);
	apacket.Pack(iat);
	apacket.Pack(grID[iat]);        
	apacket.Pack((*Cell)[iat]);
	pos[iat].putMessage(apacket);
      }
      apacket.End();
      OOMPI_COMM_WORLD[to].Send(apacket);
    }
  }

  //check if all the irecvs are completed
  recvRequests.Waitall(recvStatus);

  ParticleLayout_t::Grid_t* restrict sg(myPtcl->Lattice.getGrid(SPATIAL_GRID));
  node=0;
  int thiscell;
  for(int from=0; from<RecvCounter.size(); from++){
    int nitem = RecvCounter[from];
    if(nitem){
      OOMPI_Packed &apacket = *recvBuff[node];
      apacket >>  nitem;
      int ng(GhostOffset[from]);
      int ntot(GhostOffset[from]+NumLocalPtcls);
      for(int item=0; item<nitem; item++, ntot++, ng++) {
        GhostDataType& g(GhostList[ng]);
        apacket >> g.ID >> g.GroupID >> thiscell;
	g.R.getMessage(apacket);
	sg->addData(thiscell,ntot);
      }
      ///remove the buffer
      delete recvBuff[node]; 
      ++node;
    }
  }

}


void 
NNSpatialEngine::setMessagePatterns(const Grid_t* restrict sg) {

  //char fname[128];
  //sprintf(fname, "pattern.p%d",mynode);
  //ofstream fout(fname);
  //Lattice.printData(fout);
  //int ncontexts = mppGrid->getTotalNum();
  //const DistributedIndex& dist = Lattice.getDataSets();
  //std::copy(dist.M.begin(), dist.M.end(),ostream_iterator<int>(fout, " "));
  //fout << endl;
  
  int ncontexts(mppGrid->getTotalNum());
  int recvMessage = ncontexts-1;

  //First, create SendList
  if(SendList.empty()) {
    for(int to=0; to<ncontexts;  to++) 
      SendList.push_back(new IndexList_t);
  }

  vector<OOMPI_Packed*> recvBuff;
  OOMPI_Request_array recvRequests(recvMessage);
  OOMPI_Status_array recvStatus(recvMessage);

  int node=0;
  ///start non-blocking recv
  for(int from=0; from<ncontexts; from++) {
    if(from != myNode) {
      int nchar=sizeof(int)/sizeof(char); 
      OOMPI_Packed *apacket = new OOMPI_Packed(nchar,OOMPI_COMM_WORLD);
      recvBuff.push_back(apacket);
      recvRequests[node] = OOMPI_COMM_WORLD[from].Irecv(*apacket,MPI_ANY_TAG);
      ++node;
    }
  }


  typedef ParticleLayout_t::DistributedIndex_t IndexSet_t;
  const IndexSet_t& dist(sg->getDataSets());
  for(int to=0; to<ncontexts; to++) {
    int ij = myNode*ncontexts+to;
    if(to != myNode) {
      int nitem(0); //number of items packed to the remote node
      IndexList_t::const_iterator git(ghostcell[ij]->begin());
      IndexList_t::const_iterator git_end(ghostcell[ij]->end());
      while(git != git_end) {
        nitem += dist.size(*git); ++git;
      }
      SendList[to]->resize(nitem);
      IndexList_t::iterator sit(SendList[to]->begin());
      int item(0);
      git=ghostcell[ij]->begin();
      while(git != git_end) {
	int ic(*git);
	IndexSet_t::const_iterator dit(dist.begin(ic));
	IndexSet_t::const_iterator dit_end(dist.end(ic));
        while(dit != dit_end) {
          *sit++=*dit++; //(*SendList[to])[item++] = *dit++;
        }
        ++git;
      }
      OOMPI_COMM_WORLD[to].Send(nitem);
    }
  }
 
  RecvCounter.resize(ncontexts,0);

  recvRequests.Waitall(recvStatus);

  //collect where and how many items to recv
  int expected = 0;
  node=0;
  while(node<recvMessage) {
    OOMPI_Packed &apacket = *recvBuff[node];
    int from = recvStatus[node].Get_source();
    apacket >>  RecvCounter[from];
    if(RecvCounter[from]) expected++;
    ++node;
  }


  GhostOffset.resize(ncontexts+1,0);
  GhostOffset[0] = 0;
  for(int i=0; i<ncontexts; i++) {
    GhostOffset[i+1] = GhostOffset[i]+RecvCounter[i];
  }

  NumGhosts = GhostOffset[ncontexts];

  if(expected != NumRecvMsg) {///probably need to warn
    NumRecvMsg = expected;
  }
}

int
NNSpatialEngine::swap(Particle_t& p, int mynode) {
  return 1;
}

///*!\fn void NNSpatialEngine::sendGhostParticles(Particle_t& p)
// *\brief pack the ghost particles and send them to remote processors.
// * 
// * Either NNSpatialEngine::initialize or NNSpatialEngine::partition
// * pre-calculates the communication patterns of ghost elements
// * between the processors. sendGhostParticles pack the information 
// * of the ghost particles of the neighboring processors and
// * send them to the corresponding processor.
// * For the iat-th particle to be sent to other processors,
// * following data are packed
// * - the local index (for reverse communication by endCollection)
// * - the species index
// * - the cell index of the current position
// * - the current position
// * Note that the index iat is [0,p.getLocalNum())
// */
//void
//NNSpatialEngine::sendGhostParticles(Particle_t& p) {  
//
//  //ParticleIndex_t& cell = *(p.getIndexAttrib("cell"));
//  for(int to=0; to<SendList.size(); to++) {
//    if(SendList[to]->size()) {
//      int nitem = SendList[to]->size();
//      int nchar=sizeof(int)
//	+nitem*(3*sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
//      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
//      apacket.Start();
//      apacket.Pack(nitem);
//      for(int item=0; item<nitem; item++) {
//	int iat = SendList[to]->operator[](item);
//	apacket.Pack(iat);
//	apacket.Pack(p.GroupID[iat]);        
//	//apacket.Pack(p.Cell[iat]);
//	apacket.Pack(Cell[iat]);
//	p.R[iat].putMessage(apacket);
//      }
//      apacket.End();
//      OOMPI_COMM_WORLD[to].Send(apacket);
//
//    }
//  }
//}
//
///*!\fn void NNSpatialEngine::recvGhostParticles(Particle_t& p)
// *\brief  unpack the data sent in sendGhostParticles.
// *
// * Either NNSpatialEngine::initialize or NNSpatialEngine::partition
// * pre-calculates the communication patterns of ghost elements between
// * the processors.  GhostID is an auxiliary array to order the ghost
// * elements according to the remote processor IDs. Ordering the ghost
// * elements based on the spatial locality may improve the performance,
// * although it is not obvious. GhostID[remote+1]-GhostID[remote] is
// * the number of ghost elements from a remote
// * node. GhostID[mynode+1]-GhostID[mynode] is always zero.  
// */
//int
//NNSpatialEngine::recvGhostParticles(Particle_t& p) { 
//
//  int nghost = GhostOffset.back();
//
//  int dn = nghost - p.R.getNumGhosts();
//  if(dn) { ///not enough ghost elements, add more
//    p.R.addGhosts(dn);
//    p.GroupID.addGhosts(dn);
//  }
//
//  int nloc = p.getLocalNum();
//  int recvMessage = NumRecvMsg;
//  while(recvMessage>0) {
//    bool flag=false;
//    OOMPI_Status astatus = OOMPI_COMM_WORLD.Iprobe(MPI_ANY_TAG,flag);
//    if(flag) {
//      int from = astatus.Get_source();
//      int tag = astatus.Get_tag(); 
//      int nitem = RecvCounter[from];
//      int nchar=sizeof(int) 
//	+ nitem*(3*sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
//      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
//      OOMPI_COMM_WORLD[from].Recv(apacket,tag);
//      recvMessage--;
//      apacket >>  nitem;
//      int ng= GhostOffset[from];
//      int ntot = GhostOffset[from]+nloc;
//      for(int item=0; item<nitem; item++, ntot++, ng++) {
//	int thiscell;
//	apacket >> GhostID[ng] >> p.GroupID[ntot] >> thiscell;
//	p.R[ntot].getMessage(apacket);
//	p.Lattice.addData(thiscell,ntot);
//      }
//    }
//  }
//
////    int nloc = p.getLocalNum();
////    int nghost=0;
////    for(int inode=0; inode<remotePtcl.size(); inode++) {
////      if(inode != mynode) { 
////        nghost += remotePtcl[inode]->size();
////      }
////    }
////    p.R.addGhosts(nghost);
////    p.GroupID.addGhosts(nghost);
////    int ntot = nloc;
////    for(int inode=0; inode<remotePtcl.size(); inode++) {
////      if(inode != mynode) { 
////        list<nnMessageItem>::iterator it = remotePtcl[inode]->begin();
////        while(it != remotePtcl[inode]->end()) {
////  	p.R[ntot] = (*it).pos;
////  	p.GroupID[ntot] = (*it).species;
////  	PList[(*it).cell]->push_back(ntot);
////  	it++;ntot++;
////        }
////      }
////    }
//
//  return nghost;
//}

/*!\fn void NNSpatialEngine::startCollect(ParticlePos_t& fv)
 *\brief Overwriting the virtual function of NNEngine::startCollect.
 *
 * NNSpatialEngine explicitly distributs the data associated with the 
 * particles, thus fv is also distributed among the processors.
 * This function ensures that the memory for the ghost elements
 * are allocated and set to the default values, zero vectors,
 * before assignment.
 */
void NNSpatialEngine::startCollect(ParticlePos_t& fv) {
  ERRORMSG("NNSpatialEngine::startCollect should not be used") 
  //if(fv.getNumGhosts() < NumGhosts) {
  //  fv.clearGhosts();
  //  fv.addGhosts(NumGhosts);
  //}
  //for(int i=fv.size();i<fv.size()+NumGhosts; i++) fv[i] = 0.0;
}

/*!\fn void NNSpatialEngine::endCollect(ParticlePos_t& fv)
 *\brief Overwriting the virtual function of NNEngine::endCollect.
 *
 * The contributions to fv by remote processors are collected.
 * Since each processor knows where the ghost elements come from,
 * (see sendGhostParticles/recvGhostParticles), it can send back
 * the partial values evaluated on this procesor.
 * Once the sends are completed, each processor collect the values
 * sent from the neighboring processors.
 */
void NNSpatialEngine::endCollect(ParticlePos_t& fv) {

  ERRORMSG("NNSpatialEngine::endCollect should not be used") 
//  int nloc(fv.size());
//  for(int to=0; to<RecvCounter.size(); to++){
//    if(RecvCounter[to]) {
//      int nchar
//	= RecvCounter[to]*(sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
//      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
//      for(int ig=GhostOffset[to]; ig<GhostOffset[to+1]; ig++) {    
//	apacket << GhostList[ig].ID;
//	fv[ig+nloc].putMessage(apacket);
//      }
//      OOMPI_COMM_WORLD[to].Send(apacket);
//    }
//  }
//
//  int recvMessage = NumSendMsg;
//  int iat; 
//  SingleParticlePos_t ftmp;
//  while(recvMessage>0) {
//    bool flag=false;
//    OOMPI_Status astatus = OOMPI_COMM_WORLD.Iprobe(MPI_ANY_TAG,flag);
//    if(flag) {
//      int from = astatus.Get_source();
//      int tag = astatus.Get_tag(); 
//      int nitem = SendList[from]->size();
//      int nchar= nitem*(sizeof(int)+OHMMS_DIM*sizeof(double))/sizeof(char);
//      OOMPI_Packed apacket(nchar,OOMPI_COMM_WORLD);
//      OOMPI_COMM_WORLD[from].Recv(apacket,tag);
//      recvMessage--;
//      while(nitem) {
//	apacket >> iat;
//	ftmp.getMessage(apacket);
//        fv[iat] += ftmp;
//	--nitem;
//      }
//    }
//  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: NNSpatialEngine.OOMPI.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/

