//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <set>
#include <list>
#include <numeric>
#include <algorithm>
#include <iterator>
#include <fstream>
#include "Message/OpenMP.h"
#include "Message/Communicate.h"
#include "OhmmsPETE/OhmmsMatrix.h"
#include "Particle/NNSpatialEngine.h"
#include "Particle/PLayoutFunctions.h"
#include "Utilities/OhmmsInfo.h"
#ifdef HAVE_OOMPI
#include "oompi.h"
#endif
//#define PRINT_DEBUG
using namespace OHMMS;

NNSpatialEngine::NNSpatialEngine():
  NumSendMsg(0), CurSendMsg(0), NumRecvMsg (0), 
  CurRecvMsg(0), NumLocalPtcls(0),NumGhosts(0),
  myPtcl(0),mppGrid(0), mppAttrib(0)
{ 
  myNode = Controller->mycontext();
  LOGMSG("Creating NNSpatialEngine, distributed over mpi nodes")
}

NNSpatialEngine::~NNSpatialEngine() { }

bool
NNSpatialEngine::initialize(Particle_t& p, bool needvisitingforce) {

  Distributed[enable_mpi]=true;

  DEBUGMSG("NNSpatialEngine::initialize")
  Rmax = p.Lattice.InteractionRadius;
  RmaxSq = Rmax*Rmax;


  //save the ParticleSet
  myPtcl = &p;

  p.setMode(ParticleSet::Distributed);
  UseVisitingForce=needvisitingforce;
  if(mppGrid == 0) { //First time

    //establish the connections between connected nodes based on MPI_GRID and SPATIAL_GRID
    setRemoteConnections(p.Lattice,Rmax);

    //partition the system first
    makePartitions(p);

    MaxPairsPerPtcl= (MaxConnectedCells+1)*(MaxPtclsPerCell+1);
    if(PresetMaxNN>0) {
      MaxPairsPerPtcl = std::min(PresetMaxNN,MaxPairsPerPtcl);
    }

    LOGMSG("\tLinked cell data ")
    LOGMSG("\t  Number of particles on node " << NumLocalPtcls)
    LOGMSG("\t  Max connected cells " << MaxConnectedCells)
    LOGMSG("\t  Max numer of particles per cell " << MaxPtclsPerCell)
    PtclDist.resize(2);
    PtclDist[0]=0; PtclDist[1]=NumLocalPtcls;
    allocateAttributes(p,needvisitingforce);

    if(UseVisitingForce)
      LOGMSG("\tNNSpatialEngine::UseVisitingForce=true")
    else
      LOGMSG("\tNNSpatialEngine::UseVisitingForce=false")
  }


  //I don't think this is necessary anymore
  //setMessagePatterns(p.Lattice);
  //LOGMSG("\tNumber of maximum linked cells " << MaxConnectedCells)
  //MaxPairsPerPtcl=(MaxConnectedCells+1)*(MaxPtclsPerCell+1);
  //allocateAttributes(p,UseVisitingForce);
  return true;
}

void NNSpatialEngine::makePartitions(Particle_t& p) {

  if(p.getLocalNum() == p.getTotalNum()) {
    mppAttrib=p.getIndexAttrib("mppid");

    int np = Controller->ncontexts();
    PtclPerNode.resize(np,0);
    PtclOffset.resize(np+1,0);

    IndexList_t LocPtcl, LocCell;
    int ntot(p.getLocalNum());
    LocPtcl.reserve(ntot*2/np);
    LocCell.reserve(ntot*2/np);

    p.curR.setUnit(PosUnit::LatticeUnit);
    p.applyBC(p.R,p.curR);

    int nloc(0);
    ParticleLayout_t::Grid_t* restrict sg(p.Lattice.getGrid(SPATIAL_GRID));
    for(int iat=0; iat<ntot; ++iat) {
      int iloc = sg->loc(p.curR[iat]);
      int where = mppGrid->group(iloc);
      if(where == myNode) {
        LocPtcl.push_back(iat);
        LocCell.push_back(iloc);
        nloc++;
      }
    }

    p.select(LocPtcl);

    //copy the cell index
    if(Cell == 0) {
      Cell = new ParticleIndex_t(nloc);
    }
 
    std::copy(LocCell.begin(),LocCell.end(),Cell->begin());

    PtclOffset[myNode] = LocPtcl.size();
    int ierr = MPI_Allgather(&PtclOffset[0],np,MPI_INT,&PtclPerNode[0],np,MPI_INT,MPI_COMM_WORLD);

    PtclOffset[0]=0;
    for(int ip=0; ip<np; ip++) 
      PtclOffset[ip+1]=PtclOffset[ip]+PtclPerNode[ip];

    int iat(0);
    int gid(PtclOffset[myNode]);
    while(iat<nloc) {
      (*mppAttrib)[iat++]=gid++;
    }

    //ParticleSet::Lattice::DistributedIndex is filled
    sg->createData();
    iat=0;
    IndexList_t::const_iterator cit(LocCell.begin());
    while(iat<nloc) {
      sg->addData(*cit,iat); ++cit; ++iat;
    }

    MaxPtclsPerCell= sg->getMaxDataPerGrid();
    NumLocalPtcls=nloc;
    PtclDist.resize(2,0);
    PtclDist[1]=nloc;
  }
}

void 
NNSpatialEngine::setRemoteConnections(ParticleLayout_t& Lattice, 
				      Scalar_t rmax){

  //OmpGrid = p.Lattice.addGrid(omp_grid,p.Lattice.Grid[omp_grid]);
  mppGrid=Lattice.getGrid(MPI_GRID);
  if(mppGrid->empty()) {
    mppGrid->setGrid(Grid[MPI_GRID]);
    Lattice.initGrid(*mppGrid,*(Lattice.getGrid(SPATIAL_GRID)));
  }

  //get the pointer to the spatial grid
  const ParticleLayout_t::Grid_t* restrict sg(Lattice.getGrid(SPATIAL_GRID));

  MaxConnectedCells=Lattice.MaxConnections;

  //UniformCartesianGrid::PID will be update to reflect the mpi grid
  int ngroups(mppGrid->getTotalNum());
  int ngrids(sg->getTotalNum());
  mppGrid->resizeGroup(ngrids);

  int ctot=0;
  vector<list<int>* > groups;
  for(int ig=0; ig<ngroups; ig++) groups.push_back(new list<int>);
  int ng=0;
  for(int ig=0; ig<sg->size(0); ig++) {
    for(int jg=0; jg<sg->size(1); jg++) {
      for(int kg=0; kg<sg->size(2); kg++, ng++) {
	int iloc =  mppGrid->loc(sg->center(ig,jg,kg));
	mppGrid->group(ng) = iloc;
	groups[iloc]->push_back(ng);
      }
    }
  }

  if(ghostcell.empty()) {
    for(int ig=0;ig<ngroups*ngroups; ig++) { 
      ghostcell.push_back(new IndexList_t);
      ghostcell[ig]->reserve(ngrids);
    }
  } else {
    for(int ig=0;ig<ngroups*ngroups; ig++) { 
      ghostcell[ig]->clear();
    }
  }

  //mask array To(destination,spatial grid)
  Matrix<int> To(ngroups,ngrids);

  //iteration over the sources
  for(int ig=0; ig<ngroups; ig++) {
    //set the mask to false
    To = 0;

    //iteration over the grid points that reside on the source node
    list<int>::iterator it(groups[ig]->begin()), it_end (groups[ig]->end());
    while(it != it_end) {
      //the identiy of a spatial grid on the source node
      int ic(*it);
      //iteration over all the connected grids for the ic-th grid
      for(int isub=Lattice.first_connected(ic); 
	  isub<Lattice.last_connected(ic); 
	  isub++) {
	int thisnode = Lattice.id_connected(isub);
	To(mppGrid->group(thisnode),thisnode) = 1;
      }
      ++it;
    }

    for(int jg=0; jg<ngroups; jg++) {
      if(jg == ig) continue;
      int ij = ig+jg*ngroups;
      for(int ic=0;ic<ngrids; ic++) {
	if(To(jg,ic)) ghostcell[ij]->push_back(ic);
      }
    }
  }

  NumSendMsg = 0;    
  NumRecvMsg = 0;
  for(int jg=0; jg<ngroups; jg++) {
    if(ghostcell[myNode*ngroups+jg]->size()) NumSendMsg++;
    if(ghostcell[jg*ngroups+myNode]->size()) NumRecvMsg++;
  }

#ifdef PRINT_DEBUG
  char fname[128];
  sprintf(fname, "message.p%d",myNode);
  ofstream fout(fname);

  //Lattice.print(fout);  
  //mppGrid->printGrid(fout);

  fout << "Reverse map : spatial grid -> mpi grid " << endl;
  for(int ig=0; ig<ngrids; ig++) {
    fout << mppGrid->group(ig) << " ";
  }
  //std::copy(PID.begin(), PID.end(), ostream_iterator<int>(fout, " "));
  fout << endl;
  for(int jg=0; jg<ngroups; jg++) {
      fout << jg << " node owns ";
      std::copy(groups[jg]->begin(), groups[jg]->end(), 
		ostream_iterator<int>(fout, " "));
      fout << endl;
    }

  for(int ig=0; ig<ngroups; ig++) {
    fout << ig << " node will send data to these nodes " << endl;
    int nmessage=0;
    for(int jg=0; jg<ngroups; jg++) {
      fout << "\tRemote node = " << jg << " gets " << ghostcell[ig*ngroups+jg]->size() << " : ";
      std::copy(ghostcell[ig*ngroups+jg]->begin(),
		ghostcell[ig*ngroups+jg]->end(),
		ostream_iterator<int>(fout, " "));
      fout << endl;
      if(ig != jg) nmessage += ghostcell[ig*ngroups+jg]->size();
    }
    fout << "total message " << nmessage << endl;
  }

  for(int ig=0; ig<ngroups; ig++) {
    fout << ig << " node will get data from these nodes " << endl;
    int nmessage=0;
    for(int jg=0; jg<ngroups; jg++) {
      std::copy(ghostcell[jg*ngroups+ig]->begin(),
                ghostcell[jg*ngroups+ig]->end(), 
                ostream_iterator<int>(fout, " "));
      fout << endl;
      if(ig != jg) nmessage += ghostcell[jg*ngroups+ig]->size();
    }
    fout << "total message " << nmessage << endl;
  }
#endif /*PRINT_DEBUG*/

  for(int jg=0; jg<ngroups; jg++) delete groups[jg];
}


bool 
NNSpatialEngine::putExtra(xmlNodePtr cur){
  return true;
}

void 
NNSpatialEngine::getOptions(int argc, char** argv) {
  
  int retargc = 1;
  char **retargv = new char *[argc];
  retargv[0] = argv[0];
  int i=1;
  while(i<argc) {
    string aword(argv[i]);
    if(aword == "--mpi") {
      for(int idim=0; idim<OHMMS_DIM; idim++) 
	Grid[0][idim] = atoi(argv[++i]);
    } else if(aword == "--omp") {
      for(int idim=0; idim<OHMMS_DIM; idim++) 
	Grid[1][idim] = atoi(argv[++i]);
    } else if(aword == "--grid") {
      for(int idim=0; idim<OHMMS_DIM; idim++) 
	Grid[2][idim] = atoi(argv[++i]);
    } else if(aword == "--node") {
      myNode = atoi(argv[++i]);
    } else {
      retargv[retargc++]  = argv[i];
    }
    ++i;
  }
  
  argc = retargc;
  for (i = 0; i < retargc; ++i) {argv[i] = retargv[i];}
  delete [] retargv;
}


//int
//NNSpatialEngine::scatter(Particle_t& p, int mynode) {
//
//  int maxat = 0;
//  //if(!p.Distributed) {
//  maxat = p.partition(mppGrid,mynode);
//  FairDivide(p.getLocalNum(),omp_get_max_threads(),PtclDist);
//  //}
//
//  setMessagePatterns(p.Lattice);
//
//#ifdef PRINT_DEBUG
//  char fname[128];
//  sprintf(fname, "scatter.p%d",mynode);
//  ofstream fout(fname);
//  int ncontexts = Controller->ncontexts();
//  for(int jg=0; jg<ncontexts; jg++) {
//    if(SendList[jg]->size()) {
//      fout << "To " << jg << endl;
//      std::copy(SendList[jg]->begin(), SendList[jg]->end(),
//		ostream_iterator<int>(fout, " "));
//      fout << endl;
//    }
//  }
//  for(int jg=0; jg<ncontexts; jg++) {
//    fout << "From " << jg << " node " << RecvCounter[jg] << endl;
//  }
//
//  for(int iat=0; iat<p.getLocalNum(); iat++) {
//    fout << p.ID[iat] << " " << p.GroupID[iat] << " " 
//	 << p.Lattice.toCart(p.R[iat]) << endl;
//  }
//  fout << "============================== " << endl;
//#endif //PRINT_DEBUG
//
//
//  //return *max_element(nsub_g.begin(), nsub_g.end());
//  return maxat;
//}

void NNSpatialEngine::debugmsg(int i) {
}
  //#endif
  ///clean up the temporary data

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: NNSpatialEngine.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/
