//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Utilities/OhmmsInfo.h"
#include "Particle/NNSpatialOMPEngine.h"
#ifdef HAVE_LIBMETIS
#include "Particle/MetisGraphPartition.h"
#endif
#include "Particle/PLayoutFunctions.h"
#include "Message/OpenMP.h"
#include <fstream>
using namespace OHMMS;

NNSpatialOMPEngine::NNSpatialOMPEngine(){ //:OmpGrid(0){
  setID(ThisMethod);
  UseLayout = true;
  LOGMSG("Creating NNSpatialOMPEngine, distributed over OpenMP threads")
}

NNSpatialOMPEngine::~NNSpatialOMPEngine() { 
}

/** prepare the spatial grid to perform linked-cell calculations
 *@param p the particle whose neighbor lists are evaluated by this engine.
 *@param rmax the maximum interaction radius
 *@return true NNRegiGridEngine uses the cell index of the current configuration
 * 
 *- make a linked-cell list with the spatial grid
 *- apply boundary conditions to the current position p.R
 *- create the data sets for each grid
 *- update the layout information, such as cell indices with p.R
 *- partition the local particles over the number of threads
 */ 
bool
NNSpatialOMPEngine::initialize(Particle_t& p, bool needvisitingforce) {

  Distributed[enable_omp]=true;

  DEBUGMSG("NNSpatialOMPEngine::initialize")
  Rmax = p.Lattice.InteractionRadius;
  RmaxSq = Rmax*Rmax;

  UseVisitingForce=needvisitingforce;

  if(MaxConnectedCells != p.Lattice.MaxConnections) { 

    //OmpGrid = p.Lattice.addGrid(omp_grid,p.Lattice.Grid[omp_grid]);
    ParticleLayout_t::Grid_t* ompGrid=p.Lattice.getGrid(OMP_GRID);
    if(ompGrid->empty()) {
      ompGrid->setGrid(Grid[OMP_GRID]);
      p.Lattice.initGrid(*ompGrid,*(p.Lattice.getGrid(SPATIAL_GRID)));
    }
    //OmpGrid = &(p.Lattice.getGrid(omp_grid));
    //if(OmpGrid.empty()) {
    //  OmpGrid = p.Lattice.addGrid(omp_grid,p.Lattice.Grid[omp_grid]);
    //  p.Lattice.initGrid(*OmpGrid,p.Lattice);
    //}
    //OmpGrid->createData();

    Index_t nattrib =  PairAttrib.size();
    //for(Index_t i=nattrib; i< OmpGrid->getNumDataSets(); ++i) PairAttrib.push_back(0);
    for(Index_t i=nattrib; i< ompGrid->getNumDataSets(); ++i) PairAttrib.push_back(0);
    FirstTouch = !(p.hasAttrib(openmp_tag));
    OmpAttrib = p.getIndexAttrib(openmp_tag);

    MaxConnectedCells = p.Lattice.MaxConnections;
    p.curR.setUnit(PosUnit::LatticeUnit);
    p.applyBC(p.R,p.curR);
    MaxPtclsPerCell = createGridData(p.Lattice,p.R,p.curR);

    MaxPairsPerPtcl= (MaxConnectedCells+1)*(MaxPtclsPerCell+1);
    if(PresetMaxNN>0) {
      MaxPairsPerPtcl = std::min(PresetMaxNN,MaxPairsPerPtcl);
    }

    LOGMSG("\tLinked cell data")
    LOGMSG("\t  Max connected cells " << MaxConnectedCells)
    LOGMSG("\t  Max numer of particles per cell " << MaxPtclsPerCell)
    if(UseVisitingForce)
      LOGMSG("\tNNSpatialOMPEngine::UseVisitingForce=true")
    else
      LOGMSG("\tNNSpatialOMPEngine::UseVisitingForce=false")
  }

  return true;
}

NNSpatialOMPEngine::Index_t
NNSpatialOMPEngine::createGridData(const ParticleLayout_t& lattice, 
    const ParticlePos_t& pin, ParticlePos_t& pu) {

  LOGMSG("\tNNSpatialOMPEngine::createGridData")

  const ParticleLayout_t::Grid_t* ompGrid(lattice.getGrid(OMP_GRID));
  Index_t nparts=ompGrid->getNumDataSets();

  //First evenly divide them
  FairDivide(pin.size(),nparts,PtclDist);
  NumLocalPtcls.resize(nparts);
  Grid2Node.resize(lattice.getGrid(SPATIAL_GRID)->getTotalNum());
  Cell = new ParticleIndex_t(pin.size());
  OmpPartsA.resize(nparts,0);
  OmpPartsB.resize(nparts,0);
  GhostList.resize(nparts,0);
  ConnectedNode.resize(nparts,0);

  std::vector<Index_t> npmaxpergrid(nparts);

  for(Index_t myid=0; myid<nparts; ++myid) {
    ParticleLayout_t::const_iterator oit(ompGrid->beginData(myid));
    ParticleLayout_t::const_iterator oit_end(ompGrid->endData(myid));
    while(oit != oit_end) { 
      Grid2Node[*oit]=myid; ++oit;
    }
  }
  
//#pragma omp parallel for schedule(static,1)
//  for(omp_int_t myid=0; myid<nparts; ++myid) {
#pragma omp parallel 
  {
    omp_int_t myid=omp_get_thread_num();

    //sg: reference to the spatical grid
    const ParticleLayout_t::Grid_t* restrict sg(lattice.getGrid(SPATIAL_GRID));

    Index_t ndataset = sg->getTotalNum();
    Index_t nat = pin.size();
    OmpPartsA[myid] = new OMPIndex_t();
    OmpPartsB[myid] = new OMPIndex_t();
    OmpPartsA[myid]->create(ndataset);
    OmpPartsB[myid]->create(ndataset);
    OmpPartsA[myid]->M.resize(ndataset+1);
    OmpPartsB[myid]->M.resize(ndataset+1);


    //use masks to find the on-grid and ghost-grid
    std::vector<int> gmask(ndataset,-2);
    ParticleLayout_t::const_iterator oit(ompGrid->beginData(myid));
    ParticleLayout_t::const_iterator oit_end(ompGrid->endData(myid));
    while(oit != oit_end) { 
      gmask[*oit]=myid;++oit;
    }
    Index_t nghost = 0;
    oit=ompGrid->beginData(myid);
    while(oit != oit_end) { 
      Index_t ic=(*oit);
      Index_t nc = lattice.first_connected(ic);
      Index_t nc_last = lattice.last_connected(ic);
      while(nc!= nc_last){
        Index_t jc=lattice.id_connected(nc);
        if(gmask[jc] == -2) { 
          gmask[jc] = -1; ++nghost;
        }
        ++nc;
      }
      ++oit;
    }

    //check the ghost cells
    GhostListType *gcell = new GhostListType;
    set<Index_t>* conns = new set<Index_t>;
    for(Index_t ig=0; ig<ndataset; ++ig)
      if(gmask[ig] == -1) {//(*gcell)[item++] = ig;
        (*gcell)[ig] = new std::vector<Index_t>;
        conns->insert(Grid2Node[ig]);
      }

    OMPIndex_t* restrict new_ind = OmpPartsA[myid];

    Index_t nsub = 0;
    register SingleParticlePos_t pos;
    register Tensor_t G = lattice.G;
    for(int iat=0; iat<pin.size(); ++iat) {
      int iloc = sg->loc(pu[iat]);
      if(gmask[iloc] == myid) {
        (*Cell)[iat]=iloc;
        (*OmpAttrib)[iat]=myid;
        new_ind->add(iloc,iat);
        ++nsub;
      }
    }

    std::size_t npc=0;
    for(Index_t ic=0; ic<ndataset; ++ic) {
      npc = std::max(npc,new_ind->size(ic));
    }

    //assign the local values to the data members
    NumLocalPtcls[myid] = nsub;
    npmaxpergrid[myid] = npc;
    GhostList[myid] = gcell;
    ConnectedNode[myid] = conns;

    if(!PairAttrib[myid]) PairAttrib[myid]= new PtclPairList_t;
    PairAttrib[myid]->set(myid,0,nsub);
    Index_t npc_i = (MaxConnectedCells+1)*(npc+1);
    if(PresetMaxNN>0) npc = std::min(PresetMaxNN,npc_i);
    PairAttrib[myid]->resizePairs(nsub,npc_i);

    //update the indices and ghost lists
    nsub=0;
    new_ind->M[0]=0;
    for(Index_t ic=0; ic<ndataset; ++ic) {
      new_ind->M[ic+1] = new_ind->M[ic]+new_ind->size(ic);
      OMPIndex_t::const_iterator it(new_ind->begin(ic));
      OMPIndex_t::const_iterator it_end(new_ind->end(ic));
      while(it != it_end) {
        PairAttrib[myid]->I[nsub++]=*it;++it;
      }
    }
  }

#pragma omp parallel 
  {
    omp_int_t myid=omp_get_thread_num();
    Index_t ntot = OmpPartsA[myid]->M.back();
    const OMPIndex_t* ghost_ind = 0;
    GhostListType::iterator git(GhostList[myid]->begin());
    GhostListType::iterator git_end(GhostList[myid]->end());
    while(git != git_end) {
      Index_t ig = (*git).first;
      Index_t remoteloc = Grid2Node[ig];
      ghost_ind = OmpPartsA[remoteloc];
      OMPIndex_t::const_iterator it(ghost_ind->begin(ig));
      OMPIndex_t::const_iterator it_end(ghost_ind->end(ig));
      vector<Index_t>* restrict glist = (*git).second;
      glist->clear();
      glist->push_back(ntot);
      while(it != it_end) {
        glist->push_back(*it); ++ntot; ++it;
      }
      ++git;
    }
    PairAttrib[myid]->Visitors = ntot;
  }

  CopyA = true;
  Index_t nloc_max = 0;
  for(Index_t myid=0; myid<nparts; ++myid) 
    nloc_max = std::max(nloc_max,npmaxpergrid[myid]);
  return nloc_max;
}

NNSpatialOMPEngine::Index_t 
NNSpatialOMPEngine::updateGridData(const ParticlePos_t& pos,  ParticlePos_t& pu,
                                   ParticleLayout_t& lattice) {

  PLayoutFunc::applyBC(lattice,pos,pu);

//#pragma omp parallel for schedule(static,1)
//  for(omp_int_t myid=0; myid<NumThreads; ++myid) {
#pragma omp parallel
  {
    omp_int_t myid=omp_get_thread_num();

    const OMPIndex_t* old_ind = 0;
    OMPIndex_t* new_ind =0;
    if(CopyA) {
      old_ind = OmpPartsA[myid];
      new_ind = OmpPartsB[myid];
    } else {
      old_ind = OmpPartsB[myid];
      new_ind = OmpPartsA[myid];
    }

    new_ind->clear();
    Index_t nsub = 0;
    Index_t ndataset=old_ind->getNumDataSets();

    //sg: reference to the spatical grid
    const ParticleLayout_t::Grid_t* sg(lattice.getGrid(SPATIAL_GRID));
    for(Index_t ig=0; ig<ndataset; ++ig) {
      OMPIndex_t::const_iterator it(old_ind->begin(ig));
      OMPIndex_t::const_iterator it_end(old_ind->end(ig));
      while(it != it_end) {
        int iat = (*it);
        int iloc = sg->loc(pu[iat]);
        if(Grid2Node[iloc] == myid) {
          (*Cell)[iat]=iloc;
          (*OmpAttrib)[iat]=myid;
          new_ind->add(iloc,iat);
          nsub++;
        }
        ++it;
      }
    }

    GhostListType::const_iterator git(GhostList[myid]->begin());
    GhostListType::const_iterator git_end(GhostList[myid]->end());
    while(git != git_end) {
      Index_t ig = (*git).first;
      Index_t remoteloc = Grid2Node[ig];
      if(CopyA) {
        old_ind = OmpPartsA[remoteloc];
      } else {
        old_ind = OmpPartsB[remoteloc];
      }

      OMPIndex_t::const_iterator it(old_ind->begin(ig));
      OMPIndex_t::const_iterator it_end(old_ind->end(ig));
      while(it != it_end) {
        int iat = (*it);
        int iloc = sg->loc(pu[iat]);
        if(Grid2Node[iloc] == myid) {
          (*Cell)[iat]=iloc;
          (*OmpAttrib)[iat]=myid;
          new_ind->add(iloc,iat);
          ++nsub;
        }
        ++it;
      }       
      ++git;
    }

    NumLocalPtcls[myid]=nsub;
    PairAttrib[myid]->resizePairs(nsub,PairAttrib[myid]->getMaxNadj());

    new_ind->M[0]=0;
    Index_t nat=0;
    for(Index_t ic=0; ic<new_ind->getNumDataSets(); ic++) {
      new_ind->M[ic+1] = new_ind->M[ic]+new_ind->size(ic);
      OMPIndex_t::const_iterator it(new_ind->begin(ic));
      OMPIndex_t::const_iterator it_end(new_ind->end(ic));
      while(it != it_end) {
        PairAttrib[myid]->I[nat++]=*it;++it;
      }
    }
  }

  CopyA = (CopyA)? false:true;

//#pragma omp parallel 
//  {
//    int myid = omp_get_thread_num();
////#pragma omp for schedule(static,1) nowait
////    for(int myid=0; myid<np; myid++) {
//      int ntot=0;
//      if(CopyA) {
//        ntot = OmpPartsA[myid]->M.back();
//      } else {
//        ntot = OmpPartsB[myid]->M.back();
//      }
//      const OMPIndex_t* ghost_ind = 0;
//      GhostListType::iterator gjt = GhostList[myid]->begin();
//      while(gjt != GhostList[myid]->end()) {
//        int ig = (*gjt).first;
//        int remoteloc = Grid2Node[ig];
//        if(CopyA) {
//          ghost_ind = OmpPartsA[remoteloc];
//        } else {
//          ghost_ind = OmpPartsB[remoteloc];
//        }
//        OMPIndex_t::const_iterator it = ghost_ind->begin(ig);
//        vector<int>* restrict glist = (*gjt).second;
//        glist->clear();
//        glist->push_back(ntot);
//        while(it != ghost_ind->end(ig)) {
//          glist->push_back(*it); ntot++; it++;
//        }
//        gjt++;
//      }
//
//      PairAttrib[myid]->Visitors = ntot;
//  }
  return 1;
}

void NNSpatialOMPEngine::metisGridPartition(ParticleLayout_t& lattice) {

#ifdef HAVE_LIBMETIS
  typedef std::vector<int> IndexList_t;
  int ng = lattice.ncontexts();
  IndexList_t wgt(ng),gid(ng);

  ParticleLayout_t::Grid_t& ompGrid(*lattice.getGrid(OMP_GRID));
  int np = ompGrid.getTotalNum();

  for(int myid=0; myid<np; ++myid) {
    const OMPIndex_t* old_ind = 0;
    if(CopyA) 
      old_ind = OmpPartsA[myid];
    else 
      old_ind = OmpPartsB[myid];
    for(int ic=0; ic<old_ind->getNumDataSets(); ++ic) {
      wgt[ic] += old_ind->size(ic);
    }
  }

  MetisGraphPartition<IndexList_t> graph(np);
  graph.partition(lattice.c_offset,lattice.c_id, wgt,gid);
  ompGrid.clearData();

  for(int ig=0; ig<ng; ++ig) {
    ompGrid.addData(gid[ig],ig);
  }

  //need to remapping
#pragma omp parallel
  {
    int myid = omp_get_thread_num();
    OMPIndex_t* new_ind = 0;
    if(CopyA) {
      new_ind = OmpPartsB[myid];
    } else {
      new_ind = OmpPartsA[myid];
    }

    const ParticleLayout_t::Grid_t* OmpGrid(lattice.getGrid(OMP_GRID));
    new_ind->clear();
    ParticleLayout_t::const_iterator git(OmpGrid->beginData(myid));
    ParticleLayout_t::const_iterator git_end(OmpGrid->endData(myid));
    while(git != git_end) {
      int ig = (*git);
      int old_thread = Grid2Node[ig];
      const OMPIndex_t* old_ind = 0;
      if(CopyA) {
        old_ind = OmpPartsA[old_thread];
      } else {
        old_ind = OmpPartsB[old_thread];
      }
      OMPIndex_t::const_iterator it(old_ind->begin(ig));
      OMPIndex_t::const_iterator it_last(old_ind->end(ig));
      while(it != it_last) {
        new_ind->add(ig,*it); 
        ++it;
      }
      ++git;
    }
  }

  //update the reverse map Grid2Node[ig] returns the thread id
  for(int ig=0; ig<ng; ++ig) {
    Grid2Node[ig] = gid[ig];
  }

  CopyA = (CopyA)? false:true;

#endif

  //sync();
  //PtclDist[0] = 0;
  //for(int i=0; i<NumLocalPtcls.size(); i++) {
  //  PtclDist[i+1] = PtclDist[i]+NumLocalPtcls[i];
  //}
}

void
NNSpatialOMPEngine::collectForce(ParticlePos_t& f, int ip) {

  //Serial bottleneck
  //const PtclPairList_t& nnlist = *(PairAttrib[ip]);
  //const PtclPairList_t::VectorList_t& fv = nnlist.Fv;
  //GhostListType::const_iterator jt = GhostList[ip]->begin();
  //while(jt != GhostList[ip]->end()) {
  //  const std::vector<int>& glist = *((*jt).second);
  //  int offset=glist[0];
  //  for(int iL=1; iL<glist.size(); iL++,offset++) {
  //    f[glist[iL]]+=fv[offset];
  //  }
  //  jt++;
  //}

  const PtclPairList_t& nnlist =  *(PairAttrib[ip]);
  const ParticlePos_t& fv = nnlist.Fv;
  for(Index_t iL=0; iL<nnlist.LocalNum; iL++) f[nnlist.I[iL]]=fv[iL];

  set<Index_t>::const_iterator it(ConnectedNode[ip]->begin());
  set<Index_t>::const_iterator it_end(ConnectedNode[ip]->end());
  while(it != it_end) {
    Index_t remoteloc = (*it);
    const ParticlePos_t& rfv = PairAttrib[remoteloc]->Fv;
    GhostListType::const_iterator jt(GhostList[remoteloc]->begin());
    GhostListType::const_iterator jt_end(GhostList[remoteloc]->end());
    while(jt != jt_end) {
      if(Grid2Node[(*jt).first] == ip) {
        const std::vector<Index_t>& glist = *((*jt).second);
        Index_t offset=glist[0];
        for(Index_t iL=1; iL<glist.size(); ++iL,++offset) {
          f[glist[iL]]+=rfv[offset];
        }
      } 
      ++jt;
    }
    ++it;
  }
}

void 
NNSpatialOMPEngine::evalPairsPerThread(const ParticleLayout_t& lattice,
    const ParticlePos_t& pos, const ParticleIndex_t& grID, 
    PtclPairList_t& myppl) {

  typedef OMPIndex_t::const_iterator const_iterator;
  Index_t myid(myppl.myID);
  const OMPIndex_t* old_ind = 0;
  Index_t ntot(0);
  if(CopyA) {
    old_ind = OmpPartsA[myid];
    ntot = OmpPartsA[myid]->M.back();
  }
  else
  {
    old_ind = OmpPartsB[myid];
    ntot = OmpPartsB[myid]->M.back();
  }

  Index_t nat(ntot);
  const OMPIndex_t* ghost_ind = 0;
  GhostListType::iterator gjt(GhostList[myid]->begin());
  GhostListType::iterator gjt_end(GhostList[myid]->end());
  while(gjt != gjt_end) {
    Index_t ig = (*gjt).first;
    Index_t remoteloc = Grid2Node[ig];
    if(CopyA) {
      ghost_ind = OmpPartsA[remoteloc];
    } else {
      ghost_ind = OmpPartsB[remoteloc];
    }
    OMPIndex_t::const_iterator it(ghost_ind->begin(ig));
    OMPIndex_t::const_iterator it_end(ghost_ind->end(ig));
    vector<Index_t>* restrict glist = (*gjt).second;
    glist->clear();
    glist->push_back(ntot);
    while(it != it_end) {
      glist->push_back(*it); ++it;++ntot;
    }
    ++gjt;
  }

  PairAttrib[myid]->Visitors = ntot;

  myppl.reset();

  const_iterator jc_first, jc_last;

  for(Index_t iL=0; iL < nat; iL++) {
    Index_t iat=myppl.I[iL];
    Index_t ic=(*Cell)[iat];
    Index_t jL(old_ind->M[ic]);
    //search particles on the same grid
    const_iterator it(old_ind->begin(ic));
    const_iterator it_end(old_ind->end(ic));
    while(it != it_end) {
      Index_t jat=*it;
      if(jat != iat) {
        SingleParticlePos_t dr(lattice.toCart(pos[jat]-pos[iat]));
        Scalar_t rsq(dot(dr,dr));
        if(rsq< RmaxSq && rsq > RminSq) {
          myppl.addPair(iL,jat,jL,grID[jat],sqrt(rsq),dr);
        }
      }//exclude itself
      ++it;++jL;
    }

    //search particles on the connected grids
    for(Index_t nc=lattice.first_connected(ic); nc<lattice.last_connected(ic);nc++) {
      Index_t jc = lattice.id_connected(nc);
      Index_t where = Grid2Node[jc];
      SingleParticlePos_t del = lattice.bc(nc);
      if(where == myid) {
        jL = old_ind->M[jc];
      } else {
        GhostListType::iterator git(GhostList[myid]->find(jc));
        jL = (*git).second->operator[](0);
      }
      if(CopyA) {
        jc_first = OmpPartsA[where]->begin(jc);
        jc_last =OmpPartsA[where]->end(jc);
      } else {
        jc_first=OmpPartsB[where]->begin(jc);
        jc_last=OmpPartsB[where]->end(jc);
      }
      while(jc_first != jc_last) {
        Index_t jat = *jc_first;
        SingleParticlePos_t dr(lattice.toCart(pos[jat]-pos[iat]) + del);
        Scalar_t rsq(dot(dr,dr));
        if(rsq< RmaxSq && rsq > RminSq) {
          myppl.addPair(iL,jat,jL,grID[jat],sqrt(rsq),dr);
        }
        ++jc_first; ++jL;
      }
    }
  }
 
  myppl.checkPairs(UseVisitingForce);
}

bool NNSpatialOMPEngine::putExtra(xmlNodePtr cur){
  return true;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: NNSpatialOMPEngine.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 *******************************************************************/
