//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Message/OpenMP.h"
#include <math.h>
#ifdef USE_METIS
extern "C" {
#include <metis.h>
}
#endif

#include <iostream>
#include <fstream>
#include "Particle/ParticleOperators.h"
#include "Particle/NNEngine.h"
#include "CrystalSymmetry/KPointBase.h"
using namespace OHMMS;
void 
OHMMS::resetPhase(const ParticleSet* ptcl, 
                  ComplexMatrix_t& phi, 
		  int kstart, int kend) {
   
  typedef ParticleSet::Scalar_t            Scalar_t;
  typedef ParticleSet::SingleParticlePos_t SingleParticlePos_t;

  int npe(ptcl->getNNEngine()->size());
  vector<int> nntot(npe+1);

  nntot[0] = 0;
  for(int i=0; i<npe; i++) {
    nntot[i+1] = nntot[i] + ptcl->getNNEngine()->getPairAttrib(i).getTotNadj();
  }

  //phi.resize(ptcl->KSet->size(), nntot[npe]);
  phi.resize(kend-kstart, nntot[npe]);

#pragma omp parallel
  {
#ifdef ENABLE_OPENMP
    int myID = omp_get_thread_num();
#else
    int myID = 0;
#endif
    const ParticleSet::PtclPairList_t& nnlist(ptcl->getNNEngine()->getPairAttrib(myID));
    const KPointBase* kset(ptcl->getKSet());

    //int nkpt = ptcl->KSet->size();
    int nntot0 = nntot[myID];
    for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
      ParticleSet::PtclPairList_t::const_iterator 
        nit(nnlist.begin(iL)), nit_end(nnlist.end(iL));
      while(nit != nit_end) {
	SingleParticlePos_t r((*nit).displ);
      //for(int nni=0; nni< nnlist.nadj(iL); nni++, nn++, nntot0++){
	//SingleParticlePos_t r(nnlist[nn].displ);
	//for(int ik=0; ik<nkpt; ik++) {
	for(int ik=kstart, nk=0; ik<kend; ik++,nk++) {
	  Scalar_t kr(dot(r,kset->k_cart(ik)));
	  phi[nk][nntot0] = complex<Scalar_t>(cos(kr),sin(kr));
	}
        ++nit;++nntot0;
      }
    }
  }

}


extern void MetisPtclPartitionOMP(ParticleSet& p, int& numpart, int glevel);

void 
OHMMS::distributeParticle(ParticleSet& p, int& numpart, int glevel) {

  /////////////////////////////////////////////////////////
#if defined(USE_METIS) && defined(ENABLE_OPENMP)

  int NumThreads(p.getNNEngine()->size());
  if(NumThreads == 1) return;
  typedef PtclPairList_t::PairData_t    PairData_t;

  int n(p.getLocalNum());
  vector<int> M_copy(n+1,0);
  int n0(0);
  int nat(0);
  numpart = NumThreads;

  for(int inode=0; inode<NumThreads; inode++) {
    const PtclPairList_t& nnlist = p.nnListEngine->getPairAttrib(inode);
    //for(int nloc=1; nloc<=myAtoms[inode][2]; nloc++) {
    for(int nloc = 0; nloc <  nnlist.getLocalNum(); nloc++,nat++) {      
      M_copy[nat+1] = M_copy[nat] + nnlist.nadj(nloc);
	//M_copy[nat] = nnlist.M[nloc] + n0;
    }
    n0 += nnlist.getTotNadj();
  }

  vector<int> J_copy(M_copy[n]);
  int nn=0;
  for(int inode=0; inode<NumThreads; inode++) {
    const PtclPairList_t& nnlist = p.nnListEngine->getPairAttrib(inode);
    for(int j=0; j<nnlist.getTotNadj(); j++, nn++) {
      J_copy[nn] = nnlist[nn].J;
    }
  }

  /////////////////////////////////////////////////////////////
  //graph partition by Metis
  /////////////////////////////////////////////////////////////
  int wgtflag = 0;//no-weight is used
  int numflag = 0;//c-style
  int metisopt[5]; metisopt[0] = 0;//using default
  int volume;

  ParticleIndex_t& ompid = *p.getIndexAttrib("ompid");
  METIS_PartGraphRecursive(&n,&(M_copy[0]),&(J_copy[0]), NULL,NULL,
			   &wgtflag,&numflag,&numpart,metisopt,
			   &volume,&(ompid[0]));

  // id is used to save the local index to a new grobal index
  vector<int> num(NumThreads+1,0);
  vector<int> id(n);
  for(int i=0; i<n; i++){    
    int partid = ompid[i]+1;
    id[i] = num[partid];
    num[partid]++;
  }
  for(int ic=1; ic<num.size(); ic++)  
    num[ic] += num[ic-1];
  num[0]=0;
  for(int ic=0; ic<num.size(); ic++) 
    p.Lattice.dGrid[glevel]->PtclDist[ic] = num[ic];

  ParticleIndex_t ionidtmp(p.GroupID);
  ParticlePos_t ptmp(p.R);
  ParticleIndex_t revmap(n);
  for(int i=0; i<n; i++) {
    int newid = id[i] + num[ompid[i]];
    p.R[newid] = ptmp[i];
    p.GroupID[newid] = ionidtmp[i];
    id[i] = newid;     // map old -> new
    revmap[newid] = i; // map new -> old
  }

//    cout << p.getLocalNum() << endl << endl;
//    if(p.R.InUnit) {
//      for(int i=0; i<p.getLocalNum(); i++) {
//        cout << "Si" << ompid[i] << " " << p.Lattice.toCart(p.R[i]) << endl;
//      }
//    } else {
//      for(int i=0; i<p.getLocalNum(); i++) {
//        cout << "Si" << ompid[i] << " " << p.R[i] << endl;
//      }
//    }
#endif
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: ParticleOperators.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

