//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Utilities/OhmmsInfo.h"
#include "Particle/ParticleSet.h"
#include "ParticleBase/ParticleUtility.h"
#include "Particle/NNEngine.h"
#include "CrystalSymmetry/KPointBase.h"
#include <bitset>

namespace OHMMS {

  ParticleSet::ParticleSet():
    Mode(Serial),
    PeriodicSystem(true), 
    NeedPartition(true), 
    UseGridLayout(false), 
    KeepSumBC(false),
    PFile("unknown"),
    nnListEngine(0), 
    KSet(0)
  { 
    initCommonAttributes();
  }

  ParticleSet::ParticleSet(const SpeciesSet& species):
    Mode(Serial),
    PeriodicSystem(true), 
    NeedPartition(true), 
    UseGridLayout(false), 
    KeepSumBC(false),
    PFile("unknown"),
    mySpecies(species),
    nnListEngine(0), 
    KSet(0)
  { 
    initCommonAttributes();
  }


  ParticleSet::ParticleSet(const ParticleSet& ptclin):
    Mode(ptclin.Mode),
    NeedPartition(true), 
    mySpecies(ptclin.getSpeciesSet()),
    nnListEngine(0), 
    KSet(0)
  {
    initCommonAttributes();
    assign(ptclin);
  }

  ParticleSet::~ParticleSet() { 
    if(KSet) delete KSet;
  }

  void ParticleSet::initCommonAttributes() {
    //set the type and object name
    F.setTypeName(ParticleTags::postype_tag);
    F.setObjName(ParticleTags::force_tag);
    V.setTypeName(ParticleTags::postype_tag);
    V.setObjName(ParticleTags::velocity_tag);
    dR.setTypeName(ParticleTags::postype_tag);
    dR.setObjName(ParticleTags::displacement_tag);
    E.setTypeName(ParticleTags::scalartype_tag);
    E.setObjName(ParticleTags::energy_tag);

    RestartList.push_back(&ID);
    RestartList.push_back(&GroupID);
    RestartList.push_back(&R);

    addAttribute(dR);
  }

  void ParticleSet::clearAttributes() {

    LOGMSG("\tClear dynamically added particle attributes")

    //check if dynamic attributes are used
    bool hasf = hasAttrib(ParticleTags::force_tag);
    bool hasv = hasAttrib(ParticleTags::velocity_tag);

    //myAttribList contains dynamically created attributes, 
    //while others are just pointers
    for(int i=0; i<myAttribList.size(); ++i) {delete myAttribList[i];}

    AttribList.clear();
    INDEX.clear();
    VAL.clear();
    POS.clear();
    //TENZOR.clear();
    myAttribList.clear();
    ConfigList.clear();

    ///add basic attributes
    addAttribute(ID);
    addAttribute(GroupID);
    addAttribute(R);
    if(hasf) addDynamicAttributes(!hasv);
  }

  void ParticleSet::addDynamicAttributes(bool forceonly) {
    addAttribute(F);
    addAttribute(E);
    if(!forceonly) {
      addAttribute(V); 
    } else {
      cout << "Add only force and energy " << endl;
    }
  }

  void ParticleSet::addRestartAttrib(const string& aname) {
    bool doesnotexist = true;
    size_type i =0;
    size_type nconf = RestartList.size();
    while(doesnotexist && i<nconf) {
      if(RestartList[i++]->objName() == aname) 
        doesnotexist = false;
    }
    if(doesnotexist) {
      map<string,OhmmsObject*>::iterator it(AttribList.find(aname));
      if(it != AttribList.end()) {
        RestartList.push_back((*it).second);
      }
    }
  }

  void ParticleSet::begin_node(std::ostream& os) const {
    os << "<" << root_tag 
       << " name=\"" << myName 
       << "\" size=\"" << getLocalNum() << "\">" << std::endl;
  }

  void ParticleSet::end_node(std::ostream& os) const {
    os << "</" << root_tag << ">" << std::endl;
  }


  bool ParticleSet::get(std::ostream& os) const { 
    os << "Lattice " << endl;
    Lattice.print(os);
    for(size_type i=0; i<getTotalNum(); i++) {
      os << ID[i] << " " << R[i] << endl;
    }
    return true;
  }

  bool ParticleSet::put(std::istream&) { 
    return true;
  }

  void ParticleSet::reset() { }

  bool ParticleSet::put(xmlNodePtr cur){
    /*
#if defined(HAVE_LIBXML2)
    cur = cur->xmlChildrenNode;
    while(cur != NULL) {
      string a((const char*)(cur->name));
      if(a == "particlelayout") {
        if(nnListEngine) {
	  WARNMSG("\tDeleting an existing NNEngine")
	    delete nnListEngine;
        }
        nnListEngine = NNEngineBuilder::create(cur);
      }
      cur = cur->next;
    }
#endif
*/
    return true;
  }

  void ParticleSet::addPositionAttrib(ParticlePos_t& pos) {

    std::string pname = pos.objName();
    bool canaddpos = true;
    size_type i =0;
    size_type nconf = ConfigList.size();
    while(canaddpos && i<nconf) {
      if(ConfigList[i++]->objName() == pname) canaddpos = false;
    }

    DEBUGMSG("Adding " << pos.objName() << " to the attribute list to apply BC")
    if(canaddpos) {
      ConfigList.push_back(&pos);      
      if(pname != position_tag) RestartList.push_back(&pos);
    }
  }

  void ParticleSet::rescale(Scalar_t ascale) {

    vector<bool> cunit(ConfigList.size(),false);
    for(size_type ic=0; ic<ConfigList.size(); ic++) {
      if(ConfigList[ic]->getUnit() == PosUnit::CartesianUnit) {
	cunit[ic]=true;
	//Lattice.convert2Unit(*ConfigList[ic]);
        convert2Unit(*ConfigList[ic]);
      }
    }

    Tensor_t newlat = ascale*Lattice.R;
    Lattice.update(newlat);

    for(size_type ic=0; ic<ConfigList.size(); ic++) {
      //if(cunit[ic]) Lattice.convert2Cart(*ConfigList[ic]);
      if(cunit[ic]) convert2Cart(*ConfigList[ic]);
    }

    if(nnListEngine) nnListEngine->rescale(ascale);
  }

  void ParticleSet::setLayoutUpdateStride(int astride){
    if(nnListEngine) nnListEngine->setStride(astride);
  }   

  /** Update particle layout with a new supercell
   *
   *Performs similar tasks to applyBConds but the supercell has been changed.
   */
  void ParticleSet::updateLayout(const Tensor_t& newlat) {
    Lattice.update(newlat);
    if(KSet) KSet->LatRef.set(Lattice.R);
  }

  /** Set the mySpecies
   */
  void ParticleSet::setSpeciesSet(const SpeciesSet& sp) {
    WARNMSG("The species set is overwritten.")
    mySpecies=sp;
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 366 $   $Date: 2005-07-22 20:42:26 -0500 (Fri, 22 Jul 2005) $
 * $Id: ParticleSet.cpp 366 2005-07-23 01:42:26Z jnkim $ 
 ***************************************************************************/

