//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file ParticleSet.h
 *@brief Declare a specialized particle class for atomistic simulations.
 */
#ifndef OHMMS_PARTICLESET_H
#define OHMMS_PARTICLESET_H

#include "Particle/ParticleTraits.h"
#include "OhmmsData/OhmmsElementBase.h"
#include "ParticleBase/PtclPairListBase.h"
#include "Utilities/SpeciesSet.h"

namespace OHMMS {

  class NNEngine;
  class KPointBase;

  /** Specialized particle class for simulations with interatomic potentials.
   *
   * Derived from ParticleBase<PtclOnLatticeTraits>.
   * The ParticleLayout class represents a supercell with/without periodic boundary
   * conditions. The ParticleLayout class also takes care of spatial decompositions
   * for efficient evaluations for the interactions with a finite cutoff.
   */
  class ParticleSet:  
    public ParticleBase<PtclOnLatticeTraits>,
    public OhmmsElementBase
 {

 public:
   
   enum {Serial=0, 
	 Mixed, 
	 Distributed};
    
   typedef PairDistData<OHMMS_PRECISION,OHMMS_DIM,long>  PairData_t;
   typedef PtclPairListBase<PairData_t,OHMMS_PRECISION,OHMMS_DIM> PtclPairList_t;
  
   /**
    *@defgroup DynamicAttribs Particle attributes for dynamics
    *@{
    * Force \f$\equiv -\nabla V\f$*/
   ParticlePos_t   F;
   /* Velocity */
   ParticlePos_t   V;
   /* Carteisan displacement for fast distance evaluations */
   ParticlePos_t   dR;
   /* Atomic energy */
   ParticleScalar_t E;
   /*@}*/

   vector<OhmmsObject*>  RestartList;

   ParticleSet();
   ParticleSet(const ParticleSet&);
   ParticleSet(const SpeciesSet& species);
   virtual ~ParticleSet();

   //void assign(const Base_t&);
   void update(int i=0);
   
   /** Add an attribute to which BoundaryConditions should be applied.
    *\param pos ParticlePos_t that is governed by boundary conditions
    */
   void addPositionAttrib(ParticlePos_t& pos);

   /** Add a named attribute to the restart attribute list
    * @param aname name of the attribute to be added to the record list
    */
   void addRestartAttrib(const string& aname);
   
   /** Add F, E, V and dR to the list of ParticleAttribute
    *
    * V and dR are added if forceonly is false.
    */
   void addDynamicAttributes(bool forceonly=false);

   /** clear dynamic attributes and keep only those of the base class */
   void clearAttributes();
   
   /** apply boundary conditions to the attributes belonging to ConfigList
    * @param updateCurR if true, ParticleSet::curR is updated.
    * @param latticeunit if true, ParticleSet::curR is in LatticeUnit.
    *
    * curR is a reserved ParticlePos_t to hold temporary data for complex
    * operations for the attributes. If updateCurR == true, at the end
    * of the operations to bring back all the particle attributes associated
    * with the position variables, ParticleSet::R is copied to ParticleSet::curR
    * of the requested unit.
    */
   void applyBConds(bool updateCurR, bool latticeunit = true);

   /** update the ParticleLayout with the interaction ranges.
    * @param int_rad interaction radius
    * @param con_rad connectivity radius
    *
    * Often, it is safe to use different values for the physical interaction
    * range (int_rad) and the connectivity range for the grid (con_rad).
    * The layout class uses con_rad to build the grid connectivity,
    * while uses int_rad to determine the interactions.
    */
   void updateLayout(Scalar_t int_rad, Scalar_t con_rad);

   /** update the ParticleLayout when the unitcell has changed.
    * @param newlat the tensory describing the Lattice
    *
    * Delegate to Lattice to update the content.
    */
   void updateLayout(const Tensor_t& newlat);

   void setLayoutUpdateStride(int astride);
   void rescale(Scalar_t ascale);

   void convert(const ParticlePos_t& pin, ParticlePos_t& pout);
   void convert2Unit(const ParticlePos_t& pin, ParticlePos_t& pout);
   void convert2Cart(const ParticlePos_t& pin, ParticlePos_t& pout);
   void convert2Unit(ParticlePos_t& pout);
   void convert2Cart(ParticlePos_t& pout);

   void applyBC(const ParticlePos_t& pin, ParticlePos_t& pout);
   void applyBC(ParticlePos_t& pos);
   void applyBC(const ParticlePos_t& pin, ParticlePos_t& pout, size_type first, size_type last);

   void getMinimumDisplacement(ParticlePos_t& displ);

   //@{implementation of the virtual functions of OhmmsElementBase
   bool get(std::ostream& ) const;
   bool put(std::istream& );
   bool put(xmlNodePtr cur);
   void reset();
   void begin_node(std::ostream& os) const;
   void end_node(std::ostream& os) const;
   //@}
   
   //@{get/set function for a common file root
   void setRoot(const string& a) { PFile=a;}
   const string& getRoot() const { return PFile;}
   //@}
   
   //@{get functions for NNEngine*
   inline NNEngine* getNNEngine() { return nnListEngine;}
   inline const NNEngine* getNNEngine() const { return nnListEngine;}
   inline void setNNEngine(NNEngine* eng) { nnListEngine = eng;}
   //@}
   
   inline const KPointBase* getKSet() const { return KSet;}
   inline KPointBase* getKSet() { return KSet;}
   inline void setKSet(KPointBase* kset) {KSet=kset;}

   inline SpeciesSet& getSpeciesSet() { return mySpecies;}
   inline const SpeciesSet& getSpeciesSet() const { return mySpecies;}
   //this has to go away!!!
   void setSpeciesSet(const SpeciesSet& sp);

   ///set the mode
   inline void setMode(int imode) { Mode = imode;}
   
   void makeMove(size_type i, const SingleParticlePos_t& d);
   void restoreMove(size_type i);
   
#ifdef HAVE_MPI
   void startCollect(ParticlePos_t& fv);
   void endCollect(ParticlePos_t& fv);
   Scalar_t collect(Scalar_t x);
   void select(const vector<Index_t>& selected);
#else
   inline void startCollect(ParticlePos_t& fv) {}
   inline void endCollect(ParticlePos_t& fv){}
   inline Scalar_t collect(Scalar_t x){ return x;}
   inline void select(const vector<Index_t>& selected) { }
#endif

   
   protected:
   ///mode of operation
   int Mode;
   
   ///true, if any dimension is periodic
   bool PeriodicSystem;
   
   ///boolean to request re-partitioning of the particles on the grid
   bool NeedPartition;
   
   ///flag to determine if the grid-partition is used
   bool UseGridLayout;
   
   bool KeepSumBC;
   
   string PFile;
   
   SpeciesSet mySpecies;

   ///Reference to the NearestNeighborBase
   NNEngine* nnListEngine;
   
   ///Reference to a KPointBase
   KPointBase *KSet;

   ///A list of ParticlePos_t which the boundary conditions are applied to
   vector<ParticlePos_t*> ConfigList;


   void initCommonAttributes();
  };  
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 362 $   $Date: 2005-07-13 13:32:27 -0500 (Wed, 13 Jul 2005) $
 * $Id: ParticleSet.h 362 2005-07-13 18:32:27Z jnkim $ 
 *************************************************************************/
