// -*- C++ -*-
/***************************************************************************
 *
 * The POOMA Framework
 * 
 * This program was prepared by the Regents of the University of
 * California at Los Alamos National Laboratory (the University) under
 * Contract No.  W-7405-ENG-36 with the U.S. Department of Energy (DOE).
 * The University has certain rights in the program pursuant to the
 * contract and the program should not be copied or distributed outside
 * your organization.  All rights in the program are reserved by the DOE
 * and the University.  Neither the U.S.  Government nor the University
 * makes any warranty, express or implied, or assumes any liability or
 * responsibility for the use of this software
 *
 * Visit http://www.acl.lanl.gov/POOMA for more details
 *
 ***************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// MDParticle: public Var
//   contains a SimpleParticle that are updated by propagator
//
///////////////////////////////////////////////////////////////////////////////
#include "PoomaUI/MDEvents.h"
#include "MDParticle/SimpleParticle.h"
//MOD 11112000
#include "MDParticle/LocalHMDParticle.h"
#include "Potentials/PotentialBase.h"
#include "Utility/BondOrderFunctors.h"

char* SetMDEvent::help_msg() {
  return
  "\n events\n\n"
  " syntax: events [--rcut/--ecut/--y2cut  value --adjustcm on/off]\n\n"
  "   Set an event detector. \n"
  "   Default = --rcut 1 --adjustcm off\n"
  "   --rcut AA = use displacement in AA.\n"
  "   --ecut eV = use atomic energy  eV.\n"
  "   --y2cut value = use y2.\n"
  "   --adjustcm on/off = adjust center of mass.\n\n";
}

int SetMDEvent::action(int argc, char **argv) {
  
  ui->varlist[Name]->set(argc,argv);
  return 0;
}
/////////////////////////////////////////////////////////////////////
void MDEvents::setup() {

  Rcut = 1.0e0; Counter = 0; AdjustCM = false;
  Ecut = 0.1e0;
  RefConfig.push_back(&RefPos);
  RefConfig.push_back(&RefEne);
  d_Event = DISPLACEMENT;
}

int MDEvents::set(int argc, char **argv) {
  int i=0;
  while(i < argc) {
  
    if(!strcasecmp(argv[i], "--rcut")) {
      d_Event = DISPLACEMENT;
      Rcut = atof(argv[++i]);
    }
    if(!strcasecmp(argv[i], "--ecut")) {
      d_Event = ENERGY;
      Ecut = atof(argv[++i]);
    }
    if(!strcasecmp(argv[i], "--y2cut")) {
      d_Event = Y2;
      Ecut = atof(argv[++i]);
    }
    if(!strcasecmp(argv[i],"--adjustcm")) { 
      if(!strcasecmp(argv[++i],"on")) AdjustCM = true;
      else                            AdjustCM = false;
    }
    i++;
  }
  RcutSq = Rcut*Rcut;
  cout << " cuttoff radius = " << RcutSq << endl;
  return 0;
}

void MDEvents::print(ostream &os) {

    os.setf(ios::left,ios::adjustfield);
    os << name() << " ";
    if(d_Event == DISPLACEMENT) os << " --rcut " << Rcut << " ";
    if(d_Event == ENERGY) os << " --ecut " << Ecut << " ";
    if(d_Event == Y2) os << " --y2cut " << Ecut << " ";
    os << " --adjustcm " ;
    if(AdjustCM) os << " on ";
    else         os << " off ";
}

void MDEvents::initialize(SimpleParticle* p, PotentialBase* pb, const string& name, bool addbc) {

  PtclRef = p;
  PotRef = pb;
  RefPosFile = name + ".ref";

  if(RefPos.size() < PtclRef->getLocalNum()) {
    LocalNum = PtclRef->getLocalNum();
    RefPos.create(LocalNum);
    DeltaR.create(LocalNum);
    RefEne.create(LocalNum);
    Ene.create(LocalNum);
    AccumBC.create(LocalNum); 
    // initialization
    RefEne = 0.0e0;
    Ene = 0.0e0;
  }
  ofstream fxmol(RefPosFile.c_str());
  fxmol.close();
}

bool MDEvents::detectEvents(MDEvents::ParticlePos_t& newPos,MDEvents::Scalar_t& drmax) {

  double tmp = 0.0e0;
  drmax = 0.0e0;
  bool selected = false;

  if(d_Event == DISPLACEMENT) {

    SingleParticlePos_t newcm = 0.0e0,dr, accumbc = 0.0e0;
    int i=0;

    if(AdjustCM) {
      for(i=0; i<LocalNum; i++) {
        newcm += newPos[i];
        DeltaR[i] = newPos[i]-RefPos[i];
        dr = DeltaR[i];
        PtclRef->applyBoxConds(dr); 
        // return the change, all in the unit of supercell
        accumbc += (dr - DeltaR[i]);
      }        
      double dninv = 1.0e0/static_cast<double>(LocalNum);
      newcm = dninv*(newcm+accumbc) - RefCM;

      i = 0;
      while(i < LocalNum && !selected) {
        dr = DeltaR[i] - newcm;
        PtclRef->applyBoxConds(dr);
        tmp = PtclRef->Property.Lattice.Dot(dr,dr);
        drmax = (tmp>drmax)? tmp: drmax;
        if(drmax > RcutSq) selected = true;
        i++;
      }
    } else {
      i=0;
      while(i < LocalNum && !selected) {
        dr = newPos[i]-RefPos[i];
        PtclRef->applyBoxConds(dr); 
        tmp = PtclRef->Property.Lattice.Dot(dr,dr);
        drmax = (tmp>drmax)? tmp: drmax;
        if(drmax > RcutSq) selected = true;
        i++;
      }
    }

    if(drmax > RcutSq) selected = true;
  }

  if(d_Event == ENERGY) {
    Ene = 0.0e0;
    //double etmp = PotRef->evalEnergy(*PtclRef, newPos, Ene);
    //Ene = PotRef->ELocal;
    int i = 0;
    do {
      //drmax = fabs(RefEne[i] - Ene[i]);
      drmax = fabs(RefEne[i] - PotRef->ELocal[i]);
      if(drmax > Ecut) selected = true;
      i++;
    } while(i < LocalNum && !selected);
  }

  if(d_Event == Y2) {
    PotRef->getPairlist(*PtclRef, newPos);  
    double etmp = BondOrderY2::evaluate(*(PotRef->getPairRef()),Ene);


    int i = 0;
    do {
      drmax = fabs(RefEne[i] - Ene[i]);
      if(drmax > Ecut) selected = true;
      i++;
    } while(i < LocalNum && !selected);

  }   
  return selected;
}
//////////////////////////////////////////////////////////////////////////////
void MDEvents::setRefConfig(MDEvents::ParticlePos_t& newPos, int ntrans){

  // expect positions in the supercell unit
  Ene = RefEne; 
  RefPos = newPos;
  RefEne = PotRef->ELocal;

  if(ntrans == 0)  AccumBC = 0.0e0;
  else {
    PtclRef->apply_bconds(newPos);
    AccumBC += (RefPos - newPos);
    RefPos = newPos;
  }  

  if(AdjustCM) {
    RefCM = 0.0e0;
    for(int i=0; i<LocalNum; i++) RefCM += RefPos[i];
    double dninv = 1.0e0/static_cast<double>(LocalNum);
    RefCM *= dninv;
  }
  if(d_Event == Y2) {
    PotRef->getPairlist(*PtclRef, RefPos);  
    double etmp = BondOrderY2::evaluate(*(PotRef->getPairRef()),RefEne);
  }
//   else {
//    PtclRef->toCart(RefPos,newPos);
//    double etmp = PotRef->evalEnergy(*PtclRef, newPos, RefEne);
//  }
}

void MDEvents::setRefConfig(MDEvents::ParticlePos_t& newPos, 
                            MDEvents::ParticlePos_t& delbc,int ntrans){

  Ene = RefEne; // save Reference Energy
  RefPos = newPos; 
  RefEne = PotRef->ELocal;

  //LocalHMDParticle* hmdptcls = dynamic_cast<LocalHMDParticle*>(PtclRef);
  //hmdptcls->Boosted = false;
  //int nboost=0;
  //if(hmdptcls) {
  //  cout << "boosted atoms " ;
  //  for(int i=0; i<hmdptcls->getLocalNum(); i++) {
  //    if(PotRef->Coord[i] < 3.75 || PotRef->Coord[i] > 4.25) {
  //      cout << i << " ";
  //     hmdptcls->Boosted[i] = true;
  //      nboost++;
  //    }
  //  }
  //cout << endl;

  if(ntrans == 0)  AccumBC = 0.0e0;
  else             AccumBC += delbc;

  RefCM = 0.0e0;
  for(int i=0; i<LocalNum; i++) RefCM += RefPos[i];
  double dninv = 1.0e0/static_cast<double>(LocalNum);
  RefCM *= dninv;

  if(d_Event == Y2) {
    PotRef->getPairlist(*PtclRef, RefPos);  
    double etmp = BondOrderY2::evaluate(*(PotRef->getPairRef()),RefEne);
  }
// else {
//    PtclRef->toCart(RefPos,newPos);
//    double etmp = PotRef->evalEnergy(*PtclRef, newPos, RefEne);
//  }
}

void MDEvents::saveRefConfig(int frame){

  if(frame == 0) Ene = RefEne;
  double etot = 0.0e0;
  for(int i=0; i<RefPos.size(); i++) etot += RefEne[i];

  ofstream fxmol(RefPosFile.c_str(),ios::app);

  fxmol.setf(ios::fixed, ios::floatfield);
  fxmol << RefPos.size() << endl;
  fxmol << "TRANS = " << frame << " Epot = " << etot << endl;
  for(int i=0; i<RefPos.size(); i++) {
    // very temporary
    if(fabs( RefEne[i] - Ene[i] ) > Ecut) fxmol << "Ga ";
    else                                  fxmol << "Si "; 
    SingleParticlePos_t rtmp = PtclRef->Property.Lattice.toCart(RefPos[i]);
    fxmol << setw(12) << setprecision(4) << rtmp[0]
          << setw(12) << setprecision(4) << rtmp[1]
          << setw(12) << setprecision(4) << rtmp[2]
          << setw(12) << setprecision(4) << RefEne[i]
          << setw(12) << setprecision(4) << AccumBC[i][0]
          << setw(12) << setprecision(4) << AccumBC[i][1]
          << setw(12) << setprecision(4) << AccumBC[i][2]
          << endl;
  }
  fxmol.close();
}



void MDEvents::saveRefConfig(MDEvents::Scalar_t tmd){

  if(tmd < 1.0e-1) saveRefConfig(0);
  else             saveRefConfig(1);
  Counter++;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * POOMA_VERSION_ID: $Id: MDEvents.cpp 151 2004-11-02 23:51:14Z jnkim $ 
***************************************************************************/
