//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "ParticleIO/HDFParticleIO.h"
#include "ParticleBase/ParticleUtility.h"
#include "Utilities/OhmmsInfo.h"
#include "Utilities/SimpleParser.h"
using namespace OHMMS;

#if defined(HAVE_LIBHDF5)
#include "ParticleIO/HDFParticleAttrib.h"

struct H5PtclAttributeType {
  int inunit;
  char type_str[16];
  char name_str[16];
  H5PtclAttributeType() {}
  H5PtclAttributeType(int u, const string& tname, const string& oname) {
    inunit = u;
    strncpy(type_str, tname.c_str(), tname.size());
    type_str[tname.size()]='\0';
    strncpy(name_str, oname.c_str(), oname.size());
    name_str[oname.size()]='\0';
  }
};

bool HDFParticleIO::read(hid_t agroup) {

  typedef Particle_t::ParticleIndex_t ParticleIndex_t;
  typedef Particle_t::ParticleScalar_t ParticleScalar_t;
  typedef Particle_t::ParticlePos_t ParticlePos_t;

  //process group attributes
  vector<H5PtclAttributeType> alist;
  int natom = 0;
  hid_t dataset = H5Dopen(agroup, "size"), dataspace;
  hid_t ret = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &natom);
  H5Dclose(dataset);

  //create a temporary char array that contains the list of species
  hsize_t ns;
  dataset = H5Dopen(agroup, "speciesArray");
  dataspace = H5Dget_space(dataset);
  H5Sget_simple_extent_ndims(dataspace);
  H5Sget_simple_extent_dims(dataspace, &ns, NULL);
  char *check = new char[ns];  
  ret = H5Dread(dataset, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, check);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  vector<string> slist;
  parsewords(check,slist);
  delete [] check;  
  SpeciesSet& species = ref_.getSpeciesSet();
  for(int i=0; i<slist.size(); i++) {
    species.getSpeciesID(slist[i].c_str());
  }
  
  hid_t atype = H5Tcopy (H5T_C_S1);
  ret = H5Tset_size (atype, 16);
  hid_t s1_tid = H5Tcreate (H5T_COMPOUND, sizeof(H5PtclAttributeType));
  ret = H5Tinsert(s1_tid, "condition", HOFFSET(H5PtclAttributeType, inunit), H5T_NATIVE_INT);
  ret = H5Tinsert(s1_tid, "datatype", HOFFSET(H5PtclAttributeType, type_str), atype);
  ret = H5Tinsert(s1_tid, "name", HOFFSET(H5PtclAttributeType, name_str), atype);
  
  dataset = H5Dopen(agroup, "list");
  if(dataset>=0) {
    dataspace = H5Dget_space(dataset);
    hsize_t dims_out;
    H5Sget_simple_extent_ndims(dataspace);
    H5Sget_simple_extent_dims(dataspace, &dims_out, NULL);
    alist.resize(dims_out);
    ret = H5Dread(dataset, s1_tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, &alist[0]);
  }
  
  H5Tclose(atype);
  H5Sclose(dataspace);
  H5Dclose(dataset);
  
  if(natom <0) {
    ERRORMSG("Invalid number of particles")
    H5Gclose(agroup);
    return false;
  }

  ref_.create(natom);
  
  for(int ia=0; ia<alist.size(); ia++) {
    int utype =alist[ia].inunit;
    string otype(alist[ia].type_str);
    string oname(alist[ia].name_str);
    if(otype == scalartype_tag) { 
      ParticleScalar_t* stmp = 0;
      if(oname == energy_tag) 
        stmp = &(ref_.E);
      else 
        stmp = ref_.getScalarAttrib(oname);
      if(stmp->size()<ref_.getLocalNum()) stmp->resize(ref_.getLocalNum());
      stmp->setUnit(utype);
      HDFAttribIO<ParticleScalar_t> hdf_s(*stmp);
      hdf_s.read(agroup,oname.c_str());
    } else if(otype == postype_tag) {
      if(oname == displacement_tag) continue;
      ParticlePos_t* ptmp =0;
      if(oname == force_tag) 
        ptmp = &(ref_.F);
      else if(oname == velocity_tag) 
        ptmp = &(ref_.V);
      else 
        ptmp = ref_.getVectorAttrib(oname);
      if(ptmp->size()<ref_.getLocalNum()) ptmp->resize(ref_.getLocalNum());
      ptmp->setUnit(utype);
      HDFAttribIO<ParticlePos_t> hdf_p(*ptmp);
      hdf_p.read(agroup,oname.c_str());
    } else if(otype == indextype_tag) { 
      ParticleIndex_t& itmp = *(ref_.getIndexAttrib(oname.c_str()));
      itmp.setUnit(utype);
      HDFAttribIO<ParticleIndex_t> hdf_i(itmp);
      hdf_i.read(agroup,oname.c_str());
    }
  }
  return true;
}
bool HDFParticleIO::write(hid_t agroup) {

  typedef Particle_t::size_type size_type;
  typedef Particle_t::ParticleIndex_t ParticleIndex_t;
  typedef Particle_t::ParticleScalar_t ParticleScalar_t;
  typedef Particle_t::ParticlePos_t ParticlePos_t;
  hid_t dataspace, dataset,ret;
  hsize_t dim = 1;
  int natt=0;
  //Particle_t::PAListIterator it = ref_.first_attrib();
  vector<H5PtclAttributeType> alist;
  vector<OhmmsObject*>::iterator it(ref_.RestartList.begin());
  vector<OhmmsObject*>::iterator it_end(ref_.RestartList.end());
  while(it != it_end) {
  //while(it != ref_.last_attrib()) {
  //  OhmmsObject* ooref= (*it).second;
    OhmmsObject* ooref= (*it);
    size_type t_id = ref_.getAttribType(ooref->typeName());
    size_type o_id = ooref->id();
    int utype = 0;
    if(t_id == PA_IndexType) {
      ParticleIndex_t& itmp = *(ref_.getIndexAttrib(o_id));
      utype = itmp.getUnit();
      HDFAttribIO<ParticleIndex_t> hdf_i(itmp);
      hdf_i.write(agroup,ooref->objName().c_str());
    } else if(t_id == PA_ScalarType) {
      ParticleScalar_t& stmp = *(ref_.getScalarAttrib(o_id));
      utype = stmp.getUnit();
      HDFAttribIO<ParticleScalar_t> hdf_s(stmp);
      hdf_s.write(agroup,ooref->objName().c_str());
    } else if(t_id == PA_PositionType) {
      ParticlePos_t& ptmp = *(ref_.getVectorAttrib(o_id));
      utype = ptmp.getUnit();
      HDFAttribIO<ParticlePos_t> hdf_p(ptmp);
      hdf_p.write(agroup,ooref->objName().c_str());
    } else if(t_id == PA_TensorType) {
      WARNMSG("Not working yet")
    }
    alist.push_back(H5PtclAttributeType(utype,ooref->typeName(), ooref->objName()));;
    ++it;++natt;
  }

  //Add the list of particle attributes
  hid_t atype = H5Tcopy (H5T_C_S1);
  ret = H5Tset_size (atype, 16);
  hid_t s1_tid = H5Tcreate (H5T_COMPOUND, sizeof(H5PtclAttributeType));
  ret = H5Tinsert(s1_tid, "condition", HOFFSET(H5PtclAttributeType, inunit), H5T_NATIVE_INT);
  ret = H5Tinsert(s1_tid, "datatype", HOFFSET(H5PtclAttributeType, type_str), atype);
  ret = H5Tinsert(s1_tid, "name", HOFFSET(H5PtclAttributeType, name_str), atype);
  dim = natt;
  dataspace  = H5Screate_simple(1, &dim, NULL);
  dataset = H5Dcreate(agroup, "list", s1_tid, dataspace, H5P_DEFAULT);
  ret = H5Dwrite(dataset, s1_tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, &(alist[0]));
  ret = H5Tclose(atype);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  size_type natom = ref_.getLocalNum();
  dim = 1;
  dataspace = H5Screate_simple(1, &dim, NULL);
  dataset = H5Dcreate(agroup, "size", H5T_NATIVE_INT, dataspace, H5P_DEFAULT);
  ret =  H5Dwrite(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT,&natom);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  const SpeciesSet& species = ref_.getSpeciesSet();
  //make a char array which contains the species names separated by space
  std::stringstream s;
  s << species.Name[0];
  for(int i=1; i<species.getTotalNum(); i++) s << " " << species.Name[i];
  s << '\0';
  dim = s.str().size();
  dataspace = H5Screate_simple(1, &dim, NULL);
  dataset = H5Dcreate(agroup, "speciesArray", H5T_NATIVE_CHAR, dataspace, H5P_DEFAULT);
  ret =  H5Dwrite(dataset, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT,s.str().c_str());
  H5Dclose(dataset);
  H5Sclose(dataspace);

  return true;
}
#else
bool HDFParticleIO::read(hid_t agroup) {
  return false;
}
bool HDFParticleIO::write(hid_t agroup) {
  return false;
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: HDFParticleIO.1.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
