//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002,2005- David A. Richie and Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   David A. Richie
//   Department of Physics, Ohio State University
//   174 W 18th Avenue
//   Columbus, OH 43210
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file MCFRecordParticleAttrib.h
 * @brief Declaration of a Record engine for a ParticleAttribute using MCF format
 *
 * Relies on  components from the STORM Scientific Wavelet Package (v3.1.0)
 */
#ifndef OHMMS_DCFPARTICLEATTRIBUTE_H
#define OHMMS_DCFPARTICLEATTRIBUTE_H

#include "Utilities/OhmmsInfo.h"
#include "OhmmsData/OhmmsElementBase.h"
#include "OhmmsData/RecordProperty.h"
#include "Particle/ParticleSet.h"

//BEGIN ADDED BY DR
#ifndef STORM_USEF
#define STORM_USEF
#endif
#ifndef USE_STD
#define USE_STD
#endif
#ifndef STORM_x86_LINUX 
#define STORM_x86_LINUX 
#endif
#include "Rtmra.h"
#include "MCFsbofstream.h"
#ifdef USE_DPMESSAGE
#include "DPmessage.h"
#endif
//END ADDED BY DR

namespace OHMMS {

  /** Base class for a recorder of a particle attribute using MCF  library
   *
   * Template parameters
   * - T1 type of the physical data
   * - T2 type of the compressed data
   *
   * Ready for parallelization with FirstIndex and LastIndex.
   * NumElements is the data per an attribute. 
   */
  template<class T1 = double, class T2 = double>
  struct MCFRecordAttrib {

    typedef T2 value_type;

    MCFRecordAttrib():Use_descr(false),FirstIndex(0),LastIndex(0), NumElements(1),
                      mra_mmax(12), dcf_thresh(1), m_mag(1.0), m_mcfofs(0){ }

    explicit MCFRecordAttrib(int first, int last, int n=1):
      Use_descr(false),FirstIndex(first),LastIndex(last), NumElements(n),
      mra_mmax(12),dcf_thresh(1), m_mag(1.0), m_mcfofs(0) { }

    virtual ~MCFRecordAttrib(){if(m_mcfofs) delete m_mcfofs;}

    void resetRecord(const string& fileroot) {
      dcf_file = fileroot;
      dcf_file.append("\0");
      m_rtmra.remalloc(NumElements*(LastIndex-FirstIndex),mra_mmax+1);
      if(!m_mcfofs) {
        m_mcfofs = new MCFsbofstream< Rtmra<value_type,array> >(dcf_file);
        //Ohmms is 1st app to use MCF, congratulations
        m_mcfofs->set_author(1);
        m_mcfofs->set_mpart(0,mra_mmax,0,mra_mmax,0,mra_mmax);
        m_mcfofs->set_bplen(2,sizeof(value_type),2,2);
        m_mcfofs->threshold(dcf_thresh);
        if (Use_descr) m_mcfofs->report();
        char tmpbuf[128];
        sprintf(tmpbuf,"NumOfData %d NumOfElement %d First %d Last %d Magnification %f",
            LastIndex-FirstIndex, NumElements, FirstIndex,LastIndex,m_mag);
        m_mcfofs->logmessage(tmpbuf);
      }
    }

    void finalizeRecord() {
      m_mcfofs->final(m_rtmra); 
      delete m_mcfofs;
      m_mcfofs = 0;
    }

    bool putRecord(xmlNodePtr cur) {
      const xmlChar* mptr = xmlGetProp(cur,(const xmlChar*)"mmax");
      const xmlChar* tptr = xmlGetProp(cur,(const xmlChar*)"thresh");
      const xmlChar* gptr = xmlGetProp(cur,(const xmlChar*)"factor");
      if(mptr)  mra_mmax=atoi((const char*)mptr);
      if(tptr)  dcf_thresh=static_cast<T2>(atof((const char*)tptr));
      if(gptr)  m_mag=static_cast<T1>(atof((const char*)gptr));
      LOGMSG("\t mmax="<<mra_mmax << " threshhold=" << dcf_thresh << " magnification=" << m_mag)
      return true;
    }

    bool Use_descr;
    int FirstIndex;
    int LastIndex;
    int NumElements;
    int mra_mmax;
    T1 m_mag;
    value_type dcf_thresh;
    string dcf_file;
    Rtmra<value_type,array> m_rtmra;
    MCFsbofstream< Rtmra<value_type,array> > *m_mcfofs;
  };

  class MCFRecordScalarAttrib: public RecordProperty, private MCFRecordAttrib<OHMMS_PRECISION,double>
  {

    public:

      typedef ParticleSet::ParticleScalar_t InData_t;

      MCFRecordScalarAttrib(const InData_t& a);
      MCFRecordScalarAttrib(const InData_t& a, int first, int last);
      ~MCFRecordScalarAttrib();

      void reset(const char* fileroot);

      void report(int iter);

      void finalize();

      bool put(xmlNodePtr cur);
    private:
      const InData_t& ref_;
  };

  class MCFRecordPosAttrib: public RecordProperty, private MCFRecordAttrib<OHMMS_PRECISION,double>
  {

    public:
      typedef ParticleSet::ParticlePos_t InData_t;

      MCFRecordPosAttrib(const InData_t& a);
      MCFRecordPosAttrib(const InData_t& a, int first, int last);
      ~MCFRecordPosAttrib();

      void reset(const char* fileroot);

      void report(int iter);

      void finalize();

      bool put(xmlNodePtr cur);
    private:
      const InData_t& ref_;
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 367 $   $Date: 2005-07-22 20:42:51 -0500 (Fri, 22 Jul 2005) $
 * $Id: MCFRecordParticleAttrib.h 367 2005-07-23 01:42:51Z jnkim $ 
 ***************************************************************************/
