//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 David A. Richie and Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   David A. Richie
//   Department of Physics, Ohio State University
//   174 W 18th Avenue
//   Columbus, OH 43210
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// class DCFParticleIO
//
// Records using Xmol & DCF format in Progress mode.  Relies on 
// components from the STORM Scientific Wavelet Package (v3.1.0)
// which is (c) Copyright 1995-2002 David A. Richie.
//
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <bitset>
using std::bitset;

#include "ParticleIO/DCFParticleIO.h"
#include "ParticleIO/ParticleInputTemplate.h"
#include "Utilities/OhmmsInfo.h"

void DCFParticleIO::record(int iter) {
  if(ioMode == PROGRESS) {
    DEBUGMSG("Performing DCFParticleIO::record()")
    if(!initialized) { setup(); initialized = true;}
    recordDCF(iter); 
  } else {
    ERRORMSG("record for INPUT/OUPUT mode is not allowed.\n")
  }
}

void DCFParticleIO::setup() {

  attribMode.reset();

  stride = 1;

  bool qdescr = false;

  posrtmra_mmax = 8;	// 256 step superblocks, rather small
  ertmra_mmax = 8;	// 256 step superblocks, rather small
  frtmra_mmax = 8;	// 256 step superblocks, rather small
  posdcfofs_thresh = 1;	// default is overcompression, almost nothing written
  edcfofs_thresh = 1;	// default is overcompression, almost nothing written
  fdcfofs_thresh = 1;	// default is overcompression, almost nothing written

  char tmpbuf[80];

  int i=0; 
  while(i<options.size()) {
    if(options[i] == "--E") attribMode.set(E);
    else if(options[i] == "--Q") attribMode.set(Q);
    else if(options[i] == "--F") attribMode.set(F);
    else if(options[i] == "--V") attribMode.set(V);
    else if(options[i] == "--stride") stride = atoi(options[++i].c_str());

    else if(options[i] == "--mmax") {
       posrtmra_mmax = atoi(options[++i].c_str());
       ertmra_mmax = posrtmra_mmax;
       frtmra_mmax = posrtmra_mmax;
    }
    else if(options[i] == "--POSmmax") {
       posrtmra_mmax = atoi(options[++i].c_str());
    }
    else if(options[i] == "--Emmax") {
       ertmra_mmax = atoi(options[++i].c_str());
    }
    else if(options[i] == "--Fmmax") {
       frtmra_mmax = atoi(options[++i].c_str());
    }
    else if(options[i] == "--thresh") {
       posdcfofs_thresh = atof(options[++i].c_str());
       edcfofs_thresh = posdcfofs_thresh;
       fdcfofs_thresh = posdcfofs_thresh;
    }
    else if(options[i] == "--POSthresh") {
       posdcfofs_thresh = atof(options[++i].c_str());
    }
    else if(options[i] == "--Ethresh") {
       edcfofs_thresh = atof(options[++i].c_str());
    }
    else if(options[i] == "--Fthresh") {
       fdcfofs_thresh = atof(options[++i].c_str());
    }
    else if(options[i] == "--descr") { qdescr = true; }

    i++;
  }

// Setup Rtmra variables and associated DCFsbofstream:
  posrtmra.remalloc(OHMMS_DIM*MyPtcl.getLocalNum(),posrtmra_mmax+1);
  posdcfofs.author(1); //Ohmms is 1st app to use DCF, congratulations

  sprintf(posdcfofs_descript,"%s.pos\0",FileRoot.c_str());
  cout << posdcfofs_descript << endl;

  posdcfofs.filenamebase(posdcfofs_descript);
  posdcfofs.mmax(posrtmra_mmax);
  posdcfofs.bplen_sizeb(2);	// 2 bytes used to encode size of mpacket
  posdcfofs.bplen_n(2);		// 2 bytes used to encode n-tag
  posdcfofs._enc.quantcoef(posdcfofs_thresh);
  posdcfofs._enc.amax(OHMMS_DIM*MyPtcl.getLocalNum()-1);
  posdcfofs._enc.bplen_a(2);	// 2 bytes used to encode array-element index
  if (qdescr) posdcfofs.descr();
  if (attribMode.any()) {
    if((attribMode[E] && MyPtcl.hasAttrib(ENERGY_TAG))||
       (attribMode[Q] && MyPtcl.hasAttrib(VAL_CHARGE_TAG))) {
      ertmra.remalloc(MyPtcl.getLocalNum(),ertmra_mmax+1);
      edcfofs.author(1); //Ohmms is 1st app to use DCF, congratulations
      sprintf(edcfofs_descript,"%s.e\0",FileRoot.c_str());
      edcfofs.filenamebase(edcfofs_descript);
      edcfofs.mmax(ertmra_mmax);
      edcfofs.bplen_sizeb(2);	// 2 bytes used to encode size of mpacket
      edcfofs.bplen_n(2);	// 2 bytes used to encode n-tag
      edcfofs._enc.quantcoef(edcfofs_thresh);
      edcfofs._enc.amax(MyPtcl.getLocalNum()-1);
      edcfofs._enc.bplen_a(2);// 2 bytes used to encode array-element index
      if (qdescr) edcfofs.descr();
    }
    if((attribMode[F] && MyPtcl.hasAttrib(FORCE_TAG))
       ||(attribMode[V] && MyPtcl.hasAttrib(VELOCITY_TAG))) {
      frtmra.remalloc(3*MyPtcl.getLocalNum(),frtmra_mmax+1);
      fdcfofs.author(1); //Ohmms is 1st app to use DCF, congratulations
      sprintf(fdcfofs_descript,"%s.f\0",FileRoot.c_str());
      fdcfofs.filenamebase(fdcfofs_descript);
      fdcfofs.mmax(frtmra_mmax);
      fdcfofs.bplen_sizeb(2);	// 2 bytes used to encode size of mpacket
      fdcfofs.bplen_n(2);	// 2 bytes used to encode n-tag
      fdcfofs._enc.quantcoef(fdcfofs_thresh);
      fdcfofs._enc.amax(3*MyPtcl.getLocalNum()-1);
      fdcfofs._enc.bplen_a(2);// 2 bytes used to encode array-element index
      if (qdescr) fdcfofs.descr();
    }
  }

  // cartesian coordinate to follow
  if(MyPtcl.hasAttrib(TRAJECTORY_TAG)) {
    p_ref = MyPtcl.getVectorAttrib(TRAJECTORY_TAG);
  } else {
    p_ref = &(MyPtcl.R);
  }

  if(attribMode.any()) {
    if(attribMode[E]) {
      if(MyPtcl.hasAttrib(ENERGY_TAG)) 
        e_ref = MyPtcl.getScalarAttrib(ENERGY_TAG);
      else {
	ERRORMSG("Energy is not found among the Particle Attributes.\n");
	attribMode.flip(E);
      }
    }

    if(attribMode[Q]) {
      if(MyPtcl.hasAttrib(VAL_CHARGE_TAG)) 
        e_ref = MyPtcl.getScalarAttrib(VAL_CHARGE_TAG);
      else 
      {
        ERRORMSG("Charge is not found among the Particle Attributes.\n");
        attribMode.flip(Q);
      }
    }

    if(attribMode[F]) {
      if(MyPtcl.hasAttrib(FORCE_TAG)) 
         f_ref = MyPtcl.getVectorAttrib(FORCE_TAG);
      else 
	{
	  ERRORMSG("Force is not found among the Particle Attributes.\n");
	  attribMode.flip(F);// cannot perform
	}
    }

    if(attribMode[V]) {
      if(MyPtcl.hasAttrib(VELOCITY_TAG)) 
        f_ref = MyPtcl.getVectorAttrib(VELOCITY_TAG);
      else
	{
	  ERRORMSG("Velocity is not found among the Particle Attributes.\n");
	  attribMode.flip(V);// cannot perform
	}
    }
  }

  //calculate the integer value
  DCFMode = (attribMode[E] || attribMode[Q])
    + (attribMode[F] || attribMode[V])*2;

  if(Species.empty()) {
    SpeciesBase* species = SpeciesCollection::getSpecies();
    int nspecies = species->getTotalNum();
    Species.reserve(nspecies);
    for(int i=0; i<nspecies; i++) {
      Species.push_back(species->Name[i]);
    }
  }

//write information aout system to DCF log file:
  sprintf(tmpbuf,"NATOM %d",MyPtcl.getLocalNum());
  posdcfofs.logmessage(tmpbuf);
  for(int i=0;i<MyPtcl.getLocalNum();i++) { 
     sprintf(tmpbuf,"ATOM %d %s",i,Species[MyPtcl.IonID[i]]);
     posdcfofs.logmessage(tmpbuf);
  }
  if (attribMode.any()) {
    if((attribMode[E] && MyPtcl.hasAttrib(ENERGY_TAG))||
      (attribMode[Q] && MyPtcl.hasAttrib(VAL_CHARGE_TAG))) {
       sprintf(tmpbuf,"NATOM %d",MyPtcl.getLocalNum());
       edcfofs.logmessage(tmpbuf);
       for(int i=0;i<MyPtcl.getLocalNum();i++) { 
          sprintf(tmpbuf,"ATOM %d %s",i,Species[MyPtcl.IonID[i]]);
          edcfofs.logmessage(tmpbuf);
       }
    }
    if((attribMode[F] && MyPtcl.hasAttrib(FORCE_TAG))
      ||(attribMode[V] && MyPtcl.hasAttrib(VELOCITY_TAG))) {
       sprintf(tmpbuf,"NATOM %d",MyPtcl.getLocalNum());
       fdcfofs.logmessage(tmpbuf);
       for(int i=0;i<MyPtcl.getLocalNum();i++) { 
          sprintf(tmpbuf,"ATOM %d %s",i,Species[MyPtcl.IonID[i]]);
          fdcfofs.logmessage(tmpbuf);
       }
    }
  }

}


void DCFParticleIO::recordDCF(int iter) {

  DEBUGMSG("DCFParticleOut::init() is completed.\n");
  int nat = MyPtcl.getLocalNum();

  ParticlePos_t& bc= *(MyPtcl.getVectorAttrib("sumbc"));

  if(p_ref->InUnit) {
    int ii=0;
    for(int i=0;i<nat;i++) { 
      SingleParticlePos_t pos = MyPtcl.Lattice.toCart(p_ref->operator[](i));
      posrtmra[ii++]=pos[0]+bc[i][0];
      posrtmra[ii++]=pos[1]+bc[i][1];
      posrtmra[ii++]=pos[2]+bc[i][2];
    }
  } else {
    int ii=0;
    for(int i=0;i<nat;i++) { 
      posrtmra[ii++]=(*p_ref)[i][0]+bc[i][0];
      posrtmra[ii++]=(*p_ref)[i][1]+bc[i][1];
      posrtmra[ii++]=(*p_ref)[i][2]+bc[i][2];
    }
  }
  posrtmra++;
  posdcfofs << posrtmra;

  if(attribMode[E] || attribMode[Q] ) {
    for(int i=0;i<nat;i++) { 
      ertmra[i]=(*e_ref)[i];
    }
    ertmra++;
    edcfofs << ertmra;
  }

  if(attribMode[F] || attribMode[V]) {
    int ii=0;
    for(int i=0;i<nat;i++) { 
      frtmra[ii++]=(*f_ref)[i][0];
      frtmra[ii++]=(*f_ref)[i][1];
      frtmra[ii++]=(*f_ref)[i][2];
    }
    frtmra++;
    fdcfofs << frtmra;
  }
}


void DCFParticleIO::finalize() {

  int nat = MyPtcl.getLocalNum();

  posdcfofs.final(posrtmra);

  if(attribMode[E] || attribMode[Q] ) { edcfofs.final(ertmra); }

  if(attribMode[F] || attribMode[V]) { fdcfofs.final(frtmra); }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: DCFParticleIO.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

