//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

// #ifdef __GNUC__
// #include "Utilities/OhmmsBitset.h"
// #else
#include <bitset>
using std::bitset;
//#endif
#include <iostream>
#include <fstream>
using namespace std;
#include "ParticleIO/HDFParticleIO.h"
#include "Utilities/OhmmsInfo.h"
#include "ParticleBase/ParticleUtility.h"
#include "ParticleIO/ParticleInputTemplate.h"
#include "ParticleIO/HDFParticleAttrib.h"
using namespace OHMMS;


HDFParticleIO::HDFParticleIO(ThisParticle_t& p, int i, const char* aname):
ParticleIOBase<Scalar_t,OHMMS_DIM>(p,i,aname) { 
  HDFMode.reset();
}

void HDFParticleIO::record(int iter) {

  if(ioMode == PROGRESS) {

    DEBUGMSG("Performing HDFParticleIO::record()")

    if(!initialized) { 
      setup(); 
      initialized = true;
    }

    if(rec_lattice) {
      ofstream fmaster(metafile, ios::app);
      fmaster << "<frame id = " << iter << " >" << endl;
      MyPtcl.Lattice.print(fmaster,0);
      fmaster << "</frame>" << endl;
      fmaster.close();
    }

    //For each file, DATASET is connected to the current iteration 
    char iframe[10];
    sprintf(iframe,"%08d",iter);
    for(int item=0; item<pos_list.size(); item++) {
      hid_t h1 = H5Fopen(pos_file[item].c_str(), H5F_ACC_RDWR, H5P_DEFAULT);
      HDFAttribIO<ParticlePos_t> hdf_r(*pos_list[item]);
      hdf_r.write(h1,iframe);
      H5Fclose(h1);
    }

    for(int item=0; item<val_list.size(); item++) {
      hid_t h1 = H5Fopen(val_file[item].c_str(), H5F_ACC_RDWR, H5P_DEFAULT);
      HDFAttribIO<ParticleScalar_t> hdf_s(*val_list[item]);
      hdf_s.write(h1,iframe);
      H5Fclose(h1);
    }

    if(!initialized) {  }
  } else {
    ERRORMSG("record for INPUT/OUPUT mode is not allowed.\n")
  }
}

void HDFParticleIO::setup() {
  char fname[256];

  sprintf(metafile, "%s.rec.meta", FileRoot.c_str());
  ofstream fmaster(metafile);

  rec_lattice = false;
  ////////////////////////////////////////////////
  // always add trejectory to record
  // A propagator may a ParticlePos_t with TRAJECTORY_TAG
  // is created and added to the ParticleAttrib list.
  // The reason is to avoid complications due to boundary conditions.
  // If not added to the list, ParticleBase::R is written
  ////////////////////////////////////////////////
  // format is <attrib> typename objectname hdf-datafile </attrib>
  ////////////////////////////////////////////////
  fmaster << "<mdrecord>" << endl;
  sprintf(fname,"%s.R.h5",FileRoot.c_str());
  if(MyPtcl.hasAttrib(TRAJECTORY_TAG)) {
    pos_list.push_back(MyPtcl.getVectorAttrib(TRAJECTORY_TAG));
    fmaster << "<attrib> " << POSSTR << " " << TRAJECTORY_TAG 
	    << " " << fname << " </attrib>"
	    << endl;
  } else {
    pos_list.push_back(&(MyPtcl.R));
    fmaster << "<attrib> " << POSSTR << " " << POSITION_TAG 
	    << " " << fname << " </attrib>"
	    << endl;
  }

  pos_file.push_back(fname);
  hid_t	h1 = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  H5Fclose(h1);

  ////////////////////////////////////////////////
  //add other optional data to record
  ////////////////////////////////////////////////
  int i=0;
  while(i<options.size()) {
    if(options[i] == "--rec_lattice") {
      rec_lattice = true;
    } else if(options[i] == "--E") {
      if(!HDFMode[0]) {
	if(MyPtcl.hasAttrib(ENERGY_TAG)) { 
	  val_list.push_back(MyPtcl.getScalarAttrib(ENERGY_TAG));
	  sprintf(fname,"%s.E.h5",FileRoot.c_str());
	  val_file.push_back(fname);        
	  h1 = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	  H5Fclose(h1);
	  fmaster << "<attrib> " << SCALARSTR  << " " << ENERGY_TAG 
		  << " " << fname << " </attrib>"
		  << endl;
	} else {
	  ERRORMSG("Energy is not found among the Particle Attributes.\n");
	}
	HDFMode.set(0);
      }
    } else if(options[i] == "--Q") {
      if(!HDFMode[1]) {
	if(MyPtcl.hasAttrib(VAL_CHARGE_TAG)) { 
	  val_list.push_back(MyPtcl.getScalarAttrib(VAL_CHARGE_TAG));
	  sprintf(fname,"%s.Q.h5",FileRoot.c_str());
	  val_file.push_back(fname);        
	  h1 = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	  H5Fclose(h1);
	  fmaster << "<attrib> " << SCALARSTR  << " " << VAL_CHARGE_TAG 
		  << " " << fname << " </attrib>"
		  << endl;
	} else {
	  ERRORMSG("Charge is not found among the Particle Attributes.\n");
	}
	HDFMode.set(1);
      }
    } else if(options[i] == "--F") {
      if(!HDFMode[2]) {
	if(MyPtcl.hasAttrib(FORCE_TAG)) { 
	  pos_list.push_back(MyPtcl.getVectorAttrib(FORCE_TAG));
	  sprintf(fname,"%s.F.h5",FileRoot.c_str());
	  pos_file.push_back(fname);        
	  h1 = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	  H5Fclose(h1);
	  fmaster << "<attrib> " << POSSTR  << " " << FORCE_TAG 
		  << " " << fname << " </attrib>"
		  << endl;
	} else {
	  ERRORMSG("Force is not found among the Particle Attributes.\n");
	}
	HDFMode.set(2);
      }
    } else if(options[i] == "--V") {
      if(!HDFMode[3]) {
	if(MyPtcl.hasAttrib(VELOCITY_TAG)) { 
	  pos_list.push_back(MyPtcl.getVectorAttrib(VELOCITY_TAG));
	  sprintf(fname,"%s.V.h5",FileRoot.c_str());
	  pos_file.push_back(fname);        
	  h1 = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	  H5Fclose(h1);
	  fmaster << "<attrib> " << POSSTR  << " " << VELOCITY_TAG 
		  << " " << fname << " </attrib>"
		  << endl;
	} else {
	  ERRORMSG("Velocity is not found among the Particle Attributes.\n");
	}
	HDFMode.set(3);
      }
    }
    i++;
  }

}


void HDFParticleIO::read(istream& fpos) {

  hid_t dataset, dataspace, ret;
  hid_t hdfFile = H5Fopen(FileRoot.c_str(), H5F_ACC_RDWR, H5P_DEFAULT);

  int natom = 0;
  dataset = H5Dopen(hdfFile, "atnum");
  ret = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &natom);
  H5Dclose(dataset);

  MyPtcl.create(natom);

  vector<string> slist;
  do {
    getwords(slist,fpos);
    if(slist[1] == SCALARSTR) {
      ParticleScalar_t& stmp = *(MyPtcl.getScalarAttrib(slist[2].c_str()));
      HDFAttribIO<ParticleScalar_t> hdf_s(stmp);
      hdf_s.read(hdfFile,slist[2].c_str());
      if(slist[3] == "0") stmp.InUnit = false;
      else                stmp.InUnit = true;
    } else if(slist[1] == POSSTR) {
      ParticlePos_t& ptmp = *(MyPtcl.getVectorAttrib(slist[2].c_str()));
      HDFAttribIO<ParticlePos_t> hdf_p(ptmp);
      hdf_p.read(hdfFile,slist[2].c_str());
      if(slist[3] == "0") 
	ptmp.InUnit = false;
      else 	
	ptmp.InUnit = true;

    } else if(slist[1] == INDEXSTR) {
      ParticleIndex_t& itmp = *(MyPtcl.getIndexAttrib(slist[2].c_str()));
      HDFAttribIO<ParticleIndex_t> hdf_i(itmp);
      hdf_i.read(hdfFile,slist[2].c_str());
      if(slist[3] == "0") itmp.InUnit = false;
      else                itmp.InUnit = true;
    }
  } while(slist[0] != "</ptclset>");


  ///////////////////////////////////////////////////////  
  // apply boundary conditions to the trajectory
  ///////////////////////////////////////////////////////  
  if(MyPtcl.hasAttrib(TRAJECTORY_TAG)) {

    ParticlePos_t& R0 = *(MyPtcl.getVectorAttrib(TRAJECTORY_TAG));
    R0.InUnit = false;
    MyPtcl.R.InUnit = false;

    wrapAroundBox(MyPtcl.Lattice, R0, MyPtcl.R);

    if(MyPtcl.hasAttrib("sumbc")) {// translate exists, accumulate it
      ParticlePos_t& del = *(MyPtcl.getVectorAttrib("sumbc"));
      del += R0 - MyPtcl.R;
    } else { // first time translation is calculated
      ParticlePos_t& del = *(MyPtcl.getVectorAttrib("sumbc"));
      del = R0 - MyPtcl.R;
    }
    R0 = MyPtcl.R;
    MyPtcl.R.InUnit = true;
    convert(MyPtcl.Lattice, R0, MyPtcl.R);
  }

  MyPtcl.R.InUnit = true;

  H5Fclose(hdfFile);
}

void HDFParticleIO::write() {

  if(!initialized) {
    sprintf(metafile, "%s.meta", FileRoot.c_str());
    sprintf(datafile, "%s.h5", FileRoot.c_str());
    initialized = true;
  }

  // writing a meta file
  ofstream ftbmd(metafile); // always overwrite
  ftbmd << "<?xml version=\"1.0\"?>" << endl;
  ftbmd << "<ParticleSet>" << endl;

  ftbmd << "<unitcell>" << endl;
  MyPtcl.Lattice.print(ftbmd,1);
  ftbmd << "</unitcell>" << endl;
  ftbmd << "<ptclset oname=\"" << MyPtcl.objName() << "\">" << endl;
  ftbmd << "\t<format> hdf </format> " << endl;
  ftbmd << "\t<name> " << datafile << " </name>" <<  endl;
  for(int i=0; i<MyPtcl.getNumAttrib(); i++) {
    OhmmsObject& ooref=MyPtcl.getAttrib(i);
     ftbmd << "\t<attrib> " << ooref.typeName() << " " 
	   << ooref.objName() << " ";
     ooref.write_data(ftbmd); 
     ftbmd  << " </attrib> " << endl;
  }
  ftbmd << "</ptclset>" << endl;
  ftbmd << "</ParticleSet>" << endl;
  ftbmd.close();


  // creat HDF file handler
  hid_t hdfFile = 
    H5Fcreate(datafile, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);

  hsize_t dim = 1;
  int natom = MyPtcl.getLocalNum();
  hid_t dataspace  = H5Screate_simple(1, &dim, NULL);
  hid_t dataset =  
  H5Dcreate(hdfFile, "atnum", H5T_NATIVE_INT, dataspace, H5P_DEFAULT);
  hid_t ret = 
      H5Dwrite(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT,&natom);
  H5Dclose(dataset);
  H5Sclose(dataspace);

//   // write IonID and R first
//   // Do not write ID since ID only has internal functions
//   HDFParticleAttrib<int> hdf_ionid(MyPtcl.IonID);
//   hdf_ionid.write(hdfFile,MyPtcl.IonID.objName());

//   // write IonID and R first
//   // Do not write ID since ID only has internal functions
//   HDFParticleAttrib<int> hdf_id(MyPtcl.ID);
//   hdf_id.write(hdfFile,MyPtcl.ID.objName());

//   HDFParticleAttrib<SingleParticlePos_t> hdf_r(MyPtcl.R);
//   hdf_r.write(hdfFile,MyPtcl.R.objName());

  // write all the extra ParticleAttrib's added to the ParticleList
  for(int i=0; i<MyPtcl.getNumAttrib(); i++) {
    OhmmsObject& ooref=MyPtcl.getAttrib(i);
    if(!strcmp(ooref.typeName(), POSSTR)) {
      HDFAttribIO<ParticlePos_t> hdf_p(*(MyPtcl.getVectorAttrib(i)));
      hdf_p.write(hdfFile,ooref.objName());
    } else if(!strcmp(ooref.typeName(), SCALARSTR)) {
      HDFAttribIO<ParticleScalar_t> hdf_s(*(MyPtcl.getScalarAttrib(i)));
      hdf_s.write(hdfFile,ooref.objName());
    }  else if(!strcmp(ooref.typeName(), INDEXSTR)) {
      HDFAttribIO<ParticleIndex_t> hdf_i(*(MyPtcl.getIndexAttrib(i)));
      hdf_i.write(hdfFile,ooref.objName());
    }
  }
  H5Fclose(hdfFile);

}

void HDFParticleIO::finalize() {

  ofstream fmaster(metafile, ios::app);
  fmaster << "</mdrecord>" << endl;
  fmaster.close();
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: HDFParticleIO.old.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

