//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 David A. Richie and Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   David A. Richie
//   Department of Physics, Ohio State University
//   174 W 18th Avenue
//   Columbus, OH 43210
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <stdio.h>
#include <vector>
#include <algorithm>
using namespace std;

#include "ParticleIO/ParticleOutputFactory.h"
#include "ParticleIO/XYZParticleIO.h"
#include "ParticleIO/TbmdParticleIO.old.h"
#ifdef USE_STORM
#include "ParticleIO/DCFParticleIO.h"
#endif
// #ifdef USE_HDF5
// #include "ParticleIO/HDFParticleIO.h"
// #endif
#include "Utilities/OhmmsInfo.h"

int ParticleOutputFactory::save_stride = 100;
int ParticleOutputFactory::rec_stride = 100;

vector<string> ParticleOutputFactory::options;

vector<ParticleIOBase<Scalar_t, OHMMS_DIM>* > 
ParticleOutputFactory::save_engine;

vector<ParticleIOBase<Scalar_t, OHMMS_DIM>* > 
ParticleOutputFactory::rec_engine;


void ParticleOutputFactory::set(int argc, char **argv) {
  for(int i=0; i<argc; i++) options.push_back(argv[i]);
}

void ParticleOutputFactory::set(vector<string>& arglist) {
  options.insert(options.end(), arglist.begin(), arglist.end());
}

/*! \fn int ParticleOuputFactory::add(p, opt)
 *  \param p A particle set
 *  \param opt vector<string> containing options
 *  \return an integer for the requested Output Engine
 *  \brief Unlike, Input, multiple output engines can coexist.
 *  the ID returned by add function should be used with update for
 *  consistent output.
 */

int ParticleOutputFactory::add(Particle_t& P,  const char* fileroot) {

  //assign the default stride to dump a simulation
  //   MDParameters* myParam = GlobalMDParameters::getMDParameters();
  //   save_stride = myParam->SaveStride;


  int id = save_engine.size();
  //Ouput engines come as pairs
  // save_engines -> overwrite 
  // rec_engines -> append

  //default save engine
// #ifdef USE_HDF5
//   save_engine.push_back(new HDFParticleIO(P,ParticleIO_t::OUTPUT,fileroot));
// #else
  save_engine.push_back(new TbmdParticleIO(P,ParticleIO_t::OUTPUT,fileroot));
// #endif
 
  vector<string>::iterator it = options.begin(); 
  while(it != options.end()) {
    if((*it) == "--rec_stride" || (*it) == "--rec") {
      rec_stride = atoi((*(++it)).c_str());
    } else if((*it) == "--save_stride" || (*it) == "--save") {
      save_stride = atoi((*(++it)).c_str());
      //myParam->SaveStride = save_stride ;
    } 
    it++;
  }

  it = find(options.begin(), options.end(), "ptcl_out");
  int omode = 2; // ParticleIO_t::PROGRESS;
  if(it != options.end()) {
    it++;
    if((*it) == "--xyz") {
      rec_engine.push_back(new XYZParticleIO(P,omode,fileroot));
      rec_engine[id]->set(it, options.end());
      LOGMSG("\tCreating xyz output engine.")
      return id;
    } else if((*it) == "--tbmd") {
      LOGMSG("\tCreating TBMD output engine.")
      rec_engine.push_back(new TbmdParticleIO(P,omode,fileroot));
      rec_engine[id]->set(it, options.end());
      return id;
    } 
// #ifdef USE_HDF5
//     else if((*it) == "--hdf") {
//       LOGMSG("\tCreating HDF output engine.")
//       rec_engine.push_back(new HDFParticleIO(P,omode,fileroot));
//       rec_engine[id]->set(it, options.end());
//       return id;
//     }
// #endif
#ifdef USE_STORM
/*
    else if((*it) == "--xyzdcf") {
      LOGMSG("\tCreating xyzDCF output engine.")
      rec_engine.push_back(new XYZDCFParticleIO<Scalar_t,DIM>(P,omode,fileroot));
      rec_engine[id]->set(it, options.end());
      return id;
    }
*/
    else if((*it) == "--dcf") {
      LOGMSG("\tCreating DCF output engine.")
      rec_engine.push_back(new DCFParticleIO(P,omode,fileroot));
      rec_stride = 1;
      rec_engine[id]->set(it, options.end());
      return id;
    }
#endif
  }

#ifdef __GNUC__
// #ifdef USE_HDF5
//   rec_engine.push_back(new HDFParticleIO(P,ParticleIO_t::PROGRESS,fileroot));
// #else
  rec_engine.push_back(new TbmdParticleIO(P,ParticleIO_t::PROGRESS,fileroot));
// #endif
  rec_stride = 1000; // assign a big number not to record much
  rec_engine[id]->set(options.begin(), options.end());
#else
  rec_engine.push_back(NULL);
#endif
  return id;
}

/*! \fn bool ParticleOuputFactory::update(int id, int iter)
 *  \param id index for an output engine
 *  \param iter current iteration number
 *  \return true if ParticleOutputBase::update(iter) is successful.
 */
bool ParticleOutputFactory::update(int id, int iter) {

  if(id>=0 && (id<save_engine.size())) {

    if(rec_engine[id]) { // ignore if output engine is not requested
      if(rec_stride ==1 || iter%rec_stride == 0)  
	rec_engine[id]->record(iter);
    } 

    if(save_stride == 1 || iter%save_stride == 0)  
      save_engine[id]->write();
    return true;
  } else 
    return false;
}


bool ParticleOutputFactory::finalize(int id) {

  if(id>=0 && (id< rec_engine.size())) {
    if(rec_engine[id]) 
      rec_engine[id]->finalize(); 
    return true;
  } 
  return false;
}

bool ParticleOutputFactory::remove() {

  // clean up all the output engines
  for(int i=0; i<rec_engine.size(); i++) {
    if(rec_engine[i]) delete rec_engine[i];
    if(save_engine[i]) delete save_engine[i];
  }

  rec_engine.erase(rec_engine.begin(), rec_engine.end());
  save_engine.erase(save_engine.begin(), save_engine.end());

  return true;
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ParticleOutputFactory.cpp 151 2004-11-02 23:51:14Z jnkim $
 ***************************************************************************/
