//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

/************************************************************************
 Classes required for MEAM potential by T. Lenosky et al.
 Original code is written by Thomas Lenosky, LLNL, 12/8/99
 Modified for PoomaMD by Jeongnim Kim 12/8/99
 Modified for AlloyEAM by Richard Hennig 01/2002
 ************************************************************************/
#include <math.h>
#include "Potentials/CubicSplineType.h"
#include "Utilities/SimpleParser.h"
using namespace std;

SplineType::SplineType(const SplineType& a){
  Npt = a.Npt;
  Yp1 = a.Yp1;
  Ypn = a.Ypn;
  X = a.X;
  Y = a.Y;
  Y2 = a.Y2;
}

SplineType& SplineType::operator=(const SplineType& a){
  Npt = a.Npt;
  Yp1 = a.Yp1;
  Ypn = a.Ypn;
  X = a.X;
  Y = a.Y;
  Y2 = a.Y2;
  return *this;
}


void SplineType::readSpline(istream& is){
  /*Read s from f, using format common with fwrite_spline(f,s) */

  vector<string> slist;
  while ((getwords(slist, is))==0);
  Npt = atoi(slist[0].c_str());
  Yp1 = atof(slist[1].c_str());
  Ypn = atof(slist[2].c_str());

  X  = vector<value_type>(Npt+1);
  Y  = vector<value_type>(Npt+1);
  Y2 = vector<value_type>(Npt+1);

  X[0] = 0.0e0; Y[0] = 0.0e0; Y2[0] = 0.0e0;

  int i=1;
  while (i<=Npt) {
    int nw = getwords(slist,is);
    switch (nw) {
    case 0: getwords(slist,is); break;
    case 2: 
      X[i]  = atof(slist[0].c_str());
      Y[i]  = atof(slist[1].c_str());
      Y2[i] = 0.0E0;
      i++;
      break;
    case 3:
      X[i]  = atof(slist[0].c_str());
      Y[i]  = atof(slist[1].c_str());
      Y2[i] = atof(slist[2].c_str());
      i++;
      break;
    default:
      cerr << "Something wrong with the splines in the potential file\n";
      cerr << "Exit ungracefully";
      exit(-1);
    }
  }
}

void SplineType::readSpline(istream& is, value_type econv){
  /* Read spline for AlloyEAM from file */
  int Type;
  value_type X1, Xn;

  is >> Type;

  switch(Type) {
  case 1:
    is >> Npt;
    is >> X1 >> Xn;
    is >> Yp1 >> Ypn;

    Yp1 *= econv;
    Ypn *= econv;
    X  = vector<value_type>(Npt+1);
    Y  = vector<value_type>(Npt+1);
    Y2 = vector<value_type>(Npt+1);

    X[0] = 0.0e0; Y[0] = 0.0e0; Y2[0] = 0.0e0;

    for(int i=1; i<= Npt;i++) {
      X[i]  = (double) (i-1)/(double) (Npt-1)*(Xn-X1)+X1;
      is >> Y[i];
      Y[i] *= econv;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  case 2:
    is >> Npt;
    is >> Yp1 >> Ypn;

    Yp1 *= econv;
    Ypn *= econv;

    X  = vector<value_type>(Npt+1);
    Y  = vector<value_type>(Npt+1);
    Y2 = vector<value_type>(Npt+1);

    X[0] = 0.0e0; Y[0] = 0.0e0; Y2[0] = 0.0e0;

    for(int i=1; i<= Npt;i++) {
      is >> X[i];
      is >> Y[i];
      Y[i] *= econv;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  }

  /////////////////////////////////////////
  // important: getSpline is done after reading data
  /////////////////////////////////////////
  getSpline();
}

void SplineType::getSpline() {
  //This is just a modified version of the Numerical Recipes routine spline, 
  //which generates spline coefficients given a set of spline parameters.
  int i,k;
  double p,qn,sig,un;
  vector<double> u(X.size()+1);

  if (Yp1 > 0.99e30)
    Y2[1]=u[1]=0.0;
  else {
    Y2[1] = -0.5;
    u[1]=(3.0/(X[2]-X[1]))*((Y[2]-Y[1])/(X[2]-X[1])-Yp1);
  }
  for (i=2;i<=Npt-1;i++) {
    sig=(X[i]-X[i-1])/(X[i+1]-X[i-1]);
    p=sig*(Y2[i-1])+2.0;
    Y2[i]=(sig-1.0)/p;
    u[i]=(Y[i+1]-Y[i])/(X[i+1]-X[i]) - (Y[i]-Y[i-1])/(X[i]-X[i-1]);
    u[i]=(6.0*u[i]/(X[i+1]-X[i-1])-sig*u[i-1])/p;
  }
  if (Ypn > 0.99e30) qn=un=0.0;
  else {
    qn=0.5;
    un=(3.0/(X[Npt]-X[Npt-1]))*(Ypn-(Y[Npt]-Y[Npt-1])/(X[Npt]-X[Npt-1]));
  }
  Y2[Npt]=(un-qn*u[Npt-1])/(qn*(Y2[Npt-1])+1.0);
  for(k=Npt-1;k>=1;k--) Y2[k]= (Y2[k]*(Y2[k+1]))+u[k];
}



//// return f(r)
//SplineType::value_type 
//SplineType::splint(SplineType::value_type r, int extype) const {
//
//  const value_type oneoversixth=1.0/6.0;
//  int klo,khi,k;
//  value_type h,b,a;
//  if(r>X[1] && r < X[Npt]) { //NR part
//     klo=1;
//     khi=Npt;
//     while (khi-klo > 1) {
//       k=(khi+klo) >> 1;
//       if (X[k] > r) khi=k;
//       else klo=k;
//     }
//     h=X[khi]-X[klo];
//     //if (h<1.e-12) {
//     // // ERRORMSG("splint called with x = %lf\n",x);
//     //};
//     value_type hinv=1.0/h;
//     a=(X[khi]-r)*hinv;
//     b=(r-X[klo])*hinv;
//     return a*Y[klo]+b*Y[khi]+
//         (a*(a*a-1.e0)*Y2[klo]+b*(b*b-1.0e0)*Y2[khi])*(h*h)*oneoversixth;
//  } else { // Extrapolation past endpoints
//    if(r<=X[1]) {
//      if(extype == 0)  return Y[1] + Yp1 * (r - X[1]);
//      else  { //Fit 1/x and 1/x/x termination
//        value_type rc,yc,derc;
//        rc=X[1];
//        yc=Y[1];
//        derc=Yp1;
//        b = -rc*rc*rc*derc - rc*rc*yc;
//        a = rc*yc - b/rc;
//        return (a+b/r)/r;
//      }
//    }
//    else{
//      return Y[Npt] + Ypn * (r - X[Npt]);
//    }
//  }
//}
//
//// return val = f(r), deriv = df(r)/dr
//SplineType::value_type 
//SplineType::splint(SplineType::value_type r, 
//		   SplineType::value_type& val,
//		   int extype) const {
//
//  const value_type oneoversixth=1.0/6.0;
//  const value_type one=1.0e0, three=3.0e0;
//
//  value_type deriv=0;
//  val=0;
//  register int klo,khi,k;
//  register value_type h,b,a;
//  if(r>X[1] && r < X[Npt]) { //NR part
//     klo=1;
//     khi=Npt;
//     while (khi-klo > 1) {
//       k=(khi+klo) >> 1;
//       if (X[k] > r) khi=k;
//       else klo=k;
//     }
//     h=X[khi]-X[klo];
//     //if (h<1.e-12) {
//     //  //ERRORMSG("splint called with r = " << r << endl);
//     //};
//     value_type hinv=one/h;
//     a=(X[khi]-r)*hinv;
//     b=(r-X[klo])*hinv;
//     value_type aa=a*a, bb=b*b, hh=h*h;
//     val=a*Y[klo]+b*Y[khi]+
//         (a*(aa-one)*Y2[klo]+b*(bb-one)*Y2[khi])*(hh)*oneoversixth;
//     deriv = -(hinv)*(Y[klo]-Y[khi] + (hh*oneoversixth)*((three*aa-one)*Y2[klo] -
//              (three*bb-one)*Y2[khi]) );
//  } else { // Extrapolation past endpoints
//    if(r<=X[1]) {
//      if(extype == 0) {
//        deriv = Yp1;
//        val =Y[1] + deriv * (r - X[1]);
//      }
//      else  { //Fit 1/x and 1/x/x termination
//        value_type rc,yc,derc;
//        rc=X[1];
//        yc=Y[1];
//        derc=Yp1;
//
//        b = -rc*rc*rc*derc - rc*rc*yc;
//        a = rc*yc - b/rc;
//
//        deriv = -(a+2.0e0*b/r)/r/r;
//        val = (a+b/r)/r;
//      }
//    }
//    else{
//      deriv = Ypn;
//      val= Y[Npt] + deriv * (r - X[Npt]);
//    }
//  }
//  return deriv;
//}

/////////////////////////////////////////////
CubicSplineRegGrid::CubicSplineRegGrid(const CubicSplineRegGrid& a ) {
  Npt = a.Npt;
  Rmin = a.Rmin;
  Rmax = a.Rmax;
  dr = a.dr;
  drinv = drinv2;
  Y = a.Y;
  Yp = a.Yp;
  Ypp = a.Ypp;
}

CubicSplineRegGrid& 
CubicSplineRegGrid::operator=(const CubicSplineRegGrid& a ) {
  Npt = a.Npt;
  Rmin = a.Rmin;
  Rmax = a.Rmax;
  dr = a.dr;
  drinv = drinv2;
  Y = a.Y;
  Yp = a.Yp;
  Ypp = a.Ypp;
  return *this;
}


void CubicSplineRegGrid::resize(const int n) {
  Npt = n;
  Y = vector<value_type>(n+1); 
  Yp = vector<value_type>(n+1); 
  Ypp = vector<value_type>(n+1); 
}

CubicSplineRegGrid::value_type 
CubicSplineRegGrid::splint(value_type x0) const {
  //if(x0 < Rmin || x0 > Rmax) return 0.0;
  int i = static_cast<int>((x0-Rmin)*drinv);
  value_type xx = (x0 - static_cast<value_type>(i)*dr - Rmin)*drinv;
  return Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0);
}


CubicSplineRegGrid::value_type 
CubicSplineRegGrid::splint(value_type x0, value_type& yval) const { 
  //if(x0 < Rmin || x0 > Rmax) { yval = 0.0; return 0.0;}
  int i = static_cast<int>((x0-Rmin)*drinv);
  value_type xx = (x0 - static_cast<value_type>(i)*dr - Rmin)*drinv;
  yval = Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0);
  return (Yp[i] + Ypp[i]*(2*xx-1.0))*drinv;
}

void 
CubicSplineRegGrid::spline(const int n, value_type x1, value_type x2, vector<value_type>::iterator y) {
  resize(n);
  Rmin = x1; Rmax = x2;
  dr = (x2-x1)/static_cast<value_type>(n);  
  drinv = 1.0/dr; drinv2 = drinv*drinv;
  for(int i=0; i<=n; i++)  Y[i]   = y[i];
  for(int i=0; i< n; i++)   Yp[i]  = Y[i+1]-Y[i];
  Yp[n] = 0.0;
  for(int i=0; i<n-1; i++) Ypp[i] = 0.5*(Y[i+2]+Y[i]-2.0*Y[i+1]);
  Ypp[n-1] = 0.0; Ypp[n] = 0.0;
}


void 
CubicSplineRegGrid::readSpline(istream& is, value_type econv)  {

  /* Read spline for AlloyEAM from file */
  int Type, npt;
  value_type X1, Xn, Yp1, Ypn;
  vector<value_type> yin;

  is >> Type;
  is >> npt;
  yin.resize(npt);

  switch(Type) {
  case 1:
    is >> X1 >> Xn;
    is >> Yp1 >> Ypn;

    for(int i=0; i< npt;i++) {
      is >> yin[i];
      yin[i] *= econv;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  case 2:
    is >> Yp1 >> Ypn;
    
    for(int i=0; i< npt;i++) {
      is >> Xn;
      is >> yin[i];
      yin[i] *= econv;
      if(i == 0) X1 = Xn;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  }

  spline(npt-1, X1, Xn, yin.begin());
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 308 $   $Date: 2005-02-21 09:21:44 -0600 (Mon, 21 Feb 2005) $
 * $Id: CubicSplineType.cpp 308 2005-02-21 15:21:44Z jnkim $ 
 ***************************************************************************/

