//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/************************************************************************
 Classes required for MEAM potential by T. Lenosky et al.
 ************************************************************************/
#ifndef OHMMS_CUBIC_SPLINE_TYPE_H
#define OHMMS_CUBIC_SPLINE_TYPE_H


#include <iostream>
#include <vector>

/*! \class SplineType
 * \brief spline functions on a finite regular or irregular r points
 *
 * suitable for small number of spline points or irregular strides between r's
 * \li Original code is written by Thomas Lenosky, LLNL, 12/8/99
 * \ii Modified for PoomaMD by Jeongnim Kim 12/8/99
 * \li Modified for AlloyEAM by Richard Hennig 01/2002
 * \li Modified by Jeongnim Kim 07/26/2002
 */
struct SplineType {

  typedef double value_type;
  std::vector<value_type> X;
  std::vector<value_type> Y;
  std::vector<value_type> Y2;
  value_type Yp1;
  value_type Ypn;
  int Npt;
  int iset1,isetn,ider1,idern;

  SplineType(){ }
  SplineType(const SplineType&);
  SplineType& operator=(const SplineType&);

  void readSpline(std::istream& is);
  void readSpline(std::istream& is, value_type econv);
  void getSpline();

  inline value_type rmax() const { return X[Npt];}
  inline value_type rmin() const { return X[1];}

  inline value_type splint(value_type r, int extype= 0) const {
  const value_type oneoversixth=1.0/6.0;
  int klo,khi,k;
  value_type h,b,a;
  if(r>X[1] && r < X[Npt]) { //NR part
     klo=1;
     khi=Npt;
     while (khi-klo > 1) {
       k=(khi+klo) >> 1;
       if (X[k] > r) khi=k;
       else klo=k;
     }
     h=X[khi]-X[klo];
     //if (h<1.e-12) {
     // // ERRORMSG("splint called with x = %lf\n",x);
     //};
     value_type hinv=1.0/h;
     a=(X[khi]-r)*hinv;
     b=(r-X[klo])*hinv;
     return a*Y[klo]+b*Y[khi]+
         (a*(a*a-1.e0)*Y2[klo]+b*(b*b-1.0e0)*Y2[khi])*(h*h)*oneoversixth;
  } else { // Extrapolation past endpoints
    if(r<=X[1]) {
      if(extype == 0)  return Y[1] + Yp1 * (r - X[1]);
      else  { //Fit 1/x and 1/x/x termination
        value_type rc,yc,derc;
        rc=X[1];
        yc=Y[1];
        derc=Yp1;
        b = -rc*rc*rc*derc - rc*rc*yc;
        a = rc*yc - b/rc;
        return (a+b/r)/r;
      }
    }
    else{
      return Y[Npt] + Ypn * (r - X[Npt]);
    }
  }
  }
  inline value_type splint(value_type r, value_type& val, int extype = 0) const {
  const value_type oneoversixth=1.0/6.0;
  const value_type one=1.0e0, three=3.0e0;

  value_type deriv=0;
  val=0;
  register int klo,khi,k;
  register value_type h,b,a;
  if(r>X[1] && r < X[Npt]) { //NR part
     klo=1;
     khi=Npt;
     while (khi-klo > 1) {
       k=(khi+klo) >> 1;
       if (X[k] > r) khi=k;
       else klo=k;
     }
     h=X[khi]-X[klo];
     //if (h<1.e-12) {
     //  //ERRORMSG("splint called with r = " << r << endl);
     //};
     value_type hinv=one/h;
     a=(X[khi]-r)*hinv;
     b=(r-X[klo])*hinv;
     value_type aa=a*a, bb=b*b, hh=h*h;
     val=a*Y[klo]+b*Y[khi]+
         (a*(aa-one)*Y2[klo]+b*(bb-one)*Y2[khi])*(hh)*oneoversixth;
     deriv = -(hinv)*(Y[klo]-Y[khi] + (hh*oneoversixth)*((three*aa-one)*Y2[klo] -
              (three*bb-one)*Y2[khi]) );
  } else { // Extrapolation past endpoints
    if(r<=X[1]) {
      if(extype == 0) {
        deriv = Yp1;
        val =Y[1] + deriv * (r - X[1]);
      }
      else  { //Fit 1/x and 1/x/x termination
        value_type rc,yc,derc;
        rc=X[1];
        yc=Y[1];
        derc=Yp1;

        b = -rc*rc*rc*derc - rc*rc*yc;
        a = rc*yc - b/rc;

        deriv = -(a+2.0e0*b/r)/r/r;
        val = (a+b/r)/r;
      }
    }
    else{
      deriv = Ypn;
      val= Y[Npt] + deriv * (r - X[Npt]);
    }
  }
  return deriv;
  }
};

/*! \class CubicSplineRegGrid
 * \brief non-template version of ScaleFunctors::CubicSpineDiff<T>
 * \note Suitable for regular stride and many spline points
 */
struct CubicSplineRegGrid{

  typedef double value_type;
  int Npt;
  value_type   Rmin, Rmax, dr, drinv, drinv2;
  std::vector<value_type> Y, Yp, Ypp;

  CubicSplineRegGrid() { }
  CubicSplineRegGrid(const CubicSplineRegGrid& a );

  CubicSplineRegGrid& operator=(const CubicSplineRegGrid& a );

  void resize(const int n);

  inline value_type rmax() const { return Rmax;}
  inline value_type rmin() const { return Rmin;}

  //@return f(r)
  value_type splint(value_type x0) const;    

 //@return f'(r) and val = f(r)
  value_type splint(value_type x0, value_type& yval) const;

  void spline(const int n, value_type x1, value_type x2, 
	      std::vector<value_type>::iterator y);
  void readSpline(std::istream& is, value_type econv);
};


#endif


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 308 $   $Date: 2005-02-21 09:21:44 -0600 (Mon, 21 Feb 2005) $
 * $Id: CubicSplineType.h 308 2005-02-21 15:21:44Z jnkim $ 
 ***************************************************************************/
