//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Richard Hennig and Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "OhmmsData/FileUtility.h"
#include "Potentials/EAMPotential.h"
#include "Utilities/SimpleParser.h"
#include "Utilities/OhmmsInfo.h"
#include "ParticleBase/ParticleMessage.h"

namespace OHMMS {


  /*! \fn EAMPotential::EAMPotential()
   *  \brief Default constructor of EAMPotential
   */
  EAMPotential::EAMPotential(): EAMSet(0), mySpecies(0){
    CollectVisitingForces = true;
    myName = "EAM";
  }

  EAMPotential::~EAMPotential() {
    if(EAMSet) delete EAMSet;
  }

  bool EAMPotential::get(ostream& os) const { 
    return true;
  }

  bool EAMPotential::put(istream&) {return true;}

  void EAMPotential::reset() {}


  bool EAMPotential::put(xmlNodePtr cur){
    //get the file name
    mySpecies = &(myPtcl->getSpeciesSet());

    string potfile("EAMPOT");

#if defined(HAVE_LIBXML2)
    if(xmlHasProp(cur, (const xmlChar *) "file")) {
      potfile = (const char*)(xmlGetProp(cur, (const xmlChar *) "file"));
    }
    cur = cur->children;
    while(cur != NULL) {
      string cname((const char*)cur->name);
      if(cname == "filename") {
	putContent(potfile,cur);
      }
      cur = cur->next;
    }
#endif
    string pformat = getExtension(potfile);
    if(pformat == "xml") {
      return parseXML(potfile);       
    } else {

      LOGMSG("\tEAMPotential::Data file: " << potfile)
      Rmax = 0.0;
      size_type imass = mySpecies->addAttrib("mass");
      size_type irmax = mySpecies->addAttrib("rmax");
      
      ifstream fin(potfile.c_str());
      if(!fin) { 
	ERRORMSG(potfile << " is not found in this directory.")
	ERRORMSG("Exiting very ungracefully.")
	exit(1);
      }
      DEBUGMSG("Read species parameters from file "<< potfile)
      vector<string> slist;
      int nspecies = 0;
      do {
	while ((getwords(slist, fin))==0);
	if (slist[0] == "species") {
	  size_type newid =mySpecies->addSpecies(slist[1].c_str());
	  (*mySpecies)(imass,newid)=atof(slist[2].c_str());
	  nspecies++;
	  DEBUGMSG("New Species: " << slist[1].c_str() << " (" << newid << ") " << "m= " << (*mySpecies)(imass,newid))
	}
      } while(slist[0] != "<potential>");

      DEBUGMSG("Number of species: " << nspecies)
      if(EAMSet == 0)
        EAMSet = new EAMParam<Scalar_t>(potfile.c_str(), nspecies);
      else {
        EAMSet->readEAMParam(potfile.c_str(),nspecies);
        EAMSet->check(nspecies);
      }

      Rmax = max(Rmax, EAMSet->rcutmax());
      LOGMSG("\tMaximum interaction radius : " << Rmax << " AA")

      for (int i=0; i<nspecies; i++) (*mySpecies)(irmax,i) = Rmax*0.5;
    }
    return true;
  }

  bool EAMPotential::parseXML(const string& fname) {

#if defined(HAVE_LIBXML2)
    xmlDocPtr doc=NULL;
    xmlNsPtr ns=NULL;
    doc = xmlParseFile(fname.c_str());
    if (doc == NULL) {
      ERRORMSG(fname << " does not exist or is not a valid xml file ")
      xmlFreeDoc(doc);
      return false;
    }

    xmlNodePtr cur = xmlDocGetRootElement(doc);
    if(xmlStrcmp(cur->name,(const xmlChar*)"eam")) {
      ERRORMSG(fname << " does not have root node, eam.")
      xmlFreeDoc(doc);
      return false;
    }

    ///process attributes: set the unit of the input data
    double econv = 1.0;
    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "condition") {
	if (vname == "K")   econv = 8.617385692E-5;
	else if (vname == "eV")  econv = 1.0;
	else if (vname == "erg") econv = 6.241506363E11;
	LOGMSG("\tEAM Potential file is in " <<vname << " unit. Conversion factor " << econv)
      } 
      att = att->next;
    }
    
    xmlNodePtr cur_save = cur;

    //check the species first
    mySpecies->put(cur);

    size_type nspecies=mySpecies->getTotalNum();
    if(EAMSet == 0)
      EAMSet =  new EAMParam<Scalar_t>;

    EAMSet->phi.resize(nspecies*nspecies);
    EAMSet->rho.resize(nspecies);
    EAMSet->f.resize(nspecies);
    cur = cur_save->children;
    while(cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "potential") {
	size_type i,j;
	string ptype("none");
	att = cur->properties;
	while(att != NULL) {
	  string aname((const char*)(att->name));
	  if(aname == "A") {
	    i = mySpecies->addSpecies((const char*)(att->children->content));
	  } else if(aname == "B") {
	    j = mySpecies->addSpecies((const char*)(att->children->content));
	  } else if(aname == "type") {
	    ptype =(const char*)(att->children->content); 
	  }
	  att = att->next;
	}
	std::istringstream
	  stream((const char*)(xmlNodeListGetString(cur->doc, cur->xmlChildrenNode, 1)));
	if(ptype == "pair") {
	  if(i>j) std::swap(i,j);
	  EAMSet->phi[i*nspecies+j].readSpline(stream, econv);
	  EAMSet->phi[i*nspecies+j].getSpline();
	} else if(ptype == "dens") {
	  EAMSet->rho[i].readSpline(stream, 1.0);
	  EAMSet->rho[i].getSpline();
	} else if(ptype == "embed") {
	  EAMSet->f[i].readSpline(stream, econv);
	  EAMSet->f[i].getSpline();
	}
      }
      cur = cur->next;
    }
    EAMSet->check(nspecies);

    Rmax = EAMSet->rcutmax();
    size_type irmax = mySpecies->addAttrib("rmax");
    LOGMSG("\tThe number of species : " << nspecies)
    LOGMSG("\tSet Rmax: " << Rmax)
    for (size_type i=0; i<nspecies; i++) (*mySpecies)(irmax,i) = Rmax*0.5;
    //free xml doc and tree
    xmlFreeDoc(doc);
    return true;
#endif
  }


//  void EAMPotential::initialize() {
//
//    if(EAMSet.empty()) {
//
//      string potfile("EAMPOT");
//
//      Rmax = 0.0;
//      int imass = mySpecies->addAttrib("mass");
//      int irmax = mySpecies->addAttrib("rmax");
//
//      LOGMSG("\tData file: " << potfile.c_str())
//
//      ifstream fin(potfile.c_str());
//      if(!fin) { 
//	ERRORMSG(potfile.c_str() << " is not found in this directory")
//	  ERRORMSG("Exiting very ungracefully: EAMPotential::initialize()")
//	  exit(1);
//      }
//
//      vector<string> slist;
//      int nspecies = 0;
//      do {
//	while ((getwords(slist, fin))==0);
//	if (slist[0] == "species") {
//	  int newid =mySpecies->addSpecies(slist[1].c_str());
//	  (*mySpecies)(imass,newid)=atof(slist[2].c_str());
//	  nspecies ++;
//	  cerr << "New Species: " << slist[1].c_str() << " (" << newid
//	       << ") " << "m= " << (*mySpecies)(imass,newid) <<"\n";
//	}
//      } while(slist[0] != "<potential>");
//
//      cerr << "Number of species: " << nspecies << "\n";
//      EAMSet.push_back(new EAMParam<Scalar_t>(potfile.c_str(), nspecies));
//
//      Rmax = max(Rmax, EAMSet[0]->rcutmax());
//      cerr << "Set Rmax: " << Rmax <<"\n";
//      for (int i=0; i<nspecies; i++) (*mySpecies)(irmax,0) = Rmax*0.5;
//    }
//  }

}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: EAMPotential.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

