//////////////////////////////////////////////////////////////////
// (c) Copyright 2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

/**@file FastPotential.cpp
 * @brief Define FastPotential class.
 */
#include "Potentials/ShortRangePotential.h"
#include "Potentials/FastPotential.h"
#include "Particle/NNEngine.h"
#include "Utilities/OhmmsInfo.h"
namespace OHMMS {
  
      ///destructor
  FastPotential::~FastPotential() {
    if(srp_impl) delete srp_impl;
  }
  bool FastPotential::get(ostream&) const { return true;}
  bool FastPotential::put(istream&) {return true;}
  bool FastPotential::put(xmlNodePtr cur){
    PotXmlNode=cur; 
    return true;
  }
  void FastPotential::reset() {}

  void FastPotential::setParticleSet(Particle_t* ptclset) {
    if(myPtcl == ptclset) return;
    myPtcl=ptclset;
    srp_impl->setParticleSet(ptclset);
    if(PotXmlNode) {srp_impl->put(PotXmlNode); PotXmlNode=0;}
    GridRadius = Rmax = srp_impl->Rmax;
    nnListEngine = myPtcl->getNNEngine();
    if(nnListEngine == 0) {
      nnListEngine = NNEngineBuilder::create(LayoutXmlNode);
    } else {
      ERRORMSG("\tUsing an existing layout ")
    }
    Rmax *= nnListEngine->Rscale;
    myPtcl->setNNEngine(nnListEngine);
    myPtcl->updateLayout(Rmax,GridRadius);
    nnListEngine->initialize(*myPtcl,CollectVisitingForces);
  }

  void FastPotential::resetParticleSet(Particle_t* ptclset) {
    myPtcl=ptclset;
    srp_impl->resetParticleSet(ptclset);
    nnListEngine=myPtcl->getNNEngine();
  }

  void 
  FastPotential::setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell) {
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
    if(wrap) myPtcl->applyBConds(true,myPtcl->curR.getUnit());
  }

  void 
  FastPotential::setCurrentPos(ParticlePos_t& pos) {
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
  }

  void 
  FastPotential::setCurrentPos(const ParticlePos_t& pos, 
      const ParticlePos_t& del) {
    nnListEngine->updatePairs(*myPtcl,del);
  }

  PotentialBase::Scalar_t FastPotential::evalEnergy() {

    Scalar_t esum(0.0);
#pragma omp parallel reduction(+:esum)
    {
      omp_int_t ip=omp_get_thread_num();
      esum += srp_impl->evalEnergyNN(nnListEngine->getPairAttrib(ip),1.0);
    }//end-of-omp
    return nnListEngine->collect(esum);
  }

  PotentialBase::Scalar_t
  FastPotential::evalForce(ParticlePos_t& f, unsigned fmode) {
    f=0.0;
#if defined(HAVE_MPI) || defined(ENABLE_OPENMP)
    if(nnListEngine->UseVisitingForce)
      return evalForceMultiple(f,fmode);
    else
      return evalForcePairOnly(f,fmode);
#else
      return evalForcePairOnly(f,fmode);
#endif
  }

  PotentialBase::Scalar_t
  FastPotential::evalForcePairOnly(ParticlePos_t& f, unsigned fmode) {
    unsigned stressmode=(fmode&1);
    Scalar_t esum(0.0);
#pragma omp parallel reduction(+:esum)
    {
      omp_int_t ip=omp_get_thread_num();
      if(fmode&2) {
        nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(ip));
      }
      esum += srp_impl->evalForceNN(nnListEngine->getPairAttrib(ip),f,stressmode);
    }

    Stress=0.0;
    for(size_type ip=0; ip<nnListEngine->size(); ip++) {
      Stress += nnListEngine->getPairAttrib(ip).Stress;
    }
    return nnListEngine->collect(esum);
  }

  PotentialBase::Scalar_t
  FastPotential::evalForceMultiple(ParticlePos_t& f, unsigned fmode) {
    unsigned stressmode=fmode&1;
    Scalar_t esum = 0.0;
#pragma omp parallel reduction(+:esum)
    {
      omp_int_t ip=omp_get_thread_num();
      if(fmode&2) 
        nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(ip));

      ParticlePos_t& fv=nnListEngine->getPairAttrib(ip).Fv;
      fv=0.0;
      esum += srp_impl->evalForceNN(nnListEngine->getPairAttrib(ip),fv,stressmode);

      ////This may not be the optimal
      //int first(nnListEngine->first_id);
      //int last(nnListEngine->last_id);
      //for(int iat=first; iat<last; iat++) f[iat]=0.0;

      nnListEngine->collectForce(f,omp_get_thread_num());

    }

#pragma omp parallel
    {
      nnListEngine->collectForce(f,omp_get_thread_num());
    }

    if(stressmode) {
      Stress = 0.0;
      for(int myid=0; myid<nnListEngine->size(); myid++) {
        Stress += nnListEngine->getPairAttrib(myid).Stress;
      }
    }
    return nnListEngine->collect(esum);
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 363 $   $Date: 2005-07-13 13:33:18 -0500 (Wed, 13 Jul 2005) $
 * $Id: FastPotential.cpp 363 2005-07-13 18:33:18Z jnkim $ 
 ***************************************************************************/
