//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002, 2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_GENERIC_TWOBODY_SHORTRANGEPOTENTIAL_H
#define OHMMS_GENERIC_TWOBODY_SHORTRANGEPOTENTIAL_H

#include "Potentials/ShortRangePotential.h"

namespace OHMMS {

  template<class RadFunc>
  struct GenericTwoBodySRP: public ShortRangePotential {

    typedef RadFunc                    RadFunc_t;
    typedef PtclPairList_t::PairData_t PairData_t;

    bool SingleBond;
    Scalar_t PairFactor;
    Matrix<int> Bond;
    vector<RadFunc*> Phi;

    GenericTwoBodySRP():SingleBond(true),PairFactor(0.5e0){
      CollectVisitingForces=false;
      Bond.resize(1,1);
    }

    ~GenericTwoBodySRP() {
      for(int i=0; i<Phi.size(); i++) delete Phi[i];
    }

    /// implementing virtual functions for OhmmsElementBase
    bool get(ostream&) const { return true;}
    bool put(istream&) { return true;}
    bool put(xmlNodePtr cur) { return true;}
    void reset() {}

    void setParticleSet(Particle_t* ptclset) { 
      myPtcl=ptclset;
      const SpeciesSet& s(myPtcl->getSpeciesSet());
      Bond.resize(s.getTotalNum(),s.getTotalNum());
      Bond=-1;
      if(s.getTotalNum()>1) SingleBond=false;
    }

    void resetParticleSet(Particle_t* ptclset) { 
      myPtcl=ptclset;
    }

    inline Scalar_t 
    evalForceNN(PtclPairList_t& nnlist, ParticlePos_t& f0, 
        unsigned stressmode) const {
      if(SingleBond) 
        return evalForceNNSingle(nnlist,f0,stressmode);
      else
        return evalForceNNMultiple(nnlist,f0,stressmode);
    }

    inline Scalar_t 
    evalEnergyNN(const PtclPairList_t& nnlist,Scalar_t rscale=1.0) const {
      if(SingleBond) 
        return evalEnergyNNSingle(nnlist);
      else
        return evalEnergyNNMultiple(nnlist);
    }

    /** add a pair potential for itype-jtype bond
     */
    void addPairPotential(int itype, int jtype, RadFunc* phi) {
      if(Bond(itype,jtype) < 0) { //uninitialized pair
        int n(Phi.size());
        Bond(itype,jtype)=n; 
        //i-j is identical to j-i
        if(itype != jtype && Bond(jtype,itype)<0) {
          Bond(jtype,itype)=n; 
        }
        Phi.push_back(phi);
      }
    }

    Scalar_t 
    evalForceNNSingle(PtclPairList_t& nnlist, ParticlePos_t& f0, 
        unsigned stressmode) const;

    Scalar_t 
    evalForceNNMultiple(PtclPairList_t& nnlist, ParticlePos_t& f0, 
        unsigned stressmode) const;

    Scalar_t evalEnergyNNSingle(const PtclPairList_t& nnlist) const;

    Scalar_t evalEnergyNNMultiple(const PtclPairList_t& nnlist) const;
  };

  template<class RadFunc>
  inline ShortRangePotential::Scalar_t 
  GenericTwoBodySRP<RadFunc>::evalForceNNSingle(PtclPairList_t& nnlist, 
      ParticlePos_t& f0, unsigned stressmode) const {

    ParticleScalar_t& q(myPtcl->E);
    Scalar_t rep, sumloc = 0.0e0,psum = 0.0e0;
    Scalar_t drep;
    Tensor_t stressloc;
    const RadFunc& phi(*Phi[0]);
    for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
      int iat(nnlist.I[iL]);
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
        Scalar_t sep(cur.sep);
        if(sep < Rmax) {
          //LenoskTB/GSP/SplineTB had dR=directional cosine
          //dist = nnlist.R[nn];
          //drep = 0.5e0*Phi(dist,rep);
          //rep *= 0.5e0*dist;
          //SingleParticlePos_t df = drep*nnlist.dR[nn];
          //f0[iat] += 2*df;
          //if(stressmode) {
          //  stressloc -=dist*outerProduct(df,nnlist.dR[nn]);
          //}
          drep = PairFactor*phi(sep,rep)/sep;
          SingleParticlePos_t df(drep*cur.displ);
          f0[iat] += 2.0*df;
          if(stressmode) {
            stressloc -= outerProduct(df,cur.displ);    
          }
          rep *= PairFactor*sep;
          q[iat] += rep;
          sumloc += rep;
        }
        ++nit;
      }
    }
    nnlist.Stress+=stressloc;
    return sumloc;
  }

  template<class RadFunc>
  inline ShortRangePotential::Scalar_t 
  GenericTwoBodySRP<RadFunc>::evalForceNNMultiple(PtclPairList_t& nnlist, 
      ParticlePos_t& f0, unsigned stressmode) const {

    const ParticleIndex_t& grID(myPtcl->GroupID);
    ParticleScalar_t& q(myPtcl->E);

    Scalar_t rep, sumloc = 0.0e0,psum = 0.0e0;
    Scalar_t drep;
    Tensor_t stressloc;
    for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
      int iat(nnlist.I[iL]);
      int speciesI(grID[nnlist.I[iL]]);
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
        Scalar_t sep(cur.sep);
        if(sep < Rmax) {
          drep = PairFactor/sep*Phi[Bond(speciesI,cur.Group)]->operator()(sep,rep);
          SingleParticlePos_t df(drep*cur.displ);
          f0[iat] += 2.0e0*df;
          if(stressmode) {
            stressloc -= outerProduct(df,cur.displ);    
          }
          rep *= PairFactor*sep;
          q[iat] += rep;
          sumloc += rep;
        }
        ++nit;
      }
    }
    nnlist.Stress+=stressloc;
    return sumloc;
  }

  template<class RadFunc>
  inline ShortRangePotential::Scalar_t 
  GenericTwoBodySRP<RadFunc>::evalEnergyNNSingle(const PtclPairList_t& nnlist) 
  const {
    Scalar_t esumloc = 0.0e0;
    int nnloc(0);  
    const RadFunc& phi(*Phi[0]);
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        Scalar_t sep((*nit).sep);
        if(sep<Rmax) {
          esumloc += PairFactor*phi(sep);
        }
        ++nit;
      }
    }
    return esumloc;
  }

  template<class RadFunc>
  inline ShortRangePotential::Scalar_t 
  GenericTwoBodySRP<RadFunc>::evalEnergyNNMultiple(const PtclPairList_t& nnlist) 
  const {
    const ParticleIndex_t& grID(myPtcl->GroupID);
    Scalar_t esumloc = 0.0e0;
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      int speciesI(grID[nnlist.I[iL]]);
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
        if(cur.sep<Rmax) {
          esumloc += PairFactor*Phi[Bond(speciesI,cur.Group)]->operator()(cur.sep);
        }
        ++nit;
      }
    }
    return esumloc;
  }
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 398 $   $Date: 2006-01-28 14:40:52 -0600 (Sat, 28 Jan 2006) $
 * $Id: GenericTwoBodySRP.h 398 2006-01-28 20:40:52Z jnkim $ 
 ***************************************************************************/
