//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "OhmmsData/OhmmsElementBase.h"
#include "Potentials/KeatingParam.h"
#include "Utilities/OhmmsInfo.h"
#include "Utilities/SimpleParser.h"
using namespace std;

KeatingParam::KeatingParam(SpeciesSet& species):
  mySpecies(species),
  three_over_eight(0.375e0),
  tetra(-0.333333333333333333333e0),
  econv(1.0e0/16.0217733e0),
  nSpecies(0),Rmax(0){
  
  }

void 
KeatingParam::readKeatingParam(const char *fname) {

  int irmax = mySpecies.addAttrib("rmax");

#if defined(HAVE_LIBXML2) 
  xmlDocPtr doc = xmlParseFile(fname);
  xmlNsPtr ns;
  xmlNodePtr cur;

  // Check the document is of the right kind
  cur = xmlDocGetRootElement(doc);
  if (cur == NULL) {
    fprintf(stderr,"empty document\n");
    xmlFreeDoc(doc);
    return;
  }

  if (xmlStrcmp(cur->name, (const xmlChar *) "KeatingParam")) {
    fprintf(stderr,"document of the wrong type, root node != KeatingParam");
    xmlFreeDoc(doc);
    return;
  }

  cur = cur->xmlChildrenNode;
  xmlNodePtr cursave = cur;
  int nSecies = mySpecies.getTotalNum();
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *)"species")) {
      int newid 
	=mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "name")));

      nSpecies = max(newid, nSpecies);

      xmlNodePtr subcur = cur->xmlChildrenNode;
      while(subcur != NULL) {
	int irmax = mySpecies.addAttrib((const char*)(subcur->name));
	mySpecies(irmax,newid) 
	  = atof((const char*)(xmlNodeListGetString(doc, subcur->xmlChildrenNode, 1)));
	subcur=subcur->next;
      }
    }
    cur = cur->next;
  }

  nSpecies++;

  initializeParam(nSpecies);// allocate the data
  cur = cursave;
  while(cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *)"twobody")) {
      int iat = mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "nameA")));
      int jat = mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "nameB")));
      int ib = doublet(iat,jat);
      double r0in;
      xmlNodePtr subcur = cur->xmlChildrenNode;
      while(subcur != NULL) {
	if(!xmlStrcmp(subcur->name, (const xmlChar *)"BL")) {
	  r0in 
           = atof((const char*)(xmlNodeListGetString(doc, subcur->xmlChildrenNode, 1)));
	}
	if(!xmlStrcmp(subcur->name, (const xmlChar *)"alpha")) {
	  alpha0[ib]
           = atof((const char*)(xmlNodeListGetString(doc, subcur->xmlChildrenNode, 1)));
	}
	subcur =subcur->next;
      }
      r0[ib] = r0in;
      Rmax = max(r0in*1.1,Rmax);          
      r0sq[ib] = r0in*r0in;
      
      // NOTE: alpha = (3/8 * alpha0/r0^2)/2  and eV unit
      alpha[ib] = econv*three_over_eight*alpha0[ib]/r0sq[ib]/2.0e0;
      if(iat != jat) {
	int jd = doublet(jat,iat);
	r0[jd]     = r0[ib];
	r0sq[jd]   = r0sq[ib];
	alpha0[jd] = alpha0[ib];
	alpha[jd]  = alpha[ib];
      }
      mySpecies(irmax,iat) = r0in;
      mySpecies(irmax,jat) = r0in;
    }
    if (!xmlStrcmp(cur->name, (const xmlChar *)"threebody")) {

      int iat = mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "nameA")));
      int jat = mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "nameB")));

      int kat = mySpecies.addSpecies((const char*)(xmlGetProp(cur, (const xmlChar *) "nameC")));
      int ib = triplet(iat,jat,kat);// index of this triplet
      int ij = doublet(iat,jat);
      int jk = doublet(jat,kat);
      
      xmlNodePtr subcur = cur->xmlChildrenNode;
      while(subcur != NULL) {
	if(!xmlStrcmp(subcur->name, (const xmlChar *)"BA")) {
	  string t 
	    = (const char*)(xmlNodeListGetString(doc, subcur->xmlChildrenNode, 1));
	  if(isdigit(t[1])) {
	    theta0[ib] = atof(t.c_str());
	  } else {
	    if(t == "tetra") theta0[ib] = tetra;
	  }
	  //else 
	  //  theta0[ib] = atof(t.c_str());
	}
	if(!xmlStrcmp(subcur->name, (const xmlChar *)"beta")) {
	  beta0[ib]
           = atof((const char*)(xmlNodeListGetString(doc, subcur->xmlChildrenNode, 1)));
	}
	subcur =subcur->next;
      }

      theta[ib] = theta0[ib]*r0[ij]*r0[jk];
      //\note beta = [3/8 * beta0/(r01*r02)] eV unit
      beta[ib] = econv*three_over_eight*beta0[ib]/r0[ij]/r0[jk];
      
      if(iat != kat) { // i-j-k == k-j-i 
	int kd = triplet(kat,jat,iat);
	theta0[kd] = theta0[ib];
	theta[kd]  = theta[ib];
	beta0[kd]  = beta0[ib];
	beta[kd]   = beta[ib];
      } else {// i-j-i == j-i-j
	int jd = triplet(jat,iat,jat);
	theta0[jd] = theta0[ib];
	theta[jd] =  theta[ib];
	beta0[jd] =  beta0[ib];
	beta[jd] =   beta[ib];
      }
    }
    cur = cur->next;
  }

  xmlFreeDoc(doc);


#else
  int imass = SpeciesCollection::addAttrib("mass");

  ifstream fin(fname);
  int nwords;
  Rmax = 0.0e0; // always initialize it by 0.0e0
  nSpecies = 0;
  bool tocontinue = true;
  if(fin) {
    vector<string> slist;
    do {
      nwords = getwords(slist, fin);
      if(nwords) tocontinue = (slist[0] != "<Keating>");
    } while(tocontinue);

    tocontinue = true;
    do {
      nwords = getwords(slist, fin);
      if(nwords) {
        if(slist[0] == "species") {
          int newid =SpeciesCollection::addSpecies(slist[1].c_str());
          mySpecies(imass,newid)=atof(slist[2].c_str());
          nSpecies = max(newid, nSpecies);
        } else if (slist[0] == "</Keating>") {
          tocontinue = false;
	}
      }
    } while(tocontinue);

    nSpecies++;
    initializeParam(nSpecies);// allocate the data

    fin.seekg(0,ios::beg);
    tocontinue = true;
    do {
      nwords = getwords(slist, fin);
      if(nwords) {
        if(slist[0] == "bond") {
          int iat =SpeciesCollection::addSpecies(slist[1].c_str());
          int jat =SpeciesCollection::addSpecies(slist[2].c_str());
          int ib = doublet(iat,jat);
          double r0in = atof(slist[3].c_str());
          r0[ib] = r0in;
          Rmax = max(r0in*1.1,Rmax);          
          r0sq[ib] = r0in*r0in;
          alpha0[ib] = atof(slist[4].c_str());

          // NOTE: alpha = (3/8 * alpha0/r0^2)/2  and eV unit
          alpha[ib] = econv*three_over_eight*alpha0[ib]/r0sq[ib]/2.0e0;
	  if(iat != jat) {
	    int jd = doublet(jat,iat);
	    r0[jd]     = r0[ib];
	    r0sq[jd]   = r0sq[ib];
	    alpha0[jd] = alpha0[ib];
	    alpha[jd]  = alpha[ib];
	  }
          mySpecies(irmax,iat) = r0in;
          mySpecies(irmax,jat) = r0in;
	} else if(slist[0] == "triplet") {
          int iat =SpeciesCollection::addSpecies(slist[1].c_str());
          int jat =SpeciesCollection::addSpecies(slist[2].c_str());
          int kat =SpeciesCollection::addSpecies(slist[3].c_str());

          int ib = triplet(iat,jat,kat);// index of this triplet
          int ij = doublet(iat,jat);
          int jk = doublet(jat,kat);
          
	  if(isdigit(slist[4][0])) {//real value
            theta0[ib] = atof(slist[4].c_str());
	  } else { // string value, to add more
	    if(slist[4] == "tetra") {
	      theta0[ib] = tetra;
	    }
	  }

          theta[ib] = theta0[ib]*r0[ij]*r0[jk];
          beta0[ib] = atof(slist[5].c_str());

          //\note beta = [3/8 * beta0/(r01*r02)] eV unit
          beta[ib] = econv*three_over_eight*beta0[ib]/r0[ij]/r0[jk];

	  if(iat != kat) { // i-j-k == k-j-i 
	    int kd = triplet(kat,jat,iat);
	    theta0[kd] = theta0[ib];
	    theta[kd]  = theta[ib];
	    beta0[kd]  = beta0[ib];
	    beta[kd]   = beta[ib];
	  } else {// i-j-i == j-i-j
	    int jd = triplet(jat,iat,jat);
	    theta0[jd] = theta0[ib];
	    theta[jd] =  theta[ib];
	    beta0[jd] =  beta0[ib];
	    beta[jd] =   beta[ib];
	  }

	} else if (slist[0] == "</Keating>") {
          tocontinue = false;
	}
      }
   } while(tocontinue);

  } else {
     ERRORMSG(" KeatingParam file does not exist." << endl);
  }
#endif

  //print(cout);

  return;

}

void KeatingParam::print(ostream& o) const {

  o << "Species information " << endl;
  mySpecies.get(o);
  o << "Pair-interaction parameters " << endl;
  int m = 0;
  for(int i=0; i<nSpecies; i++) {
    for(int j=0; j<nSpecies; j++, m++) {
      o << i << "-" << j 
	<< setw(15) << r0[m] 
	<< setw(15) << alpha[m] << endl;
    }
  }

  o << "Three-body-interaction parameters " << endl;
  m = 0;
  for(int i=0; i<nSpecies; i++) {
    for(int j=0; j<nSpecies; j++) {
      for(int k=0; k<nSpecies; k++, m++) {
	o << i << "-" << j << "-" << k 
	  << setw(15) << theta0[m] 
	  << setw(15) << beta0[m] << endl;
      }
    }
  }

}

void KeatingParam::initializeParam(int n) {
  int n2 = n*n;
  r0 = vector<value_t>(n2,1.0e10); 
  r0sq = vector<value_t>(n2,1.0e10);
  alpha0 = vector<value_t>(n2,0.0e0);
  alpha = vector<value_t>(n2,0.0e0);

  n2 = n*n*n;
  theta0 = vector<value_t>(n2,tetra);
  theta = vector<value_t>(n2,tetra);
  beta0 = vector<value_t>(n2,0.0e0);
  beta = vector<value_t>(n2,0.0e0);
}

//  void KeatingParam::addTwobodyParam(const string& a, const string& b, 
//                                     value_t r, value_t alp) {
//    vmap<string,unsigned>::iterator it= SpeciesID.find(a);
//    vmap<string,unsigned>::iterator jt= SpeciesID.find(b);

//    unsigned i,j;  
//    if(it == SpeciesID.end()) {
//      ERRORMSG(a << " does not belong to any species" << endl);
//      return;
//    } else {
//      i = (*it).second;
//    }
//    if(jt == SpeciesID.end()) {
//      ERRORMSG(b << " does not belong to any species" << endl);
//      return;
//    } else {
//      j = (*jt).second;
//    }
//    int id = i*nSpecies+j;
//    r0[id] = r;
//    alpha0[id] = alp;
//    r0sq[id] = r*r;
//    // NOTE: alpha = (3/8 * alpha0/r0^2)/2  and eV unit
//    alpha[id] = econv*three_over_eight*alp/r0sq[id]/2.0e0;

//    // update the cutoffradius
//    Rsqmax = (Rsqmax < r0sq[id])? r0sq[id]: Rsqmax;

//    if(i != j) {
//      int jd = j*nSpecies+i;
//      r0[jd]     = r0[id];
//      r0sq[jd]   = r0sq[id];
//      alpha0[jd] = alpha0[id];
//      alpha[jd]  = alpha[id];
//    }
//  }
 
//  void KeatingParam::addThreebodyParam(const string& a, const string& b, 
//  				     const string& c, const string& what , 
//                                       value_t be){
//    if(strcasecmp(what.c_str(), "tetra") == 0) {
//      addThreebodyParam(a,b,c,tetra,be);
//    } else {
//      addThreebodyParam(a,b,c,atof(what.c_str()),be);
//    }    
//  }
//  void KeatingParam::addThreebodyParam(const string& a, const string& b, 
//  				     const string& c, value_t the , value_t be){
//    vmap<string,unsigned>::iterator it= SpeciesID.find(a);
//    vmap<string,unsigned>::iterator jt= SpeciesID.find(b);
//    vmap<string,unsigned>::iterator kt= SpeciesID.find(c);
//    unsigned i,j,k;  
//    if(it == SpeciesID.end()) {
//      ERRORMSG(a << " does not belong to any species" << endl);
//      return;
//    } else {
//      i = (*it).second;
//    }
//    if(jt == SpeciesID.end()) {
//      ERRORMSG(b << " does not belong to any species" << endl);
//      return;
//    } else {
//      j = (*jt).second;
//    }
//    if(kt == SpeciesID.end()) {
//      ERRORMSG(b << " does not belong to any species" << endl);
//      return;
//    } else {
//      k = (*kt).second;
//    }
//    // i,j,k maps species

//    int id = nSpecies*(i*nSpecies+j)+k;
//    int ij = nSpecies*i+j, jk = nSpecies*k+j;
//    theta0[id] = the;
//    theta[id] = the*r0[ij]*r0[jk];
//    beta0[id] = be;
  
//    // NOTE: alpha = [3/8 * beta0/(r01*r02)]/2 (counting again), eV unit
//    beta[id] = econv*three_over_eight*be/r0[ij]/r0[jk];

//    if(i != k) { // i-j-k == k-j-i 
//      int kd = nSpecies*(k*nSpecies+j)+i;
//      theta0[kd] = theta0[id];
//      theta[kd]  = theta[id];
//      beta0[kd]  = beta0[id];
//      beta[kd]   = beta[id];
//    }
//    if(i == k) { // i-j-i == j-i-j 
//      int jd = nSpecies*(j*nSpecies+i)+j;
//      theta0[jd] = theta0[id];
//      theta[jd] =  theta[id];
//      beta0[jd] =  beta0[id];
//      beta[jd] =   beta[id];
//    }

//  }   

//  void KeatingParam::print(ostream& o) const {
//    o << "Parameters of Keating-type potentials" << endl;
//    o << "    Number of species " << nSpecies << endl;

//    vmap<string,unsigned>::const_iterator it=SpeciesID.begin(),jt,kt;
//    for(; it != SpeciesID.end(); ++it) {
//      o << "    species " << (*it).first << "  " << (*it).second << endl;  
//    }
//    it=SpeciesID.begin();
//    int i,j,k;
//    for(; it != SpeciesID.end(); ++it) {
//      i = (*it).second;
//      int is = i*nSpecies;
//      for(jt = SpeciesID.begin(); jt != SpeciesID.end(); ++jt) {
//        j = (*jt).second;
//        o << "    twobody " << (*it).first << " " << (*jt).first <<
//              "  "  << r0[is+j] << "  " << alpha0[is+j] << endl; 
//      }
//    }
//    it=SpeciesID.begin();
//    for(; it != SpeciesID.end(); ++it) {
//      i = (*it).second;
//      for(jt = SpeciesID.begin(); jt != SpeciesID.end(); ++jt) {
//        j = (*jt).second;
//          for(kt = SpeciesID.begin(); kt != SpeciesID.end(); ++kt) {
//            k = (*kt).second;
//            int is = nSpecies*(nSpecies*i+j)+k;
//            o << "    threebody " 
//               << (*it).first << " " << (*jt).first << " " << (*kt).first << " "
//               << theta[is] << "  " << beta[is] << "  " << is << endl; 
//          }
//      }
//    }
//  }

/*
char* readLine(char *s, int max, istream &fp){
  char ch;
  int i = 0;
  while ( fp.get(ch) && !( ch == '\n' || ch ==';') )
  {
    if ( ch == '\\' ) // line continuation character
    {
      // check if backslash is followed by a newline
      fp.get(ch);
      if ( ch == '\n' )
      {
        // backslash followed by newline, do nothing
      }
      else
      {
        // backslash not followed by newline
        if ( i < max - 1 )
          s[i++] = '\\';
        if ( i < max - 1 )
          s[i++] = ch; //
      }
    }
    else
    {
      if (i < max - 1)
        s[i++] = ch;
    }
  }
  if (max > 0) s[i] = '\0';  // add terminating NULL

  if ( !(ch == '\n' || ch == ';' ) )
    return NULL;             // return NULL for end of file
  return s;
}
unsigned parsewords(char *inbuf, vector<string>& slist){
  const char* token = "=, \t\n";

  char *tmpstr = new char[strlen(inbuf)+1];
  strcpy(tmpstr, inbuf);
  slist.erase(slist.begin(), slist.end());
  
  int num = 0;
  char* tokenp = strtok(tmpstr, token);
  while(tokenp && tokenp[0] != '#') {
    num++;
    slist.push_back(string(tokenp));
    tokenp = strtok(0,token);
  }

  delete [] tmpstr;
  return num;

}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 310 $   $Date: 2005-03-01 17:01:46 -0600 (Tue, 01 Mar 2005) $
 * POOMA_VERSION_ID: $Id: KeatingParam.cpp 310 2005-03-01 23:01:46Z jnkim $ 
 ***************************************************************************/

