//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Potentials/KeatingPotential.h"
#include "Potentials/KeatingParam.h"
#include "Particle/NNEngine.h"
#include "Utilities/SimpleParser.h"
#include "Utilities/OhmmsInfo.h"
using namespace OHMMS;

KeatingPotential::Scalar_t 
KeatingPotential::evalEnergyNN(const PtclPairList_t& nnlist, Scalar_t rscale) const {
  
  typedef PtclPairList_t::PairData_t    PairData_t;

  vector<SingleParticlePos_t> dr(nnlist.getMaxNadj());
  vector<int> stype(nnlist.getMaxNadj());
  Scalar_t esum(0.0e0);
  for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
    size_type iat(nnlist.I[iL]);
    size_type num_nn(nnlist.nadj(iL));
    size_type n2(0);
    int itype(myPtcl->GroupID[iat]);
    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      Scalar_t r(cur.sep);
      if(r < Rmax) {
	int jtype(cur.Group);
	esum += VSet->f2(VSet->doublet(itype,jtype),r*r);
        stype[n2]=jtype;
        dr[n2]=cur.displ;
        ++n2;
      }
      ++nit;
    }
    for(size_type nnj=0; nnj<n2-1; nnj++) {
      int jtype(stype[nnj]);
      for(size_type nnk=nnj+1; nnk<n2; nnk++) {
        int i_jik(VSet->triplet(jtype,itype,stype[nnk]));
        esum += VSet->f3(i_jik,dot(dr[nnj],dr[nnk]));
      }
    }
  }
  //cout << "Repulsive = " << rep <<  " " << esum << " " << esum + rep << endl;
  return esum;
}

KeatingPotential::Scalar_t 
KeatingPotential::evalForceNN(PtclPairList_t& nnlist, 
			      ParticlePos_t& fv, 
			      unsigned stressmode) const {
  typedef PtclPairList_t::PairData_t    PairData_t;
  ParticleScalar_t& q = myPtcl->E;

  Scalar_t  esum(0.0);
  Tensor_t stressloc(0.0);
  vector<size_type> Loc(nnlist.getMaxNadj());
  vector<size_type> stype(nnlist.getMaxNadj());

  for(size_type iL = 0; iL < nnlist.getLocalNum(); iL++) {
    size_type iat(nnlist.I[iL]);
    int itype(myPtcl->GroupID[iat]);
    Scalar_t elocal(0.0);
    SingleParticlePos_t fon(0.0);
    size_type nnloc(0),n2(0);
    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      Scalar_t r(cur.sep);
      if(r<Rmax) {
        Scalar_t rinv(1.0/r);
        int jtype(cur.Group);
        int bond_ij(VSet->doublet(itype,jtype));
        SingleParticlePos_t dcos(rinv*cur.displ);
        Scalar_t phi;
        SingleParticlePos_t df(VSet->df2(bond_ij,r,phi)*dcos);
        elocal += phi; // energy calculated in
        fon += 2.0*df;
        if(stressmode) stressloc -= outerProduct(df,cur.displ);
        Loc[n2]=nnloc;
        stype[n2]=jtype;
        ++n2;
      }
      ++nit;++nnloc;
    }

    nit = nnlist.begin(iL);
    for(int nnj=0; nnj<n2-1; nnj++) {
      int jtype(stype[nnj]);
      const PairData_t& v2j(*(nit+Loc[nnj]));
      SingleParticlePos_t dr_ij(v2j.displ);
      for(size_type nnk=nnj+1; nnk<n2; nnk++) {
	int i_jik(VSet->triplet(jtype,itype,stype[nnk]));
        const PairData_t& v2k(*(nit+Loc[nnk]));

	SingleParticlePos_t dr_ik(v2k.displ);
	Scalar_t lcos(dot(dr_ij,dr_ik));
	elocal += VSet->f3(i_jik,lcos);
	Scalar_t df3(VSet->df3(i_jik,lcos)); // beta*2*(r_ij*r_ik-beta)

	SingleParticlePos_t dfj(df3*dr_ik);
	SingleParticlePos_t dfk(df3*dr_ij);

	if(stressmode) {
	  stressloc -= outerProduct(dfj, dr_ij);
	  stressloc -= outerProduct(dfk, dr_ik);
	}

	fon += (dfj+dfk);
	fv[v2j.Where]-=dfj;
	fv[v2k.Where]-=dfk;
      }
    }
    fv[iL] += fon;
    q[iat] = elocal;
    esum += elocal;
  }

  nnlist.Stress = stressloc;
  return esum;
}

void KeatingPotential::analize(const PtclPairList_t& nnlist,
			       size_type first, size_type last) { 

  ParticleScalar_t& exx = *(myPtcl->getScalarAttrib("exx"));
  ParticleScalar_t& eyy = *(myPtcl->getScalarAttrib("eyy"));
  ParticleScalar_t& ezz = *(myPtcl->getScalarAttrib("ezz"));
  ParticleScalar_t& exy = *(myPtcl->getScalarAttrib("exy"));
  ParticleScalar_t& eyz = *(myPtcl->getScalarAttrib("eyz"));
  ParticleScalar_t& ezx = *(myPtcl->getScalarAttrib("ezx"));

  typedef PtclPairList_t::PairData_t    PairData_t;

  vector<size_type> Loc(nnlist.getMaxNadj());
  vector<size_type> stype(nnlist.getMaxNadj());
  for(size_type iL = 0; iL < nnlist.getLocalNum(); iL++) {
    size_type iat(nnlist.I[iL]);
    int itype(myPtcl->GroupID[iat]);
    Tensor_t stressloc(0.0);
    size_type n2(0);
    //for(int nnj = 0; nnj< nnlist.nadj(iL); nnj++, nnloc++) {
    //  const PairData_t& cur = nnlist[nnloc];
    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    int nnloc(0);
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      Scalar_t r(cur.sep);
      if(r<Rmax) {
        Scalar_t rinv(1.0/r);
        int jtype(cur.Group);
        int bond_ij(VSet->doublet(itype,jtype));
        SingleParticlePos_t dcos(rinv*cur.displ);
        Scalar_t phi;
        SingleParticlePos_t df(VSet->df2(bond_ij,r,phi)*dcos);
        stressloc -= outerProduct(df,cur.displ);
        Loc[n2]=nnloc;
        stype[n2]=jtype;
        ++n2; 
      }
      ++nit;++nnloc;
    }

    nit = nnlist.begin(iL);
    for(size_type nnj=0; nnj<n2-1; nnj++) {
      int jtype(stype[nnj]);
      //const PairData_t& v2j = nnlist[Loc[nnj]];
      const PairData_t& v2j(*(nit+Loc[nnj]));

      SingleParticlePos_t dr_ij(v2j.displ);
      for(size_type nnk=nnj+1; nnk<n2; nnk++) {
	int i_jik(VSet->triplet(jtype,itype,stype[nnk]));
        const PairData_t& v2k(*(nit+Loc[nnk]));
        //const PairData_t& v2k = nnlist[Loc[nnk]];
	SingleParticlePos_t dr_ik(v2k.displ);
	Scalar_t lcos(dot(dr_ij,dr_ik));
	Scalar_t df3(VSet->df3(i_jik,lcos)); // beta*2*(r_ij*r_ik-beta)
	SingleParticlePos_t dfj(df3*dr_ik);
	SingleParticlePos_t dfk(df3*dr_ij);

	stressloc -= outerProduct(dfj, dr_ij);
	stressloc -= outerProduct(dfk, dr_ik);
      }
    }
    exx[iat]=stressloc(0,0);
    exy[iat]=stressloc(0,1);
    ezx[iat]=stressloc(0,2);
    eyy[iat]=stressloc(1,1);
    eyz[iat]=stressloc(1,2);
    ezz[iat]=stressloc(2,2);
  }
}

/*
KeatingPotentialOutput::KeatingPotentialOutput(KeatingPotential& kp): 
  ref_(kp){ 
}

void KeatingPotentialOutput::reset(const char* fileroot) {
  FileName = fileroot;
  FileName.append(".vff");
}
    
void KeatingPotentialOutput::report(int iter) {

  if(iter!=-1) return;

  typedef KeatingPotential::ParticleScalar_t ParticleScalar_t;
  //Register particle attributes
  ParticleScalar_t* exx =ref_.myPtcl->getScalarAttrib("exx");
  ParticleScalar_t* eyy =ref_.myPtcl->getScalarAttrib("eyy");
  ParticleScalar_t* ezz =ref_.myPtcl->getScalarAttrib("ezz");
  ParticleScalar_t* exy =ref_.myPtcl->getScalarAttrib("exy");
  ParticleScalar_t* eyz =ref_.myPtcl->getScalarAttrib("eyz");
  ParticleScalar_t* ezx =ref_.myPtcl->getScalarAttrib("ezx");

#pragma omp parallel 
  {
    int myID = omp_get_thread_num();
    const NNEngine *nnListEngine = ref_.myPtcl->nnListEngine;
    int first = nnListEngine->first_id(myID);
    int last = nnListEngine->last_id(myID);
    int nitem = last-first;

    ref_.analize(nnListEngine->getPairAttrib(myID),first,last); //,*atomicStress[myID]);
  }//end-of-omp

}
    
void KeatingPotentialOutput::finalize() {


}
    
bool KeatingPotentialOutput::put(xmlNodePtr cur){
  cout << "Calling KeatingPotentialOutput::put(xmlNodePtr cur)" << endl;
  return true;
}
*/

//KeatingPotential::Scalar_t 
//KeatingPotential::evalForceNN(const PtclPairList_t& nnlist, 
//			      int first, int last,
//			      ParticlePos_t& f0, ParticleScalar_t& q, 
//			      Tensor_t& stressloc, unsigned stressmode){
//  
//  int nnloc = 0;
//  Scalar_t  esum=0.0;
//  int nnmax = 2*nnlist.getMaxNadj();
//  vector<SingleParticlePos_t> dcos(nnmax);
//  vector<Scalar_t> dist(nnmax);
//  vector<int> J(nnmax);
//
//  for(int iat=first, iL = 0; iat < last; iat++, iL++) {
//
//    int num_nn = nnlist.nadj(iL);
//    int itype = myPtcl->GroupID[iat];
//    Scalar_t elocal = 0.0;
//    for(int nnj = 0; nnj< num_nn; nnj++, nnloc++) {
//      dcos[nnj] = nnlist.dR[nnloc];
//      dist[nnj]  = nnlist.R[nnloc];
//      J[nnj]  = nnlist.iadj(iL,nnj);
//    }
//
//    for(int nnj=0; nnj<num_nn; nnj++) {
//      int jat = J[nnj];
//      int jtype = myPtcl->GroupID[jat];
//      int bond_ij = VSet->doublet(itype,jtype);
//
//      Scalar_t phi;
//      SingleParticlePos_t df = VSet->df2(bond_ij, dist[nnj], phi)*dcos[nnj];
//      elocal += phi; // energy calculated in
//       
//      if(stressmode) {stressloc -= dist[nnj]*outerProduct(df, dcos[nnj]);}
//      f0[iat] += 2.0*df;
//
//      SingleParticlePos_t dr_ij = dist[nnj]*dcos[nnj];
//       
//      for(int nnk=nnj+1; nnk<num_nn; nnk++) { // j<k
//	int kat = J[nnk];
//	int ktype = myPtcl->GroupID[kat];
//	// j-i-k triplet index
//	int i_jik = VSet->triplet(jtype,itype,ktype);             
//
//	SingleParticlePos_t dr_ik = dist[nnk]*dcos[nnk];
//	Scalar_t lcos = dot(dr_ij,dr_ik);
//	elocal += VSet->f3(i_jik,lcos);
//	Scalar_t df3 = VSet->df3(i_jik,lcos); // beta*2*(r_ij*r_ik-beta)
//	 
//	SingleParticlePos_t dfj = df3*dr_ik;
//	SingleParticlePos_t dfk = df3*dr_ij;
//	 
//	if(stressmode) {
//	  stressloc -= outerProduct(dfj, dr_ij);
//	  stressloc -= outerProduct(dfk, dr_ik);
//	}
//	f0[iat] += (dfj+dfk);
//	f0[jat] -= dfj;
//	f0[kat] -= dfk;
//      }
//    }
//    q[iat] = elocal;
//    esum += elocal;
//  }
//  return esum;
//}



/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: KeatingPotential.1.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
