//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003 by Jeongnim Kim and Thomas Lenosky
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Potentials/MEAMPotential.h"
#include "Potentials/MEAMParam.h"
#include "ParticleBase/ParticleMessage.h"
#include "Utilities/OhmmsInfo.h"
using namespace OHMMS;


MEAMPotential::MEAMPotential(): mySpecies(NULL){
  CollectVisitingForces = true;
  myName = "MEAM";
}

MEAMPotential::MEAMPotential(int argc, char** argv): 
mySpecies(NULL){
  CollectVisitingForces = true;
  myName = "MEAM";
}

MEAMPotential::~MEAMPotential() {
  for(int i=0; i<MEAMSet.size(); i++) delete MEAMSet[i];
}

void MEAMPotential::readMEAMPOT(const string& fname) {

  size_type imass   = mySpecies->addAttrib("mass");
  size_type irmax   = mySpecies->addAttrib("rmax");

  ifstream fin(fname.c_str());
  vector<string> slist;
  bool tofindend = true;
  int nspecies = 0;
  while(tofindend) {
    getwords(slist, fin);
    if(slist[0] == "species") {
      size_type newid =mySpecies->addSpecies(slist[1].c_str());
      (*mySpecies)(imass,newid)=atof(slist[2].c_str());
      //MEAMSet.push_back(new MEAMParam(slist[3].c_str()));
      MEAMSet.push_back(new MEAMParam(fname.c_str()));
      Rmax = max(Rmax, MEAMSet[nspecies]->rcutmax());
      (*mySpecies)(irmax,newid) = Rmax*0.5;
    } else if (slist[0] == "end") {
      tofindend = false;
    }
  }
  LOGMSG("\tmaxmimum interaction range = " << Rmax << " AA");
  cout << "Size of two-body " << sizeof(MEAM2Body<double,3>) << endl;
  cout << "Size of three-body " << sizeof(MEAM3Body<double>) << endl;
}


MEAMPotential::Scalar_t 
MEAMPotential::evalEnergyNN(const PtclPairList_t& nnlist, Scalar_t rscale) const {

  typedef PtclPairList_t::PairData_t    PairData_t;
  // storing three-body function
  size_type max_est = nnlist.getMaxNadj();
  vector<Scalar_t> sf_f(max_est);
  vector<SingleParticlePos_t> sf_dcos(max_est);
  Scalar_t r, esum(0.0);
  for(size_type iL=0; iL<nnlist.getLocalNum(); iL++) {
    int n3(0);
    Scalar_t emaX(0.0);
    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      r = cur.sep;
      // check all the nearest atoms within a cutoff of phi
      if(r < MEAMSet[0]->clsgnecut) {
	esum += 0.5e0*MEAMSet[0]->phi.splint(r); //phi(r),double counting
	
	if(r < MEAMSet[0]->clsrhonecut) {
	  emaX += MEAMSet[0]->xrho[0]->splint(r); 
	}
	if(r < MEAMSet[0]->clstricut1) {
	  sf_f[n3]  = MEAMSet[0]->xf[0]->splint(r); 
          Scalar_t rinv(1.0/r);
	  sf_dcos[n3] = rinv*cur.displ;
	  n3++;
	}
      }
      ++nit;
    } // while(nit != nit_end)
    
    // calculating x for U(x) for the embedded atom term
    // NOTE that terms from rho have been already added to emaX
    for(int nj=0; nj < n3-1; nj++) {
      for(int nk=nj+1; nk < n3; nk++) {
	Scalar_t lcos = dot(sf_dcos[nj], sf_dcos[nk]);
	emaX += (sf_f[nj]*sf_f[nk]*MEAMSet[0]->xg[0]->splint(lcos));
      }
    }
    
    esum += MEAMSet[0]->xu[0]->splint(emaX); // adding u(x) to elocal
  }
  return esum; // return the energy
}

bool MEAMPotential::get(ostream& os) const { 
  return true;
}
bool MEAMPotential::put(istream&) {return true;}
void MEAMPotential::reset() {

}

bool MEAMPotential::put(xmlNodePtr cur){

  if(myPtcl ==0) {
    ERRORMSG("A particle set needs to be assigned before initialization")
    return false;
  }

  string PotFile("MEAMPOT");
  if(cur != NULL) {
    const xmlChar* fptr=xmlGetProp(cur, (const xmlChar *)"file");
    if(fptr == NULL) {
      fptr=xmlGetProp(cur, (const xmlChar *)"src");
    }
    if(fptr) PotFile = (const char*)fptr;
    else {
      cur = cur->children;
      while(cur != NULL) {
        string cname((const char*)cur->name);
        if(cname == "filename") {
          putContent(PotFile,cur);
        }
        cur = cur->next;
      }
    }
  }

  mySpecies = &(myPtcl->getSpeciesSet());
  readMEAMPOT(PotFile);
  return true;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: MEAMPotential.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

