//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file PotentialBase.h
 * @brief Declaration of the base class for a Potential
 */
#ifndef OHMMS_POTENTIALBASE_H
#define OHMMS_POTENTIALBASE_H

#include "OhmmsData/OhmmsElementBase.h"
#include "OhmmsData/RecordProperty.h"
#include "OhmmsPETE/OhmmsMatrix.h"
#include "Particle/ParticleSet.h"

namespace OHMMS {

  /** Abstract class. Derived classes implement inter-atomic potential.
   *
   * Potential classes are not templated, i.e., they only work with the Particle_t defined in Particle/ParticleSet.h
   */
  class PotentialBase: public OhmmsElementBase {

  public:

    typedef ParticleSet                     Particle_t;
    typedef Particle_t::size_type           size_type;
    typedef Particle_t::Scalar_t            Scalar_t;
    typedef Particle_t::SingleParticlePos_t SingleParticlePos_t;
    typedef Particle_t::Tensor_t            Tensor_t;

    typedef Particle_t::ParticleIndex_t     ParticleIndex_t;
    typedef Particle_t::ParticleScalar_t    ParticleScalar_t;
    typedef Particle_t::ParticlePos_t       ParticlePos_t;

    typedef Particle_t::PtclPairList_t      PtclPairList_t;
    typedef Matrix<Scalar_t>                DynMatrix_t;

    ///default constructor
    PotentialBase();   

    ///destructor
    virtual ~PotentialBase();

    /** Assign a particle set, virtual function for composite pattern
     *\param ptclset the ParticleSet this potential operates on
     */
    virtual void setParticleSet(Particle_t* ptclset) {myPtcl=ptclset;}

    /** Return the pointer to the particle set
     */
    Particle_t* getParticleSet() {return myPtcl;}

    /** evalute a total energy of a given configuration
     *  \return Total potential energy for a given configuration myPtcl->R
     */
    virtual Scalar_t evalEnergy() = 0;

    /** evaluate forces and particle energies
     *\param fv ParticlePos_t for forces calculated.
     *\param fmode action mode
     *\return Total potential energy for a given configuration myPtcl->R
     */
    virtual Scalar_t evalForce(ParticlePos_t& fv, unsigned fmode=0) = 0;


    /** Obtain analytic Hessian (2nd derivate of potentials) for phonon spectrum.
     *\param dm Dynamic Matrix DynMatrix_t
     *\return true if dynamic matrix calculations are done by potential 
     *
     *Provide an emypty function which return 0.0e0 if the 
     *2nd derivatives are not readily available.
     */
    virtual bool evalDynamicMatrix(DynMatrix_t&) = 0;

    /** Set the current position to be pos
     *@param pos position array
     */
    virtual void setCurrentPos(ParticlePos_t& pos)=0;

    /** Set the current position to be pos
     *@param pos position array
     *@param wrap bool, if true, call myPtcl->applyBConds()
     *@param update bool, if true (default), update the cell information.
     */
    virtual void setCurrentPos(ParticlePos_t& pos, bool wrap,
        bool updatecell=true)=0;
    
    /** Set the current position to be pos and set the displacement vector dr
     *@param pos position array
     *@param dr  displacement array
     */
    virtual void setCurrentPos(const ParticlePos_t& pos, 
        const ParticlePos_t& dr)=0;

    /** A derived class can provide a specialized RecordProperty.
     *@return RecordProperty 
     *
     *Simply return a pointer to a report engine for
     *the physical quantities associated with the potential.
     *Normally, myRecorder = 0.
     */
    RecordProperty* getRecorder() { return myRecorder;}

    ///return the stress tensor
    Tensor_t getStress() { return Stress;}

    ///returns the maximum interaction radius
    Scalar_t getRmax() const { return Rmax;}

    ///sets the maximum interaction radius
    void setRmax(Scalar_t rmax) { Rmax = rmax;}

    bool collectVisitingForces() const { return CollectVisitingForces;}

    ///reset the ParticleSet 
    virtual void resetParticleSet(Particle_t* ptclset) = 0;

    xmlNodePtr getLayoutXmlNode() { return LayoutXmlNode;}

  protected:

    ///true, if the visiting forces need to be assembled
    bool CollectVisitingForces;

    /// Maximum radius for the interaction calculated 
    Scalar_t Rmax;

    /// Maximum radius to connect the grids
    Scalar_t GridRadius;

    /// Total stress
    Tensor_t Stress;

    /// Reference to a particle set
    Particle_t* myPtcl;
  
    ///Reference to the NearestNeighborBase
    NNEngine* nnListEngine;

    /// An engine to record potential-dependent quantities
    RecordProperty* myRecorder;

    /// xml node to define layout
    xmlNodePtr PotXmlNode;
    /// xml node to define layout
    xmlNodePtr LayoutXmlNode;

    private:
    /// disable copy constructor
    PotentialBase(const PotentialBase&) { }
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 383 $   $Date: 2005-09-05 15:30:56 -0500 (Mon, 05 Sep 2005) $
 * $Id: PotentialBase.h 383 2005-09-05 20:30:56Z jnkim $ 
 ***************************************************************************/
