//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/** @file PotentialCollection.1.cpp
 * @brief Declare member functions with large computations.
 *
 * OpenMP/MPI directives are used by these member functions.
 * Some compilers have trouble compiling large cpp files and
 * it is not a bad idea to compile multiple sizable cpp files.
 */
#include <numeric>
#include "Potentials/PotentialCollection.h"
#include "Potentials/ShortRangePotential.h"
#include "Particle/NNEngine.h"
#include "Message/OpenMP.h"
#include "Utilities/OhmmsInfo.h"

namespace OHMMS {


  PotentialCollection::Scalar_t 
  PotentialCollection::evalEnergy() {

    Scalar_t esum = 0.0e0;
    for(int l=0; l<H_long.size(); l++) 
      esum += H_long[l]->evalEnergy();

#pragma omp parallel 
    {
      int np = nnListEngine->size();
#pragma omp for nowait 
      for(int myID=0; myID<np; myID++) {
        NNEngine *nnListEngine = myPtcl->nnListEngine;
        Scalar_t e=0.0;
        for(int s=0; s<H_short.size(); s++) {
  	  e += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID),1.0);
        }
        nnListEngine->getPairAttrib(myID).Energy=e;
      }
    }//end-of-omp
   
    for(int myID=0; myID<nnListEngine->size(); myID++) {
      esum += nnListEngine->getPairAttrib(myID).Energy;
    }

    return nnListEngine->collect(esum);
  }


  PotentialCollection::Scalar_t
  PotentialCollection::evalForce(ParticlePos_t& f0, unsigned fmode) {

    Scalar_t esum = 0.0;
#pragma omp parallel reduction(+:esum)
    {
      int np = nnListEngine->size();
      unsigned stressmode=fmode&1;
#pragma omp for
      for(int myid=0; myid<np; myid++) {

        if(stressmode&2) 
          nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(myid));

        ParticlePos_t& fv=nnListEngine->getPairAttrib(myid).Fv;
        int nv = nnListEngine->getPairAttrib(myid).Visitors;
        for(int iL=0; iL<nv; iL++) fv[iL]=0.0;
        // nnListEngine->getPairAttrib(myid).Fv[iL]=0.0;

        Scalar_t e=0;
        for(int s=0; s<H_short.size(); s++) {
          e += H_short[s]->evalForceNN(nnListEngine->getPairAttrib(myid),fv, stressmode);
          //e += H_short[s]->evalForceNN(nnListEngine->getPairAttrib(myid),f0,stressmode);
        }
        nnListEngine->getPairAttrib(myid).Energy = e;
        esum = e;
      }

#pragma omp for 
      for(int myid=0; myid<np; myid++) {
        nnListEngine->collectForce(f0,myid);
      }
    }

    if(stressmode) {
      Stress = 0.0;
      for(int myid=0; myid<nnListEngine->size(); myid++) {
        Stress += nnListEngine->getPairAttrib(myid).Stress;
      }
    }

    PotEnergy = esum;
    return esum;
  }

  bool PotentialCollection::evalDynamicMatrix(DynMatrix_t&) {
    return false;
  }
  
  PotentialCollection::Scalar_t 
  PotentialCollection::evalForceAndPressure(ParticlePos_t& f, 
					    ParticleScalar_t& q, 
					    Tensor_t& stress, 
					    Scalar_t& dedv) {


    Scalar_t esum = evalForce(f,q,stress,1), eplus=0.0,eminus=0.0;

    const Scalar_t dilation =    1.001000000000000;
    const Scalar_t contraction = 0.999000000000000;

#pragma omp parallel reduction(+:eplus,eminus)
    {
      const NNEngine *nnListEngine = myPtcl->nnListEngine;
      int myID = omp_get_thread_num();
      for(int s=0; s<H_short.size(); s++) {
	eplus += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID), dilation);
      }
      for(int s=0; s<H_short.size(); s++) {
	eminus += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID), contraction);
      }
    }//end-of-omp

    //FIX IT FOR MPI
    //eplus = myPtcl->getNNEngine()->collect(eplus);
    //eminus = myPtcl->getNNEngine()->collect(minus);
    Scalar_t dV = (pow(dilation,3)-pow(contraction,3))*myPtcl->Lattice.Volume;
    dedv = (eplus-eminus)/2.0/dV;
    return esum;
  }  

  PotentialCollection::Scalar_t
  PotentialCollection::getEnergy() {
    Scalar_t esum = 0.0;
    for(int myID=0; myID<nnListEngine->size(); myID++) {
      esum += nnListEngine->getPairAttrib(myID).Energy;
    }
    return esum;
  }

  void 
  PotentialCollection::setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell) {
    //First update ParticleSet::curR to be inside the unit cell
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(myPtcl->Lattice,myPtcl->curR, myPtcl->ID);
    myPtcl->applyBConds(true,myPtcl->curR.getUnit());
  }
  
  void 
  PotentialCollection::setCurrentPos(ParticlePos_t& pos) {
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(myPtcl->Lattice,myPtcl->curR, myPtcl->ID);
  }
  
  void 
  PotentialCollection::setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& dr) {
    nnListEngine->updatePairsPerThread(dr,
        nnListEngine->getPairAttrib(omp_get_thread_num()));
    //nnListEngine->updatePairs(dr);
  }

}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 310 $   $Date: 2005-03-01 17:01:46 -0600 (Tue, 01 Mar 2005) $
 * $Id: PotentialCollection.OMP.cpp 310 2005-03-01 23:01:46Z jnkim $ 
 ***************************************************************************/

