//////////////////////////////////////////////////////////////////
// (c) Copyright 2004- by Jeongnim Kim
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_SWPARAMETERS_H
#define OHMMS_SWPARAMETERS_H

namespace OHMMS {
  /** storage for the temporary data for two- and three-body interactions
   */
  template<class T>
  struct SWPair {
    int J;
    T sep, rinv, g,dg;
    TinyVector<T,3> dcos;
    SWPair(){}
  };

  /*!\class SWParam
   * \brief Parameters specific for SW potentials. Reduced to pair-wise form
   */
   template<class T>
   struct SWParam {
     int p,q;
     T A0, B0, a0, A, B, a; 
     T Sigma, Gamma, Epsilon, Lambda0, Lambda;
     SWParam() { // initialized by Silicon

       p=4;q=0;
       Epsilon = 2.315; // rescale for the cohesive energy of the diamond 4.63 eV
       //Epsilon = 2.1672; // origial value
       Sigma = 2.09510e0;
       Gamma = 1.20e0;
       A0 = 7.0496;
       B0 = 0.60222;
       a0=1.8;
       Lambda0 = 21.0;
       reset();
       /*
       A = 94.6804405e0; // A = 189.360881e0, double counting
       B = 8.159861385e0;// B = 16.31972277e0, double counting
       a = 3.77118e0;          // a*sigma a = 1.8
       Lambda = sqrt(48.61499998e0);
       */
     }

     SWParam(const SWParam<T>& swA, const SWParam<T>& swB) {
       p = swA.p; q = swA.q;
       Gamma = swA.Gamma;
       a0=swA.a0;
       Epsilon = sqrt(swA.Epsilon*swB.Epsilon);
       Sigma = (swA.Sigma+swB.Sigma)*0.5;
       A0 = sqrt(swA.A0*swB.A0);
       B0 = sqrt(swA.B0*swB.B0);
       Lambda0 = sqrt(swA.Lambda0*swB.Lambda0);
       reset();
     }

     inline void reset() {
       A = Epsilon*A0*B0*pow(Sigma,p)*0.5;
       B = Epsilon*A0*pow(Sigma,q)*0.5;
       a = a0*Sigma;
       Lambda = sqrt(Epsilon*Lambda0);
     }

     SWParam(xmlNodePtr cur) {
       put(cur);
     }

     ~SWParam(){}

     void put(xmlNodePtr cur) {
       ParameterSet param;
       param.add(Epsilon,"epsilon","eV");
       param.add(Sigma,"sigma","AA");
       param.add(Gamma,"gamma","none");
       param.add(Lambda0,"lambda","none");
       param.add(A0,"A","none");
       param.add(B0,"B","none");
       param.add(a0,"a","none");
       param.add(p,"p","int");
       param.add(q,"q","int");
       param.put(cur);
       reset();
     }

     inline T evaluate(T r, T rinv, T& g3) const {
       T rmainv= 1.0/(r - a);
       T tmp1 = A*pow(rinv,p);   // (B/r)^p
       T tmp2 = B*pow(rinv,q);   // (1/r)^q
       T tmp3 = exp(Sigma*rmainv);    // exp(sigma/(r-a))
       g3 =  Lambda*pow(tmp3,Gamma);
       return tmp3*(tmp1-tmp2);
     }

     bool inside(T r) const { return r<a;}

     inline T evaluate(int j, T r, const TinyVector<T,3>& dr, SWPair<T>& v2, TinyVector<T,3>& df) 
       const {

       T rinv(1.0e0/r);
       T rmainv(1.0e0/(r - a));
       T twoa(A*pow(rinv,p));   // A*(1/r)^p
       T twob(B*pow(rinv,q));   // B*(1/r)^q
       T fc(exp(Sigma*rmainv)); // exp(sigma/(r-a))
       T e2((twoa-twob)*fc);
       T dtwoa(twoa*rinv*p);
       T dtwob(twob*rinv*q);
       T dfc(Sigma*rmainv*rmainv);   // sigma/(r-a)^2
       T dV2j((dtwob-dtwoa)*fc - e2*dfc);
       T gtmp(Lambda*pow(fc,Gamma));

       v2.J=j;
       v2.sep=r;
       v2.rinv = rinv;
       v2.g = gtmp;
       v2.dg = -gtmp*dfc*Gamma;//fc*sig*gam/(r-a)^2
       v2.dcos = rinv*dr;
       df = dV2j*v2.dcos;
       return e2;
     }
   };
}

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 308 $   $Date: 2005-02-21 09:21:44 -0600 (Mon, 21 Feb 2005) $
 * $Id: SWParam.h 308 2005-02-21 15:21:44Z jnkim $ 
 ***************************************************************************/
