//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Richard Hennig and Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifdef USE_OPENMP
#include <omp.h>
#endif

#include <map>
#include <vector>
using namespace std;
#include "Potentials/AlloyEAMPotential.h"
#include "Potentials/CubicSplineType.h"
//#include "ScaleFunctors/CubicSplineDiff.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"
#include <string.h>

/*! \class AlloyEAMParam
 *  \brief Contain spline parameters for AlloyEAM potentials.
 *         Contains various splines for a AlloyEAM potential
 *  \note Using SplineType defined in CubicSplineType.h
 */
struct AlloyEAMParam {
  typedef double Scalar_t;
  Scalar_t clsphircut;    //!<Range of pair potential
  Scalar_t clsrhorcut;    //!<Range for rho(r)
  Scalar_t clsfrhocut;    //!<Range for f(rho)
  int      nspecies;          //Number of species

  ///////////////////////////////////////
  // Original spline function
  ///////////////////////////////////////
  //typedef SplineType ScaleFunction_t;
  ///////////////////////////////////////

  typedef CubicSplineRegGrid ScaleFunction_t;

  ////////////////////////////////////////////
  //If inline works well, this might work better
  ////////////////////////////////////////////
  //typedef CubicSplineDiff<Scalar_t> ScaleFunction_t;
  ////////////////////////////////////////////


  vector<ScaleFunction_t> phi, f, rho;  // SplineTypeAlloyEAM phi, f, rho;

  AlloyEAMParam(const char *fname, int num) {
    readAlloyEAMParam(fname, num);

    clsphircut = phi[0].rmax(); //X[phi[0].Npt];
    clsrhorcut = rho[0].rmax(); //X[rho[0].Npt];
    clsfrhocut = f[0].rmin(); // f[0].X[1];
    for (int i=1; i<num; i++) {
      clsphircut = max(clsphircut, phi[i].rmax());
      clsrhorcut = max(clsrhorcut, rho[i].rmax());
      clsfrhocut = min(clsfrhocut, f[1].rmin());
    }
  }
  void readAlloyEAMParam(const char *fname, int num);

  inline Scalar_t rcutmax() { 
    return (clsphircut > clsrhorcut)? clsphircut : clsrhorcut;
  }
  inline int bond(int i, int j) const {
    return i*nspecies+j;
  }
};

/*! \fn AlloyEAMPotential::AlloyEAMPotential()
 *  \brief Default constructor of AlloyEAMPotential
 */
AlloyEAMPotential::AlloyEAMPotential(): EAMSet(NULL){
  Name = "EAM";
}

/*! \fn AlloyEAMPotential::AlloyEAMPotential(ThisParticle_t* ptclset)
 *  \param ptclset A pointer to a OHMMS::Particle_t
 */
AlloyEAMPotential::AlloyEAMPotential(ThisParticle_t* ptclset)
  :PotentialBase(ptclset), EAMSet(NULL)
{
  Name = "EAM";
}

/*! \fn AlloyEAMPotential::AlloyEAMPotential(int argc, char** argv) 
 *  \param argc number of arguments
 *  \param argv list of arguments
 */
AlloyEAMPotential::AlloyEAMPotential(int argc, char** argv): EAMSet(NULL) 
{
  Name = "EAM";
  int i=1;
  while(i<argc) {options.push_back(argv[i++]);}
}

AlloyEAMPotential::AlloyEAMPotential(vector<string>& arglist):EAMSet(NULL)
{
  options.insert(options.end(), arglist.begin(), arglist.end());
}

AlloyEAMPotential::~AlloyEAMPotential() {
  if(EAMSet) delete EAMSet;
}

void AlloyEAMPotential::initialize() {

  string potfile("AlloyEAMPOT");
  int i=0;
  while(i<options.size()) {
    if(options[i] == "--potfile") { potfile = options[++i];}
    i++;
  }

  Rmax = 0.0;
  SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
  int imass = SpeciesCollection::addAttrib("mass");
  int irmax = SpeciesCollection::addAttrib("rmax");

  cerr << "Data file: " << potfile.c_str() << "\n";

  ifstream fin(potfile.c_str());
  if(!fin) { 
    cerr << potfile.c_str() << " is not found in this directory.\n";
    cerr << "Exiting very ungracefully.\n";
    exit(1);
  }
  cerr << "Read species parameters from file "<< potfile.c_str() << "\n";

  vector<string> slist;
  int nspecies = 0;
  do {
    while ((getwords(slist, fin))==0);
    if (slist[0] == "species") {
      int newid =SpeciesCollection::addSpecies(slist[1].c_str());
      (*mySpecies)(imass,newid)=atof(slist[2].c_str());
      nspecies ++;

      DEBUGMSG("New Species: " << slist[1].c_str() << " (" << newid << ") " << "m= " << (*mySpecies)(imass,newid))
    }
  } while(slist[0] != "<potential>");

  DEBUGMSG("Number of species: " << nspecies)

  if(EAMSet) {
    delete EAMSet;
  }
  EAMSet = new AlloyEAMParam(potfile.c_str(), nspecies);

  Rmax = max(Rmax, EAMSet->rcutmax());

  DEBUGMSG("Set Rmax: " << Rmax )

  for (int i=0; i<nspecies; i++) (*mySpecies)(irmax,0) = Rmax*0.5;
}

Scalar_t 
AlloyEAMPotential::e(const ParticlePos_t& R, int i, int myID) 
{
//    const PtclPairList_t* nnlist = nnListEngine->getPairAttrib();
//    int nn = nnlist->nadj(i);
//    Scalar_t  r;
//    Scalar_t  rho = 0.0e0, elocal =0.0e0;
//    SingleParticlePos_t dr;

//    for(int jL=0; jL<nn; ++jL) 
//      {
//        int j = nnlist->iadj(i,jL);  //atomic index of jL-th nn of i

//        if(R.InUnit) 
//  	dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
//        else
//  	dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

//        r = sqrt(myPtcl->Lattice.Dot(dr,dr));

//        // check all the nearest atoms within a cutoff of phi
//        if (r<EAMSet->clsphircut) {
//  	elocal += 
//  	  0.5*EAMSet->phi[EAMSet->bond(myPtcl->IonID[i], myPtcl->IonID[j])].splint(r,0);
//  	//phi(r),double counting
//        }

//        if (r<AlloyEAMSet[0]->clsrhorcut) {
//  	rho += EAMSet[0]->rho[myPtcl->IonID[j]].splint(r,0);
//        }
//      }

//    elocal += EAMSet[0]->f[myPtcl->IonID[i]].splint(rho,0);

//    return elocal; // return the energy
  return 0.0;
}

Scalar_t 
AlloyEAMPotential::evalEnergy() {

  //NOTE PotentialBase::evalEnergy interface is modified 09-27-2001
  //PotentialBase::Scalar_t 
  //AlloyEAMPotential::evalEnergy(SimpleParticle& inSys) {

  getPairList(); // using a pair-list


  Scalar_t  esum =0.0e0;

#if defined(SGI) && defined(USE_OPENMP)
#pragma omp parallel reduction(+:esum)
 {
    int myID = omp_get_thread_num();
#else
#pragma omp parallel for reduction(+:esum)
  for(int myID=0; myID<nnListEngine->size(); myID++)  {
#endif

    const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

    Scalar_t  r, elocal=0, rho;
    int nnloc = 0;

    for(int iL = 0; iL < nnlist.getLocalNum(); iL++) {

      int iG = nnlist.I[iL]; // get the index, physical index
      int itype = myPtcl->IonID[iG];

      int num_nn = nnlist.nadj(iL);
      //emaX = 0.0e0;
      rho = 0.0e0;
      for(int nni = 0; nni< num_nn; nni++, nnloc++) {

        int jG = myPtcl->ID[nnlist.iadj(iL,nni)];
	int jtype = myPtcl->IonID[jG];

	//if(inSys.IonID[j] == AlloyEAMSpecies) continue;
	r = nnlist.R[nnloc];

	// check all the nearest atoms within a cutoff of phi
	if (r<EAMSet->clsphircut) {
	  elocal += 
	    0.5*EAMSet->phi[EAMSet->bond(itype,jtype)].splint(r);
	  //phi_ij(r), double counting
	}

	if (r<EAMSet->clsrhorcut) {
	  rho += EAMSet->rho[jtype].splint(r);
	}
      }
      elocal += EAMSet->f[itype].splint(rho); 
    }
    esum += elocal;
  }
  return esum; // return the energy
}


Scalar_t 
AlloyEAMPotential::evalForce(ParticlePos_t& f0, ParticleScalar_t& q,
                         Tensor_t& stress, unsigned stressmode) {
  //NOTE PotentialBase::evalForce change
  //PotentialBase::Scalar_t 
  //AlloyEAMPotential::evalForce(SimpleParticle& inSys, 
  //			ParticlePos_t& f0, ParticleScalar_t& q,
  //			Tensor_t& stress, unsigned stressmode) {

  DEBUGMSG("Entering AlloyEAMPotential::evalForce.\n")

  getPairList(); // using a pair-list

  Scalar_t esum = 0.0e0;
  Scalar_t p00 = 0.0e0, p11=0.0e0, p22 = 0.e0, p01 = 0.0, p02=0.0, p12 = 0.0;

#if defined(SGI) && defined(USE_OPENMP)
#pragma omp parallel reduction(+:esum,p00,p11,p22,p01,p02,p12)
 {
    int myID = omp_get_thread_num();
#else
#pragma omp parallel for reduction(+:esum,p00,p11,p22,p01,p02,p12)
  for(int myID=0; myID<nnListEngine->size(); myID++)  {
#endif

    const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

    // storing three-body function
    int max_est = 2*nnlist.getTotNadj()/nnlist.getLocalNum();
    vector<int> srho_id(max_est);
    vector<Scalar_t> srho_r(max_est);
    vector<Scalar_t> srho_rho(max_est);
    vector<Scalar_t> srho_drho(max_est);
    vector<SingleParticlePos_t> srho_dr(max_est);

    Scalar_t esumloc = 0.0;
    Tensor_t stressloc;

    map<int,SingleParticlePos_t> fv;   //visiting forces on non-local particles 
    int first = nnListEngine->first_id(myID);
    int last  = nnListEngine->last_id(myID);
    int nnloc = 0;

    Scalar_t r, val, deriv;

    for(int iL = 0; iL < nnlist.getLocalNum(); iL++) {

      int iG = nnlist.I[iL]; // get the index, physical index
      int itype = myPtcl->IonID[iG];

      int n2 =0;
      Scalar_t elocal = 0.0;
      int num_nn = nnlist.nadj(iL);

      for(int nni = 0; nni< num_nn; nni++, nnloc++) {

	int jg = nnlist.iadj(iL,nni);
        int jG = myPtcl->ID[jg];
	int jtype = myPtcl->IonID[jG];

        r = nnlist.R[nnloc];
	//if(inSys.IonID[j] != AlloyEAMSpecies) continue; 
	// going thru all the nn atoms within a cutoff of phi
	if(r < EAMSet->clsphircut) {

	  deriv = EAMSet->phi[EAMSet->bond(itype,jtype)].splint(r, val);

	  // phi(r)
	  SingleParticlePos_t dr = nnlist.dR[nnloc];
	  elocal += val*0.5e0;                     // double counting
	  SingleParticlePos_t df = 0.5e0*deriv*dr; // double counting
	  //cout << r << " phi(r)= " << val << "  phi'(r)=" << deriv << "\n";
	  
	  f0[iG] += 2*df;
	  //f0[jG] -= df;// OpenMP

	  //cout << df << "\n";

	  if(stressmode) { stress -= r*outerProduct(df, dr);}

	  if(r < EAMSet->clsrhorcut) { 
	    // save data associated with rho(r) 
	    deriv = EAMSet->rho[jtype].splint(r, val);
	    srho_id[n2] = jg;      // ID
	    srho_r[n2] = r;       // distance
	    srho_rho[n2] = val;   // rho(r)
	    srho_drho[n2] = deriv;// d rho(r)/ dr
	    srho_dr[n2] = dr;     // (r[j]-r[i])/r
	    n2++;
	  }
	}
      }

      // calculating x for U(x) for the embedded atom term
      Scalar_t emaX = 0.e0;
      for(int nj=0; nj < n2; nj++) {
	emaX += srho_rho[nj];
      }

      // Done with emaX, calculate the EMA term
      // du/dx, used for a force prefactor

      Scalar_t dux = EAMSet->f[itype].splint(emaX, val);
      elocal +=val; 

      for(int nj=0; nj < n2; nj++) { // forces from drho
	int jg = srho_id[nj];
	int jG = myPtcl->ID[jg];

	SingleParticlePos_t df = dux * srho_drho[nj] * srho_dr[nj];
	f0[iG] += df;

#ifdef USE_OPENMP
        // Forgot to add - (unary) operator, substract later for fv
	map<int,SingleParticlePos_t>::iterator it;
	if(jg < first || jg >= last) {//jg logically belongs to other node
	  it = fv.find(jG); 
	  if(it == fv.end()) 
	    fv[jG] = df;
	  else  
	    (*it).second += df;
	} else { // jG logically belongs to this node
	  f0[jG] -= df;
	}
#else
	f0[jG] -= df;
#endif

	//cout << df << "\n";

	if(stressmode) { 
	  stressloc -= srho_r[nj]*outerProduct(df, srho_dr[nj]);
	}
      }

      q[iG] = elocal;
      esumloc += elocal;
    }

    esum += esumloc;
    if(stressmode) {
      p00 += stressloc(0,0);
      p11 += stressloc(1,1);
      p22 += stressloc(2,2);
      p01 += stressloc(0,1);
      p02 += stressloc(0,2);
      p12 += stressloc(1,2);
    }
#pragma omp critical
    {    
#ifdef USE_OPENMP
      map<int,SingleParticlePos_t>::iterator it = fv.begin();
      while(it != fv.end()) {
	f0[(*it).first] -= (*it).second;
	it++;
      }
#endif
    }
  }

  if(stressmode)
    stress += Tensor_t(p00,p01,p02,-p01,p11,p12,-p02,-p12,p22);

  // cout << "Force= " << f0[1][0] << "\n";
  // cout << "  Pressure= " << psum << "\n";
  // cout << "  Energy= " << esum << "\n";
  return esum;
}

bool
AlloyEAMPotential::evalDynamicMatrix(DynMatrix_t&) {

  return false;

}


void AlloyEAMParam::readAlloyEAMParam(const char *fname, int nspec) {
  Scalar_t econv = 1.0;
  nspecies = nspec;

  ifstream fin(fname);
  if(!fin) { 
    cerr << fname << " is not found in this directory.\n";
    cerr << "Exiting very ungracefully.\n";
    exit(1);
  }
  cerr << "Read potential parameter from file "<< fname << "\n";

  phi = vector<ScaleFunction_t>(nspecies*nspecies);
  rho = vector<ScaleFunction_t>(nspecies);
  f   = vector<ScaleFunction_t>(nspecies);

  vector<string> slist;

  do {
    while ((getwords(slist, fin))==0);

    if (slist[0] == "<eunit>") {
      if (slist[1] == "K")   econv = 8.617385692E-5;
      if (slist[1] == "eV")  econv = 1.0;
      if (slist[1] == "erg") econv = 6.241506363E11;
      cerr << "Energy units: " << slist[1].c_str() 
	   << " (factor: " << econv << ")\n";
    }

    if (slist[0] == "<potential>") {
      if (slist[1] == "pair") {
	int i = atoi(slist[2].c_str())-1;
	int j = atoi(slist[3].c_str())-1;
	cerr << "Read pairpotential " << i << " " << j << "\n";
	if (i>j) {int k=i; i=j; j=k;}

	phi[i*nspecies+j].readSpline(fin, econv);
	//phi[i*nspecies+j].getSpline();
	if (i!=j) phi[j*nspecies+i] = phi[i*nspecies+j];
      }
      if (slist[1] == "dens") {
	int i = atoi(slist[2].c_str())-1;
	cerr << "Read density function " << i << "\n";

	rho[i].readSpline(fin, 1.0);
	//rho[i].getSpline();
      }
      if (slist[1] == "embed") {
	int i = atoi(slist[2].c_str())-1;
	cerr << "Read embedding function " << i << "\n";

	f[i].readSpline(fin, econv);
	//f[i].getSpline();
      }
    }
  } while(slist[0] != "end");

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: AlloyEAMPotentialOMP.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

