//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/CGPropagator.h"
#include "ParticleBase/ParticleUtility.h"
#include "Propagators/ConstraintBase.h"
using namespace OHMMS;

/** constructor
 *
 *  Three data members are 
 * - CG : congugate gradients
 * - G0 : previous gradients
 * These members are not added to the attribute list of ParticleSet.
 */
CGPropagator::CGPropagator(): 
  CG(postype_tag,"cg"), 
  G0(postype_tag,"g0"), 
  FlexibleCell(false),
  NumLineMin(0)
{

  CGmode = 0; //Polak-Ribiere
  deltaEmax = 1.0e-6;
  gdotgmax = 1e-6;
  fmax = 1e-3;
  gdotg_ind = E.add("Dot(F,F)");
  gdoth_ind = E.add("Dot(CG,F)");

  m_param.add(deltaEmax,"energytol","none");
  m_param.add(gdotgmax,"gradtol","none");
  m_param.add(fmax,"forcetol","none");
}

void 
CGPropagator::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;

  if(!Initialized) { 

    initializeBase(ptclin,potin,makecopy);

    //Used to use CartesianUnit for displacement calculations
    //Instead, force myPtcl->R to be in LatticeUnit
    myPtcl->convert2Unit(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addRestartAttrib(force_tag);

    CG.resize(myPtcl->getLocalNum());

    if(CGmode == 0) {
      G0.resize(myPtcl->getLocalNum());
      LOGMSG("\t Using Polak-Ribiere CG minizmiation")
    } else {
      LOGMSG("\t Using Fletcher-Reeve CG minizmiation")
    } 
    Initialized = true; 
  }

  RestartCG = true;
  //applyBConds(true);
  d_pe_save = 0.0;
}

void 
CGPropagator::setTemperature(bool rescale, Scalar_t tref) {
  //do nothing
} 

bool CGPropagator::propagate(Interaction_t * pot) {

  E[NOW] += 1;
 
  ParticlePos_t& F = myPtcl->F;

  if(RestartCG) { 
    applyBConds(true);
    E[POT_ENERGY] = myPot->evalForce(myPtcl->F,1);
    myConstraint->apply(myPtcl->F);

    E[FREE_ENERGY] = E[POT_ENERGY];

    gdotg = Dot(F,F);
    gdotg0 = gdotg;
    gdoth  = gdotg;
    gamma  = 0.0e0;
    CG     = F;
    // copy F to G0 required for <g_new|g_old>
    if(CGmode == 0)  G0=F; 

    RestartCG = false;
    DEBUGMSG(" Starting energy = " << E[FREE_ENERGY] << endl);
    //Set large number
    //myPtcl->setLayoutUpdateStride(1000);
  }

  //energy evaluations use only the displacement
  lambdasum = 0.0;

  // Perform lineminimization
  if(!lineoptimization()) {
    WARNMSG("\tCGPropagator::Lineminimization has failed. Lambda = " << Lambda)
    RestartCG = true;
    return false;
  }

  lambdasum = 0.0;
  for(int i=0; i<myPtcl->getLocalNum(); i++) {
    myPtcl->R[i] += Lambda*myPtcl->Lattice.toUnit(CG[i]);
  }

  //Calculate the displacement for interactions
  //myPot->setCurrentPos(myPtcl->R);
  E[POT_ENERGY] = myPot->evalForce(F,2);
  myConstraint->apply(F);

  E[FREE_ENERGY] = E[POT_ENERGY];
  gdotg0 = gdotg;
  Scalar_t fx=0.0;
  gdotg  = Dot(F,F,fx);
  E[gdotg_ind] = gdotg;

  if(sqrt(fx) < fmax) {
     LOGMSG(" has reached force criteria max|F|" << sqrt(fx) << endl);
     return false; // forces converged
  }

  if(gdotg < gdotgmax) {
     LOGMSG(" has reached G_norm criteria " << endl);
     return false; // forces converged
  }


  if(CGmode == 0) { // Polak-Ribiere (defualt)
    gdoth = Dot(F,G0);
    E[gdoth_ind] = gdoth;
    gamma = (gdotg - gdoth)/gdotg0; 
    G0 = F; // copy current gradient
  } else { //Fletcher-Reeve
    gamma = gdotg/gdotg0;
  }

  if(fabs(gamma) < deltaEmax){
    LOGMSG(" has reached Gamma criteria " << endl);
    return false; // forces converged
  }

  if(gamma > 0.1e0) {
    DEBUGMSG(" Gamma " << gamma << " is too big: restart CG ")
    RestartCG = true;
  }

  Scalar_t dx = fabs(d_pe_save - E[POT_ENERGY])/static_cast<Scalar_t>(myPtcl->getTotalNum());

  if(dx <= deltaEmax) {
    LOGMSG(" has reached energy tolerance criteria " << endl);
    return false; //
  }
  
  d_pe_save = E[POT_ENERGY];

  CG = F + gamma * CG; // update conjugate gradients

  Counter++;
  return true;
}

void CGPropagator::applyBConds(bool wrapping) 
{
  myPot->setCurrentPos(myPtcl->R,wrapping);
}

/** evalaute energy for R + dl*CG 
 *@param dl a proposed move along the Conjuate Gradient direction
 *@return the energy for the proposed position
 *
 *R and CG are fixed vectors during a line minimization.
 *CGPropagator utilizes an efficient update method of the NNEngine without re-calculating
 *the list. This is justfied since the lineminimization involves very small displacement
 *from the current position. It is important to calculate the displacement accurately
 *for accurate neighbor relations. 
 */
CGPropagator::Return_t
CGPropagator::Func(CGPropagator::Return_t dl) {

  NumLineMin++;
  ParticlePos_t& dR = *myPtcl->getVectorAttrib(displacement_tag);
  double dlx = dl -lambdasum;
  dR = dlx*CG;
  myPot->setCurrentPos(myPtcl->R,dR);
  lambdasum = dl;
  return myPot->evalEnergy();
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: CGPropagator.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
