//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim and Yaojun Du
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/////////////////////////////////////////////////////////////////
#include "Propagators/DimerPropagator.h"
#include "Potentials/PotentialBase.h"
#include "Utilities/RandomFunctors.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

//!< Constructor.
DimerPropagator::DimerPropagator() : 
  R_plus(postype_tag,"r+"), 
  R_minus(postype_tag,"r-"), 
  deltaR(postype_tag,"dr"),
  RC(postype_tag,"rc"), CG(postype_tag,"cg"),
  Force_1(postype_tag,"f1"), Force_2(postype_tag,"f2"),
  SPF(postype_tag,"spf"),
  dV(postype_tag,"dv"),      
  DimerParaForce(postype_tag,"dpf"), 
  DimerPerpForce(postype_tag,"dvf"),
  Para_Unit_Vec(postype_tag,"puv" ),
  Perp_Unit_Vec(postype_tag,"vuv" ),                        
  ForceRef(postype_tag,"fr"), 
  ParaVecRef(postype_tag,"pvr"),
  D_Para(postype_tag, "dpara"), D_Normal(postype_tag, "dnormal"),
  Fold(postype_tag, "fold"),Gold(postype_tag, "gold"),
  CGT(postype_tag, "cgt"), CGTu(postype_tag, "cgtu")
{
  Dt = 1.0;
  DimerDist  = 0.001; //Distance between Repelica_1 and center in unit of A
  d_theta  = 2.0e-3; //Finite difference rotation step size in rad for CG 
  MaxMove  = 0.1;    //The Maximum allowed step size for dimer translation
  MaxNR    = 1;      //Maximum number of rotation steps per translation step
  MaxCom   = 0.01;  //Maximum value for each component of force for stopping dimer 
  MaxFN    = 0.001;    //Mag. of the reotational force ar which the rotation stops
  dR       = 0.001;    //Finite difference translation step size for CG translation
  deltaEmax = 1.0e-6;
  RandomMove = 0.01;   //MAG of random move 
  ReShuffle = false;  //Flag for random move
  tp = -1.0;          //default tp=-1, R_2 is reported in the last frame

  PE_plus = E.add("E(plus)");
  PE_minus = E.add("E(minus)");
  PE_center = E.add("E(center)");
  magDF = E.add("Dot(F,F)");
  fn12diff = E.add("FN1 - FN2");   
  fn12sum = E.add("FN1 + FN2");   
  magPF = E.add("ParallelF");
  magTF = E.add("TotalF");
  curvature = E.add("Curvature");

  addParameter(ReShuffle,"shuffle","bool");
  addParameter(Dt,"timestep","fsec");
  addParameter(T_ref,"temperature","K");
  addParameter(DimerDist,"dimerdistance","none");//sep
  addParameter(d_theta,"delta_theta","none");//dth
  addParameter(MaxNR,"maxrotation","none");//maxnr,int
  addParameter(MaxMove,"maxdispl","AA");//maxmove
  addParameter(MaxCom,"forcetol","none");
  addParameter(MaxFN,"rforcetol","none");//maxfn
  addParameter(dR,"lambda","none");//dR
  addParameter(RandomMove,"random_move","none");//mag
  addParameter(deltaEmax,"energytol","none");

}


/*! \fn bool DimerPropagator::propagate(PotentialBase* pot)
 *  \return true if a new search should continue
 *          false if a saddle point is found
 */
bool 
DimerPropagator::propagate(PotentialBase* pot) {
  

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);

  if(!Initialized) {
    myPtcl->addAttribute(R_plus);
    myPtcl->addAttribute(R_minus);
    adjust(false,false);
    PotRef      = pot;      //save a refernce  
    Initialized = true;   
    Restart_Rotation = true;  //for the first rotation within a chosen plane
    FirstTranslation = true;  //for the first CG translation
    pot->set(myPtcl);        //assign a particle set to PotentialBase::myPtcl
    myPtcl->updateR(myPtcl->R); // adjust has set myPtcl->R = RC;
    E[PotEnergy] = pot->evalEnergy();
    d_pe_save = E[FreeEnergy] = E[PotEnergy];
    if(ReShuffle){ LOGMSG("\t Random displacement are applied"); }
    if(tp<0){ LOGMSG("\t Replica_2 are reported in the last frame");}
    else{LOGMSG("\t Replica_1 are reported in the last frame");}

    return true;
  }

  int  i = 0;
  do {
    if(!RotateDimer()) break;
    i++;
  } while(i<=MaxNR);

  
  SPF = 0.5 * (Force_1 + Force_2);
  //Scalar_t d1 = sqrt(Dot(SPF, SPF));   
  //Scalar_t d2 = sqrt(Dot(DimerParaForce, DimerParaForce));
  Scalar_t d1 = Dot(SPF, SPF);   
  Scalar_t d2 = Dot(DimerParaForce, DimerParaForce);

  E[fn12diff] = sqrt(gdotg);
  E[fn12sum] = sqrt(d1);
  E[magPF] = sqrt(d2);
  E[magTF] = sqrt(d1+d2);

  //cout<<"FN1 - FN2 = "<<sqrt(gdotg)<<" FN1 + FN2 = "<<sqrt(d1)<<" Parallel Force "<<sqrt(d2);
  //cout<<" Total Force "<<sqrt(d1 + d2)<<endl;
  
  CGTranslateDimer();
  Counter++;                  // increment the internal counter  
  E[Now] += Dt; 

  Get_Energy();  
  Tensor_t stress; 

  V = RC-myPtcl->R;
  myPtcl->applyBC(RC,myPtcl->R);
  myPtcl->convert(myPtcl->R,RC);
  PotRef->setCurrentPos(RC,false);
  E[PotEnergy] =  PotRef->evalForce(F,0);

  evalKineticEnergy();
  /*If each component of F is  less than  0.001 eV/Ang, 
    i.e. MaxCom, the dimer is considered converged.*/
  //for(int j=0; j < static_cast<double>(myPtcl->getTotalNum()); j++){
  for(int j=0; j < myPtcl->getTotalNum(); j++){
    for(int k=0; k<3; k++){
      if(fabs(F[j][k]) > MaxCom ){
	d_pe_save = E[PotEnergy]; //d_pe;
	Restart_Rotation = true;   
	return true;
      }
    }
  }
  
  if(E[curvature] > 0.0) return true;
  
  //I don't think this is every useful
  //myPtcl->convert(myPtcl->R,R0);
  //if(tp<0.0) {
  //  R0 -= 100.0*deltaR;
  //} else {
  //  R0 += 100.0*deltaR;
  //}
  //myPtcl->convert(R0,myPtcl->R);

  //   convert(myPtcl->Lattice, myPtcl->R, R0);
  //   if(tp<0.0){R0 -= 100.0 * deltaR;}
  //   else{R0 += 100.0 * deltaR;}
  //   convert(myPtcl->Lattice, R0, myPtcl->R);
  //   cout<<" Dimer stopped: saddle point has been found "<<endl; 
  return false;
}




/*!\fn void DimerPropagator::TranslateDimer()
 * \brief Section 2.C. Translating the dimer (Eq. 20)
 */
void DimerPropagator::TranslateDimer() {
  
  // get a reference for the velocity
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  //  curvature = ((PEnergy_1 + PEnergy_2) - 2.0 * PEnergy_C)/(DimerDist * DimerDist);
  
  if(E[curvature] < 0.0) {
    DEBUGMSG("Sign(curvature) is -")
    //SPF =  ForceRef  - 2.0 * DimerParaForce;
     SPF = 0.5 * (Force_1 + Force_2) - DimerParaForce;
  }
  else{
    DEBUGMSG("Sign(curvature) is +")
    SPF =  -1.0 * DimerParaForce;
  }
  
  dV = Dt*SPF;
  
   for(int j=0; j < myPtcl->getTotalNum(); j++){
    for(int k=0; k<3; k++){
      if((SPF[j][k] *  V[j][k]) < 0.0 ){
	V[j][k] = 0.0;
      }
    }
  }


  Scalar_t dv_dot_v = Dot(dV,V);
  //  if(dv_dot_v > 0.0) {
  //   cout<<"Sign(dv_dot_v) is +"<<"  ";
  Scalar_t d = 1.0 + dv_dot_v/Dot(dV, dV);
  V = d*dV; 
  //  if((curvature < 0.0) && (sqrt(Dot(DimerParaForce, DimerParaForce)) < 1) )
  // V = dV;
  // }
  /*  else{
      cout<<"Sign(dv_dot_v) is -"<<"  ";                 
      V = dV;
      }*/
  
  //tp = Dot(V, V);
  //cout<<"Mag of V"<<"    "<<sqrt(tp)<<endl;
  //Scalar_t tmp = sqrt(Dot(V*Dt, V*Dt));

  Scalar_t tmp = Dt*sqrt(Dot(V, V));
  if(tmp > MaxMove) { 
    dV = V;
    normalize(dV);        //instead of NormalizeVec
    RC = RC + MaxMove * dV;  
    cout<<"a BIG move"<<endl; 
  } 
  /*
    else if(curvature < 0){
    dV = V;
    normalize(dV);        //instead of NormalizeVec
    RC = RC + MaxMove * dV; 
    }*/ 
  else    
    RC  =  RC + Dt*V; 
}


void DimerPropagator::CGTranslateDimer() {
  
  Scalar_t FrP1, FrP2, CR, FrP;
  
  if(E[curvature] < 0.0) {
    DEBUGMSG("Sign(curvature) is -")
    SPF = 0.5 * (Force_1 + Force_2) - DimerParaForce;
  }
  else{
    DEBUGMSG("Sign(curvature) is +")
    SPF =  -1.0 * DimerParaForce;
  }
  
  if(FirstTranslation){
    FirstTranslation = false;
    Gam   = 0.0;
    Fold  = SPF;
    Gold  = SPF;
  }
  else{
    Scalar_t d1 = Dot(SPF, SPF);
    Scalar_t d2 = Dot(SPF, Fold);
    Gam  = (d1 - d2)/Dot(Fold, Fold);
  }
  
  CGT = SPF + Gam * Gold;
  CGTu = CGT;
  normalize(CGTu);
  
  Fold = SPF;
  Gold = CGT;
  
  FrP1 = Dot(SPF, CGTu);
  
  //R0 = RC;             //save RC
  RC = RC  + dR*CGTu;
  Get_Force(false);
  
  if(E[curvature] < 0.0) {
    //cout<<"Sign(curature) is -"<<"  ";
    SPF = 0.5 * (Force_1 + Force_2) - DimerParaForce;
  }
  else{
    //cout<<"Sign(curature) is +"<<"  ";                
    SPF =  -1.0 * DimerParaForce;
  }
  
  FrP2 = Dot(SPF, CGTu);
  CR   = (FrP1 - FrP2)/dR;
  FrP = (FrP1 + FrP2)/2.0;
  
  if(CR < 0.0){
    DEBUGMSG(" CR is -.")
    RdR = MaxMove; 
  } 
  else{
    DEBUGMSG(" CR is +.")
    RdR = FrP/CR;
    if(fabs(RdR) > MaxMove){
      if(RdR < 0.0) RdR =  -1.0 * MaxMove;
      else RdR = MaxMove;
    }
    RdR = RdR - dR/2.0;
  }
  
  //RC = R0;  //restore RC after a finite difference step
  RC = RC + CGTu * RdR;
}

bool DimerPropagator::RotateDimer() {
  
  if(Restart_Rotation) { 
    
    Get_Force(true); //Para_Unit_Vec & Forces are calculated 
    Get_Energy();    //d_pe is calculated 
    
    gdotg  = Dot(DimerPerpForce, DimerPerpForce);
    gdotg0 = gdotg;
    gdoth  = gdotg;
    gamma  = 0.0e0;
    
    CG     = DimerPerpForce;
    Perp_Unit_Vec = CG;
    normalize(Perp_Unit_Vec); 
    
    Restart_Rotation = false;
  }
  
  // Perform minimization: Rotate replica_1 to new position
  if(!angle_optimization()) {
    LOGMSG("has no need for rotation ")
    return false;
  }

  Scalar_t costheta = cos(Delta_theta)*DimerDist;  
  Scalar_t sintheta = sin(Delta_theta)*DimerDist;  

  //R0 = RC + costheta * Para_Unit_Vec + sintheta * Perp_Unit_Vec;
  //deltaR = R0 - RC;
  deltaR = costheta * Para_Unit_Vec + sintheta * Perp_Unit_Vec;
  Get_Force(true); //Para_Unit_Vec & Forces are calculated
  Get_Energy();    //d_pe is calculated 

  gdotg0 = gdotg;
  gdotg  = Dot(DimerPerpForce, DimerPerpForce);
  
  if(sqrt(gdotg) < MaxFN) {
    return false;  // forces converged
  }
  
  gdoth = Dot(DimerPerpForce, CG);
  //  gamma = (gdotg - gdoth)/gdotg0;   //CG method
  gamma = 0.0;                        //SD method
 
  
  Scalar_t dx = 
    fabs(d_pe_save - E[PotEnergy])/static_cast<double>(myPtcl->getTotalNum());
  /*
  if(dx <= deltaEmax*0.001) {
    cout<<" has reached tolerance energy criteria in rotation " << endl;
    return false; //
    }*/
  
  d_pe_save = E[PotEnergy];//d_pe

  // update conjugate gradients
  //! set the direction of PREVIOUS CG perpendicular to NEW Para_Unit_Vec
  Scalar_t d= -Dot(Perp_Unit_Vec, Para_Unit_Vec);
  Perp_Unit_Vec = Perp_Unit_Vec + d* Para_Unit_Vec;
  normalize(Perp_Unit_Vec);
   
  d = sqrt(Dot(CG,CG))*gamma;
  CG = DimerPerpForce + d*Perp_Unit_Vec; 
  Perp_Unit_Vec = CG;
  normalize(Perp_Unit_Vec); 

  return true;
}


bool 
DimerPropagator::angle_optimization(){
  //////////////////////////////////////////////////////////
  //Using references to avoid allocating more
  // D_Para and D_Normal are strictly local to this function.
  // So are ForceRef and ParaVecRef
  //ParticlePos_t   D_Para(postype_tag,"dp"), D_Normal(postype_tag,"dn");
  ///////////////////////////////////////////////////////////
  // ParticlePos_t& D_Para = ForceRef;
  //ParticlePos_t& D_Normal = ParaVecRef;

  Scalar_t        FdotTh_1, FdotTh_2, Cth, Fth;
  
  FdotTh_1 =  Dot(DimerPerpForce, Perp_Unit_Vec); 
  Scalar_t costheta = cos(d_theta)*DimerDist;
  Scalar_t sintheta = sin(d_theta)*DimerDist;

  //R0 = RC + costheta * Para_Unit_Vec + sintheta * Perp_Unit_Vec;
  //deltaR  = R0 - RC; 
  deltaR = costheta * Para_Unit_Vec + sintheta * Perp_Unit_Vec;
  D_Para   = deltaR;
  normalize(D_Para);

  Scalar_t d = -Dot(Perp_Unit_Vec, D_Para);
  D_Normal = Perp_Unit_Vec + d* D_Para;
  normalize(D_Normal);
  Get_Force(false);                    //Only Foreses are calculated

  FdotTh_2 =  Dot(DimerPerpForce, D_Normal);

  Cth      = (FdotTh_1 - FdotTh_2)/d_theta;
  Fth      = 0.5 * (FdotTh_1 + FdotTh_2);
  Delta_theta = 0.5 * atan(2.0 * (Fth/Cth)) + 0.5 * d_theta;

  if(Cth < 0.0) 
    Delta_theta = Delta_theta + 0.5 * 3.1415927;  // Delta_theta += PI/22

  //if(Delta_theta <= 1e-9) 
  //  return true;
  //else  
  return true;

}

/** evalute energies of the dimers and the center configuration
 */
void DimerPropagator::Get_Energy(){
  R_plus  = RC + deltaR;
  myPtcl->updateR(R_plus,false);
  E[PE_plus]= PotRef->evalEnergy();

  R_minus  = RC - deltaR;
  myPtcl->updateR(R_minus,false);
  E[PE_minus]= PotRef->evalEnergy();

  myPtcl->updateR(RC,false); 
  E[PE_center]= PotRef->evalEnergy();
  E[PotEnergy]=E[FreeEnergy] = E[PE_center];
}

/** evalate forces on the dimers
 *@param update_para_unit_vec if true, normalize the dimer forces 
*/
void DimerPropagator::Get_Force(bool update_para_unit_vec){

  Tensor_t      stress;  
  // getting a reference to the local energy
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);

  R_plus  = RC + deltaR;
  PotRef->setCurrentPos(R_plus,false);
  E[PE_plus] = PotRef->evalForce(Force_1,0);// PEnergy_1 = PotRef->evalForce(Force_1, Eloc, stress, 0);

  R_minus  = RC - deltaR;
  PotRef->setCurrentPos(R_minus,false);
  E[PE_minus] = PotRef->evalForce(Force_2,0);// PEnergy_2 = PotRef->evalForce(Force_2, Eloc, stress, 0);

  //get unit vector parallel to the dimer
  ParaVecRef = deltaR;
  normalize(ParaVecRef);

  //get curvature
  Scalar_t fd1 = Dot(Force_1, ParaVecRef);
  Scalar_t fd2 = Dot(Force_2, ParaVecRef);
  E[curvature]  =  (fd2 - fd1)/(2.0 * DimerDist );

  //get the average force parallel to the dimer
  ForceRef = 0.5 * (Force_1 + Force_2);
  Scalar_t d = Dot(ForceRef,  ParaVecRef);
  DimerParaForce = d* ParaVecRef;

  //get  force perpendicular to the dimer
  Scalar_t d1 = - Dot(Force_1, ParaVecRef);
  Scalar_t d2 = - Dot(Force_2, ParaVecRef);
  Force_1         = Force_1 + d1* ParaVecRef;
  Force_2         = Force_2 + d2* ParaVecRef;
  DimerPerpForce  = Force_1 - Force_2;

  if(update_para_unit_vec) Para_Unit_Vec = ParaVecRef;
}


void DimerPropagator::adjust(bool trescale, bool wrapping) 
{

  resetBase(); 
  // reset, time coefficient with new or old parameter
  Counter = -1;

  if(!Initialized) {
    ////////////////////////////////////////
    // data members of ParticlePos_t are not added to the ParticleAttributes
    // Need to allocate for a given size
    ////////////////////////////////////////
    int nat = myPtcl->getLocalNum();
    CG.resize(nat);
    R_plus.resize(nat);
    R_minus.resize(nat);
    deltaR.resize(nat);
    RC.resize(nat);
    Force_1.resize(nat);
    Force_2.resize(nat);
    SPF.resize(nat);
    dV.resize(nat);
    DimerParaForce.resize(nat);
    DimerPerpForce.resize(nat);
    Para_Unit_Vec.resize(nat);
    Perp_Unit_Vec.resize(nat);
    ForceRef.resize(nat);
    ParaVecRef.resize(nat);
    D_Para.resize(nat);
    D_Normal.resize(nat);
    Fold.resize(nat);
    Gold.resize(nat);
    CGT.resize(nat);
    CGTu.resize(nat);

    //set RC to the input position
    myPtcl->Lattice.applyBC(myPtcl->R, RC);  
    myPtcl->R.setUnit(RC.getUnit());
    myPtcl->R = RC;
  
    //ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
    //V = 0.0;      //set accumulated velocity as zero 
    d_pe_save= 1.0e6;
    LOGMSG("\t ftol is set to: "<<MaxCom)
    LOGMSG("\t MaxMove is set to: "<<MaxMove)
    LOGMSG("\t separation is set to: "<<DimerDist)
    LOGMSG("\t finite difference for translation: "<<dR)
    LOGMSG("\t finite difference for rotation: "<<d_theta)
  }

  ///////////////////////////////////////////////////////////
  //A. Du wrote:
  //deltaR   = RandomVector;                    // pseudo code
  ///////////////////////////////////////////////////////////
  RandomVector<SingleParticlePos_t, RandomGenerator_t> rvg(Random);
  RandomSequence<ParticlePos_t, RandomGenerator_t>::apply(ForceRef, rvg);
  RandomSequence<ParticlePos_t, RandomGenerator_t>::apply(deltaR, rvg);
  normalize(deltaR);                    

  //Scalar_t d=0.1*DimerDist;
  deltaR *= DimerDist;
  if(ReShuffle){
    LOGMSG("\treshuffle data with a random displacement "<<RandomMove)
    RandomSequence<ParticlePos_t, RandomGenerator_t>::apply(ParaVecRef, rvg);
    normalize(ParaVecRef);
    ParaVecRef *= RandomMove;
    RC += ParaVecRef;
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: DimerPropagator.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
