//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim and Yaojun Du
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/****************************************************************************
 * Implemented by Yaojun Du on June 15 2002
 * Modified by Jeongnim Kim on June 18 2002
 *
 * DimerPropagator finding the saddle point using Dimer method
 * The Dimer includes two replicas allowing us to detect curvatures in 
 * the phase space. Lowest curvature mode is searched using Modifed 
 * CG method.
 * The CG method is applied to dimer rotation. However, more presicely,
 * the SD method is applied to dimer rotation, since there is only one
 * ratation for one translation. When the dimer is closing to the saddle
 * point,
 * and it moves slowly, more rotations per distance may be expected. Thus,
 * when the dimer is approaching the saddle point, the trajectory is 
 * much more following the lowest curvature mode than the one when the dimer
 * is far from the saddle point. 
 * Quick Min is applied to dimer translation.
 *
 *  \ref Graeme Henkelman and Hannes Jonsson, 
 * "A dimer method of finding saddle points on high dimensional potential
 *  surfaces using only first derivaties", 
 *  Journal of Chemical Physics, Vol 111, 7010 (1999).
 *
 ***************************************************************************/


#ifndef OHMMS_DIMERPROPAGATOR_H
#define OHMMS_DIMERPROPAGATOR_H

#include "Propagators/PropagatorBase.h"

namespace OHMMS {

  class DimerPropagator: public PropagatorBase { // using Dimer algorithm

  public:

    DimerPropagator();
    ~DimerPropagator() { }
 
    // implementing virutal functions
    bool propagate(PotentialBase*);
    void adjust(bool trescale, bool wrapping);
    void applyBConds();

  protected:

    bool RotateDimer();
    bool angle_optimization();   //similar to lineoptimization for CG
    void TranslateDimer();
    void CGTranslateDimer();
    void Get_Force(bool update_para_unit_vec);
    void Get_Energy();
  
  protected:

    PotentialBase  *PotRef;
  
    // Distance between two replicas are 2*Delta_R 
    // E[PE_plus] -> E of replica_1  Force_1 -> force exerted on replica_1
    // E[PE_minus] -> E of replica_2  Force_2 -> force exerted on replica_2
    // E[PE_center] -> E of middle-point */
    int  PE_plus, PE_minus, PE_center, 
      magDF, fn12diff, fn12sum, magPF,  magTF, magV,
      curvature;

    // Other control parameters
    bool     Restart_Rotation, ReShuffle; 
    int      MaxNR;
    Scalar_t Dt, T_ref;
    Scalar_t Delta_theta, d_theta, DimerDist;
    Scalar_t deltaEmax, MaxMove, RandomMove, MaxFN, MaxCom;
    Scalar_t d_pe_save, gdotg, gdotg0, gdoth, gamma; 


    /* R1_0 -> replica_1  R2_0 -> replica_2  RC->midpoint
       deltaR -> relative position of replica_1 with repect 
       to midpoint, i.e. R1_0=RC+R1_1 and R2_0=RC-R1_1.
       if R1_1 and RC are given, the postion of dimer is 
       complete 
       ! The initialization of dimer is performed as following
       1> initialize RC by  mySys->R    = RC;
       2> initialize R1_1 with a random ParticlePos_t. */
  
     
    ///conjugate-gradients
    ParticlePos_t  CG;
    ///positoins of the center dimer
    ParticlePos_t  RC;
    ///dimer vector \f$2\delta R = R_{1} - R{2}\f$
    ParticlePos_t  deltaR;
    ///RC+deltaR
    ParticlePos_t  R_plus;
    ///RC-deltaR
    ParticlePos_t  R_minus;

    ParticlePos_t  Force_1, Force_2, SPF, dV;
    ParticlePos_t  DimerParaForce, DimerPerpForce, Para_Unit_Vec, Perp_Unit_Vec; 
    ParticlePos_t  ForceRef,ParaVecRef;
    ParticlePos_t  D_Para, D_Normal;

    //Variables for CG translation of the dimer
    bool            FirstTranslation;
    Scalar_t        Gam, dR, RdR, tp;
    ParticlePos_t   Fold, Gold, CGT, CGTu;
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: DimerPropagator.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/



