//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/EPNVerlet.h"
#include "Potentials/PotentialBase.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

//<constructor
EPNVerlet::EPNVerlet() :T_ref(0.0), P_ref(0.0), Dt(1.0), VolMass(1.0),
                        dS(postype_tag,"ds"), S2(postype_tag,"s2"), Friction(0.0)
{
  pressure_ind = E.add("Pressure");
  pv_pot = E.add("P*V");
  pv_kin = E.add("K_L");

  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
  m_param.add(P_ref,"pressure","GPa");
  m_param.add(VolMass,"volmass","GPa");
}

void 
EPNVerlet::initialize(Particle_t* ptclin, PotentialBase* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  
  Scalar_t tmp = Friction*Dt*0.5e0;
  c0 = 2.0e0/(1.0e0+tmp);
  c1 = 1.0e0 - c0;
  c2 = Dt*Dt/(1.0e0+tmp);
  oneover2dt = 0.5e0/Dt;

  if(!Initialized) {

    initializeBase(ptclin,potin,makecopy);

    myPtcl->addAttribute(scalartype_tag, energy_tag);

    ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
    ParticlePos_t& S1 = *myPtcl->getVectorAttrib("s1");
    ParticlePos_t& Scaled_Force = *myPtcl->getVectorAttrib("sf");

    d0 = c0;
    d1 = c1;
    d2 = c2;  

    P_ext.diagonal(P_ref);
    cout << "\nP_ext = \n" << P_ext << "\n";

    //////////////////////////////////////////////////////
    // data members of ParticlePos_t are not added to the 
    // ParticleAttributes  Need to allocate for a given size
    ///////////////////////////////////////////////////////
    int nat = myPtcl->getLocalNum();
    dS.resize(nat);
    S2.resize(nat);
    //Scaled_Force.resize(nat);


    // The velocity must be in cartesian coordinates
    V.setUnit(PosUnit::CartesianUnit);

    myPtcl->convert2Unit(myPtcl->R);
    S1.setUnit(PosUnit::LatticeUnit);
    S2.setUnit(PosUnit::LatticeUnit);
    dS.setUnit(PosUnit::LatticeUnit);
    Scaled_Force.setUnit(PosUnit::LatticeUnit);

    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addPositionAttrib(S1);

    h0 = myPtcl->Lattice.R;
    h1=h0;
    h2=h0;   
    G0 = myPtcl->Lattice.M;
    Initialized = true;
  }

  /*! set the velocity to a reference temperature only when the clock is zero    
   * When the clock E[NOW] > 0, it assumes that the simulation restarts from
   * the previous configuration.
   */
  if(ParameterModified && T_ref>zeroKelvin) 
    setTemperature(ParameterModified,T_ref);
  else
    applyBConds(true);
}

void 
EPNVerlet::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticlePos_t& S1 = *myPtcl->getVectorAttrib("s1");

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
    myPtcl->convert2Unit(myPtcl->R,S1);
    myPtcl->convert2Unit(V,dS);
    S2 = S1 + Dt*dS;
    myPtcl->convert2Unit(S2,myPtcl->R);
  } else {
    V = 0.0;
    dS = 0.0;
    myPtcl->convert2Unit(myPtcl->R);
    S1 = myPtcl->R;
  }

  applyBConds(true);
}


bool  EPNVerlet::propagate(PotentialBase* pot) {

  if(!Initialized) initialize(myPtcl,pot,false);

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  ParticlePos_t& S0 = myPtcl->R;
  ParticlePos_t& S1 = *myPtcl->getVectorAttrib("s1");
  ParticlePos_t& Scaled_Force = *myPtcl->getVectorAttrib("sf");

  Tensor_t stress=0;
  myPot->setCurrentPos(S0,false);
  E[POT_ENERGY] = myPot->evalForce(F,1);
  DEBUGMSG("stress = " << stress);
  evalAccel(F);

  Get_Sigma_Tensor();
  Get_Tot_stress(stress);

  // 1. Atomic motion
  Get_Scaled_Force();
  S2 = S1;
  S1 = S0;
  S0 = c0 * S1 + c1 * S2 + c2 * Scaled_Force;
  dS = oneover2dt*(S0-S2);

  myPtcl->convert(dS, V);
  evalKineticEnergy();
  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY];

  // 2. Lattice motion
  h2 = h1;
  h1 = h0;
  G1 = G0;

  h0 = d0 * h1 + d1 * h2 + d2 * Tot_stress;
  dh = (h0 - h2) * oneover2dt; 

  // p*V term for energy
  const Scalar_t onethird = 0.33333333333333333333333333;
  E[pressure_ind] = trace(PiTensor)*onethird;
  E[pv_pot] = -E[pressure_ind]*myPtcl->Lattice.Volume;
  E[pv_kin] = VolMass*0.5*trace(dot(transpose(dh),dh));

  E[FREE_ENERGY] += E[pv_pot];
  E[FREE_ENERGY] += E[pv_kin];

  // Update the atom coordinates, lattice vectors, and the metric
  myPtcl->updateLayout(h0);
  G0 = myPtcl->Lattice.M;
  dG = 2*(G0 - G1) * oneover2dt;

  //update the current position: no need to apply BConds
  myPtcl->updateR(S0);

  Counter++; // increment the internal counter

  E[NOW] += Dt;

  return (E[TEMPERATURE] > KE_min_K);
}

void EPNVerlet::Get_Sigma_Tensor()
{
  //   SingleParticlePos_t a, b, c, ap, bp, cp;
  //   a[0] = myPtcl->Lattice.R(0,0);
  //   a[1] = myPtcl->Lattice.R(0,1);
  //   a[2] = myPtcl->Lattice.R(0,2);
  //   b[0] = myPtcl->Lattice.R(1,0);
  //   b[1] = myPtcl->Lattice.R(1,1);
  //   b[2] = myPtcl->Lattice.R(1,2);
  //   c[0] = myPtcl->Lattice.R(2,0);
  //   c[1] = myPtcl->Lattice.R(2,1);   
  //   c[2] = myPtcl->Lattice.R(2,2);
  //cross(b, c); cross(c,a); cross(a,b)
  SingleParticlePos_t ap = cross(myPtcl->Lattice.a(1),myPtcl->Lattice.a(2));
  SingleParticlePos_t bp = cross(myPtcl->Lattice.a(2),myPtcl->Lattice.a(0));
  SingleParticlePos_t cp = cross(myPtcl->Lattice.a(0),myPtcl->Lattice.a(1));
  
  SigmaTensor = Tensor_t(ap[0], ap[1], ap[2], bp[0], bp[1], bp[2], cp[0], cp[1], cp[2]);

}


void EPNVerlet::Get_Tot_stress(Tensor_t stress)
{
  SingleParticlePos_t v;
  Tensor_t  stressloc=0, tmp;
  Scalar_t  Vinv;

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  Vinv = 1.0 / myPtcl->Lattice.Volume;

  for(int i=0; i<myPtcl->getLocalNum(); i++) {
    stressloc += mass[ myPtcl->GroupID[i]] * outerProduct(V[i], V[i]);
  }

  PiTensor = Vinv * (stressloc + stress);
  tmp = PiTensor - P_ext;
  Tot_stress =  (1.0 / VolMass) * dot(tmp, SigmaTensor);
}


void EPNVerlet::Get_Scaled_Force()
{
  Tensor_t tmp;
  
  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticlePos_t& Scaled_Force = *myPtcl->getVectorAttrib("sf");

  tmp = inverse(G0);
  tmp = dot(tmp, dG);

  myPtcl->convert(F, Scaled_Force);
  myPtcl->convert(V, dS);

  for(int i=0; i<myPtcl->getLocalNum(); i++) 
    Scaled_Force[i] -= dot(tmp, dS[i]);

}


void EPNVerlet::applyBConds(bool wrapping) {

  if(wrapping) myPtcl->applyBConds(); 
  myPtcl->updateR(myPtcl->R);
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: EPNVerlet.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
