//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/LeapFrogPropagator.h"
#include "Potentials/PotentialBase.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

LeapFrogPropagator::LeapFrogPropagator():
ToApplyBConds(false),  T_ref(0.0), Dt(1.0), Vp(postype_tag,"vp")
{ 
}


bool 
LeapFrogPropagator::propagate(PotentialBase* pot) {

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  ParticlePos_t& Vm = *myPtcl->getVectorAttrib("vm");
  Tensor_t stress;

  if(!Initialized) {
    adjust(false,false);
    pot->set(myPtcl); // assign a particle set to PotentialBase::myPtcl
    Initialized = true; 

    E[PotEnergy] = pot->evalForce(F, Eloc, stress, 0);
  }

  ParticlePos_t& R0 = myPtcl->R;
  Vp = Vm+Dtover2*F;
  R0 += Dt*Vp;

  //apply bconds on R0 due to the random V and F
  if(ToApplyBConds)  applyBConds();

  V = 0.5*(Vp+Vm);
  Vm = Vp;

  myPtcl->convert(R0, myPtcl->R);
  myPtcl->update(); 

  myPtcl->advance(); // SimpleParticle::Counter is incremented
  stress = 0;
  E[PotEnergy] = pot->evalForce(F, Eloc, stress, 0);
  evalAccel(F);

  evalKineticEnergy();
  E[FreeEnergy] = E[PotEnergy] + E[KineticEnergy];

  Counter++; // increment the internal counter
  bool moving = (E[Temperature] > zeroKelvin);

  E[Now] += Dt;
  return moving;
}

void LeapFrogPropagator::adjust(bool trescale, bool wrapping) 
{

  resetBase();

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  Dtover2 = Dt/2.0;

  if(!Initialized) {

    ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
    ParticlePos_t& R0 = *myPtcl->getVectorAttrib(trajectory_tag);

    ///previous velocity is required to restart vm(inus)
    ParticlePos_t& Vm = *myPtcl->getVectorAttrib("vm");

    ///forward velocity is local to LeapFrogPropagator
    Vp.resize(V.size());

    ///////////////////////////////////
    // these values are always in cartesian coordinates
    ///////////////////////////////////
    myPtcl->convert2Cart(myPtcl->R);
    Vp.setUnit(PosUnit::CartesianUnit);
    Vm.setUnit(PosUnit::CartesianUnit);

    //if(fabs(E[Now]) < zeroKelvin && T_ref > zeroKelvin) {
    if(fabs(E[Now]) < zeroKelvin) {
      ///needs to assign Velocity, if not initialized
      if(T_ref>zeroKelvin && E[Temperature]<zeroKelvin) { 
	setVelocity(T_ref);
      }
      Vm = -0.5*V;
    }
    myPtcl->convert(myPtcl->R, R0);
  }

  // scale temperature with the new parameters
  if(trescale) setTemperature(trescale,T_ref);

  if(Initialized && wrapping) {
    // defer applying boundary conditions,
    // since the forces are evaluated after adding random vectors
    ToApplyBConds = true;
  }
}

void 
LeapFrogPropagator::applyBConds(){

   ParticlePos_t& R0 = *myPtcl->getVectorAttrib(trajectory_tag);
   ParticlePos_t t(R0.size());
   //R1 is temporary data and does not affect dynamics
   wrapAroundBox(myPtcl->Lattice, R0, t);

   //records the jumps due to the boundary conditions
   ParticlePos_t& del = *myPtcl->getVectorAttrib("sumbc");
   del += R0-t; // accumulate the change
   R0   = t;    // copy vectors in a supercell to the trejectory

   // update the current position
   //convert(myPtcl->Lattice, R0, myPtcl->R);

   // request re-parition
   myPtcl->NeedPartition = true;
   ToApplyBConds = false;
}

bool LeapFrogPropagator::get(ostream& os) const {
  os << m_name << " --temperature " << T_ref
     << " --mdclok " << E[Now] 
     << " --timestep " << Dt << endl;
  return true;
}

bool LeapFrogPropagator::put(istream& is) {
  return true;
}

void LeapFrogPropagator::reset() {

}



#if defined(HAVE_LIBXML2)

bool LeapFrogPropagator::put(xmlNodePtr cur){

  m_name = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
  xmlNodePtr cursave = cur;

  cur = cur->xmlChildrenNode;
  while(cur != NULL) {

    if ((!xmlStrcmp(cur->name, (const xmlChar *)"parameter"))) {

      string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
      if(aname == "temperature") {
        putContent(T_ref,cur);
      } else if(aname == "mdclock") {
        putContent(E[Now],cur);
      } else if(aname == "timestep") {
        putContent(Dt,cur);
      } 
    }
    cur = cur->next;
  }
  return true;
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: EVNLeapFrog.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
