//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/EVNVelocityVerlet.h"
#include "ParticleBase/ParticleUtility.h"
#include "Message/OpenMP.h"
#include <limits>
using namespace OHMMS;

EVNVelocityVerlet::EVNVelocityVerlet(): T_ref(0.0), Dt(1.0)
{ 
  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
}
/** Implement the virtual function
 *
 *The steps follow
 *- initialize the coefficients for the EOM
 *- initialize myPtcl
 *- initialize velocity and position-dependent variables 
 */
void 
EVNVelocityVerlet::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  Counter = 0;
  Dtover2 = Dt/2.0;
  DtDtover2 = Dt*Dt/2.0;

  initializeBase(ptclin,potin,makecopy);

  /**
   * \if Initialized == false
   * add ParticleSet::R  to the ConfigList so that Boundary Conditions can be applied during the simulations.
   * \endif
   */
  if(!Initialized) { 
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addRestartAttrib(velocity_tag);
    myPtcl->addRestartAttrib(force_tag);
    Initialized = true;
  }

  dtover2m=minv;
  for(int i=0; i<dtover2m.size(); i++) dtover2m[i]*=Dtover2;

  /**
   * \if E[NOW]<=zeroKelvin && T_ref>zeroKelvin
   * set the velocity to the target temperature.
   *  \else
   * apply Boundary Conditions to start a simulation block.
   *  \endif
   */
  if(ParameterModified) {
    setTemperature(RescaleTemperature,T_ref);
  } else  
    applyBConds(true);

  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,0);

  //if(E[NOW]<numeric_limits<Scalar_t>::epsilon()) {
  //  //zero is used to indicate that NN list is already evalauted.
  //  myPot->evalForce(myPtcl->F, 0);
  //}

  RescaleTemperature = false;
}

void 
EVNVelocityVerlet::setTemperature(bool rescale, Scalar_t tref) {

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  if(rescale) {
    LOGMSG("\tEVNVelocityVerlet::setTemperature T_ref = " << T_ref << " and rescale the current velocity ")
  }
  //if tref>0, overwrites the target temperature
  if(tref>zeroKelvin) T_ref = tref;
  if(T_ref>zeroKelvin) 
    scaleTemperature(rescale,T_ref);
  else
    myPtcl->V = 0.0;

  ScaledV = 1.0;
  applyBConds(true);
} 

bool EVNVelocityVerlet::propagate(Interaction_t* pot) {

  //if(!Initialized) initialize(myPtcl,pot,false);
  if(Counter == UpdateStride) {
    Counter=0;
    myPot->setCurrentPos(myPtcl->R);
  }

#pragma omp parallel
  {
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];

    const ParticleIndex_t& gid = myPtcl->GroupID;
    const ParticlePos_t& F = myPtcl->F;
    ParticlePos_t& V = myPtcl->V;
    ParticlePos_t& dR = myPtcl->dR;

    //r(t+dt) = r(t) + dt v(t) + dt*dt/2 a = r(t)+dt*(v(t)+dt/2*a) = r(t)+dt*v(t+dt/2)
    //v(t+dt/2) = v(t) + dt/2 *a(t)
#pragma ivdep
    for(int iat=first; iat<last; iat++) {
      V[iat]+= dtover2m[gid[iat]]*F[iat];
      dR[iat]=Dt*V[iat];
      //dR[iat]=Dt*V[iat]+DtDtover2*F[iat];
      //V[iat] += Dtover2*F[iat];
    }
  }

  //2 to update the NN list 
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,2); 

  Scalar_t ke=0.0;
#pragma omp parallel reduction(+:ke)
  {
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];

    const ParticlePos_t& dR = myPtcl->dR;
    for(int iat=first; iat<last; iat++) {
      myPtcl->R[iat] += dR[iat];
    }

    const ParticleIndex_t& gid = myPtcl->GroupID;
    const ParticlePos_t& F = myPtcl->F;
    ParticlePos_t& V = myPtcl->V;
    Scalar_t ke_part(0.0);
    for(int iat=first; iat<last; iat++) {
      int itype = gid[iat];
      V[iat] += dtover2m[itype]*F[iat];
      ke_part += mass[itype]*dot(V[iat],V[iat]);
    }
    ke+=ke_part;
  }

  E[KINETIC_ENERGY] = 0.5*ke;
  E[TEMPERATURE] = TwoOverDOF*eV_to_Kelvin*E[KINETIC_ENERGY];
  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY];

  Counter++;
  E[NOW] += Dt;

  return (E[TEMPERATURE] > KE_min_K);
}

void EVNVelocityVerlet::applyBConds(bool wrapping) 
{

  myPot->setCurrentPos(myPtcl->R, wrapping);
  Counter = 0;
}



/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: EVNVelocityVerlet.OMP.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
