//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/LatticeRelaxation.h"
#include "Propagators/CGPropagator.h"
#include "Propagators/NoPropagator.h"
#include "ParticleBase/ParticleUtility.h"
#include "Propagators/ConstraintBase.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
using namespace OHMMS;

/** Argument passed to Interaction_t::setCurrentPos
 *
 * Any update in this object uses a fixed position 
 * in the lattice unit and it is safe to use the cell
 * information compiled by the engine that moves the
 * particle positions.
 */
const bool DoNotUpdateCell = false;

/** default constructor
 */
LatticeRelaxation::LatticeRelaxation(): FixedPtclPosition(false), 
			UseCG(true),
			Dt(2.0), 
			Friction(0.01)
{

  CGmode = 0; //Polak-Ribiere
  deltaEmax = 1.0e-6;
  gdotgmax = 1e-6;
  fmax = 1e-3;
  hfactor_AA = 10;
  MaxDispl = 0.2; 

  gdotg_ind = E.add("Dot(F,F)");
  gdoth_ind = E.add("Dot(CG,F)");
  pv_ind = E.add("PV");
  exx_ind = E.add("exx");
  eyy_ind = E.add("eyy");
  ezz_ind = E.add("ezz");
  exy_ind = E.add("exy");
  eyz_ind = E.add("eyz");
  ezx_ind = E.add("ezx");

  m_param.add(UseCG,"cg","bool");
  m_param.add(FixedPtclPosition,"fixedpos","bool");
  m_param.add(MinCGSteps,"minsteps","int");
  m_param.add(hfactor_AA,"pressurefactor","AA");

  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");

  //stopping criterion
  m_param.add(MaxDispl,"maxdispl","AA");
  m_param.add(deltaEmax,"energytol","none");
  m_param.add(gdotgmax,"gradtol","none");
  m_param.add(fmax,"forcetol","none");

  //add Lattice-variable
  VC.addInputParameters(m_param);
}

/** destructor
 *
 * Delete the propagator for particle position.
 */
LatticeRelaxation::~LatticeRelaxation() {
  delete PtclPropagator;
}


void 
LatticeRelaxation::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;

  VC.setConstraints();

  if(!Initialized) { 

    initializeBase(ptclin,potin,makecopy);

    if(FixedPtclPosition) {
      PtclPropagator = new NoPropagator();
      myPtcl->convert2Unit(myPtcl->R);
    } else {
      PtclPropagator = new CGPropagator();
    }

    PtclPropagator->initialize(ptclin,potin,false);
    Initialized = true; 
  }

  hfactor = hfactor_AA*sqrt(myPtcl->Lattice.Volume)/3.0;
  cout << "hfactor (AA) " << hfactor_AA << " " << hfactor << endl;

  //Steepest-decent method uses the variables differently
  if(!UseCG) {
    //usual time coefficients for the Verlet algorithm
    Scalar_t tmp = Friction*Dt*0.5e0;
    c0 = 2.0e0/(1.0e0+tmp);
    c1 = 1.0e0 - c0;
    c2 = Dt*Dt/(1.0e0+tmp);

    tH = myPtcl->Lattice.R;
    //H[0]
    Ho = transpose(tH);
    //H[-dt]
    Hcg = Ho; 

    //H[-2*dt]
    Hg0 = Ho;
    Sigma = myPtcl->Lattice.G;

    myPot->setCurrentPos(myPtcl->R,false,false);
    evalForceAndStress(myPtcl->F);
  }

  d_pe_save = 0.0;
}

void 
LatticeRelaxation::setTemperature(bool rescale, Scalar_t tref) {
  PtclPropagator->setTemperature(rescale,tref);
} 

bool LatticeRelaxation::propagate(Interaction_t * pot) {

  if(!Initialized) {initialize(myPtcl,pot,false);}
 
  E[NOW] += 1;
 
  bool moving = PtclPropagator->propagate(myPot);

  if(UseCG) {
    moving |= relaxByCG();
  } else {
    moving |= relaxBySD();
  }

  if(UseCG && !moving) {
    LOGMSG("Complete the relaxation with " << NumLineMin << " line minimization")
  }
  return moving;
}

bool LatticeRelaxation::relaxBySD() {

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);

  Hg0 = Hcg;
  Hcg = Ho;
  Ho = c0*Hcg+c1*Hg0+c2*Hg;
  Hg0 -= Ho;

  tH = transpose(Ho);
  myPtcl->updateLayout(tH);

  myPot->setCurrentPos(myPtcl->R,false,DoNotUpdateCell);
  evalForceAndStress(F);
  E[FREE_ENERGY] = E[POT_ENERGY]+E[pv_ind];
  Counter++; // increment the internal counter
  E[NOW] += Dt;

  return (Hg0.max()>fmax);
}

bool LatticeRelaxation::relaxByCG() {

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);

  if(RestartCG) { 
    applyBConds(true);
    Sigma = myPtcl->Lattice.G;
    evalForceAndStress(F);
    E[FREE_ENERGY] = E[POT_ENERGY]+E[pv_ind];
    tH = myPtcl->Lattice.R;
    Ho = transpose(tH);
    Hcg = Hg;
    Hg0 = Hg;

    RestartCG = false;
    myPtcl->setLayoutUpdateStride(1000);
  }

  // Perform lineminimization
  if(!lineoptimization()) {
    RestartCG = true;
    return false;
  }

  Tensor_t dh = Lambda*Hcg;
  Scalar_t dl = dh.max();
  if(dl>MaxDispl) {
    ERRORMSG("Too big Lambda = " << Lambda)
    return false;
  }

  Ho += dh;
  tH = transpose(Ho);
  myPtcl->updateLayout(tH);
  myPot->setCurrentPos(myPtcl->R,false,DoNotUpdateCell);
  evalForceAndStress(F);

  E[FREE_ENERGY] = E[POT_ENERGY]+E[pv_ind];
  gdotg0 = gdotg;
  gdotg = OhmmsBlas::dot(D2,Hg.data(),Hg.data());
  E[gdotg_ind] = gdotg;

  E[KINETIC_ENERGY] = VC.P_res*pressure_to_kbar;

  if(converged(VC.Stress_res.max())) {
    return false;
  }

  if(CGmode == 0) { // Polak-Ribiere (defualt)
    gdoth = OhmmsBlas::dot(D2,Hg.data(),Hg0.data());
    E[gdoth_ind] = gdoth;
    gamma = (gdotg - gdoth)/gdotg0; 
    Hg0 = Hg;
  } else { //Fletcher-Reeve
    gamma = gdotg/gdotg0;
  }

  if(fabs(gamma) < deltaEmax) return false;
  if(gamma > 0.1e0) {
    DEBUGMSG(" Gamma is too big: restart CG" << gamma << endl);
    RestartCG = true;
    return true;
  }

  Scalar_t dx = fabs(d_pe_save - E[FREE_ENERGY])/static_cast<Scalar_t>(myPtcl->getTotalNum());
  if(dx <= deltaEmax) {
    LOGMSG(" has reached energy tolerance criteria " << endl);
    return false; //
  }
  
  d_pe_save = E[FREE_ENERGY];

  Hcg = Hg + gamma * Hcg;

  Counter++;
  return true;
}

void 
LatticeRelaxation::evalForceAndStress(ParticlePos_t& f) {

  E[POT_ENERGY] = myPot->evalForce(f,1);
  myConstraint->apply(f);
  
  Scalar_t volinv = 1.0/myPtcl->Lattice.Volume;
  VC.Stress_ind = volinv*myPot->getStress();
  VC.Stress_res = VC.Stress_ind - VC.Stress_ext;

  VC.P_res = trace(VC.Stress_res);
  
  Hg = hfactor*dot(VC.Stress_res,Sigma);

  VC.applyConstraints(Hg);

  E[exx_ind] = VC.Stress_ind(0,0);
  E[eyy_ind] = VC.Stress_ind(1,1);
  E[ezz_ind] = VC.Stress_ind(2,2);
  E[exy_ind] = VC.Stress_ind(0,1);
  E[eyz_ind] = VC.Stress_ind(1,2);
  E[ezx_ind] = VC.Stress_ind(0,2);

  E[pv_ind] = myPtcl->Lattice.Volume*VC.P_ref;
}

void LatticeRelaxation::applyBConds(bool wrapping) 
{
  PtclPropagator->applyBConds(wrapping);
  if(UseCG) {
    RestartCG = true;
  } else {
    Hcg = Ho; 
    Hg0 = Ho;
  }
}

/** evalaute energy for R + dl*CG 
 *@param dl a proposed move along the Conjuate Gradient direction
 *@return the energy for the proposed position
 *
 *R and CG are fixed vectors during a line minimization.
 *LatticeRelaxation utilizes an efficient update method of the NNEngine without re-calculating
 *the list. This is justfied since the lineminimization involves very small displacement
 *from the current position. It is important to calculate the displacement accurately
 *for accurate neighbor relations. 
 */
LatticeRelaxation::Return_t 
LatticeRelaxation::Func(LatticeRelaxation::Return_t dl) {
   NumLineMin++;
   Tensor_t dh = dl*Hcg;
   if(dh.max() > MaxDispl) return E[FREE_ENERGY]+1.0;

   Tensor_t h = transpose(Ho +dh);
   myPtcl->updateLayout(h);

   myPot->setCurrentPos(myPtcl->R,false,DoNotUpdateCell);
   Scalar_t e = myPot->evalEnergy();
   e += myPtcl->Lattice.Volume*VC.P_ref;
   myPtcl->updateLayout(tH);
   return e;
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: LatticeRelaxation.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
