//////////////////////////////////////////////////////////////////
// (c) Copyright 2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_CGLATTICE_ANYPROPAGATOR_H
#define OHMMS_CGLATTICE_ANYPROPAGATOR_H

#include "Propagators/PropagatorBase.h"
#include "Optimize/CGOptimization.h"

namespace OHMMS {

  template<class TT>
  struct LatticeVariables {

    typedef typename TT::Type_t T;

    enum {D2 = TT::Size};

    T P_res, P_ref, P_ref_kB;
    TT Stress_ext, Stress_ext_kB;
    TT Stress_ind;
    TT Stress_res;
    TT CellMask;

    inline LatticeVariables(){ 
      Stress_ext_kB = 1.1e12;
      CellMask = 1.0;
    }

    virtual ~LatticeVariables(){ }

    void setConstraints() {
      if(Stress_ext_kB(0,0)>1e12) {
	P_ref = P_ref_kB/pressure_to_kbar;
	Stress_ext = 0.0;
	Stress_ext.diagonal(P_ref);
      } else {
	T pconv = 1.0/pressure_to_kbar;
	Stress_ext = pconv*Stress_ext_kB; 
	P_ref = trace(Stress_ext);
	P_ref_kB = P_ref*pressure_to_kbar;
      }
    }

    void addInputParameters(ParameterSet& params) {
      params.add(P_ref_kB,"pressure","kBar");
      params.add(Stress_ext_kB,"stress","tensor-kBar");
      params.add(CellMask,"conditions","tensor");
    }

    inline void applyConstraints(TT& h) {
      for(int i=0;i<D2;i++) h(i) *= CellMask(i);
    }
  };

  /** Propagator for the unit cell
   */
  class 
  LatticeRelaxation: 
	     public PropagatorBase, 
	     private CGOptimization<PropagatorBase::Tensor_t>
  {

  public:

    enum {D2 = Tensor_t::Size};

    LatticeRelaxation();
    ~LatticeRelaxation();
 
    // implementing virutal functions
    void initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy = false);
    void setTemperature(bool rescale, Scalar_t tref); 
    bool propagate(Interaction_t*);
    void applyBConds(bool wrapping);

  protected:

    /** implement the virtual function of CGOptimization
     *@return the free energy with a modified lattice Ho + dl*Hcg.
     */
    Return_t Func(Return_t dl);

    /** control parameter for the lattice relaxation
     *
     * \if true
     * Conjugate minimizations are performed.
     * \else
     * Steepest decent minimizations are perfoemd.
     */
    bool UseCG;

    /** control parameter for particle positions
     * 
     * \if true,
     * The scaled positions are fixed. If true, use NoPropagator
     * \else
     * The positions are relaxed using CGPropagator.
     * \endlf
     */
    bool FixedPtclPosition;

    //internal indices to access a property of E
    size_type gdotg_ind, gdoth_ind, pv_ind;
    size_type exx_ind,eyy_ind,ezz_ind,exy_ind,eyz_ind,ezx_ind;

    //accepted parameters
    ///time step: used by the SD minimization
    Scalar_t Dt;
    ///Friction: used by the SD minimization
    Scalar_t Friction;
    ///Scaling factor for the lattice dynamics
    Scalar_t hfactor_AA;
    ///Maximum changes in the lattice component
    Scalar_t MaxDispl;

    //internal parameters
    Scalar_t lambdasum;
    Scalar_t hfactor;
    Scalar_t c0,c1,c2;

    Tensor_t Ho, tH, Sigma;
    Tensor_t Hg, Hg0, Hcg;

    /** the propagator to move the particle positions at a fixed lattice.
     */
    PropagatorBase* PtclPropagator;

    /** the stress-related variables */
    LatticeVariables<Tensor_t>  VC;

    bool relaxByCG();
    bool relaxBySD();
    void evalForceAndStress(ParticlePos_t& f);
  };
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: LatticeRelaxation.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
