//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Propagators/PropagatorBase.h"
#include "ParticleBase/ParticleUtility.h"
#include "Propagators/ConstraintBase.h"
#include "Message/OpenMP.h"
using namespace OHMMS;

void PropagatorBase::evalKineticEnergy(){

  Scalar_t kesum = 0.0e0;
#ifdef ENABLE_OPENMP
#pragma omp parallel reduction(+:kesum)
  {
    const ParticlePos_t& Vref = myPtcl->V;
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];
    for(int i=first; i<last; ++i) {
      kesum += (dot(Vref[i],Vref[i]))*mass[myPtcl->GroupID[i]];
    }
  }
  kesum *= 0.5;
#else
    const ParticlePos_t& Vref = myPtcl->V;
    Scalar_t factor = 0.5e0; // 1/2 factor for KE 
    if(MultiSpecies) {
      for(int i=0; i<myPtcl->getLocalNum(); ++i) {
        kesum += (dot(Vref[i],Vref[i]))*mass[myPtcl->GroupID[i]];
      }
      kesum *= factor;
    } else {
      for(int i=0; i<myPtcl->getLocalNum(); ++i) {
        kesum += dot(Vref[i],Vref[i]);
      }
      kesum *= factor*mass[0];
    }
#endif
    
  ///////////////
  //\todo MPI sum could be an issue
  ///////////////
  E[KINETIC_ENERGY] = myPtcl->collect(kesum);
  //E[KineticEnergy] = kesum; //d_ke = kesum;
  E[TEMPERATURE] =E[KINETIC_ENERGY]*TwoOverDOF*eV_to_Kelvin;
}

PropagatorBase::Tensor_t 
PropagatorBase::evalKineticPressure(const ParticlePos_t& vref) {

  Tensor_t ke;
#ifdef ENABLE_OPENMP
#pragma omp parallel
  {
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];
    const ParticleIndex_t& gid=myPtcl->GroupID;
    Tensor_t ke_loc;
    for(int i=first; i<last; ++i) {
      ke_loc += mass[gid[i]]*outerProduct(vref[i],vref[i]);
    }
#pragma omp critical
    {
      ke *= ke_loc;
    }
  }
#else
  if(MultiSpecies) {
    for(int i=0; i<myPtcl->getLocalNum(); ++i) {
      ke += mass[myPtcl->GroupID[i]]*outerProduct(vref[i],vref[i]);
    }
  } else {
    for(int i=0; i<myPtcl->getLocalNum(); ++i) {
      ke += outerProduct(vref[i],vref[i]);
    }
    ke *= mass[0];
  }
#endif

  Scalar_t kesum = 0.5*trace(ke);
  E[KINETIC_ENERGY] = myPtcl->collect(kesum);
  E[TEMPERATURE] = kesum*TwoOverDOF*eV_to_Kelvin;
  return ke;
}

bool PropagatorBase::setVelocity(Scalar_t temp) {

  //if(temp<zeroKelvin || E[Temperature]>zeroKelvin || E[Now] > 0.0) return false; 
#pragma omp parallel
  {
    Scalar_t dim_v = static_cast<Scalar_t>(OHMMS_DIM);
    Scalar_t vMag = sqrt(temp*Kelvin_to_eV*dim_v+1e-16);
    ParticlePos_t& Vref = myPtcl->V;
    omp_int_t ip = omp_get_thread_num();
    size_type first = LocIndex0[ip];
    size_type last = LocIndex0[ip+1];
#ifdef HAVE_LIBBOOST
    RandomVector<SingleParticlePos_t, BoostRandom<Scalar_t> > rvg(*reng[ip]);
    RandomNormSequence<ParticlePos_t, BoostRandom<Scalar_t> >::apply(Vref, rvg, first, last);
#else
    RandomVector<SingleParticlePos_t, RandomGenerator_t> rvg(Random);
    RandomNormSequence<ParticlePos_t, RandomGenerator_t>::apply(Vref, rvg, first, last);
#endif
    std::vector<Scalar_t> minv_t = minv;
    for(size_type i=0; i<minv_t.size(); i++) minv_t[i] = vMag*sqrt(minv_t[i]);
    if(MultiSpecies) 
      for(size_type iat=first; iat<last; iat++) {
        Vref[iat] *= minv_t[myPtcl->GroupID[iat]];
      }
    else {
      Scalar_t fm = minv_t[0];
      for(size_type iat=first; iat<last; iat++) {
        Vref[iat] *= fm;
      }
    }
  }

#ifdef ENABLE_OPENMP
  //Skip drift checking or freeze
  E[TEMPERATURE] = temp;
  E[KINETIC_ENERGY] = E[TEMPERATURE]/TwoOverDOF*Kelvin_to_eV;
#else
  //int LocalNum  = myPtcl->getLocalNum();
  //Scalar_t dim_v = static_cast<Scalar_t>(OHMMS_DIM);
  //Scalar_t vMag = sqrt(temp*Kelvin_to_eV*dim_v+1e-16);
  //RandomVector<SingleParticlePos_t, RandomGenerator_t> rvg(Random);
  //// generate a sequence of normalized random vectors
  //RandomNormSequence<ParticlePos_t, RandomGenerator_t>::apply(Vref, rvg);
  //if(MultiSpecies) {
  //  std::vector<Scalar_t> minv_t = minv;
  //  for(int i=0; i<mt.size(); i++) minv_t[i] = vMag*sqrt(minv_t[i]);
  //  for(int iat=0; iat< LocalNum; ++iat) 
  //    Vref[i] *= minv_t[myPtcl->GroupID[i]];
  //} else {
  //  vMag *= sqrt(minv[0]);
  //  Vref *= vMag; // scale by the 1/sqrt(m)
  //}
  // apply constraint
  myConstraint->removeDrift(myPtcl->V);
  myConstraint->apply(myPtcl->V);
  evalKineticEnergy();
  Scalar_t fac = sqrt(temp/E[TEMPERATURE]);
  myPtcl->V *= fac;
  E[KINETIC_ENERGY]*=fac*fac;
  E[TEMPERATURE]*=fac*fac;
#endif

  return true;
}

void PropagatorBase::evalAccel(ParticlePos_t& f){

  myConstraint->apply(f);
  if(MultiSpecies) {
    for(int i=0; i< f.size(); ++i)  f[i] = minv[myPtcl->GroupID[i]]*f[i];
  } else {
    f *= minv[0];
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: PropagatorBase.1.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
