//////////////////////////////////////////////////////////////////
// (c) Copyright 2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

/** @file TPLangevinBase.h
 * Base class for TP-X propagators based on Trotter factorization (LeapFrog algorithm).
 */
#ifndef OHHMS_TPLANGEVINBASE_H
#define OHHMS_TPLANGEVINBASE_H

#include "Particle/ParticleSet.h"
#include "Utilities/RandomGenerator.h"

namespace OHMMS {

  /** Base class to perform TP(N,mu) simulation using Langevin thermostat
   *
   * Declare and initialize common variables for TP(N,mu) propagators.
   */
  struct TPLangevinBase {

    ///constructor
    TPLangevinBase();

    ///virtual destructor
    virtual ~TPLangevinBase();

    typedef ParticleSet::Scalar_t T;

    ///index for the internal pressure
    std::size_t pressure_ind;

    ///index for PV
    std::size_t pv_ind;

    //index for the kinetic energy of the pressure
    std::size_t pk_ind;

    std::size_t volume_ind;

    ///reference temperature
    T T_ref;

    ///External pressure in kB
    T P_ref_kB;

    ///Frequency of the volume fluctuation \f$\omega_b\f$ in 1/fsec
    T VolFreq;

    ///Pressure from potential \f$\sum_{i}{\bf r}_i\cdot{\bf f}_i\f$ (eV)
    T P_p;

    ///Pressure from kinetic f\$\sum_{i}{\bf r}_i\cdot{\bf f}_i\f$ (eV)
    T  P_k;

    ///Induced pressure (P_p+P_k)/(DV)- dEdVol eV/AA^3
    T P_ind;

    ///External pressure in eV/AA^3
    T P_ref;

    ///time step in fsec
    T Dt;

    ///friction for the particle velocities in fsec-1
    T Friction;

    ///friction for the lattice momentum in fsec-1
    T VolFriction;

    ///Dynamic variables associated with the volume
    T Volume;

    ///The mass of the lattice
    T VolMass, VolMassInv;

    ///D/degrees-of-freedom
    T DoverDOF;

    ///dt/2
    T dtover2;

    ///(dt/2)/VolMass
    T dtover2W;

    ///dt/VolMass
    T dtoverW;

    ///Coefficient for Friction*dt/2
    T chi_vel;

    ///Coefficient for VolFriction*dt/2
    T chi_vol;

    ///Coefficient for (1.0-dt*chi_vel)
    T cf_vel;

    ///Coefficient for (1.0-dt*chi_vol)
    T cf_vol;

    ///Gaussian norm for dynamic variables
    T SigmaVol;

    ///boolean to turn on random fluctuations to the velocities
    bool AddNoise;

    std::vector<T> SigmaV;
    ParticleSet::ParticlePos_t Vr;

    GaussianRandom randomSource;
    void setSigma(T Temp, const std::vector<T>& mi); 
    void initParameters(RecordNamedProperty<T>& properties,ParameterSet& params);
    void checkParameters(const std::string& name); 
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: TPLangevinBase.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
