//////////////////////////////////////////////////////////////////
// (c) Copyright 2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TPNHydroVerlet.h"
#include "Propagators/VerletPropagator.h"
#include "Potentials/PotentialBase.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

TPNHydroVerlet::TPNHydroVerlet()
  : T_ref(0.0), P_ref(0.0), Dt(1.0), 
    Vol(1.0), VolV(0.0),VolV1(0.0),VolF(0),VolMass(10000.), 
    Chi(0.0),ChiV(0.0), ChiV1(0.0),ChiMass(100000.)
{
  pressure_ind = E.add("Pressure");
  pv_ind = E.add("P*dV");

  m_param.add(Dt,"timestep","fsec");

  m_param.add(T_ref, "temperature","K");
  m_param.add(P_ref_kB,"pressure","kB");

  //Arbitrary units at this point
  m_param.add(VolMass,"volmass","au");
  m_param.add(ChiMass,"Q","au");

  m_param.add(Vol,"volume","AA^3");
  m_param.add(VolV,"volume-vel","AA^3");

  m_param.add(Chi,"chi","au");
  m_param.add(ChiV,"chi-vel","au");
}

void 
TPNHydroVerlet::initialize(Particle_t* ptclin, PotentialBase* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;

  KE_ref = Kelvin_to_eV*T_ref;
  P_ref = P_ref_kB/pressure_to_kbar;

  dtover2 = Dt*0.5;
  dtsqover2 = Dt*Dt*0.5;

  initializeBase(ptclin,potin,makecopy);

  if(!Initialized) {

    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);

    R1.setUnit(PosUnit::CartesianUnit);
    R1.resize(myPtcl->R.size());

    F1.setUnit(PosUnit::CartesianUnit);
    F1.resize(myPtcl->R.size());
    Initialized = true;
  }

  /*! set the velocity to a reference temperature only when the clock is zero    
   * When the clock E[NOW] > 0, it assumes that the simulation restarts from
   * the previous configuration.
   */
  if(ParameterModified) {
    Vol = myPtcl->Lattice.Volume;
    if(RescaleTemperature) {
      Chi = 0; ChiV = 0.0; ChiV1 = 0.0;
      VolV = 0.0; VolV1 = 0.0; VolF = 0.0;
    }
    if(T_ref>zeroKelvin) setTemperature(RescaleTemperature,T_ref);
  } else
    applyBConds(true);

  DoverDOF = OHMMS_DIM/DOF;
  TwoPlusDoverDOF = 2.0+DoverDOF;
  ParameterModified = false;
}

void 
TPNHydroVerlet::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
  } else {
    V=0.0;
  }

  ScaledV = 1.0;
  applyBConds(true);
}


bool 
TPNHydroVerlet::propagate(PotentialBase* pot) {

  if(!Initialized) initialize(myPtcl,pot,false);

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  ParticlePos_t& dR = *myPtcl->getVectorAttrib(displacement_tag);
  ParticlePos_t& R = myPtcl->R;
  const Scalar_t oneoverdim = 1.0/OHMMS_DIM;

  Scalar_t twoKE = 2.0e0*E[KINETIC_ENERGY];

  OneOverDV=oneoverdim/Vol;
  E[pressure_ind] = (P_v + twoKE)*OneOverDV-dEdVol;

  //Force on the volume
  VolF = OHMMS_DIM*Vol*(E[pressure_ind] -  P_ref)+DoverDOF*twoKE;

  //Acceleration on the thermostat (don't know why but am sticking to the paper)
  ChiA = (twoKE+VolMass*VolV*VolV-(DOF+1.0)*KE_ref)/ChiMass;

  cout << "P_ind VolF, ChiA " << E[pressure_ind] << " " << VolF << " " << ChiA << endl;

  //the changes of Volume and thermostat variable
  Scalar_t dVol = VolV*Dt+(VolF/VolMass-VolV*ChiV)*dtsqover2;
  Scalar_t dChi = ChiV*Dt+ChiA*dtsqover2;
  
  //R += V*Dt+(F-ChiV*V-pr_factor*V)*dtsqover2;
  Scalar_t vfact = Dt-ChiV*dtsqover2-TwoPlusDoverDOF*dtsqover2;  
  cv_scale = exp(dVol);
  cout << "Volume scale " << cv_scale << endl;
  Scalar_t c1 = vfact*cv_scale;
  Scalar_t c2 = cv_scale*dtsqover2;

  //Save the old position
  R1 = R;
  //Update the position
  R = cv_scale*R1+c1*V+c2*F;
  Chi += dChi;

  //Scale the lattice by (Vol(t)/Vol(0))^(1/D)
  Scalar_t Vol0 = Vol;
  Vol += dVol;
  Scalar_t vscale =pow(Vol/Vol0, oneoverdim);
  Tensor_t newlat = vscale*myPtcl->Lattice.R;
  myPtcl->updateLayout(newlat);
  evalForceAndPressure(F1,Eloc);

  //R1 = V +(F-V*ChiV-(TwoPlusDoverDOF)*V*VolV)*Dt/2 + F1*Dt/2;   
  vscale = 1.0-dtover2*(ChiV+TwoPlusDoverDOF*VolV);
  R1 = vscale*V+dtover2*(F+F1);

  //update the velocity iteratively
  updateVelocities(V);

  F = F1;

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  ///pressure criterion
  return  (E[TEMPERATURE] > KE_min_K );
}

void TPNHydroVerlet::updateVelocities(ParticlePos_t& V) {

  Scalar_t vchi0 = ChiV1+2.0*ChiA*Dt;
  Scalar_t vchi = vchi0;

  Scalar_t vvol0 = VolV1+2.0*VolF/VolMass*Dt;
  Scalar_t vvol = vvol0;

  Scalar_t vvol_fixed = VolV+(VolF/VolMass-VolV*ChiV)*dtover2;
  Scalar_t vchi_fixed = ChiV+ChiA*dtover2;
  Scalar_t gamma = 1.0/cv_scale;
  Scalar_t cv = 1.0/(gamma+(vchi+TwoPlusDoverDOF*vvol)*dtover2);  
  V = cv*R1;
  evalKineticEnergy();

  bool notconverged = true;
  double twoKE = 2.0*E[KINETIC_ENERGY];

  //Save to the next step
  ChiV1 = ChiV;
  VolV1 = VolV;

  while(notconverged) {

    Scalar_t p_ind = (P_v + twoKE)*OneOverDV-dEdVol;
    Scalar_t fvol = OHMMS_DIM*Vol*(p_ind-P_ref)+DoverDOF*twoKE;    
    //Scalar_t fvol = P_v+(1.0+DoverDOF)*twoKE;
    Scalar_t achi = (twoKE+VolMass*vvol*vvol-(DOF+1.0)*KE_ref)/ChiMass;

    vchi = vchi_fixed + achi*dtover2;
    vvol = vvol_fixed+(fvol/VolMass-vchi*vvol)*dtover2;

    cv = 1.0/(gamma+(vchi+TwoPlusDoverDOF*vvol)*dtover2);  
    V = cv*R1;
    evalKineticEnergy();
    twoKE = 2.0*E[KINETIC_ENERGY];

    if(fabs((vchi-vchi0)/vchi0)<1e-6) notconverged = false;
    vchi0 = vchi;
    vvol0 = vvol;
  }

  //VolF = P_v+(1.0+DoverDOF)*twoKE;
  //ChiA = (twoKE+VolMass*vvol*vvol-(DOF+1.0)*KE_ref)/ChiMass;
  VolV = vvol;
  ChiV = vchi;
}

void 
TPNHydroVerlet::applyBConds(bool wrapping) {

  myPot->setCurrentPos(myPtcl->R,wrapping);

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  evalForceAndPressure(F,Eloc);
}

void TPNHydroVerlet::evalForceAndPressure(ParticlePos_t& F, ParticleScalar_t& Eloc) {

  myPot->setCurrentPos(myPtcl->R,false);
  Tensor_t stress;
  E[POT_ENERGY] = myPot->evalForceAndPressure(F,Eloc,stress,dEdVol);

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 273 $   $Date: 2005-01-17 16:13:53 -0600 (Mon, 17 Jan 2005) $
 * $Id: TPNHydroVerlet.cpp 273 2005-01-17 22:13:53Z jnkim $ 
 ***************************************************************************/
