//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TVNLangevin.h"
#include "ParticleBase/ParticleUtility.h"
#include "Message/OpenMP.h"
#include <limits>
using namespace OHMMS;

TVNLangevin::TVNLangevin(): T_ref(0.0), Dt(1.0), Friction(0.05)
{ 
  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
}

void 
TVNLangevin::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  Counter = 0;

  if(!Initialized) {
    initializeBase(ptclin,potin,makecopy);
    setSigma(T_ref, Dt, Friction);
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addRestartAttrib(velocity_tag);
    myPtcl->addRestartAttrib(force_tag);
    Initialized = true;
  }
  
  if(ParameterModified) 
    setTemperature(RescaleTemperature,T_ref);
  else
    applyBConds(true);

  if(E[NOW]<numeric_limits<Scalar_t>::epsilon()) {
    myPot->evalForce(myPtcl->F, 0);
  }

//#ifdef HAVE_LIBMCF
//  if(recorder.empty()) recorder.resize(omp_get_max_threads(),0);
//#pragma omp parallel
//  {
//    int ip = omp_get_thread_num();
//    int first = LocIndex0[ip];
//    int last = LocIndex0[ip+1];
//    if(recorder[ip]==0) {
//      recorder[ip]=new MCFRecordPosAttrib(myPtcl->R,first,last);
//    }
//    char fname[64];
//    sprintf(fname,"mcftest.%03d",ip);
//    recorder[ip]->reset(fname);
//  }
//#endif
  ParameterModified = false;
}

void 
TVNLangevin::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  if(T_ref>zeroKelvin) {
    setSigma(T_ref, Dt, Friction);
    scaleTemperature(rescale,T_ref);
    ScaledV = 1.0;
    applyBConds(true);
  }
}

bool  TVNLangevin::propagate(Interaction_t* pot) {

  if(Counter == UpdateStride) {
    Counter = 0; 
    myPot->setCurrentPos(myPtcl->R); 
  }

#pragma omp parallel
  {                                                                    
    int ip = omp_get_thread_num();                                    
    int first = LocIndex0[ip];                                       
    int last = LocIndex0[ip+1];                                     
                                                                   
    //Random Number Generators usually are not thread-safe
    //engine to generate a set of correlated vectors, CorrelatedVectorGenerator
    RandomVector<SingleParticlePos_t, BoostRandom<Scalar_t> > CVG(*reng[ip]);
    register SingleParticlePos_t r1, r2;
    const ParticleIndex_t& gid = myPtcl->GroupID;
    ParticlePos_t::const_iterator F= myPtcl->F.begin()+first;
    ParticlePos_t::iterator V = myPtcl->V.begin()+first;
    ParticlePos_t::iterator dR = myPtcl->dR.begin()+first;
#pragma ivdep
    for(int iat=first; iat<last; iat++) {
      int itype = gid[iat];
      CVG.get(r1,r2); // correlated random vectors 
      r1*=OneOverSqrtM[itype];
      r2*=OneOverSqrtM[itype];
      SingleParticlePos_t a(minv[itype]*(*F));
      SingleParticlePos_t v(*V);
      *dR = SigmaR*r1+c1dt*v+c2dt2*a;
      *V  = CorrRVx*r1+CorrRVy*r2+c0*v+c12dt*a;
      ++F; ++V; ++dR;
    }

  }

  //2 to use new dR for force evaluation
  E[POT_ENERGY] = pot->evalForce(myPtcl->F,2); 
 
  Scalar_t ke=0.0;
//#pragma omp parallel
#pragma omp parallel reduction(+:ke)
  {                                                                    
    int ip = omp_get_thread_num();                                    
    int first = LocIndex0[ip];                                       
    int last = LocIndex0[ip+1];                                     

    ParticlePos_t::const_iterator dR = myPtcl->dR.begin()+first;
    ParticlePos_t::iterator R = myPtcl->R.begin()+first;
#pragma ivdep
    for(int iat=first; iat<last; ++iat) {
      *R += *dR; ++R; ++dR;
    }

    const ParticleIndex_t& gid = myPtcl->GroupID;
    ParticlePos_t::const_iterator F= myPtcl->F.begin()+first;
    ParticlePos_t::iterator V = myPtcl->V.begin()+first;
#pragma ivdep
    for(int iat=first; iat<last; iat++) {
      int itype = gid[iat];
      SingleParticlePos_t v(*V+c2dtoverm[itype]*(*F));
      ke += mass[itype]*dot(v,v);
      *V=v;
      ++F;++V;
    }
  }

  E[KINETIC_ENERGY] = 0.5*ke;
  E[TEMPERATURE] = TwoOverDOF*eV_to_Kelvin*E[KINETIC_ENERGY];
  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY];

  Counter++;
  E[NOW] += Dt;
  //return (E[TEMPERATURE] > KE_min_K);
  return true;
}


void TVNLangevin::applyBConds(bool wrapping) 
{
  myPot->setCurrentPos(myPtcl->R, wrapping);
  Counter = 0;
}

void
TVNLangevin::setSigma(Scalar_t Temp, Scalar_t dt, Scalar_t friction) {
     
  // if the friction is not set, use 0.01 
  if(friction < numeric_limits<Scalar_t>::epsilon()) 
    Friction = 0.05/dt;
  else 
    Friction=friction;
  Dt = dt;
  Scalar_t gamma = Friction*Dt;
  Scalar_t bi = 1/gamma;
  c0 = exp(-gamma);
  Scalar_t x = Temp*Kelvin_to_eV;
  SigmaR = Dt*sqrt(x*bi*(2.0e0-bi*(3.0e0-4.0e0*c0+c0*c0)));
  SigmaV = sqrt(x*(1.0e0-c0*c0));
  //Scalar_t sigrv = Dt*x*bi*(1.0e0-c0)*(1.0e0-c0)/SigmaR/SigmaV;
  Scalar_t sigrv = Dt*x*bi*(1.0e0-c0)*(1.0e0-c0)/SigmaR/SigmaV;
  CorrRVx = SigmaV*sigrv;
  CorrRVy = SigmaV*sqrt(1.0e0-sigrv*sigrv);

  OneOverSqrtM.resize(minv.size());
  for(int i=0; i<minv.size(); ++i) OneOverSqrtM[i]=sqrt(minv[i]);

  c1 = (1.0e0-c0)*bi; c1dt = c1*Dt; c2 = (1.0e0-c1)*bi; 
  Scalar_t c2dt = c2*Dt;  
  c2dt2 = c2dt*Dt; 
  c12dt = (c1-c2)*Dt;
  c2dtoverm.resize(minv.size());
  for(int i=0; i<minv.size(); ++i) c2dtoverm[i]=c2dt*minv[i];
  // if the friction is not set, use 0.01 
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: TVNLangevin.OMP.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
