//////////////////////////////////////////////////////////////////
// (c) Copyright 2003- by Jeongnim Kim, Benjamin Cho, Daniel Go, Alfonso Reina
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TVNNoseHoover.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

TVNNoseHoover::TVNNoseHoover():
  ToApplyBConds(false),  T_ref(0.0), Dt(1.0), Chi(0.0), Q(100.0) 
{ 
  nose_ke_ind = E.add("Nose-KE");
  nose_pe_ind = E.add("Nose-PE");

  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Q,"Q","fsec-1");
  m_param.add(Chi,"Chi","nose-var1");
  m_param.add(Theta,"Theta","nose-var2");
}

void 
TVNNoseHoover::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  Dtover2 = Dt/2.0;
  DtoverQ = Dt/Q;

  if(!Initialized) {

    initializeBase(ptclin,potin,makecopy);

    KE_ref = Kelvin_to_eV*T_ref;
    //Q *= amu_to_mass_unit;

    ///////////////////////////////////
    // these values are always in cartesian coordinates
    ///////////////////////////////////
    myPtcl->convert2Cart(myPtcl->R);

    ///only myPtcl->R has to be added to the position data to which boundary conditions apply
    myPtcl->addPositionAttrib(myPtcl->R);

    Initialized = true;
  }

  if(ParameterModified) 
    setTemperature(RescaleTemperature,T_ref);
  else
    applyBConds(true);

  ParameterModified=false;

}

void 
TVNNoseHoover::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
  } else {
    V=0.0;
  }
  applyBConds(true);
}

bool 
TVNNoseHoover::propagate(Interaction_t* pot) {


  if(!Initialized) initialize(myPtcl,pot,false);

  
  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  Tensor_t stress;

  // ************* degrees of freedom *************
  Scalar_t g = static_cast<Scalar_t>(OHMMS_DIM*myPtcl->getTotalNum()-1);
  ParticlePos_t& dR = *myPtcl->getVectorAttrib(displacement_tag);

  // ************ Eval velocity at t+dt/2 *************
  // V = a*V+b*F;     
  // V = (V+Dtover2*F)/(1.0+Dtover2*Chi);
  Scalar_t v0 = 1.0/(1.0+Chi*Dtover2);
  Scalar_t v1 = Dtover2/(1.0+Chi*Dtover2);

  V = v0*V+v1*F;

  // ************* Eval position at t+dt *************
  dR = Dt*V;
  myPtcl->R += dR; //Dt*V;

  // ************* Eval Theta at t+dt/2 *************
  Theta += Dtover2*Chi;

  // ************* Eval Chi at t+dt *************
  evalKineticEnergy();
  //Chi += DtoverQ*(2*E[KINETIC_ENERGY]-g*KE_ref);
  Chi += DtoverQ*(2*E[KINETIC_ENERGY]-g*KE_ref);
  
  // ************* Eval Theta at t+dt *************
  Theta += Dtover2*Chi;

  // ************* Update the force with R0 at t+dt *************
  myPot->setCurrentPos(myPtcl->R,dR);

  stress = 0;
  E[POT_ENERGY] = myPot->evalForce(F, Eloc, stress, 0);
  evalAccel(F);

  // ************* Eval velocity at t+dt *************
  v0 = 1.0/(1.0+Chi*Dtover2);
  v1 = Dtover2/(1.0+Chi*Dtover2);
  V = v0*V+v1*F;

  // ************* Eval Kinetic Energy at t+dt *************
  evalKineticEnergy();

  // ************* Eval Nose-KE and Nose-PE *************
  E[nose_ke_ind] = Chi*Chi*Q*0.5;
  E[nose_pe_ind] = g*KE_ref*Theta;

  // ************* Eval total energy *************
  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY] + E[nose_ke_ind] + E[nose_pe_ind];

  // increment the internal counter
  Counter++; 
  E[NOW] += Dt;

  return (E[TEMPERATURE] > KE_min_K);
}

void TVNNoseHoover::applyBConds(bool wrapping) 
{

  myPot->setCurrentPos(myPtcl->R,wrapping);

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  Tensor_t stress;
  E[POT_ENERGY] = myPot->evalForce(F, Eloc, stress, 0);
  evalAccel(F);
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 273 $   $Date: 2005-01-17 16:13:53 -0600 (Mon, 17 Jan 2005) $
 * $Id: TVNNoseHoover.cpp 273 2005-01-17 22:13:53Z jnkim $ 
 ***************************************************************************/
