//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <iostream>
#include <fstream>
using namespace std;

#include "QuantumBase/ElectronStructure.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/CommCreate.h"
#include "Utilities/XMLTemplate.h"
#include <libxml/xpath.h>

using namespace OHMMS;

ElectronStructure::ElectronStructure(): KSet(0), RefPtclSet(0), qmXmlNode(0){ 
  FixedFermiEnergy = false;
  ZeroTemperature = true;
  EvaluateDensityMatrix = false;
  Shifted = false;
  Gamma = true;

  SpinDeg = 2.0e0;      addParameter(SpinDeg,"spindeg","none");
  Sigma = 0.0e0;        addParameter(Sigma,"sigma","eV");
  Mu = 0.0e0;           addParameter(Mu,"fermi","eV");
  MuRef = 0.0e0;        addParameter(MuRef,"fermi_ref","eV");
  NumStates = -1;       addParameter(NumStates,"nstates","none");
  NumElectrons = 0;     addParameter(NumElectrons,"nel","none");
  NumCoreElectrons = 0; addParameter(NumCoreElectrons,"numofcore","none");

  BasisDim = 0;
  Homo = 0;
  NumKpts = 0;
  OwnKSet = false;

  //for(int idim=0; idim<OHMMS_DIM; idim++) kgrid[idim] = 1;
  //for(int idim=0; idim<OHMMS_DIM; idim++) korigin[idim] = 0.0;
}
   
ElectronStructure::~ElectronStructure(){ 
  if(OwnKSet) delete KSet;
}

bool ElectronStructure::resetKPoint(ParticleSet* ptclin) {

  typedef ParticleSet::SingleParticlePos_t SingleParticlePos_t;
  typedef ParticleSet::SingleParticleIndex_t SingleParticleIndex_t;

  bool kpartition = false;
  bool createdKpt=false;
  KPointBase* k=ptclin->getKSet();
  if(k == 0) {
    LOGMSG("\tCreating a kpoint set for a particle ")
    xmlXPathContextPtr context = xmlXPathNewContext(qmXmlNode->doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((const xmlChar*)"//kpoint",context);
    createdKpt=true;
    k = new KPointBase;
    k->ref_ = ptclin;
    ptclin->setKSet(k);
    //circular pointing: OK
    if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
      xmlNodePtr cur = result->nodesetval->nodeTab[0];
      if(k->put(result->nodesetval->nodeTab[0])) {
        kpartition = k->distributed();
	if(EvaluateDensityMatrix)  k->extend2FBZ();
	Gamma = false;
      }
    }
    xmlXPathFreeObject(result);
    xmlXPathFreeContext(context);
  }

  KSet=k;

  if(KSet->size()) {
    if(NumKpts < KSet->size()) {
      NumKpts = KSet->size();
      Occ.resize(NumKpts,NumStates);
      E.resize(NumKpts,NumStates);
      LOGMSG("\tThe total number of k-points " << NumKpts)
    }
  } else {
    KSet->addKpt(SingleParticlePos_t(0,0,0),1);
    Gamma = true;
  }


  Occ = 0.0e0;
  E = 0.0e0;

  if(createdKpt && Controller->master())  {
    ESOutput testout(*this);
    testout.reset("kpoints");
    testout.report(-1);
  }

  return kpartition;
}

void 
ElectronStructure::setESParameters(ParticleSet* ptclin) {

  RefPtclSet = ptclin;
  NumElectrons = 0.0;
  BasisDim = 0;
  bool reduce_nstates = false;
  int i=0;
  int first=0, last=0;
#if defined(HAVE_LIBXML2)
  ///process input tree to get the values
  if(qmXmlNode == 0) return;
  xmlNodePtr cur = qmXmlNode->xmlChildrenNode;
  while (cur != NULL) {
    string cname((const char*)(cur->name));
   if (cname == "parameter" || cname == "PARAMETER") {
      string aname =  (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
      if(aname == "nstates") {
	putContent(NumStates,cur);
	if(NumStates == 0) reduce_nstates = true;
      } else if(aname == "numel") {
	putContent(NumElectrons,cur);
      } else if(aname == "sigma") {
	putContent(Sigma,cur);
	if(Sigma>0) ZeroTemperature = false;
      } else if(aname == "eshift") {
	putContent(MuRef,cur);
	Shifted = true;
      } else if(aname == "eref") {
	putContent(MuRef,cur);
      } else if(aname == "eig_lower") {
	putContent(first,cur);
      } else if(aname == "eig_upper") {
	putContent(last,cur);
      }
    }
    cur = cur->next;
  }

#else
  ///using linear vector (worst case)
  while(i<m_options.size()) {
    if(m_options[i].find("nstates") < m_options[i].size()) {
      NumStates = atoi(m_options[++i].c_str());
    } else if(m_options[i].find("nel") < m_options[i].size()) {
      NumElectrons = atof(m_options[++i].c_str());
    } else if(m_options[i].find("sigma") < m_options[i].size()) {
      Sigma = atof(m_options[++i].c_str());
      ZeroTemperature = false;
    } else if(m_options[i].find("reduce") < m_options[i].size()) {
      reduce_nstates = true;
    } else if(m_options[i].find("eref") < m_options[i].size()) {
      MuRef = atof(m_options[++i].c_str());
      Shifted = true;
    } else if(m_options[i].find("kpoint") < m_options[i].size()) {
      Gamma = false;
      int ik=i+1;
      if(m_options[ik].find("grid") < m_options[ik].size()) {
	for(int idim=0; idim<OHMMS_DIM; idim++) {
	  kgrid[idim] = atoi(m_options[++ik].c_str());
	}
	ik++;
	while(ik<m_options.size()) {
	  if(m_options[ik].find("origin")<m_options[ik].size()) {
	    for(int idim=0; idim<OHMMS_DIM; idim++) {
	      korigin[idim] = atof(m_options[++ik].c_str());
	    }
	  }
	  ik++;
	}
      } else if(m_options[ik].find("file")<m_options[ik].size()) {
	KFile = m_options[++ik];
      } else
	KFile = m_options[++i];
    }

    i++;
  }
#endif

  //The number of states are not given. Assume that we will use the number of els.
  //if(NumStates<0) reduce_nstates=true;
  if(reduce_nstates) {
    WARNMSG("Minimum number of states determined by total el numbers")
  }

  typedef ParticleSet::ParticleScalar_t ParticleScalar_t;
  SpeciesSet& species(ptclin->getSpeciesSet());
  int qindex(species.addAttrib("q"));
  for(int iat=0; iat<ptclin->getLocalNum(); iat++) {
    NumElectrons += species(qindex,ptclin->GroupID[iat]);
  }

  if(BasisDim == 0) {
    qindex = species.addAttrib("ns");
    for(int iat=0; iat<ptclin->getLocalNum(); iat++) {
      BasisDim += static_cast<int>(species(qindex, ptclin->GroupID[iat]));
    }
  }

  if(reduce_nstates) {
    // estimate the smallest number of states to calculate
    int ns_zero = static_cast<int>(NumElectrons/SpinDeg);
    if(ZeroTemperature) { // electron temperature is not zero
      if(ns_zero < NumElectrons/SpinDeg) {
	//non-integer number of electrons: e.g., Transition metals
	// reset the temperature 1 meV and add 10% of the valence states
        if(ns_zero < 10) NumStates = ns_zero+1;
        else             NumStates = ns_zero + ns_zero/10;
        Sigma = 0.001;
      } else {
        NumStates =  ns_zero;
      }
    } else {
      // adding 10% of the valence states
      if(ns_zero < 20) NumStates = ns_zero+1;
      else             NumStates = ns_zero + ns_zero/20;

      // too many states: reset to the total dimension
      if(NumStates > BasisDim) NumStates = BasisDim;
    }
  } else {
    NumStates=(NumStates<0)?BasisDim:NumStates;
    //if(NumStates <= 0) NumStates = BasisDim;
  }

#ifdef SPINORBIT
  SpinDeg = 1.0e0;
  BasisDim *= 2;                 // doubling the basis set
  NumStates *= 2;
#endif


  if(first) {
    States.First = first;     
    if(last) 
      States.Last = last;
    else  
      States.Last = first + NumStates;

    if(States.Last>NumStates) States.Last = NumStates;
  } else {
    States.First = 1;     
    States.Last = NumStates;
  }

  LOGMSG("\tBasis Dim = " << BasisDim)
  LOGMSG("\tNumStates = " << NumStates)
  LOGMSG("\tEigen values = " << States.First << " - " << States.Last)

  setElectronStates();

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 365 $   $Date: 2005-07-14 09:03:43 -0500 (Thu, 14 Jul 2005) $
 * $Id: ElectronStructure.1.cpp 365 2005-07-14 14:03:43Z jnkim $ 
 ***************************************************************************/
