//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_QUANTUMBASE_ELECTRONSTRUCTURE_H
#define OHMMS_QUANTUMBASE_ELECTRONSTRUCTURE_H

#include "EigSolver/EigenSpaceDimension.h"
#include <complex>
#include <vector>
#include <map>
using namespace std;
#include "Particle/ParticleSet.h"
#include "CrystalSymmetry/KPointBase.h"
#include "OhmmsNumeric/Matrix.h"
#include "OhmmsData/OhmmsParameter.h"
#include "QuantumBase/ElectronStructureIO.h"

namespace OHMMS {

  struct PhononMode {

    typedef ParticleSet::Scalar_t             Scalar_t;
    typedef ParticleSet::ParticlePos_t        ParticlePos_t;
    typedef ParticleSet::SingleParticlePos_t  SingleParticlePos_t;
    int ID;
    Scalar_t Omega2;
    Scalar_t E;
    SingleParticlePos_t Q;
    ParticlePos_t Z;

    PhononMode() { }
    PhononMode(const PhononMode& p): 
      ID(p.ID), Omega2(p.Omega2), E(p.E), Q(p.Q){ Z.resize(p.Z.size());
    }
    PhononMode& operator=(const PhononMode& p) {
      ID = p.ID;
      Omega2 = p.Omega2;
      E = p.E;
      Q = p.Q;
      Z.resize(p.Z.size());
      return *this;
    }
  };

  class ElectronStructure {

    ///declare ElectronStructureIO as a friend
    friend class ESOutput;

  public:

    typedef ParticleSet::Scalar_t  real_type;
    typedef complex<real_type>     complex_type;

    bool ZeroTemperature;
    bool EvaluateDensityMatrix;
    bool Shifted;
    bool Gamma;
    int NumStates;
    int BasisDim;
    int NumKpts;
    int Homo;

    real_type SpinDeg;
    real_type NumElectrons;
    real_type NumCoreElectrons;
    real_type Mu;
    real_type MuRef;
    real_type Sigma;
    real_type SumOcc, dSumOcc;
    real_type SumEnergy, dSumEnergy;

    EigenSpaceDimension States;
    RealMatrix_t Occ, E, dOcc;
    ParticleSet* RefPtclSet;

    ElectronStructure();
    virtual ~ElectronStructure();

    void setESParameters(ParticleSet* ptclin);

    bool resetKPoint(ParticleSet* ptclin);
    const KPointBase* getKPoint() const { return KSet;}    

    void resize(int dim, real_type nels, int nstates);
    void resize();
    void setElectronStates();

    virtual real_type
    el_ph_coupling(const vector<int>& psi_i, 
                   const PhononMode& , 
		   vector<int>& psi_f, real_type del) = 0; 

    real_type getBandEnergy();

  protected:

    static const int BiIter = 50;
    static const real_type BiTol;

    bool FixedFermiEnergy ;
    bool OwnKSet;
    KPointBase* KSet;
    string KFile;
    xmlNodePtr qmXmlNode;
    int findHomoLevel();
    real_type nelec(real_type ef );
    void bisection();
    inline real_type fermi(real_type e, real_type ef) {
      real_type ex = (e - ef)/Sigma;
      if(ex > 50.0e0)       return 0.0e0;
      else if(ex < -50.0e0) return 1.0e0;
      else                  return 1.0e0/(1.0e0+exp(ex));
    }


    ///m_param[name] returns OhmmsElementBase*
    map<string,OhmmsElementBase*> m_param;

    /*!\fn template<class PDT> void addParameter(PDT& a, const char* aname)
     *\brief add a new parameter corresponding to an xmlNode <parameter/>
     */
    template<class PDT>
    void addParameter(PDT& a, const char* aname,const char* uname) {
      map<string,OhmmsElementBase*>::iterator it = m_param.find(aname);
      if(it == m_param.end()) {
        m_param[aname] = new OhmmsParameter<PDT>(a,aname,uname);
      }
    }

  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 337 $   $Date: 2005-05-26 17:51:13 -0500 (Thu, 26 May 2005) $
 * POOMA_VERSION_ID: $Id: ElectronStructure.h 337 2005-05-26 22:51:13Z jnkim $ 
 ***************************************************************************/
