//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#include "Message/OpenMP.h"
#include <math.h>
#include <vector>
#include <iostream>
using namespace std;

#include "Utilities/RandomGenerator.h"
#include "OhmmsNumeric/OhmmsBlasf.h"

template<class Matrix_t>
FermionWaveFunctions<Matrix_t>::FermionWaveFunctions() { }

template<class Matrix_t>
FermionWaveFunctions<Matrix_t>::FermionWaveFunctions(int n, int m):
  Base_t(n,m){ 
  init(n,m);
}


template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::init(int n, int m) {

  resize(n,m);

  int numthreads = omp_get_max_threads();
  LocLoc = vector<int>(numthreads+1);
  int ndim = m/numthreads;
  int res = size2()-ndim*numthreads;

  // using a simple distribution
  LocLoc[0] = 0;
  for(int i =1; i<=numthreads; i++) {
    if(i<=res)    LocLoc[i] = LocLoc[i-1] + ndim + 1;
    else    LocLoc[i] = LocLoc[i-1] + ndim;
  }
}

template<class Matrix_t>
FermionWaveFunctions<Matrix_t>::~FermionWaveFunctions() { }

template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::randomize(int iband){
  if(iband < 0) {
    for(int i=0; i<size1(); i++)  randomize(i);
  } else {
    int ii=iband*size2();
    for(int ib=0; ib<size2(); ib++) data()[ii++] = 2.0*Random()-1;
  }
}


template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::orthonormalize(int iband) {

  if(iband< 0) { // initialize everything
    for(int i=0; i<size1(); i++) orthonormalize(i);
  } else {


    for(int is=0; is<iband; is++) {
       orthogonalize(iband,is);
     }

    normalize(iband);    
  }
}

template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::print(ostream& os, int omode) {

   int ii = 0;
   for(int i=0; i<size1(); i++) {
     for(int j=0; j<size2(); j++) {
       os << data()[ii++] << " ";
     }
     os << endl;
   }
}


template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::copy(int iband, value_type* wout) const {
  for(int j=0; j<size2(); j++)
    wout[j] = operator()(iband,j);
}

template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::assign(int iband, const value_type* win) {

  for(int j=0; j<size2(); j++)
    operator()(iband,j) = win[j];
}


template<class Matrix_t>
void FermionWaveFunctions<Matrix_t>::normalize(int iband) {

  value_type* xref= data().begin()+iband*size2();
  value_type ovl = OhmmsBlas::dot(size2(), xref, xref);
  ovl = 1.0/sqrt(ovl);
  OhmmsBlas::scal(size2(),ovl,xref);


//   value_type ovl = 0.0;
// #pragma omp parallel reduction(+:ovl) 
//   {
//      int myID = omp_get_thread_num();
//      const value_type* xref = data().begin()+iband*size2();
//      int ib = LocLoc[myID];
//      const int ibf = LocLoc[myID+1];
//      for(; ib < ibf; ib++) ovl += xref[ib]*xref[ib];
//   }
//   ovl  = 1.0/sqrt(ovl);
// #pragma omp parallel firstprivate(ovl)
//   {
//      int myID = omp_get_thread_num();
//      int ib = LocLoc[myID];
//      value_type* xref = data().begin()+iband*size2();
//      const int ibf = LocLoc[myID+1];
//      for(; ib < ibf; ib++) xref[ib] *= ovl;
//   }
}

template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::orthogonalize(int iband, int is) {

  value_type* xref= data().begin()+iband*size2();
  const value_type* yref= data().begin()+is*size2();

  value_type ovl = -OhmmsBlas::dot(size2(), xref, yref);
  OhmmsBlas::axpy(size2(),ovl,yref,xref); 

//   value_type ovl = 0.0;
// #pragma omp parallel reduction(+:ovl) 
//   {
//      int myID = omp_get_thread_num();
//      const value_type* xref = data().begin()+iband*size2();
//      const value_type* yref = data().begin()+is*size2();
//      int ib = LocLoc[myID];
//      const int ibf = LocLoc[myID+1];
//      for(; ib < ibf; ib++) ovl += xref[ib]*yref[ib];
//   }
// #pragma omp parallel firstprivate(ovl)
//   {
//      int myID = omp_get_thread_num();
//      value_type* xref = data().begin()+iband*size2();
//      const value_type* yref = data().begin()+is*size2();
//      int ib = LocLoc[myID];
//      const int ibf = LocLoc[myID+1];
//      for(; ib < ibf; ib++) xref[ib] -= ovl*yref[ib];
//   }
}

//! orthogonalize a state with this, this is constant
template<class Matrix_t>
void 
FermionWaveFunctions<Matrix_t>::orthogonalize(iterator1& astate, 
					      int minband, int maxband) const {
  if(minband < 0) return;

  for(int iband =minband ; iband <= maxband; iband++) {
    const value_type* yref= data().begin()+iband*size2();

    value_type ovl = -OhmmsBlas::dot(size2(), astate, yref);
    OhmmsBlas::axpy(size2(),ovl,yref,astate); 

//     value_type ovl = 0;
// #pragma omp parallel reduction(+:ovl)
//     {
//      int myID = omp_get_thread_num();
//       const value_type* xref = data().begin()+iband*size2();
//       int ib = LocLoc[myID];
//       const int ibf = LocLoc[myID+1];
//       for(; ib < ibf; ib++) ovl += xref[ib]*astate[ib];
//     }

// #pragma omp parallel firstprivate(ovl) 
//     {
//      int myID = omp_get_thread_num();
//       const value_type* xref = data().begin()+iband*size2();
//       int ib = LocLoc[myID];
//       const int ibf = LocLoc[myID+1];
//       for(; ib < ibf; ib++) astate[ib] -= ovl*xref[ib];
//     }
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: FermionWaveFunctions.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
