//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#include "Message/OpenMP.h"
#include <math.h>
#include <vector>
#include <iostream>
using namespace std;

#include "Utilities/RandomGenerator.h"
#include "QuantumBase/FermionWaveFunctions.h"
#include "OhmmsNumeric/OhmmsBlasf.h"

FermionWaveFunctions::FermionWaveFunctions() { }
FermionWaveFunctions::FermionWaveFunctions(int n, int m):BaseType_t(n,m){ 
  init(n,m);
}


void FermionWaveFunctions::init(int n, int m) {

  if(D1 != n || D2 != m) resize(n,m);

  int numthreads = omp_get_max_threads();
  LocLoc = vector<int>(numthreads+1);
  int ndim = m/numthreads;
  int res = D2-ndim*numthreads;
  // using a simple distribution
  LocLoc[0] = 0;
  for(int i =1; i<=numthreads; i++) {
    if(i<=res)    LocLoc[i] = LocLoc[i-1] + ndim + 1;
    else    LocLoc[i] = LocLoc[i-1] + ndim;
  }
}

FermionWaveFunctions::~FermionWaveFunctions() { }

void FermionWaveFunctions::randomize(int iband){
  if(iband < 0) {
    for(int i=0; i<D1; i++)  randomize(i);
  } else {
    int ii = iband*D2;
    for(int ib=0; ib<=iband; ib++) X[ii++] = 2.0*Random()-1;
  }
}
void FermionWaveFunctions::orthonormalize(int iband) {

  if(iband< 0) { // initialize everything
    for(int i=0; i<nrow(); i++) orthonormalize(i);
  } else {


    for(int is=0; is<iband; is++) {
       orthogonalize(iband,is);
     }

    normalize(iband);    
  }
}

void FermionWaveFunctions::print(ostream& os, int omode) {

   int ii = 0;
   for(int i=0; i<D1; i++) {
     for(int j=0; j<D2; j++) {
       os << X[ii++] << " ";
     }
     os << endl;
   }
  BaseType_t dm(D1,D1);

  for(int i=0; i<D1;i++) {
    for(int j = 0; j<D1; j++) {
      T ovl = 0.0;      
      int ib = D2*i;
      int jb = D2*j;
      for(int k=0; k<D2; k++, ib++, jb++) {
	ovl += X[ib]*X[jb];
      }
      dm(i,j) = ovl;
    }
  }
  os << "Overlap matrix" << endl;
  for(int i=0; i<D1;i++) {
    for(int j = 0; j<D1; j++) {
      if(i == j && fabs(dm(i,j)-1) > 1e-10) os << "diag " << i<< endl;
      if(i != j && fabs(dm(i,j)) > 1e-10) os <<"offdig " << i << " " << j << endl;
    }

  }
}


void FermionWaveFunctions::copy(int iband, T* wout) const {

#pragma omp parallel
  {
     int myID = omp_get_thread_num();
     const T* xref = begin()+iband*D2;
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) wout[ib] = xref[ib];
  }
}

void FermionWaveFunctions::assign(int iband, const T* win) {
#pragma omp parallel
  {
     int myID = omp_get_thread_num();
     T* xref = begin()+iband*D2;
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) xref[ib] = win[ib];
  }
}

void FermionWaveFunctions::normalize(int iband) {
  T ovl = 0.0;
#pragma omp parallel reduction(+:ovl) 
  {
     int myID = omp_get_thread_num();
     const T* xref = begin()+iband*D2;
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) ovl += xref[ib]*xref[ib];
  }
  ovl  = 1.0/sqrt(ovl);

#pragma omp parallel firstprivate(ovl)
  {
     int myID = omp_get_thread_num();
     int ib = LocLoc[myID];
     T* xref = begin()+iband*D2;
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) xref[ib] *= ovl;
  }
}

void FermionWaveFunctions::orthogonalize(int iband, int is) {

  T ovl = 0.0;

#pragma omp parallel reduction(+:ovl) 
  {
     int myID = omp_get_thread_num();
     const T* xref = begin()+iband*D2;
     const T* yref = begin()+is*D2;
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) ovl += xref[ib]*yref[ib];
  }

#pragma omp parallel firstprivate(ovl)
  {
     int myID = omp_get_thread_num();
     T* xref = begin()+iband*D2;
     const T* yref = begin()+is*D2;
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) xref[ib] -= ovl*yref[ib];
  }
}

//! orthogonalize a state with this, this is constant
void 
FermionWaveFunctions::orthogonalize(WaveFunctionRef_t astate, 
				    int minband, int maxband) const {
  if(minband < 0) return;

  for(int iband =minband ; iband <= maxband; iband++) {
    T ovl = 0;
#pragma omp parallel reduction(+:ovl)
    {
     int myID = omp_get_thread_num();
      const T* xref = begin()+iband*D2;
      int ib = LocLoc[myID];
      const int ibf = LocLoc[myID+1];
      for(; ib < ibf; ib++) ovl += xref[ib]*astate[ib];
    }

#pragma omp parallel firstprivate(ovl) 
    {
     int myID = omp_get_thread_num();
      const T* xref = begin()+iband*D2;
      int ib = LocLoc[myID];
      const int ibf = LocLoc[myID+1];
      for(; ib < ibf; ib++) astate[ib] -= ovl*xref[ib];
    }
  }

//    T norm = 0.0;
//  #pragma omp parallel reduction(+:norm) 
//    {
//       int myID = omp_get_thread_num();
//       int ib = LocLoc[myID];
//       const int ibf = LocLoc[myID+1];
//       for(; ib < ibf; ib++) norm += astate[ib]*astate[ib];
//    }
//    norm  = 1.0/sqrt(norm);

//  #pragma omp parallel firstprivate(norm)
//    {
//       int myID = omp_get_thread_num();
//       int ib = LocLoc[myID];
//       const int ibf = LocLoc[myID+1];
//       for(; ib < ibf; ib++) astate[ib] *= norm;
//    }
}


////////////////////////////////////////////////////////////////
// Using omp::pfor
////////////////////////////////////////////////////////////////
//  void FermionWaveFunctions::normalize(int iband, int myid) {

//    T ovl = 0.0;
//    const T* xref = begin()+iband*D2;
//  #pragma omp parallel for reduction(+:ovl)
//    for(int ib=0; ib<D2; ++ib) {
//      ovl += xref[ib]*xref[ib];
//    }

//    ovl  = 1.0/sqrt(ovl);

//    T* xrefnew = begin()+iband*D2;
//  #pragma omp parallel for firstprivate(ovl)
//    for(int ib=0; ib < D2; ++ib) {
//      xrefnew[ib] *= ovl;
//    }      
//  }

//  void FermionWaveFunctions::orthogonalize(int iband, int is, int myid) {
//    //  T ovl = OhmmsBlas::dot(D2, begin()+iband*D2, begin()+is*D2);
//    T ovl = 0.0;

//    const T* xref = begin()+iband*D2;
//    const T* yref = begin()+is*D2;

//  #pragma omp parallel for reduction(+:ovl)
//    for(int ib = 0; ib < D2; ib++) {
//      ovl += xref[ib]*yref[ib];
//    }

//    T* xrefnew = begin()+iband*D2;

//  #pragma omp parallel for firstprivate(ovl)
//    for(int ib = 0; ib < D2; ib++) {
//      xrefnew[ib] -= ovl*yref[ib];
//    }      
//  }


////////////////////////////////////////////////////////////////
// Using omp::explicitly distributing
////////////////////////////////////////////////////////////////
//  void FermionWaveFunctions::normalize(int iband, int myid) {
//    T ovl = 0.0;
//  #pragma omp parallel reduction(+:ovl) 
//    {
//       int na = D2/omp_get_max_threads();
//       int myID = omp_get_thread_num();
//       int ib = na*myID + D2*iband;
//       int ibf = ib + na;
//       for(; ib < ibf; ib++) ovl += X[ib]*X[ib];
//    }
//    ovl  = 1.0/sqrt(ovl);

//  #pragma omp parallel firstprivate(ovl)
//    {
//       int na = D2/omp_get_max_threads();
//       int myID = omp_get_thread_num();
//       int ib = na*myID + D2*iband;
//       int ibf = ib + na;
//       for(; ib < ibf; ib++) X[ib] *= ovl;
//    }
//  }

//  void FermionWaveFunctions::orthogonalize(int iband, int is, int myid) {
//    //  T ovl = OhmmsBlas::dot(D2, begin()+iband*D2, begin()+is*D2);
//    T ovl = 0.0;

//    const T* xref = begin()+iband*D2;
//    const T* yref = begin()+is*D2;

//  #pragma omp parallel reduction(+:ovl) 
//    {
//       int na = D2/omp_get_max_threads();
//       int myID = omp_get_thread_num();
//       int ib = na*myID + D2*iband;
//       int ibf = ib + na;
//       int jb = na*myID + D2*is;
//       for(; ib < ibf; ib++, jb++) ovl += X[ib]*X[jb];
//    }
//  #pragma omp parallel firstprivate(ovl)
//    {
//       int na = D2/omp_get_max_threads();
//       int myID = omp_get_thread_num();
//       int ib = na*myID + D2*iband;
//       int ibf = ib + na;
//       int jb = na*myID + D2*is;
//       for(; ib < ibf; ib++, jb++) X[ib] -= ovl*X[jb];
//    }
//  }

////////////////////////////////////////////////////////////////
// Using blas_mp
////////////////////////////////////////////////////////////////
// void FermionWaveFunctions::normalize(int iband, int myid) {

//   T ovl = OhmmsBlas::dot(D2, begin()+D2*iband, begin()+D2*iband);
//   ovl  = 1.0/sqrt(ovl);

// #pragma omp parallel firstprivate(ovl)
//   {
//      int na = D2/omp_get_max_threads();
//      int myID = omp_get_thread_num();
//      int ib = na*myID + D2*iband;
//      int ibf = ib + na;
//      for(; ib < ibf; ib++) X[ib] *= ovl;
//   }
// }

// void FermionWaveFunctions::orthogonalize(int iband, int is, int myid) {
//   //  T ovl = OhmmsBlas::dot(D2, begin()+iband*D2, begin()+is*D2);
//   T ovl = OhmmsBlas::dot(D2, begin()+D2*iband, begin()+D2*is);
// #pragma omp parallel firstprivate(ovl)
//   {
//      int na = D2/omp_get_max_threads();
//      int myID = omp_get_thread_num();
//      int ib = na*myID + D2*iband;
//      int ibf = ib + na;
//      int jb = na*myID + D2*is;
//      for(; ib < ibf; ib++, jb++) X[ib] -= ovl*X[jb];
//   }
// }
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: FermionWaveMatrix.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
