//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifndef CUBICSPLINE_FINITEDIFF_H
#define CUBICSPLINE_FINITEDIFF_H

#include <vector>
using std::vector;

/*! \class template<class T> CubicSplineDiff
 * \brief spline functions on regular grid [Rmin, Rmax)
 */
template<class T>
class CubicSplineDiff{
public:
  int Npt;//!< number of grid points
  T   Rmin, Rmax, dr, drinv, drinv2;//!< boundary and other values for book-keeping
  vector<T> Y, Yp, Ypp; //!< values, derivatives, and the second derivatives

  //!< default constructor
  CubicSplineDiff() { }

  //!< copy constructor
  CubicSplineDiff(const CubicSplineDiff<T>& a ) {
    Npt = a.Npt;
    Rmin = a.Rmin;
    Rmax = a.Rmax;
    dr = a.dr;
    drinv = drinv2;
    Y = a.Y;
    Yp = a.Yp;
    Ypp = a.Ypp;
  }

  //!< copy operator
  CubicSplineDiff<T>& operator=(const CubicSplineDiff<T>& a ) {
    Npt = a.Npt;
    Rmin = a.Rmin;
    Rmax = a.Rmax;
    dr = a.dr;
    drinv = drinv2;
    Y = a.Y;
    Yp = a.Yp;
    Ypp = a.Ypp;
    return *this;
  }

  //inline CubicSplineDiff(const int n) { resize(n);}

  inline T rmax() const { return Rmax;}//!< return the upper bound of the grid
  inline T rmin() const { return Rmin;}//!< return the lower bound of the grid

  //!< resize the arrays
  inline void resize(const int n) {
    Npt = n;
    Y = vector<T>(n+1); 
    Yp = vector<T>(n+1); 
    Ypp = vector<T>(n+1); 
  }

  /*!\fn T operator()(T x0)
   *\param x0, the value at which a function is evaluated.
   *\return value at x0
   */
  inline T operator()(T x0) const {
    //if(x0 < Rmin || x0 > Rmax) return 0.0;
    int i = static_cast<int>((x0-Rmin)*drinv);
    T xx = (x0 - static_cast<T>(i)*dr - Rmin)*drinv;
    return Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0);
  }

  /*!\fn T splint(T x0)
   *\param x0, the value at which a function is evaluated.
   *\return value at x0
   */
  inline T splint(T x0) const {
    return this->operator()(x0);
  }

  /*!\fn T operator()(T x0, T& yval)
   *\param x0, the value at which a function is evaluated.
   *\param yval = f(x0)/x0
   *\return derivatie f'(x0)
   *\note operator() is used by SlaterKoster classes 
   */
  inline T operator()(T x0, T& yval) const{ 
    //if(x0 < Rmin || x0 > Rmax) { yval = 0.0; return 0.0;}
    int i = static_cast<int>((x0-Rmin)*drinv);
    T xx = (x0 - static_cast<T>(i)*dr - Rmin)*drinv;
    yval = (Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0))/x0;
    return (Yp[i] + Ypp[i]*(2*xx-1.0))*drinv;
  }

  /*!\fn T splint(T x0, T& yval)
   *\param x0, the value at which a function is evaluated.
   *\param yval =  f(x0)
   *\return derivatie f'(x0)
   */
  inline T splint(T x0, T& yval) const{ 
    //if(x0 < Rmin || x0 > Rmax) { yval = 0.0; return 0.0;}
    int i = static_cast<int>((x0-Rmin)*drinv);
    T xx = (x0 - static_cast<T>(i)*dr - Rmin)*drinv;
    yval = Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0);
    return (Yp[i] + Ypp[i]*(2*xx-1.0))*drinv;
  }

  /*!\fn T splint(T x0, T& yval, T& yp)
   *\param x0, the value at which a function is evaluated.
   *\param yval =  f(x0)/x0
   *\param yp =  f'(x0)
   *\return second derivatie f''(x0)
   */
  inline T operator()(T x0, T& yval, T& yp) const {
    //if(x0 < Rmin || x0 > Rmax) {yval = 0; yp = 0;return 0.0;}
    int i = static_cast<int>((x0-Rmin)*drinv);
    T xx = (x0 - static_cast<T>(i)*dr - Rmin)*drinv;
    yval = (Y[i]+Yp[i]*xx+Ypp[i]*xx*(xx-1.0))/x0;
    yp  =  (Yp[i] + Ypp[i]*(2*xx-1.0))*drinv/x0;
    return Ypp[i]*xx*2.0*drinv2;
  }

  /*!\fn void spline(const int n, T x1, T x2, T* y)
   *\param n number of points at which the values are given.
   *\param x1 assigned to Rmin
   *\param x2 assigned to Rmax
   *\param y an value array whose size is n+1
   */
  void spline(const int n, T x1, T x2, T* y) {
    resize(n);
    Rmin = x1; Rmax = x2;
    dr = (x2-x1)/static_cast<T>(n);  
    drinv = 1.0/dr; drinv2 = drinv*drinv;
    for(int i=0; i<=n; i++)  Y[i]   = y[i];
    for(int i=0; i< n; i++)  Yp[i]  = Y[i+1]-Y[i];
    Yp[n] = 0.0;
    for(int i=0; i<n-1; i++) Ypp[i] = 0.5*(Y[i+2]+Y[i]-2.0*Y[i+1]);
    Ypp[n-1] = 0.0; Ypp[n] = 0.0;
  }

  void readSpline(istream& is, T econv);
};

/*!\fn void CubicSplineDiff<T>::readSpline(istream& is, T econv)
 *\brief specialized input routine to initialize internal values
 * introduced for compatilibility to the classes Potentials/CubicSplineType.h
 */
template<class T>
void 
CubicSplineDiff<T>::readSpline(istream& is, T econv)  {

  /* Read spline for AlloyEAM from file */
  int Type, npt;
  T X1, Xn, Yp1, Ypn;
  vector<T> yin;

  is >> Type;
  is >> npt;
  yin.resize(npt);

  switch(Type) {
  case 1:
    is >> X1 >> Xn;
    is >> Yp1 >> Ypn;

    for(int i=0; i< npt;i++) {
      is >> yin[i];
      yin[i] *= econv;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  case 2:
    is >> Yp1 >> Ypn;
    
    for(int i=0; i< npt;i++) {
      is >> Xn;
      is >> yin[i];
      yin[i] *= econv;
      if(i == 0) X1 = Xn;
      // cerr << i << " " << X[i] << " " << Y[i] << "\n";
    }
    break;
  }

  spline(npt-1, X1, Xn, yin.begin());
}

#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * POOMA_VERSION_ID: $Id: CubicSplineDiff.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
