//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

/***************************************************************************
 * Scale Function proposed by 
 * L. Goodwin, A. J. Skinner and D. G. Pettifor, Europhys. Lett. {\bf 9}, 701 (1989).
 ***************************************************************************/
#ifndef OHMMS_GSKSCALEFUNCTION_H
#define OHMMS_GSKSCALEFUNCTION_H

#include <math.h>

template<class T>
struct GSPScale {

  T N, Nc;
  T Ro, Rc, R1, Rcut;

  GSPScale(T val = 1.0, T n = 2, T nc = 0, T ro = 2.352125, 
           T rc = 2.8, T r1 = 2.8, T rcut =3){
    set(val,n,nc,ro,rc,r1,rcut);
  }

  inline T operator()(T x0) const {
    if(x0 <= R1)        
      return A0*pow((Ro/x0),N)*exp(-N*pow(x0*Rcinv,Nc));
    else if(x0 < Rcut)  { 
      T rad = x0-R1;
      return c0+rad*(c1+rad*(c2+rad*c3));
    } else                
      return 0.0e0;
  }

  inline T operator()(T x0, T& yval) const {
    if(x0 <= R1) {
      T rinv = 1.0e0/x0;
      T z = pow(x0*Rcinv,Nc);
      yval =  A0*rinv*pow(Ro*rinv,N)*exp(-N*z);
      return -yval*N*(1+Nc*z);
    } else if(x0 < Rcut) {
      T rad = x0 - R1;
      yval = (c0+rad*(c1+rad*(c2+rad*c3)))/x0;
      return c1+rad*(2.0e0*c2+3.0e0*rad*c3);       
    } else {
      yval = 0.0e0;
      return 0.0e0;
    }
  } 

  inline T operator()(T x0, T& yval, T& yp)const  {
    if(x0 <= R1) {
      T rinv = 1.0e0/x0;
      T z = pow(x0*Rcinv,Nc);
      yval =  A0*pow((Ro*rinv),N)*exp(-N*z)*rinv*rinv;
      yp = -yval*N*(1+Nc*z);
      //return -(yp-yval)*N*(1+Nc*z) -yval*N*Nc*Nc*z;
      return (yval-yp)*N*(1+Nc*z)-yval*N*Nc*Nc*z;
    } else if(x0 < Rcut) {
      T rad = x0 - R1;
      yval = (c0+rad*(c1+rad*(c2+rad*c3)))/x0/x0;
      yp  = (c1+rad*(2.0e0*c2+3.0e0*rad*c3))/x0;       
      //return 2.0e0*c2 + 6*c3*rad;
      return 0.0e0;
    } else {
      yval = 0.0e0;
      return 0.0e0;
    }
  } 
  inline void set(T val, T n, T nc, T ro, T rc, T r1, T rcut);

  T A0, Rcinv;
  T c0, c1, c2, c3;     
};

template<class T>
inline 
void GSPScale<T>::set(T val, T n, T nc, T ro, T rc, T r1, T rcut){

  N = n;
  Nc = nc;
  Ro = ro;
  Rc = rc;
  R1 = r1;
  Rcut = rcut;

  // default behavior is defined
  if(Rc > 1e-3) Rcinv = 1.0e0/Rc;
  A0 =val*exp(N*pow(Ro*Rcinv,Nc));
  if(R1 < Ro) { 
    R1 = 1.3e0*Ro;
    Rcut = 1.5e0*Ro;
  }  

  T rad = R1;
  c1 = this->operator()(rad,c0); 
  c0 *= rad;// note that df returns df/r

  rad = Rcut - R1;
  c2 = -(2.0e0*c1*rad + 3.0e0*c0)/rad/rad;
  c3 =  (      c1*rad + 2.0e0*c0)/rad/rad/rad;
}

#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 264 $   $Date: 2005-01-15 16:49:17 -0600 (Sat, 15 Jan 2005) $
 * POOMA_VERSION_ID: $Id: GSPScale.h 264 2005-01-15 22:49:17Z jnkim $ 
 ***************************************************************************/
