//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifndef OHMMS_INVERSEPOWERSCALEMODEL_H
#define OHMMS_INVERSEPOWERSCALEMODEL_H
#include <math.h>

template<class T>
struct DummyScaleFunctor{ 
  virtual void set(T c, T n, T r0, T rc=-1.0e0)=0;
};

template<class T, int N>
struct InvPowerScaleN: public DummyScaleFunctor<T> {

  T R0, C, Rcut;
  InvPowerScaleN(T c=0.0, T r0=2.35, T rc=-1.0): R0(r0){
    set(c,0,r0,rc);
  }

  inline void set(T c, T n, T r0, T rc=-1.0e0) 
  {
    R0 = r0; C = c*pow(R0,N);
    if(rc<0.0e0) 
      Rcut=1.1*r0;
    else
      Rcut=rc;
  }

  inline T operator()(T r) const { 
    return C/pow(r,N);
  }

  inline T operator()(T r, T& v_over_r) const { 
    v_over_r = C/pow(r,N+1); return -N*v_over_r;
  }

  inline T operator()(T r, T& v_over_rr, T& dv_over_r) const {
    v_over_rr = C/pow(r,N+2); dv_over_r = -N*v_over_rr; return -(N+1)*dv_over_r;
  }
};


template<class T>
struct InvPowerScaleN<T,0>: public DummyScaleFunctor<T> {
  T R0, C, Rcut;
  InvPowerScaleN(T c=0.0, T r0=2.35, T rc=-1.0): R0(r0){
    set(c,0,r0,rc);
  }

  inline void set(T c, T n, T r0, T rc=-1.0e0) 
  {
    R0 = r0; C = c;
    if(rc<0.0e0) 
      Rcut=1.1*r0;
    else
      Rcut=rc;
  }

  inline T operator()(T r) const { 
    return C;
  }

  inline T operator()(T r, T& v_over_r) const { 
    v_over_r = 0.0; return C/r;
  }

  inline T operator()(T r, T& v_over_rr, T& dv_over_r) const {
    v_over_rr = 0.0; dv_over_r = 0.0; return C/(r*r);
  }
};

template<class T>
struct InvPowerScaleN<T,2>: public DummyScaleFunctor<T> {

  T R0, C, Rcut;
  InvPowerScaleN(T c=0.0, T r0=2.35, T rc=-1.0): R0(r0){
    set(c,0,r0,rc);
  }

  inline void set(T c, T n, T r0, T rc=-1.0e0) 
  {
    R0 = r0; C = c*R0*R0;
    if(rc<0.0e0) 
      Rcut=1.1*r0;
    else
      Rcut=rc;
  }

  inline T operator()(T r) const { 
    return C/(r*r);
  }

  inline T operator()(T r, T& v_over_r) const { 
    v_over_r = C/(r*r*r); return -2.0*v_over_r;
  }

  inline T operator()(T r, T& v_over_rr, T& dv_over_r) const {
    v_over_rr = C/(r*r*r*r); dv_over_r = -2.0*v_over_rr; 
    return -3.0*dv_over_r;
  }
};

template<class T>
struct InvPowerScaleN<T,-1>: public DummyScaleFunctor<T> {

  T Np, R0, C, Rcut;
  InvPowerScaleN(T c=0.0, T n=2, T r0=2.35, T rc=-1.0){
    set(c,n,r0,rc);
  }

  inline void set(T c, T n, T r0, T rc=-1.0e0) 
  {
    Np = n; R0 = r0; C = c*pow(R0,n);
    if(rc<0.0e0) 
      Rcut=1.1*r0;
    else
      Rcut=rc;
  }

  inline T operator()(T r) const { 
    return C/pow(r,Np);
  }

  inline T operator()(T r, T& v_over_r) const { 
    v_over_r = C/pow(r,Np+1); return -Np*v_over_r;
  }

  inline T operator()(T r, T& v_over_rr, T& dv_over_r) const {
    v_over_rr = C/pow(r,Np+2); dv_over_r = -Np*v_over_rr; 
    return -(Np+1)*dv_over_r;
  }
};


#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * OHMMS_VERSION_ID: $Id: InvPowerScale.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
