// -*- C++ -*-
/***************************************************************************
 * This class implements the non-orthogonal tight-binding model
 * for Transition metals proposed by Mehl and Papaconstantopoulos in PRB ...
 * to parameterize hopping terms and pair-potential.
 * Implemented by D. Trinkle July '98 on TBMD
 * Implemented by Jeongnim Kim Sep '99 
 * Rewritten by D. Trinkle April '03 to improve efficiency and ease of use
 ***************************************************************************/
#ifndef NRL_SCALEFUNCTIONS_H
#define NRL_SCALEFUNCTIONS_H

#include <math.h>

template<class T>
struct NRLOnsite {

  T A, B, C, D;
  T Bp, Cp, Dp;

  NRLOnsite(T a = 0, T b=0, T c=0, T d=0){
   set(a,b,c,d);
  }

  void set(T a, T b, T c, T d){ 
    A = a;
    B = b; C = c; D = d;
    Bp = B*2.0/3.0;
    Cp = C*4.0/3.0;
    Dp = 2*D;
  }

  //inline T operator()(T rho) {
  inline T f(T rho) const {
    const double twothirds    = 0.666666666666666666666666666666666666667;
    T rho23=pow(rho,twothirds);
    return A+rho23*(B+C*rho23) + D*rho*rho;
  }

  //inline T operator()(T rho, T& val) {
  inline T df(T rho) const {
    const double onethird    = 0.333333333333333333333333333333333333333;
    T rho13 = pow(rho,onethird);
    return Bp/rho13+rho13*Cp + Dp*rho;
  }
};

template<class T>
struct SmoothCutoff {

  T R0, Rcut, Screen;
  SmoothCutoff(T rcut=0, T screen=1) {
    set(rcut,screen);
  }

  void set(T rcut, T screen) {
    Rcut = rcut;
    R0 = Rcut - 5.0*screen;
    Screen = 1.0/screen;   
  }

  inline T operator()(T r) const {
    return 1.0/(1.0+exp((r-R0)*Screen));
  }

  inline T operator()(T r, T& v)const  {
    T tmp = exp((r-R0)*Screen);
    v = 1.0/(1.0+tmp);
    return -Screen*tmp*v*v;
  }
};


template<class T>
struct NRLDensity {

  T R0, Rcut, Screen, Lambda;
  NRLDensity(T rcut=0, T screen=1, T lambda=1) {
    set(rcut,screen,lambda);
  }

  void set(T rcut, T screen, T lambda) {
    Rcut = rcut;
    R0 = Rcut - 5.0*screen;
    Screen = 1.0/screen;   
    if(lambda > 0) Lambda = -lambda;
    else           Lambda = lambda;
  }

  inline T operator()(T r) const {

    return exp(Lambda*r)/(1.0+exp((r-R0)*Screen));
  }

  inline T operator()(T r, T& v)const  {
    T tmp = exp((r-R0)*Screen);
    v = exp(Lambda*r)/(1.0+tmp);
    return v*(Lambda-Screen*tmp*v);
  }
};


// To make these functions easier to differentiate, we use the following
// shortcut definitions:

// y = a*b*c, a = (D+r*(E+r*(F+r*Fb))), b = exp(-G*r), c = 1/(1+x).
//                                                     x = exp(l0*(r-r0))
// y' = a'bc + y*( b'/b + c'/c )
// 
// y'' = a''bc + (a'bc + y')*( b'/b + c'/c )
//         + y*( b''/b - (b'/b)^2  +  c''/c - (c'/c)^2 )
//
// Which simplifies, as
// a' = (E+r*(2*F+r*3*Fb))  and  a'' = (2*F + r*6*Fb)
// b'/b = -G                and  b''/b - (b'/b)^2 = 0
// c'/c = -c*l0*x           and  c''/c - (c'/c)^2 = l0*l0*(c*x-1)*(c*x)

// To make matters "simpler," more extensible, and more efficient,
// we make a few definitions:
// Store fc = 1/(1+x) and expgr_fc = b*c  when calc. y
// Store fc_p = c'/c = -l0*fc*x = l0*(fc-1) and expgr_fc_p = -G + fc_p
//   when calc. y'

// Then, we can write:
// y   = a*expgr_fc
// y'  = a'*expgr_fc + y*expgr_fc_p
// y'' = a''*expgr_fc + 2*y'*expgr_fc_p + y*(fc_pp - expgr_fc_p^2)
//   where fc_pp = (fc''/fc) - (fc'/fc)^2 = fc_p*(fc_p + l0)

// The advantage of this method is that the function "a" can be replaced
// relatively easily: see poly, poly_p, and poly_pp.

// In addition, the "init_spline" method has been rewritten to use the 
// values extracted from the second derivative calculation, so now the
// splines will be guaranteed to match at Rmin. --DRT


// ============================= NRLScaleFunction ======================
// Quadratic polynomial, no splines.
template<class T> 
struct NRLScaleFunction{

  NRLScaleFunction(T e0=0.0, T f0=0.0, T fbar = 0.0, T g0=0.0, T Rcut0=0.0, 
		   T screenl=1.0) {
    set(e0,f0,fbar,g0,Rcut0,screenl);
  }

  void set(T e0, T f0, T fbar0, T g0, T Rcut0, T screenl) {
    Rcut = Rcut0;               // Cutoff parameters
    screenl_inv = 1.0/screenl;
    r0 = Rcut0 - 5.0*screenl;

    E = e0;  F = f0;  Fb = fbar0;  // NRL parameters
    G = g0;                                // Note: input is already squared
  }

  // ==== y ====
  T operator()(T r){ 
    if (r > Rcut)  return 0.0;
    // Store for use by first + second derivative code... --DRT
    fc = 1./(1.0 + exp((r-r0)*screenl_inv));
    expgr_fc = exp(-G*r)*fc;
    // y   = a*expgr_fc
    return poly(r)*expgr_fc;
  }

  // ==== y' ====
  T operator()(T r, T& vrinv){
    if (r > Rcut) { vrinv = 0.0; return 0.0; }
    T y = operator()(r); // Calc's y(r) and sets up extra var's --DRT
    vrinv = y/r;         // Always the same
    fc_p = screenl_inv*(fc-1.);  // = fc'/fc, for use by second deriv.
    expgr_fc_p = -G + fc_p;      // also used by second deriv. code
    // y'  = a'*expgr_fc + y*expgr_fc_p
    return poly_p(r)*expgr_fc + y*expgr_fc_p;
  }
  
  // ==== y'' ====
  T operator()(T r, T& yrinv2, T& dyrinv){
    if (r > Rcut) {yrinv2 = 0.0; dyrinv = 0.0; return 0.0; }
    // Take care of df/dr * (1/r) and f/r^2 first...
    T y;                     // value of function
    T yp = operator()(r, y); // Sets up extra var's
    dyrinv = yp * (1/r);
    yrinv2 = y * (1/r);
    y *= r;

    T fc_pp = fc_p*(fc_p + screenl_inv); // = (fc''/fc) - (fc'/fc)^2
    // y'' = a''*expgr_fc + 2*y'*expgr_fc_p + y*(fc_pp - expgr_fc_p^2)
    return poly_pp(r)*expgr_fc + 2*yp*expgr_fc_p 
      + y*(fc_pp - expgr_fc_p*expgr_fc_p);
  }

private:
  // Cutoff parameters:
  T Rcut, r0, screenl_inv;
  // NRL functional form:
  T E, F, Fb, G;

  // Extra storage to make evaluations faster: --DRT
  T expgr_fc, expgr_fc_p;  // exp(-g*r)*fc, and "derivative"
  T fc, fc_p;              // fc, and fc'/fc

  // Polynomial function and it's two derivatives:
  inline T poly(T r)    { return E + r*(F + r*Fb);}
  inline T poly_p(T r)  { return F + 2*r*Fb;}
  inline T poly_pp(T r) { return 2*Fb;}  
};
// ============================= NRLScaleFunction ======================



// ============================ NRLScaleFunction2 ======================
// Cubic polynomial, no splines.
template<class T>
struct NRLScaleFunction2{

  NRLScaleFunction2(T d0 = 0.0, T e0=0.0, T f0=0.0, T g0 = 0.0, T h0=0.0, 
                    T Rcut0=0.0, T screenl=1.0) {
    set(d0, e0, f0, g0, h0, Rcut0,screenl);
  }

  void set(T d0, T e0, T f0, T fbar0, T g0, T Rcut0, T screenl) {
    Rcut = Rcut0;               // Cutoff parameters
    screenl_inv = 1.0/screenl;
    r0 = Rcut0 - 5.0*screenl;

    D = d0; E = e0;  F = f0;  Fb = fbar0;  // NRL parameters
    G = g0;                                // Note: input is already squared
  }

  // ==== y ====
  T operator()(T r){ 
    if (r > Rcut)  return 0.0;
    // Store for use by first + second derivative code... --DRT
    fc = 1./(1.0 + exp((r-r0)*screenl_inv));
    expgr_fc = exp(-G*r)*fc;
    // y   = a*expgr_fc
    return poly(r)*expgr_fc;
  }

  // ==== y' ====
  T operator()(T r, T& vrinv){
    if (r > Rcut) { vrinv = 0.0; return 0.0; }
    T y = operator()(r); // Calc's y(r) and sets up extra var's --DRT
    vrinv = y/r;         // Always the same
    fc_p = screenl_inv*(fc-1.);  // = fc'/fc, for use by second deriv.
    expgr_fc_p = -G + fc_p;      // also used by second deriv. code
    // y'  = a'*expgr_fc + y*expgr_fc_p
    return poly_p(r)*expgr_fc + y*expgr_fc_p;
  }
  
  // ==== y'' ====
  T operator()(T r, T& yrinv2, T& dyrinv){
    if (r > Rcut) {yrinv2 = 0.0; dyrinv = 0.0; return 0.0; }
    // Take care of df/dr * (1/r) and f/r^2 first...
    T y;                     // value of function
    T yp = operator()(r, y); // Sets up extra var's
    dyrinv = yp * (1/r);
    yrinv2 = y * (1/r);
    y *= r;

    T fc_pp = fc_p*(fc_p + screenl_inv); // = (fc''/fc) - (fc'/fc)^2
    // y'' = a''*expgr_fc + 2*y'*expgr_fc_p + y*(fc_pp - expgr_fc_p^2)
    return poly_pp(r)*expgr_fc + 2*yp*expgr_fc_p 
      + y*(fc_pp - expgr_fc_p*expgr_fc_p);
  }

private:
  // Cutoff parameters:
  T Rcut, r0, screenl_inv;
  // NRL functional form:
  T D, E, F, Fb, G;

  // Extra storage to make evaluations faster: --DRT
  T expgr_fc, expgr_fc_p;  // exp(-g*r)*fc, and "derivative"
  T fc, fc_p;              // fc, and fc'/fc

  // Polynomial function and it's two derivatives:
  inline T poly(T r)    { return D + r*(E + r*(F + r*Fb));}
  inline T poly_p(T r)  { return E + r*(2*F + 3*r*Fb);}
  inline T poly_pp(T r) { return 2*F + 6*r*Fb;}  
};
// ============================ NRLScaleFunction2 ======================


// ============================== NRLSplineScale =======================
// Quadratic polynomial, with splines
template<class T> 
struct NRLSplineScale {

  NRLSplineScale(T e0=0.0, T f0=0.0, T fbar0=0.0, T g0=0.0, 
                 T Rcut0=0.0, T screenl=1.0, T Rmin0=0.0, T sigma0=0.0) { 
    set(e0,f0,fbar0,g0,Rcut0,screenl,Rmin0,sigma0);
  }

  void set(T e0, T f0, T fbar0, T g0, T Rcut0, T screenl,
	   T Rmin0, T sigma0) {
    Rcut = Rcut0;               // Cutoff parameters
    screenl_inv = 1.0/screenl;
    r0 = Rcut0 - 5.0*screenl;

    E = e0;  F = f0;  Fb = fbar0;  // NRL parameters
    G = g0;                                // Note: input is already squared

    Rmin = Rmin0;  sigma = sigma0;    // Spline parameters
    Rend = Rmin - sigma;
    init_spline(); // Initialize spline param
  }

  // ==== y ====
  T operator()(T r){ 
    if (r > Rcut)  return 0.0;
    if (r >= Rmin) {
      // Store for use by first + second derivative code... --DRT
      fc = 1./(1.0 + exp((r-r0)*screenl_inv));
      expgr_fc = exp(-G*r)*fc;
      // y   = a*expgr_fc
      return poly(r)*expgr_fc;
    }
    // else... we need to use the spline value:
    T u = spline_u(r);
    return a + u*u*u*(b + u*c);
  }

  // ==== y' ====
  T operator()(T r, T& vrinv){
    if (r > Rcut) {
      vrinv = 0.0;
      return 0.0;
    } 

    T y = operator()(r); // Calc's y(r) and sets up extra var's --DRT
    vrinv = y/r;         // Always the same

    if (r >= Rmin) {
      fc_p = screenl_inv*(fc-1.);  // = fc'/fc, for use by second deriv.
      expgr_fc_p = -G + fc_p;      // also used by second deriv. code
      // y'  = a'*expgr_fc + y*expgr_fc_p
      return poly_p(r)*expgr_fc + y*expgr_fc_p;
    } 
    // else... we need to use the spline value:
    T u = spline_u(r);
    return u*u*(3*b+4*c*u);
  }
  
  // ==== y'' ====
  T operator()(T r, T& yrinv2, T& dyrinv){
    if (r > Rcut) {
      yrinv2 = 0.0;
      dyrinv = 0.0;
      return 0.0;
    }

    // Take care of df/dr * (1/r) and f/r^2 first...
    T y;                     // value of function
    T yp = operator()(r, y); // Sets up extra var's
    dyrinv = yp * (1/r);
    yrinv2 = y * (1/r);
    y *= r;
    
    if (r >= Rmin) {
      T fc_pp = fc_p*(fc_p + screenl_inv); // = (fc''/fc) - (fc'/fc)^2
      // y'' = a''*expgr_fc + 2*y'*expgr_fc_p + y*(fc_pp - expgr_fc_p^2)
      return poly_pp(r)*expgr_fc + 2*yp*expgr_fc_p 
	+ y*(fc_pp - expgr_fc_p*expgr_fc_p);
    }
    // else... we need to use the spline value:
    T u = spline_u(r);
    return u*(6.*b + 12.*u*c);
  }

private:
  // Cutoff parameters:
  T Rcut, r0, screenl_inv;
  // NRL functional form:
  T E, F, Fb, G;
  // Spline param at small distances:
  T Rmin, Rend, sigma; // We use the spline from Rend up to Rmin
  T a, b, c;

  // Extra storage to make evaluations faster: --DRT
  T expgr_fc, expgr_fc_p;  // exp(-g*r)*fc, and "derivative"
  T fc, fc_p;              // fc, and fc'/fc

  // Polynomial function and it's two derivatives:
  inline T poly(T r)    { return E + r*(F + r*Fb);}
  inline T poly_p(T r)  { return F + 2*r*Fb;}
  inline T poly_pp(T r) { return 2*Fb;}  


  // Spline handling -- convert r into "u" for spline:
  inline T spline_u(T r) { return ( ( r > Rend ) ? (r-Rend) : 0); }

  void init_spline () // Constructs the spline param.
  {
    // New method--completely independent of scaling function. --DRT
    T s0, sp0, spp0; // Values of scaling function + deriv. at Rmin
    // Evaluate at Rmin -- requires that we use >= comparison in
    // function evaluations...
    spp0 = operator()(Rmin, s0, sp0); // Get second deriv., first, and value
    sp0 *= Rmin;  s0 *= Rmin*Rmin;
    
    a = s0 - sp0*sigma/2. + spp0*sigma*sigma/12.;
    b = sp0/(sigma*sigma) - spp0/(3.*sigma);
    c = -sp0/(2.*sigma*sigma*sigma) + spp0/(4.*sigma*sigma);
  }
 };
// ============================== NRLSplineScale =======================


// ============================= NRLSplineScale2 =======================
// Cubic polynomial, with splines
template<class T> 
struct NRLSplineScale2 {

  NRLSplineScale2(T d0=0.0, T e0=0.0, T f0=0.0, T fbar0=0.0, T g0=0.0, 
                 T Rcut0=0.0, T screenl=1.0, T Rmin0=0.0, T sigma0=0.0) { 
    set(d0, e0,f0,fbar0,g0,Rcut0,screenl,Rmin0,sigma0);
  }

  void set(T d0, T e0, T f0, T fbar0, T g0, T Rcut0, T screenl,
	   T Rmin0, T sigma0) {
    Rcut = Rcut0;               // Cutoff parameters
    screenl_inv = 1.0/screenl;
    r0 = Rcut0 - 5.0*screenl;

    D = d0; E = e0;  F = f0;  Fb = fbar0;  // NRL parameters
    G = g0;                                // Note: input is already squared

    Rmin = Rmin0;  sigma = sigma0;    // Spline parameters
    Rend = Rmin - sigma;
    init_spline(); // Initialize spline param
  }

  // ==== y ====
  T operator()(T r){ 
    if (r > Rcut)  return 0.0;
    if (r >= Rmin) {
      // Store for use by first + second derivative code... --DRT
      fc = 1./(1.0 + exp((r-r0)*screenl_inv));
      expgr_fc = exp(-G*r)*fc;
      // y   = a*expgr_fc
      return poly(r)*expgr_fc;
    }
    // else... we need to use the spline value:
    T u = spline_u(r);
    return a + u*u*u*(b + u*c);
  }

  // ==== y' ====
  T operator()(T r, T& vrinv){
    if (r > Rcut) {
      vrinv = 0.0;
      return 0.0;
    } 

    T y = operator()(r); // Calc's y(r) and sets up extra var's --DRT
    vrinv = y/r;         // Always the same

    if (r >= Rmin) {
      fc_p = screenl_inv*(fc-1.);  // = fc'/fc, for use by second deriv.
      expgr_fc_p = -G + fc_p;      // also used by second deriv. code
      // y'  = a'*expgr_fc + y*expgr_fc_p
      return poly_p(r)*expgr_fc + y*expgr_fc_p;
    } 
    // else... we need to use the spline value:
    T u = spline_u(r);
    return u*u*(3*b+4*c*u);
  }
  
  // ==== y'' ====
  T operator()(T r, T& yrinv2, T& dyrinv){
    if (r > Rcut) {
      yrinv2 = 0.0;
      dyrinv = 0.0;
      return 0.0;
    }

    // Take care of df/dr * (1/r) and f/r^2 first...
    T y;                     // value of function
    T yp = operator()(r, y); // Sets up extra var's
    dyrinv = yp * (1/r);
    yrinv2 = y * (1/r);
    y *= r;
    
    if (r >= Rmin) {
      T fc_pp = fc_p*(fc_p + screenl_inv); // = (fc''/fc) - (fc'/fc)^2
      // y'' = a''*expgr_fc + 2*y'*expgr_fc_p + y*(fc_pp - expgr_fc_p^2)
      return poly_pp(r)*expgr_fc + 2*yp*expgr_fc_p 
	+ y*(fc_pp - expgr_fc_p*expgr_fc_p);
    }
    // else... we need to use the spline value:
    T u = spline_u(r);
    return u*(6.*b + 12.*u*c);
  }

private:
  // Cutoff parameters:
  T Rcut, r0, screenl_inv;
  // NRL functional form:
  T D, E, F, Fb, G;
  // Spline param at small distances:
  T Rmin, Rend, sigma; // We use the spline from Rend up to Rmin
  T a, b, c;

  // Extra storage to make evaluations faster: --DRT
  T expgr_fc, expgr_fc_p;  // exp(-g*r)*fc, and "derivative"
  T fc, fc_p;              // fc, and fc'/fc

  // Polynomial function and it's two derivatives:
  inline T poly(T r)    { return D + r*(E + r*(F + r*Fb));}
  inline T poly_p(T r)  { return E + r*(2*F + 3*r*Fb);}
  inline T poly_pp(T r) { return 2*F + 6*r*Fb;}  


  // Spline handling -- convert r into "u" for spline:
  inline T spline_u(T r) { return ( ( r > Rend ) ? (r-Rend) : 0); }

  void init_spline () // Constructs the spline param.
  {
    // New method--completely independent of scaling function. --DRT
    T s0, sp0, spp0; // Values of scaling function + deriv. at Rmin
    // Evaluate at Rmin -- requires that we use >= comparison in
    // function evaluations...
    spp0 = operator()(Rmin, s0, sp0); // Get second deriv., first, and value
    sp0 *= Rmin;  s0 *= Rmin*Rmin;
    
    a = s0 - sp0*sigma/2. + spp0*sigma*sigma/12.;
    b = sp0/(sigma*sigma) - spp0/(3.*sigma);
    c = -sp0/(2.*sigma*sigma*sigma) + spp0/(4.*sigma*sigma);
  }
};

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 264 $   $Date: 2005-01-15 16:49:17 -0600 (Sat, 15 Jan 2005) $
 * $Id: EDPotential.h,v 1.1 1998/12/21 16:03:21 trinkle Exp trinkle
 ***************************************************************************/
