//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef SLATERKOSTERARRAYBASE_H
#define SLATERKOSTERARRAYBASE_H

#include <vector>
namespace SlaterKosterConst {
  const double onethird     = 0.333333333333333333333333333333333333333;
  const double twothirds    = 0.666666666666666666666666666666666666667;
  const double sqrt3        = 1.732050807568877293527446341505872366943;
  const double oneoversqrt3 = 0.5773502691896257645091487805019574556475;
  const double twooversqrt3 = 1.154700538379251529018297561003914911295;
  const double oneovertwosqrt3 = 0.2886751345948128822545743902509787278238;

  enum {SK_Px =0, SK_Py =1, SK_Pz =2, 
    SK_Dxy = 3, SK_Dyz = 4, SK_Dzx = 5, SK_Dx2y2 = 6, SK_Dz2r2 = 7};
}

using namespace SlaterKosterConst;

template<class T>
class SlaterKosterBase {
public:

  virtual void getOnsite(T*)  { }
  virtual void getOffsite(T r, const T* , T* b) = 0;
  virtual void getDeriv(T r, const T*, const T [][3], T* bx, T* by, T* bz){ }
  virtual void getDeriv2(T r, const T*, const T [][3], const T [][3][3], 
                         T* bx, int i, int j) {}

  //!< Fill-in the hamiltonian matrix with the indecies
  virtual void assign(T* ){}

  //!< Reset the block dimension after all the SlaterKoster blocks are added
  virtual void reset(int nrow, int ncol) = 0;

protected:
  SlaterKosterBase(){}
  ~SlaterKosterBase() {}

  int I0, J0; // starting (row, col) used to calculate relative coordinates
              // of each block after all the blocks are assembled  
};

// aux structure K(L1,L2)
template<class T, unsigned L1, unsigned L2, class SC> 
struct SK{

};

template<class T, unsigned L1, unsigned L2, class SC>
struct SlaterKoster: public SlaterKosterBase<T>
{

};

// composite class for SlaterKosterBase
template<class T>
class SlaterKosterComp{ //: public SlaterKosterBase<T> {
private:

  std::vector<SlaterKosterBase<T>* > SKSet;
  T Rcut;
  int Nrow, Ncol, Lmax;
public:

  SlaterKosterComp():Nrow(0), Ncol(){SKSet.reserve(5);}
  ~SlaterKosterComp() { }

  inline int maxL() const { return Lmax;}
  inline int nrow() const { return Nrow;}
  inline int ncol() const { return Ncol;}

  void assign(T* b){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->assign(b);
  }

  void add(SlaterKosterBase<T>* sk) { 
    SKSet.push_back(sk); // add a SlaterKosterBase
  }

  void reset(int nrow, int ncol, int lmax) {
    Lmax = lmax;
    Nrow = nrow; Ncol = ncol;
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->reset(nrow, ncol);
  }

  inline void setRcut(T r){ Rcut = r; }
  inline T    rcut() const{ return Rcut; }

  void getOnsite(T* b){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getOnsite(b);  
  }
  void getOffsite(T r, const T* p, T* b){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getOffsite(r,p,b);  
  }
  void getDeriv(T r, const T* p, const T pd[][3], T* bx, T* by, T* bz){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getDeriv(r,p,pd,bx,by,bz);  
  }

  void getDeriv2(T r, const T* p, const T pd [][3], const T pdd[][3][3], 
                 T* bx, int d0, int d1) {
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getDeriv2(r,p,pd,pdd,bx,d0,d1);
  }

};

template<class PosType, class T>
void getPD(const PosType& p, T* pvec, int maxL) {
  pvec[0] = p[0]; pvec[1] = p[1]; pvec[2] = p[2];
  if(maxL == 2) {
    pvec[3] = p[0]*p[1]; 
    pvec[4] = p[1]*p[2]; 
    pvec[5] = p[2]*p[0];
    pvec[6] = 0.5e0*(p[0]*p[0]-p[1]*p[1]);
    pvec[7] = (3.0e0*p[2]*p[2]-1.0e0)*oneovertwosqrt3;
  }
} 
template<class PosType, class T>
void getPD(const PosType& p, T* pvec, T pd[][3], int maxL) {
  //pvec[0] = p[0]; pvec[1] = p[1]; pvec[2] = p[2];
  getPD(p,pvec,maxL);
  pd[0][0] =  1-p[0]*p[0]; pd[1][0]= -p[1]*p[0]; pd[2][0] = -p[2]*p[0];
  pd[0][1] =     pd[1][0]; pd[1][1]=1-p[1]*p[1]; pd[2][1] = -p[2]*p[1];
  pd[0][2] =     pd[2][0]; pd[1][2]=   pd[2][1]; pd[2][2] =1-p[2]*p[2];
  if(maxL == 2) {
    pd[3][0] = pd[0][0]*p[1] + p[0]*pd[1][0]; 
    pd[3][1] = pd[0][1]*p[1] + p[0]*pd[1][1]; 
    pd[3][2] = pd[0][2]*p[1] + p[0]*pd[1][2]; 

    pd[4][0] = pd[1][0]*p[2] + p[1]*pd[2][0]; 
    pd[4][1] = pd[1][1]*p[2] + p[1]*pd[2][1]; 
    pd[4][2] = pd[1][2]*p[2] + p[1]*pd[2][2]; 

    pd[5][0] = pd[2][0]*p[0] + p[2]*pd[0][0]; 
    pd[5][1] = pd[2][1]*p[0] + p[2]*pd[0][1]; 
    pd[5][2] = pd[2][2]*p[0] + p[2]*pd[0][2]; 

    pd[6][0] = p[0]*pd[0][0] - p[1]*pd[1][0];
    pd[6][1] = p[0]*pd[0][1] - p[1]*pd[1][1];
    pd[6][2] = p[0]*pd[0][2] - p[1]*pd[1][2];

    pd[7][0] = sqrt3*p[2]*pd[2][0];
    pd[7][1] = sqrt3*p[2]*pd[2][1];
    pd[7][2] = sqrt3*p[2]*pd[2][2];
  }
}

template<class PosType, class T>
void getPD(const PosType& pvec, T* p, T pd[][3], T pdd[][3][3], int maxL) {
  getPD(pvec,p,pd,maxL);

  for(int k=0; k<3; k++) 
    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[k][i][j] = -(p[k]*pd[i][j] + pd[k][j]*p[i]);

  if(maxL) {
    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[3][i][j] = pdd[0][i][j]*p[1]+p[0]*pdd[1][i][j]
                     + pd[0][i]*pd[1][j]+pd[0][j]*pd[1][i];


    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[4][i][j] = pdd[1][i][j]*p[2]+p[1]*pdd[2][i][j]
                     + pd[1][i]*pd[2][j]+pd[1][j]*pd[2][i];

    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[5][i][j] = pdd[2][i][j]*p[0]+p[2]*pdd[0][i][j]
                     + pd[2][i]*pd[0][j]+pd[2][j]*pd[0][i];

    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[6][i][j] = pd[0][i]*pd[0][j]+p[0]*pdd[0][i][j]
                     - pd[1][i]*pd[1][j]-p[1]*pdd[1][i][j];

    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[7][i][j] = sqrt3*(pd[2][i]*pd[2][j] + p[2]*pdd[2][i][j]);

  } // end UseD case

  // 1/r term is added
  for(int k=0; k<8; k++) 
    for(int i=0; i<3; i++)
      for(int j=0; j<3; j++)
        pdd[k][i][j] -= p[j]*pd[k][i];
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: SlaterKosterBase.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
