//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef SLATERKOSTERARRAY_ONSITE_BASE_H
#define SLATERKOSTERARRAY_ONSITE_BASE_H

template<class T>
class SlaterKosterOnsiteBase {
public:

  virtual void getValue(T*, T rho) = 0; // value
  virtual void getDeriv(T*, T rho) = 0; //  deriv

  //!< Fill-in the hamiltonian matrix with the indecies
  virtual void assign(T* ) = 0;

  //!< Reset the block dimension after all the SlaterKoster blocks are added
  virtual void reset(int nrow) = 0;

protected:

  SlaterKosterOnsiteBase(){}
  ~SlaterKosterOnsiteBase() {}
};

template<class T, unsigned L, class SF>
struct SKOnsite: public SlaterKosterOnsiteBase<T>
{

};

template<class T,class SF>
struct SKOnsite<T,0,SF>: public SlaterKosterOnsiteBase<T>
{
  SF* V;
  int myO;
  SKOnsite(int o, SF* sc): myO(o), V(sc){}

  void getValue(T* b, T rho) { 
    b[myO] += V->f(rho);
  }

  void getDeriv(T* db, T rho) {
    db[myO] = V->df(rho);
  }

  //!< Fill-in the hamiltonian matrix with the indecies
  void assign(T* b){ b[myO] = myO; }

  //!< Reset the block dimension after all the SlaterKoster blocks are added
  void reset(int nrow) { myO += nrow*myO;}

};

template<class T,class SF>
struct SKOnsite<T,1,SF>: public SlaterKosterOnsiteBase<T>
{
  SF* V;
  int my0, my1, my2;
  SKOnsite(int o, SF* sc): my0(o), V(sc){}

  void getValue(T* b, T rho) { 
    T tmp = V->f(rho);
    b[my0] += tmp; b[my1] += tmp; b[my2] += tmp;
  }

  void getDeriv(T* db, T rho) { 
    T dv = V->df(rho);
    db[my0] = dv; db[my1] = dv; db[my2] = dv;
  }

  //!< Fill-in the hamiltonian matrix with the indecies
  void assign(T* b){ 
    b[my0] = my0; b[my1] = my1; b[my2] = my2;
  }

  //!< Reset the block dimension after all the SlaterKoster blocks are added
  void reset(int nrow) { 
    my0 += my0*nrow;
    my1 = my0 + nrow + 1;
    my2 = my1 + nrow + 1;
  }

};


template<class T,class SF>
struct SKOnsite<T,2,SF>: public SlaterKosterOnsiteBase<T>
{
  // including splitting t2g and eg
  SF *V, *V2;
  int my0, my1, my2, my3, my4;

  SKOnsite(int o, SF* sc, SF* sc2): my0(o), V(sc), V2(sc2){}

  void getValue(T* b, T rho) { 

    T tmp = V->f(rho);
    b[my0] += tmp; b[my1] += tmp; b[my2] += tmp;
    if(V2){
      T tmp2 = V2->f(rho);
      b[my3] += tmp2; b[my4] += tmp2; 
    } else {
      b[my3] += tmp; b[my4] += tmp; 
    }
  }

  void getDeriv(T* b, T rho) { 
    T tmp = V->df(rho);
    b[my0] += tmp; b[my1] += tmp; b[my2] = +tmp;
    if(V2) {
      T tmp2 = V2->df(rho);
      b[my3] += tmp2; b[my4] += tmp2; 
    } else {
      b[my3] += tmp; b[my4] += tmp; 
    }
  }

  //!< Fill-in the hamiltonian matrix with the indecies
  void assign(T* b){ 
    b[my0] = my0; b[my1] = my1; b[my2] = my2;
    b[my3] = my3; b[my4] = my4;
  }

  //!< Reset the block dimension after all the SlaterKoster blocks are added
  void reset(int nrow) { 
    my0 +=  nrow*my0;
    my1 = my0 + nrow + 1;
    my2 = my1 + nrow + 1;
    my3 = my2 + nrow + 1;
    my4 = my3 + nrow + 1;
  }

};

//composite class for SlaterKosterOnsite
//template<class T>
template<class T, class SF>
class SlaterKosterOnsiteComp {
private:

  vector<SlaterKosterOnsiteBase<T>* > SKSet;
  int Nrow;

public:

  SlaterKosterOnsiteComp():Nrow(0){SKSet.reserve(5);}
  ~SlaterKosterOnsiteComp() { }

  inline int nrow() const { return Nrow;}
  inline int ncol() const { return Nrow;}

  void assign(T* b){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->assign(b);
  }

//   void add(SlaterKosterOnsiteBase<T>* sk)
//     SKSet.push_back(sk); // add a SlaterKosterBase
//   }

  void add(int L, int pos, SF* v, SF* v2=NULL) { 
    switch(L) {
    case(0):
      SKSet.push_back(new SKOnsite<T,0,SF>(pos,v));
      break;
    case(1):
      SKSet.push_back(new SKOnsite<T,1,SF>(pos,v));
      break;
    case(2):
      SKSet.push_back(new SKOnsite<T,2,SF>(pos,v,v2));
      break;
    }
  }

  void reset(int nrow) {
    Nrow = nrow;
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->reset(nrow);
  }

  void getValue(T* b, T rho){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getValue(b,rho);  
  }

  void getDeriv(T* b, T rho){
    for(int i=0; i<SKSet.size(); i++)
      SKSet[i]->getDeriv(b,rho);  
  }
};


#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SlaterKosterOnsite.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
