//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_SPINORBIT_H
#define OHMMS_SPINORBIT_H

#include <complex>
using namespace std;

template<class T>
class SpinOrbitBase {
public:
  virtual void getSO(complex<T>*) = 0;
  virtual void set(int,int) = 0;
  virtual void apply(const complex<T>* x, complex<T>* y) = 0;
  virtual void apply(const T* x, T* y) = 0;
  virtual void copy(complex<T>* x, int i0, int j0, unsigned int col) = 0;
  virtual void copy(complex<T>* x, int up, unsigned int col) = 0;
protected:
  SpinOrbitBase(){}
  ~SpinOrbitBase() {}
};

template<class T>
class SpinOrbitComp: public SpinOrbitBase<T> {
  vector<SpinOrbitBase<T>* > SOSet;
  int Nrow, Ncol;
public:

  SpinOrbitComp(){SOSet.reserve(5);}
  ~SpinOrbitComp() { }

  inline int nrow() const { return Nrow;}
  inline int ncol() const { return Ncol;}

  void add(SpinOrbitBase<T>* sk) { SOSet.push_back(sk); }

  void getSO(complex<T>* sob) {
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->getSO(sob);
  }

  void apply(const complex<T>* x, complex<T>* y) {
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->apply(x,y);
  }

  void apply(const T* x, T* y) {
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->apply(x,y);
  }

  void copy(complex<T>* x, int i0, int j0, unsigned int cols) {
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->copy(x,i0,j0,cols);
  }

  void copy(complex<T>* x, int up, unsigned int cols) {
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->copy(x,up,cols);
  }

  void set(int n,int m) {
    Nrow = 2*n; Ncol = 2*m;
    for(int i=0; i<SOSet.size(); i++) SOSet[i]->set(n,m);
  }
};

template<class T>
class SpinOrbitPP: public SpinOrbitBase<T> {
  T Lambda,halfL;
  complex<T> halfLi;
  int u0, u1, u2, d0, d1, d2;
  int N, N2;
public:
  SpinOrbitPP(T lam, int, int);
  void getSO(complex<T>* );
  void set(int,int);
  void apply(const complex<T>* x, complex<T>* y);
  void apply(const T* x, T* y);
  void copy(complex<T>* x, int i0, int j0, unsigned int cols);
  void copy(complex<T>* x, int up, unsigned int cols);
};

template<class T> 
SpinOrbitPP<T>::SpinOrbitPP(T lam, int i0, int j0): Lambda(lam) {
  halfL = Lambda*0.5;
  u0 = i0; u1=i0+1; u2 = i0+2;
  halfLi = complex<T>(0,halfL);
}
template<class T> 
void
SpinOrbitPP<T>::set(int n, int m) {
  N = m;
  N2 = N*2;
  d0 = u0 + N;
  d1 = u1 + N;
  d2 = u2 + N;
}

template<class T>
void SpinOrbitPP<T>::getSO(complex<T>* sob) {
  sob[u0*N2+u1] = -halfLi;
  sob[u1*N2+u0] =  halfLi;
  sob[d0*N2+d1] =  halfLi;
  sob[d1*N2+d0] = -halfLi;

  sob[u0*N2+d2] =  halfL;
  sob[u2*N2+d0] = -halfL;
  sob[d0*N2+u2] = -halfL;
  sob[d2*N2+u0] =  halfL;

  sob[u1*N2+d2] = -halfLi;
  sob[u2*N2+d1] =  halfLi;
  sob[d1*N2+u2] = -halfLi;
  sob[d2*N2+u1] =  halfLi;
//    sob[Iuu[0][1]] = complex<T>(0,-lam);
//    sob[Iuu[1][0]] = complex<T>(0,lam);

//    sob[Iud[0][2]] = complex<T>(lam,0);
//    sob[Iud[2][0]] = complex<T>(-lam,0);

//    sob[Iud[1][2]] = complex<T>(0,-lam);
//    sob[Iud[2][1]] = complex<T>(0,lam);

}

template<class T>
void 
SpinOrbitPP<T>::apply(const complex<T>* x, complex<T>* y) {
 
  y[u0] -= halfLi*x[u1];
  y[u1] += halfLi*x[u0];
  y[d0] += halfLi*x[d1];
  y[d1] -= halfLi*x[d0];

  y[u0] += halfL *x[d2];
  y[u2] -= halfL *x[d0];
  y[d0] -= halfL *x[u2];
  y[d2] += halfL *x[u0];

  y[u1] -= halfLi*x[d2];
  y[u2] += halfLi*x[d1];
  y[d1] -= halfLi*x[u2];
  y[d2] += halfLi*x[u1];
}

template<class T>
void 
SpinOrbitPP<T>::apply(const T* x, T* y) {
 
  y[2*u0]   += halfL*(x[2*u1+1]+x[2*d2]); 
  y[2*u0+1] += halfL*(x[2*d2+1]-x[2*u1]);
  y[2*u1]   += halfL*(x[2*d2+1]-x[2*u0+1]);
  y[2*u1+1] += halfL*(x[2*u0]-x[2*d2]);
  y[2*d0]   -= halfL*(x[2*d1+1]+x[2*u2]);
  y[2*d0+1] += halfL*(x[2*d1]-x[2*u2+1]);
  y[2*d1]   += halfL*(x[2*d0+1]+x[2*u2+1]);
  y[2*d1+1] -= halfL*(x[2*d0]+x[2*u2]);
  y[2*u2]   -= halfL*(x[2*d0]+x[2*d1+1]);
  y[2*u2+1] += halfL*(x[2*d1]-x[2*d0+1]);
  y[2*d2]   += halfL*(x[2*u0]-x[2*u1+1]);
  y[2*d2+1] += halfL*(x[2*u0+1]+x[2*u1]);
}

template<class T>
void 
SpinOrbitPP<T>::copy(complex<T>* x, int i0, int j0, unsigned int bigcol) {
  int del = i0*bigcol + j0; 
  x[u0*bigcol+u1+del] -= halfLi;
  x[u1*bigcol+u0+del] += halfLi;
  x[d0*bigcol+d1+del] += halfLi;
  x[d1*bigcol+d0+del] -= halfLi;
  
  x[u0*bigcol+d2+del] += halfL;
  x[u2*bigcol+d0+del] -= halfL;
  x[d0*bigcol+u2+del] -= halfL;
  x[d2*bigcol+u0+del] += halfL;
  
  x[u1*bigcol+d2+del] -= halfLi;
  x[u2*bigcol+d1+del] += halfLi;
  x[d1*bigcol+u2+del] -= halfLi;
  x[d2*bigcol+u1+del] += halfLi;
}

template<class T>
void 
SpinOrbitPP<T>::copy(complex<T>* x, int up, unsigned int bigcol) {
  //      int ushift = (ncoltot*2-iuu-1)*iuu/2;
  int ncol2 = bigcol*2;
  int iu0 = up + u0;
  int iu1 = up + u1;
  int iu2 = up + u2;
  int id0 = up + d0;
  int id1 = up + d1;
  int id2 = up + d2;
  x[(ncol2-iu0-1)*iu0/2 + iu1] -= halfLi; //x[u0*bigcol+u1+del] -= halfLi;
  x[(ncol2-id0-1)*id0/2 + id1] += halfLi; //x[d0*bigcol+d1+del] += halfLi;
  
  x[(ncol2-iu0-1)*iu0/2 + id2] += halfL;  //x[u0*bigcol+d2+del] += halfL;
  x[(ncol2-iu2-1)*iu2/2 + id0] -= halfL;  //x[u2*bigcol+d0+del] -= halfL;

  x[(ncol2-iu1-1)*iu1/2 + id2] -= halfLi; //x[u1*bigcol+d2+del] -= halfLi;
  x[(ncol2-iu2-1)*iu2/2 + id1] += halfLi; //x[u2*bigcol+d1+del] += halfLi;
}
//  template<class T>
//  void SpinOrbitPP<T>::getUU(T* uui) {
//    uui[I[0][1]] = -Lambda*0.5;
//    uui[I[1][0]] =  Lambda*0.5;
//  }                 

//  template<class T>
//  void SpinOrbitPP<T>::getUD(T* udr, T* udi) {
//    udr[I[0][2]] =  Lambda*0.5;
//    udr[I[2][0]] = -Lambda*0.5;

//    udi[I[1][2]] = -Lambda*0.5;
//    udi[I[2][1]] =  Lambda*0.5;
//  }                 
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SpinOrbitBase.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
