//////////////////////////////////////////////////////////////////
//   (c) Copyright 1998-2002,2003- Jeongnim Kim.   
// All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "SlaterKosterTB/FastSKTB.h"
#include "Particle/NNEngine.h"
#include "Utilities/Clock.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/OpenMP.h"
#include "OhmmsNumeric/SparseEigenSolver.h"
#include "OhmmsNumeric/SparseEigenSolverBuilder.h"
#include "QuantumBase/ElectronStructureIO.h"
#include "SlaterKosterTB/TBModelBase.h"
using namespace OHMMS;
#include <iostream>
#include <fstream>
#include <iomanip>

FastSKTB::FastSKTB(): thisSolver(0){ 
  setName("FastSKTB");
  myRecorder = new ESOutput(*this);
  accCPU.resize(5,0.0);
}

FastSKTB::~FastSKTB(){ 
  LOGMSG("FastSKTB timing report (in secs)")
  LOGMSG("    Setting up H/S matrices      = " << accCPU[0])
  LOGMSG("    Diagonalization              = " << accCPU[1])
  LOGMSG("    Electronic Structure energy  = " << accCPU[2])
  LOGMSG("    Evaluation of Density Matrix = " << accCPU[3])
  LOGMSG("    Force evaluation             = " << accCPU[4])
  if(thisSolver) delete thisSolver;
}

void 
FastSKTB::setEigenSolver() {
  //set gamma-point specific initialization
  Gamma = true; 
  if(!KSet) {
    KSet = new KPointBase;
    KSet->addKpt(SingleParticlePos_t(0,0,0),1);
  }
  
  NumKpts = KSet->size();
  Occ.resize(NumKpts,NumStates);
  E.resize(NumKpts,NumStates);
  Occ = 0.0e0;
  E = 0.0e0;

  // create a solver
  LOGMSG("FastSKTB::setParticeSet, creating a symmetric solver")
  thisSolver = new EigSolver_t(States,!Identity);
  thisSolver->resize(Identity);

  A.resize(BasisDim,BasisDim);
  if(!Identity) {
    DenMatList.push_back(&RhoS); 
    B.resize(BasisDim,BasisDim);
  }
}

FastSKTB::Scalar_t 
FastSKTB::solveGES(){

  bool needrho = EvaluateDensityMatrix;

  Pooma::Clock timer;
  timer.start();
  thisSolver->H.fill(0.0e0);
  if(!Identity) thisSolver->S.fill(0.0e0);

#pragma omp parallel
  {
    omp_int_t ip=omp_get_thread_num();
    if(UpdateNNList) 
      nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(ip));
    if(Identity)
      assignBlocks(myPtcl->getNNEngine()->getPairAttrib(ip),thisSolver->H);
    else
      assignBlocks(myPtcl->getNNEngine()->getPairAttrib(ip),
          thisSolver->H, thisSolver->S);
  }

  timer.stop();
  accCPU[0]+= timer.cpu_time();

  timer.start();
  if(needrho) {
    thisSolver->solveEigVectors(Identity);
  } else {
    thisSolver->solveEigValues(Identity);
  }
  timer.stop();
  accCPU[1]+= timer.cpu_time();

  timer.start();
  for(int is=0; is<E.cols(); is++)  {
    E[0][is] = thisSolver->eigval(is);
  }
  Scalar_t tbe = getBandEnergy();
  timer.stop();
  accCPU[2]+= timer.cpu_time();
  return tbe;
}

void FastSKTB::assignBlocks(const PtclPairList_t& nnlist, PackedSymMatrix_t& h) {

  int ip(nnlist.myID);
  typedef PtclPairList_t::PairData_t PairData_t;
  const ParticleIndex_t& grID(myPtcl->GroupID);
  int nloc(nnlist.getLocalNum());

  //assign the diagonal term
  Hmat.clearDiagonal(ip);
  int stride(Hmat.getBlockSize());
  SparseMatrix_t::pointer hPtr=Hmat.diag(ip,0);
  for(int iL=0; iL<nloc;iL++) {
    SKP.getSKBlocks(grID[nnlist.I[iL]], hPtr); 
    hPtr+=stride;
  }
  SKP.setOnsite(Hmat,nnlist);

  Matrix<Scalar_t> hBlock(HamIndex->MaxRowBlock,HamIndex->MaxColBlock);
  hPtr=Hmat.diag(ip,0);
  for(int iL=0; iL<nloc;iL++) {
    int iat(nnlist.I[iL]);
    int speciesI(grID[iat]);
    int nbi(HamIndex->row_block(iat));
    int ioff(HamIndex->row_loc(iat));

    SparseMatrix_t::pointer hb_ptr=hPtr;
    for(int k=0,ii=ioff; k<nbi; k++,ii++) {
      for(int kp=0,jj=ioff; kp<nbi; kp++, jj++) {
        if(kp>=k) {
          h(ii,jj) += *hb_ptr;
        }
        ++hb_ptr;
      }
    }
    hPtr+=stride;

    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    while(nit != nit_end) {
      const PairData_t& cur(*nit++);
      int jat(cur.Next);
      if(jat<iat) continue;
      hBlock=0.0;
      SKP.getSKBlocks(speciesI,cur.Group,cur.sep,cur.displ,hBlock.data());
      int nbj(HamIndex->row_block(jat));
      int joff(HamIndex->row_loc(jat));
      const Scalar_t* ob_ptr=hBlock.data();
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
	for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
	  h(ii,jj) += *ob_ptr++;
	}//kp
      }//k
    }//nit!=nit_end
  }
}

void FastSKTB::assignBlocks(const PtclPairList_t& nnlist, PackedSymMatrix_t& h,
PackedSymMatrix_t& s) {

  int ip(nnlist.myID);
  typedef PtclPairList_t::PairData_t PairData_t;
  const ParticleIndex_t& grID(myPtcl->GroupID);
  int nloc(nnlist.getLocalNum());
  Matrix<Scalar_t> hBlock(HamIndex->MaxRowBlock,HamIndex->MaxColBlock), 
    sBlock(HamIndex->MaxRowBlock,HamIndex->MaxColBlock);

  Hmat.clear(0,ip);               

  int stride(Hmat.getBlockSize());
  SparseMatrix_t::pointer hPtr=Hmat.diag(ip,0);
  SparseMatrix_t::pointer sPtr=Smat.diag(ip,0);
  //assign the diagonal term
  for(int iL=0; iL<nloc;iL++) {
    SKP.getSKBlocks(grID[nnlist.I[iL]], hPtr,sPtr);
    hPtr+=stride; sPtr+=stride;
  }
  SKP.setOnsite(Hmat,nnlist);

  hPtr=Hmat.diag(ip,0);
  sPtr=Smat.diag(ip,0);
  for(int iL=0; iL<nloc;iL++) {
    int iat(nnlist.I[iL]);
    int speciesI(grID[iat]);
    int nbi(HamIndex->row_block(iat));
    int ioff(HamIndex->row_loc(iat));

    SparseMatrix_t::pointer  hb_ptr=hPtr;
    SparseMatrix_t::pointer  sb_ptr=sPtr;
    for(int k=0,ii=ioff; k<nbi; k++,ii++) {
      for(int kp=0,jj=ioff; kp<nbi; kp++, jj++,hb_ptr++,sb_ptr++) {
        if(jj>ii) {
          h(ii,jj) = *hb_ptr;
          s(ii,jj) = *sb_ptr;
        } 
      }
    }
    hPtr+=stride;
    sPtr+=stride;

    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      int jat(cur.Next);
      if(jat<iat) continue;
      hBlock=0.0;sBlock=0.0;
      SKP.getSKBlocks(speciesI,cur.Group,cur.sep,cur.displ,
          hBlock.data(),sBlock.data());
      int nbj(HamIndex->row_block(jat));
      int joff(HamIndex->row_loc(jat));
      const Scalar_t* oh_ptr=hBlock.data();
      const Scalar_t* os_ptr=sBlock.data();
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
	for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
	  h(ii,jj) += *oh_ptr++;
	  s(ii,jj) += *os_ptr++;
	}
      }
      ++nit;
    }
  }
}

FastSKTB::Scalar_t 
FastSKTB::evalForceTB(ParticlePos_t& f0, ParticleScalar_t& q, 
  		      Tensor_t& stress, unsigned stressmode) {

  /////////////////////////////////////////////
  // always get the ground electronic structure
  /////////////////////////////////////////////
  EvaluateDensityMatrix = true;

  Scalar_t tbe = solveGES();

  vector<Scalar_t> w(Homo);
  if(ZeroTemperature || fabs(dSumOcc)<1e-16) {
    for(int is=0; is<Homo; is++)  w[is] = SpinDeg*Occ[0][is];
  } else {
    Scalar_t sigmainv = 1.0/Sigma;
    Scalar_t fac = dSumEnergy/dSumOcc;
    for(int is=0; is<Homo; is++) {
      w[is] = SpinDeg*(Occ[0][is]+(E[0][is]*sigmainv-fac)*dOcc[0][is]);
    }
  }

  Pooma::Clock timer;
  timer.start();
  Dense2SparseMatrix::product(thisSolver->Psi,thisSolver->Psi,w,Rho); 
  if(!Identity) {
    for(int is=0; is<Homo; is++) w[is] *= E[0][is];
    Dense2SparseMatrix::product(thisSolver->Psi,thisSolver->Psi,w,RhoS); // only need Rho
  }
  timer.stop();
  accCPU[3]+=timer.cpu_time();

  timer.start();
#pragma omp parallel
  {
    int np=myPtcl->getNNEngine()->size();
#pragma omp for nowait
    for(int myid=0; myid<np; myid++) {
      tbForceWithDM(myPtcl->getNNEngine()->getPairAttrib(myid), 
          f0,q,stress,stressmode);
    }
  }
  timer.stop();
  accCPU[4]+=timer.cpu_time();

  return tbe;
}

void FastSKTB::tbDynamicMatrix(DynMatrix_t& dymat) {
}

FastSKTB::Scalar_t
FastSKTB::el_ph_coupling(const vector<int>& psi_i,  
                         const PhononMode& phonon,
			 vector<int>& psi_f, Scalar_t del) {
  return 0.0;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: FastSKTB.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/
