//////////////////////////////////////////////////////////////////
// (c) Copyright 2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "SlaterKosterTB/SKParameters.h"
#include "Utilities/SimpleParser.h"
#include "Utilities/OhmmsInfo.h"

#define ENABLE_INVPOWERTB
#define ENABLE_LENOSKYTB
#define ENABLE_NRLTB
#define ENABLE_SPLINETB
#define ENABLE_GSPTB
#define ENABLE_ZBSURFACE

#if defined(ENABLE_INVPOWERTB)
#include "SlaterKosterTB/Models/InvPowerTB.h"
#endif
#if defined(ENABLE_LENOSKYTB)
#include "SlaterKosterTB/Models/LenoskyTB.h"
#endif
#if defined(ENABLE_NRLTB)
#include "SlaterKosterTB/Models/NRLTB.h"
#endif
#if defined(ENABLE_SPLINETB)
#include "SlaterKosterTB/Models/SplineTB.h"
#endif
#if defined(ENABLE_GSPTB)
#include "SlaterKosterTB/Models/GSPTB.h"
#endif
#if defined(ENABLE_ZBSURFACE)
#include "SlaterKosterTB/Models/ZincBlendSurface.h"
#endif
#include "OhmmsData/AttributeSet.h"

namespace OHMMS {

  //@{forward declaration of builder functions
  /**function to create a TBModelBase
   *@param modelname the name of a TB model
   *@param fin input stream, continue parsing using istream
   *@return a TBModelBase if the modelname is available
   */
  TBModelBase*  createTBModel(const std::string& modelname);//, std::istream& fin);

  /**function to create a TBOnsiteBase
   *@param oname the name of a TB Onsite model
   *@param fin input stream, continue parsing using istream
   *@return a TBOnsiteBase if the modelname is available
   */
  TBOnsiteBase* createTBOnsite(const std::string& oname);
  //@}

  bool SKParameters::get(ostream&) const {
    return true;
  }
  bool SKParameters::put(istream&){
    return true;
  }
  void SKParameters::reset(){
  }

  bool SKParameters::put(xmlNodePtr cur) {

    if(myPtcl == 0) {
      cout << "SKParameters do not have valid particle set!!!!!!" << endl;
      return false;
    }

    string fname("TBPARAM");
    OhmmsAttributeSet aAttrib;
    aAttrib.add(fname,"href"); 
    aAttrib.add(fname,"src"); 
    aAttrib.add(fname,"file"); 
    aAttrib.put(cur);

    cur = cur->children;
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "filename") {
	putContent(fname,cur);
      }
      cur = cur->next;
    }

    init(fname,OtherPot);

    //set the parameters to curOnsite: don't know why
    if(curOnsite) {
      curOnsite->SK_h = &SK_h;
      curOnsite->SK_s = &SK_s;
      OtherPot.push_back(curOnsite); //add curOnsite to OtherPot
    }

    Rmax = 0.0e0;
    for(int ib=0; ib<SK_h.size(); ib++) { Rmax = std::max(Rmax, SK_h[ib]->rcut()); }
    for(int ib=0; ib<SK_s.size(); ib++) { Rmax = std::max(Rmax, SK_s[ib]->rcut()); }

    CollectVisitingForces=false;
    for(int i=0; i<OtherPot.size(); i++) {
      OtherPot[i]->setParticleSet(myPtcl);
      CollectVisitingForces|= OtherPot[i]->CollectVisitingForces;
    }

    //may add OtherPot
    return true;
  }


  /** Initializations of parameters associated with SlaterKoster Parameters
   *@param fname the name of input file
   *@param pot_s a list of ShortRangePotential added by a TBModelBase
   *@return true, if the requested model is successfully created and initialized.
   *
   * Extra potentials, on-site or any short-range potential is added here.
   */
  bool
  SKParameters::init(const std::string& fname,
		     std::vector<ShortRangePotential*>& pot_s) 
  {

    std::ifstream fin(fname.c_str());
    if(!fin) {
      ERRORMSG("TB parameters are missing. File " << fname << " does not exist.")
      return false;
    }
    
    // make private references to Species
    SpeciesSet& mySpecies(myPtcl->getSpeciesSet());

    // make private references to Species
    int icharge = mySpecies.addAttrib("q");
    int ins     = mySpecies.addAttrib("ns");
    int imass   = mySpecies.addAttrib("mass");
    int irmax   = mySpecies.addAttrib("rmax");
    
    int Lmax = 0;
    vector<string> slist, vlist;
    bool tofindsk = true;
    int nspecies = 0;
    string tbmodelname,onsitetype("none");
    ///////////////////////////////////////
    // first, get the species information
    // Until, XML parser works on many platforms, use SimpleParser::getXwords
    ///////////////////////////////////////
    do {
      int nw=getXwords(slist,fin);
      if(nw == 0) continue;
      if(slist[0] == "species") {
	getXwords(vlist, fin, "/species");
	int iargc=0, newid;
	int ns = 0;
	while(iargc<vlist.size()) {
	  if(vlist[iargc] == "name") {
	    newid =mySpecies.addSpecies(vlist[++iargc].c_str());
	    iargc++;
	  } else if(vlist[iargc] == "mass") {
	    mySpecies(imass,newid)=atof(vlist[++iargc].c_str()); // -m
	    iargc++;
	  } else if(vlist[iargc] == "valence") {
	    mySpecies(icharge,newid)=atof(vlist[++iargc].c_str());//-q
	    iargc++;
	  } else if(vlist[iargc] == "orbital") {
	    int l = atoi(vlist[++iargc].c_str());
	    Lmax = max(l,Lmax);
	    ns += (2*l+1);
	    iargc++;
	  }
	  iargc++;
	}
	mySpecies(ins,newid)=ns;
	vlist.erase(vlist.begin(), vlist.end());
	nspecies++;
      } else if(slist[0] == "onsite") {
	onsitetype = slist[2];
	///////////////////////////////////////////
	// create onsite potential
	///////////////////////////////////////////
	curOnsite = createTBOnsite(onsitetype);
        if(curOnsite) {
          curOnsite->setParticleSet(myPtcl);
          curOnsite->put(fin); 
        }
      } 
      else if(slist[0] == "TBPARAM") {
	tofindsk = false;
	tbmodelname = slist[2];
      }
    } while(tofindsk && !fin.eof());
    
    /////////////////////////////////
    // bond information is reset, whenver new species are added
    /////////////////////////////////
    Bond.resize(nspecies, nspecies);
    Bond = -1; // no bond is connected
    
    SKModel = tbmodelname;
    curSK = createTBModel(tbmodelname);
    curSK->setParticleSet(myPtcl);

    if(curSK) {

      ShortRangePotential* rep = curSK->init(SK_h,SK_s,Bond,fin);
      ///model returns ShortRangePotential
      if(rep) pot_s.push_back(rep);
      for(int ia=0; ia<nspecies; ia++) {
	for(int ja=0; ja<nspecies; ja++) {
	  reset(ia,ja,
		static_cast<int>(mySpecies(ins,ia)),
		static_cast<int>(mySpecies(ins,ja)),
		Lmax);
	}
      }
      return true;
    } else {
      return false;
    }
  }

  TBModelBase* createTBModel(const std::string& mname) { //, std::istream& fin) {

    enum {harrison,lenosky,nrl,gsp,spline};
    map<string,int> model_map;
    model_map["lenosky"] = lenosky;
    model_map["nrl"] = nrl;
    model_map["gsp"] = gsp;
    model_map["spline"] = spline;
    model_map["invpower"] = harrison;

    map<string,int>::iterator it = model_map.find(mname);
    if(it == model_map.end()) {
      ERRORMSG(mname << " not available currently. Choose [lenosky|nrl|gsp|spline|invpower]")
      return false;
    }

    TBModelBase* tbmodel = NULL;
    switch((*it).second) {
#ifdef ENABLE_INVPOWERTB
    case(harrison):
      LOGMSG("\tUsing (r_0/r)^n model.")
      tbmodel = new InvPowerTB;
      break;
#endif
#ifdef ENABLE_LENOSKYTB
    case(lenosky):
      LOGMSG("\tUsing Lenksy model: cubic spline functions.")
      tbmodel = new LenoskyTB;
      break;
#endif
#ifdef ENABLE_NRLTB
    case(nrl):
      LOGMSG("\tUsing NRL model: consider using spline functions.")
      tbmodel = new NRLTB;
      break;
#endif
#ifdef ENABLE_SPLINETB
    case(spline):
      LOGMSG("\tUsing generic model: cubic spline functions")
      tbmodel = new SplineTB;
      break;
#endif
#ifdef ENABLE_GSPTB
    case(gsp):
      LOGMSG("\tUsing GSP model: (r_0/r)^n - (r_0/r_c)^n")
      tbmodel = new GSPTB;
    break;
#endif
    }

    return tbmodel;
  }

  TBOnsiteBase* createTBOnsite(const std::string& oname) {

    TBOnsiteBase* onsite = 0;

#if defined(ENABLE_ZBSURFACE)
    if(oname == "zbsurface") {
      onsite = new ZincBlendSurface;
    } 
#endif  
#if defined(ENABLE_NRLTB)
    if(oname == "nrl") {
      onsite = new NRLTBOnsite;
    }
#endif  
    //if(onsite)  onsite->put(fin); 
    
    return onsite;
  }
  
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 330 $   $Date: 2005-05-26 17:44:25 -0500 (Thu, 26 May 2005) $
 * $Id: SKParameters.IO.cpp 330 2005-05-26 22:44:25Z jnkim $ 
 ***************************************************************************/
