//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/** @file SlaterKosterTB.1.cpp
 * @brief Definitions of the member functions with high computational load
 */
#include "SlaterKosterTB/SlaterKosterTB.h"
#include "SlaterKosterTB/TBModelBase.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/Communicate.h"
#include "Message/OpenMP.h"
#include "Particle/NNEngine.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
using namespace OHMMS;

void 
SlaterKosterTB::setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell) {

  if(HamIndex == 0) createIndex();

  //distance table is re-created
  myPtcl->curR.setUnit(PosUnit::LatticeUnit);
  nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
  nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
  if(wrap) myPtcl->applyBConds(true,myPtcl->curR.getUnit());

  if(EvaluateDensityMatrix) {
    updateAll();
  } else {
    updateHandS();
  }
  //update the spare index and matrix
}

void 
SlaterKosterTB::setCurrentPos(ParticlePos_t& pos) {

  if(HamIndex == 0) createIndex();

  myPtcl->curR.setUnit(PosUnit::LatticeUnit);
  nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
  nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
  if(EvaluateDensityMatrix) {
    updateAll();
  } else {
    updateHandS();
  }
}

void 
SlaterKosterTB::setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& dr) {
  //Index does not change
  nnListEngine->updatePairs(*myPtcl,dr);
}

void
SlaterKosterTB::updateHandS() {

#pragma omp parallel
  {
    int ip=omp_get_thread_num();
    HamIndex->update(nnListEngine->getPairAttrib(ip),ip);
  }

  //Need to global operations
  HamIndex->checkBounds();

  if(Hmat.empty()) {
    vector<SparseMatrix_t*>::iterator hit(HamMatList.begin()),hit_end(HamMatList.end());
    while(hit != hit_end) { (*hit)->create(); ++hit;}
  }

#pragma omp parallel
  {
    //each block can resize and allocate freely
    int ip=omp_get_thread_num();
    vector<SparseMatrix_t*>::iterator hit(HamMatList.begin()),hit_end(HamMatList.end());
    while(hit_end != hit) { (*hit)->resize(ip); ++hit;}
  }
}
void
SlaterKosterTB::updateAll() {

#pragma omp parallel
  {
    int ip=omp_get_thread_num();
    HamIndex->update(nnListEngine->getPairAttrib(ip),ip);
    //DenIndex->update(nnListEngine->getPairAttrib(ip),ip);
    DenIndex->update(*(HamIndex->J[ip]), *(HamIndex->M[ip]), HamIndex->first_diag(ip),ip);
  }

  //Need to global operations
  HamIndex->checkBounds();
  DenIndex->checkBounds();

  if(Hmat.empty()) {
    vector<SparseMatrix_t*>::iterator hit(HamMatList.begin()),hit_end(HamMatList.end());
    while(hit != hit_end) { (*hit)->create(); ++hit;}

    vector<SparseMatrix_t*>::iterator dit(DenMatList.begin()),dit_end(DenMatList.end());
    while(dit != dit_end) { (*dit)->create(); ++dit;}
  }

#pragma omp parallel
  {
    //each block can resize and allocate freely
    int ip=omp_get_thread_num();

    vector<SparseMatrix_t*>::iterator hit(HamMatList.begin()),hit_end(HamMatList.end());
    while(hit_end != hit) { (*hit)->resize(ip); ++hit;}

    vector<SparseMatrix_t*>::iterator dit(DenMatList.begin()),dit_end(DenMatList.end());
    while(dit_end != dit) { (*dit)->resize(ip); ++dit;}
  }
}

SlaterKosterTB::Scalar_t SlaterKosterTB::evalEnergy() {

  //Set DensityMatrix to zero
  EvaluateDensityMatrix = false;
  TBEnergy = solveGES();  

  if(SKP.size()) {
    Scalar_t esum(0.0);
#pragma omp parallel reduction(+:esum)
    {
      int np=nnListEngine->size();
#pragma omp for nowait 
      for(int ip=0; ip<np; ip++) {
        esum += SKP.evalEnergyNN(nnListEngine->getPairAttrib(ip),1.0);
      }
    }//end-of-omp
    return nnListEngine->collect(esum)+TBEnergy;
  } else
    return TBEnergy;
}

SlaterKosterTB::Scalar_t 
SlaterKosterTB::evalForce(ParticlePos_t& f0, unsigned fmode) {

  EvaluateDensityMatrix = true;
  unsigned stressmode=(fmode&1);
  UpdateNNList = fmode&2;
  f0=0.0;
  Stress=0.0;

  TBEnergy = evalForceTB(f0,myPtcl->E,Stress,stressmode);

  Scalar_t extra(0.0);
  if(SKP.size()) {
    if(nnListEngine->UseVisitingForce) {
      extra += evalForceExtraV(f0,stressmode);
    }
    else
      extra += evalForceExtra(f0,stressmode);
  }

  return TBEnergy+extra;
}

SlaterKosterTB::Scalar_t 
SlaterKosterTB::evalForceExtraV(ParticlePos_t& f0, unsigned stressmode) {

  Scalar_t extra(0.0);

  int np=nnListEngine->size();
#pragma omp parallel reduction(+:extra)
  {
    omp_int_t ip=omp_get_thread_num();
    ParticlePos_t& fv=nnListEngine->getPairAttrib(ip).Fv;

    //Add the TB forces
    int first(nnListEngine->first_id(ip));
    int last(nnListEngine->last_id(ip));
    int iL(0);
    while(first != last){fv[iL++]=f0[first++];}
    extra += SKP.evalForceNN(nnListEngine->getPairAttrib(ip), fv,stressmode);
  }

#pragma omp parallel
  {
    nnListEngine->collectForce(f0,omp_get_thread_num());
  }

  for(int ip=0; ip<np; ip++) {
    Stress += nnListEngine->getPairAttrib(ip).Stress;
  }

  nnListEngine->endCollect(f0);
  return nnListEngine->collect(extra);
}

SlaterKosterTB::Scalar_t 
SlaterKosterTB::evalForceExtra(ParticlePos_t& f0, unsigned stressmode) {

  Scalar_t esum(0.0);

  int np=nnListEngine->size();
#pragma omp parallel reduction(+:esum)
  {
#pragma omp for nowait 
    for(int ip=0; ip<np; ip++) {
      esum += SKP.evalForceNN(nnListEngine->getPairAttrib(ip), f0,stressmode);
    }
  }

  for(int ip=0; ip<np; ip++) {
    Stress += nnListEngine->getPairAttrib(ip).Stress;
  }

  return nnListEngine->collect(esum);
}

/**set sparse Hamiltonian and Overlap matrices
 * @param setmode what to do with the diagonals, 0 == evaluate
 * @param nnlist the neighbor list of this thread
 * @param myID thread id
 *
 * Fill-in the values for the sparse matrix for each thread.
 * solveGES calls this function if a derived class uses Hmat and Smat,
 * the sparse matrix for Hamiltonian and overlap matrix.
 */
void 
SlaterKosterTB::setHamiltonian(int setmode, const PtclPairList_t& nnlist, int ip){

  Hmat.clear(setmode,ip);               

  typedef PtclPairList_t::PairData_t PairData_t;
  const ParticleIndex_t& grID(myPtcl->GroupID);
  int stride(Hmat.getBlockSize());

  //ignore setmode for now. Think how not to assign diagonal elements
  if(Identity) {
    SparseMatrix_t::pointer restrict hPtr=Hmat.diag(ip,0);
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      SKP.getSKBlocks(grID[nnlist.I[iL]], hPtr);
      hPtr+=stride;
    }
    SparseMatrix_t::pointer restrict hPtr_=Hmat.offdiag(ip,0);
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      int speciesI(grID[nnlist.I[iL]]);
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
        SKP.getSKBlocks(speciesI,cur.Group,cur.sep,cur.displ,hPtr_);
        ++nit; hPtr_+=stride;
      }
    }
  } else {
    Smat.clear(setmode,ip); 
    SparseMatrix_t::pointer restrict hPtr=Hmat.diag(ip,0);
    SparseMatrix_t::pointer restrict sPtr=Smat.diag(ip,0);
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      SKP.getSKBlocks(grID[nnlist.I[iL]], hPtr,sPtr);
      hPtr+=stride; sPtr+=stride;
    }
    SparseMatrix_t::pointer restrict hPtr_=Hmat.offdiag(ip,0);
    SparseMatrix_t::pointer restrict sPtr_=Smat.offdiag(ip,0);
    for(int iL=0; iL<nnlist.getLocalNum();iL++) {
      int speciesI(grID[nnlist.I[iL]]);
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
	SKP.getSKBlocks(speciesI,cur.Group,cur.sep,cur.displ,hPtr_,sPtr_);
        ++nit; hPtr_+=stride; sPtr_+=stride;
      }
    }
  }
  SKP.setOnsite(Hmat,nnlist);
}

/** perform traces to calculate forces with Rho and RhoS
  *@param nnlist neighbor-list data of this thread
  *@param f the forces to be evaluated
  *@param e the energies to be evaluated
  *@param stressloc local stress component to be added
  *@param stressmode 0 for not calculating stress and 1 otherwise
  *@return tight-bining (band) energy
  */
SlaterKosterTB::Scalar_t 
SlaterKosterTB::tbForceWithDM(const PtclPairList_t& nnlist,
			      ParticlePos_t& f, ParticleScalar_t& e, 
                              Tensor_t& stressloc, unsigned stressmode) const {

  omp_int_t myID = omp_get_thread_num();
  typedef PtclPairList_t::PairData_t PairData_t;

  const ParticleIndex_t& grID(myPtcl->GroupID);

  // temporary block matrices used for force
  int maxr = HamIndex->MaxRowBlock;
  int maxc = HamIndex->MaxColBlock;
  Matrix<Scalar_t> Tx(maxr,maxc), Ty(maxr,maxc), Tz(maxr,maxc);
  Matrix<Scalar_t> Sx(maxr,maxc), Sy(maxr,maxc), Sz(maxr,maxc);
  
  Scalar_t fx, fy, fz;
  Scalar_t esumloc=0.0;

  //density matrix does not have duplicate pair-indices, while hamiltonian has
  //Work array is used to mask valid mapping.
  //-2 indicates that the elements are not connected
  vector<int> Work(HamIndex->getTotNum(),-2);
  
  int nnd(0); //counter of non-zero blocks of the density matrix
  int nnh(0); //counter of non-zero blocks of the hamiltonian matrix 
  bool ablock;
  for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
    int iat(nnlist.I[iL]);  
    int itype(grID[iat]);
    Work[iat] = -1; //set the diagonal term as -1
    for(int nni=0; nni< DenIndex->nadj(iL, myID); nni++, nnd++){
      Work[DenIndex->iadj(iL,nni,myID)] = nnd; //set the off-diagoal locator
    }

    int nbi(HamIndex->row_block(iat));
    const Scalar_t* trho = Rho.diag(myID,iL);
    e[iat] += OhmmsBlas::dot(nbi*nbi,trho,Hmat.diag(myID,iL));

    PtclPairList_t::const_iterator nit(nnlist.begin(iL));
    int num_nn(nnlist.nadj(iL));
    for(int nni=0; nni<num_nn; nni++, nnh++) {//scan non-zero H 
      const PairData_t& cur(*nit++);
      int ij(SKP.pair_id(itype,cur.Group));
      if(ij<0) continue;
      Tx = 0.0;  Ty = 0.0; Tz = 0.0;
      if(Identity)
        ablock=SKP.getSKBlocks(ij,cur.sep,cur.displ,Tx.data(),Ty.data(),Tz.data());
      else {
        Sx = 0.0;  Sy = 0.0; Sz = 0.0;
        ablock=SKP.getSKBlocks(ij,cur.sep,cur.displ,
            Tx.data(),Ty.data(),Tz.data(),Sx.data(),Sy.data(),Sz.data());
      }
      if(!ablock) continue; //zero block, skip the rest

      int jat(cur.Next);
      int dloc(Work[jat]);

      if(dloc < -1) continue; //not connected
      if(dloc == -1) {
        trho = Rho.diag(myID,iL);
      } else if(dloc >= 0) {
        trho = Rho.offdiag(myID,dloc);
      }
      
      int nbtot(nbi*HamIndex->row_block(jat));
      fx = OhmmsBlas::dot(nbtot,trho,Tx.data());
      fy = OhmmsBlas::dot(nbtot,trho,Ty.data());
      fz = OhmmsBlas::dot(nbtot,trho,Tz.data());
      e[iat] += OhmmsBlas::dot(nbtot,trho,Hmat.offdiag(myID,nnh));

      if(!Identity) {
        if(dloc == -1) {
          trho = RhoS.diag(myID,iL);
        } else if(dloc >= 0) {
          trho = RhoS.offdiag(myID,dloc);
        }
        fx -= OhmmsBlas::dot(nbtot,trho,Sx.data());
        fy -= OhmmsBlas::dot(nbtot,trho,Sy.data());
        fz -= OhmmsBlas::dot(nbtot,trho,Sz.data());
      }
      ////////////////////////////////////////////////////////////////////
      //WARNING: SlaterKoster returns -(derivative), dR[nn]=R[jat]-R[iat]
      ////////////////////////////////////////////////////////////////////
      SingleParticlePos_t df= SingleParticlePos_t(fx,fy,fz);
      f[iat] += 2.0*df;
      
      //////////////////////////////////////////////
      //For OpenMP performance, use non-symmetric form.
      //f0[i] += df; f0[j] -= df;
      //T esum = 0.5*Trace(trho, H->offdiag(myID,iat), nbi, nbj);
      //q[i] += esum;
      //q[j] += esum;
      //////////////////////////////////////////////
      
      if(stressmode) { 
        stressloc -= outerProduct(df,cur.displ);
      }
    }//nni
    std::fill(Work.begin(),Work.end(),-2);
  } // iL
  return 0;
}

/// incomplete
bool SlaterKosterTB::evalDynamicMatrix(DynMatrix_t& dynmat) {
  /*
  int setmode = 0;
  setHamiltonian(setmode);
  tbDynamicMatrix(dynmat);
  //  for(int i=0; i<Pot_s.size(); i++)  Pot_s[i]->evalDynamicMatrix(dynmat);
  */
  return false; 
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: SlaterKosterTB.1.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/
