//////////////////////////////////////////////////////////////////
//   (c) Copyright 1998-2002, 2003- Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "SlaterKosterTB/SlaterKosterTBKPT.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/Communicate.h"
#include "Message/OpenMP.h"
#include "Particle/NNEngine.h"
#include "Particle/ParticleOperators.h"
#include "SparseMatrix/SparseMatrixOps.h"
#include "SparseMatrix/SparseIndexOps.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "OhmmsNumeric/SparseEigenSolver.h"
#include "OhmmsNumeric/SparseEigenSolverBuilder.h"
#include "ParticleBase/ParticleMessage.h"
#include "QuantumBase/ElectronStructureIO.h"
#include "SlaterKosterTB/TBModelBase.h"
using namespace OHMMS;
SlaterKosterTBKPT::SlaterKosterTBKPT(): 
  thisSolver(0), Kgid(0) 
{ 
  if(omp_get_max_threads()>1) {
    ERRORMSG("\tSlaterKosterTBKPT::tbForce does not work with OpenMP.")
    ERRORMSG("Set the OMP_NUM_THREADS=1")
  }
  setName("SKTB_kpt");
  myRecorder = new ESOutput(*this);
}

SlaterKosterTBKPT::~SlaterKosterTBKPT(){ 

  if(thisSolver) delete thisSolver;
}


void 
SlaterKosterTBKPT::setEigenSolver() {

  DenMatList.push_back(&RhoR);
  DenMatList.push_back(&RhoI);
  if(!Identity){ 
    DenMatList.push_back(&HdR);
    DenMatList.push_back(&HdI);
  }

  //create a solver
  thisSolver = createHerSolver(States);
}


/** Calculate the ground state energy
 *
 * Sparse indices are handled by setCurrentPos operators.
 */
SlaterKosterTBKPT::Scalar_t 
SlaterKosterTBKPT::solveGES(){

#pragma omp parallel
  {
    int np=Hmat.getIndex()->groups();
#pragma omp for nowait
    for(int ip=0; ip<np; ip++) {
      if(UpdateNNList) 
        nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(ip));

      int setmode=0; //clear everything
      setHamiltonian(setmode, myPtcl->getNNEngine()->getPairAttrib(ip), ip);
    }
  }

  if(resetKPoint(myPtcl)) {
    int nproc = Controller->ncontexts();
    Kgid  = Controller->mycontext();
    nK.resize(nproc);
    dnK.resize(nproc);
    KSet->partition(nproc);
    for(int ip=0; ip<nproc; ip++) {
      nK[ip] = (KSet->end(ip)- KSet->begin(ip))*NumStates;
    }
    dnK[0] = 0;
    for(int ip=0; ip<nproc-1; ip++) {
      dnK[ip+1] = dnK[ip]+nK[ip];
    }
  }
  Kstart = KSet->begin(Kgid);    
  Kend = KSet->end(Kgid);    

  resetPhase(myPtcl, Phi, Kstart, Kend);
  if(EvaluateDensityMatrix) {
    if(Identity) 
      thisSolver->getEigvectors(Hmat,Phi,E,Kstart,Kend);
    else
      thisSolver->getEigvectors(Hmat,Smat,Phi,E,Kstart,Kend);
  } else {

    if(Identity) 
      thisSolver->getEigvalues(Hmat,Phi,E,Kstart,Kend);
    else
      thisSolver->getEigvalues(Hmat,Smat,Phi,E,Kstart,Kend);
  }

#ifdef HAVE_MPI
  ///////////////////////////////////
  //Collect all the energies calculated by parallel nodes
  //Once each node has complete eigen energy information, 
  //determine i)occupation numbers and ii) fermi levels
  //Redundant parallel operations to minimize communication
  ///////////////////////////////////
  if(KSet->distributed()) {
    //\todo partition among mpi nodes to be taken care of
    MPI_Allgatherv(&(E[Kstart][0]), nK[Kgid], MPI_DOUBLE, &(E[0][0]), 
		   &(nK[0]),&(dnK[0]), MPI_DOUBLE,MPI_COMM_WORLD);
  }
#endif

  return getBandEnergy();

}

/** @warning SlaterKosterTBKPT NOT parallelized.  */
SlaterKosterTBKPT::Scalar_t 
SlaterKosterTBKPT::evalForceTB(ParticlePos_t& f0, ParticleScalar_t& q,
			       Tensor_t& stress, unsigned stressmode) {

  typedef PtclPairList_t::PairData_t PairData_t;

  EvaluateDensityMatrix = true;
  Scalar_t tbe = solveGES();
  Rho.clear(0); // total density matrix

  vector<Scalar_t> w(Homo);
  int mpi_id = Controller->mycontext();
  const ParticleIndex_t& grID(myPtcl->GroupID);

  Scalar_t sigmainv=0.0,fac=0.0;
  bool correct = !(ZeroTemperature || fabs(dSumOcc) <1e-16);
  if(correct) {
    sigmainv = 1.0/Sigma;
    fac = dSumEnergy/dSumOcc;
  }

  for(int ik=Kstart, nk=0; ik < Kend; ik++,nk++) {
    RhoR.clear(0);
    RhoI.clear(0);
    Scalar_t wk = KSet->W[ik];
    if(correct) {
      for(int is=0; is<Homo; is++) {
	w[is] = SpinDeg*(Occ[ik][is]+(E[ik][is]*sigmainv-fac)*dOcc[ik][is])*wk;
      }
    } else {
      for(int is=0; is<Homo; is++)  w[is] = SpinDeg*Occ[ik][is]*wk;
    }
    thisSolver->product(nk,nk,w,RhoR, RhoI);

    if(!Identity) {
      HdR.clear(0);
      HdI.clear(0);
      for(int is=0; is<w.size(); is++) w[is] *= E[ik][is];
      thisSolver->product(nk,nk,w,HdR,HdI);
    }

    NNEngine *nnListEngine = myPtcl->getNNEngine();
    for(int myID=0; myID<nnListEngine->size(); myID++) {

      //resize the temporary block matrices
      int maxr(HamIndex->MaxColBlock);
      int maxc(HamIndex->MaxColBlock);
      Matrix<Scalar_t> Tx(maxr,maxc), Ty(maxr,maxc), Tz(maxr,maxc);
      Matrix<Scalar_t> Sx(maxr,maxc), Sy(maxr,maxc), Sz(maxr,maxc);

      const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);
      vector<int> Work(HamIndex->getTotNum(),-2);

      Scalar_t pvec[8], pd[8][3];
      Scalar_t psumloc = 0.0;
      Tensor_t stressloc;

      Scalar_t fx, fy, fz;
      Scalar_t *trhor, *trhoi;
      int nntot = HamIndex->first_offdiag(myID);
      int nnh(0);
      int nnd(0);
      bool ablock;
      for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
        int iat(nnlist.I[iL]);
        int itype(grID[iat]);
	Work[iat] = -1;
	for(int nni=0; nni< DenIndex->nadj(iL, myID); nni++, nnd++){
	  Work[DenIndex->iadj(iL,nni,myID)] = nnd;
	}

	int nbtot(HamIndex->row_block(iat)*HamIndex->col_block(iat));
        int num_nn(nnlist.nadj(iL));

	//accumulate the total density matrix
	OhmmsBlas::axpy(nbtot,RhoR.diag(myID,iL),Rho.diag(myID,iL));
        PtclPairList_t::const_iterator nit(nnlist.begin(iL));
        for(int nni=0; nni<num_nn; nni++, nnh++, nntot++) {
          const PairData_t& cur(*nit++);
	  int ij(SKP.pair_id(itype,cur.Group));
	  if(ij<0) continue;
	  Tx = 0.0;  Ty = 0.0; Tz = 0.0;
	  if(Identity) {
	    ablock = SKP.getSKBlocks(ij,cur.sep,cur.displ,
	        		     Tx.data(), Ty.data(), Tz.data());
          } else {
	    Sx = 0.0;  Sy = 0.0; Sz = 0.0;
	    ablock = SKP.getSKBlocks(ij,cur.sep,cur.displ,
	        		     Tx.data(), Ty.data(), Tz.data(),
	        		     Sx.data(), Sy.data(), Sz.data());
	  }
	  if(!ablock) continue;
	  int jat(cur.Next);
	  int dloc = Work[jat];
	  if(dloc < -1) continue;

	  if(dloc == -1) {
	    trhor = RhoR.diag(myID,iL);
	    trhoi = RhoI.diag(myID,iL);
	  } else if(dloc >= 0) {
	    trhor = RhoR.offdiag(myID,dloc);
	    trhoi = RhoI.offdiag(myID,dloc);
	  }

	  nbtot = HamIndex->row_block(jat)*HamIndex->col_block(jat);
	  Scalar_t phir(Phi[nk][nntot].real());
	  Scalar_t phii(Phi[nk][nntot].imag());

	  fx = phir*OhmmsBlas::dot(nbtot,trhor,Tx.data())
	    +  phii*OhmmsBlas::dot(nbtot,trhoi,Tx.data());
	  fy = phir*OhmmsBlas::dot(nbtot,trhor,Ty.data())
	    +  phii*OhmmsBlas::dot(nbtot,trhoi,Ty.data());
	  fz = phir*OhmmsBlas::dot(nbtot,trhor,Tz.data())
	    +  phii*OhmmsBlas::dot(nbtot,trhoi,Tz.data());


	  q[iat] +=  phir*OhmmsBlas::dot(nbtot,trhor,Hmat.offdiag(myID,nnh))
	    +       phii*OhmmsBlas::dot(nbtot,trhoi,Hmat.offdiag(myID,nnh));
	  if(!Identity) {
	    if(dloc == -1) {
	      trhor = HdR.diag(myID,iL);
	      trhoi = HdI.diag(myID,iL);
	    } else if(dloc >= 0) {
	      trhor = HdR.offdiag(myID,dloc);
	      trhoi = HdI.offdiag(myID,dloc);
	    }
	    fx -= phir*OhmmsBlas::dot(nbtot,trhor,Sx.data())
	      +   phii*OhmmsBlas::dot(nbtot,trhoi,Sx.data());
	    fy -= phir*OhmmsBlas::dot(nbtot,trhor,Sy.data())
	      +   phii*OhmmsBlas::dot(nbtot,trhoi,Sy.data());
	    fz -= phir*OhmmsBlas::dot(nbtot,trhor,Sz.data())
	      +   phii*OhmmsBlas::dot(nbtot,trhoi,Sz.data());
	  }

	
	  ////////////////////////////////////////////////////////////////// 
	  // WARNING: SK returns  -(derivative),  dR[nn] = R[jat]-R[iat]
	  ////////////////////////////////////////////////////////////////// 
	  SingleParticlePos_t df= SingleParticlePos_t(fx,fy,fz);
	  f0[iat] += 2.0e0*df;
	  //f0[jat] -= df;
	  if(stressmode) { 
	    stressloc -= cur.sep*outerProduct(df,cur.displ);
	  }
	}//nni
        std::fill(Work.begin(),Work.end(),-2);
      }// iL
    
      if(stressmode) {
        stress += stressloc;
      }
    }// end-of-omp
  }// k-sum
  
  NNEngine *nnListEngine = myPtcl->getNNEngine();
  for(int myID=0; myID<nnListEngine->size(); myID++) {
    const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);
    const SparseIndex_t& dIndex = *(Rho.getIndex());
    for(int iL=0; iL<dIndex.num_diag(myID); iL++) {
      int iG = nnlist.I[iL];// get the physical index to store forces
      int nbtot = dIndex.row_block(iG)*dIndex.col_block(iG);
      q[iG] +=  OhmmsBlas::dot(nbtot,Rho.diag(myID,iL),Hmat.diag(myID,iL));
    }
  }

  //\todo: will encapsulate it better: collect(f0); collect(q);
  if(KSet->distributed()) {
    gsum(f0,Kgid);
    gsum(q,Kgid);
  }
  return tbe;
}
void SlaterKosterTBKPT::tbDynamicMatrix(DynMatrix_t& ) {
}

SlaterKosterTBKPT::Scalar_t
SlaterKosterTBKPT::el_ph_coupling(const vector<int>& psi_i, 
                                  const PhononMode& phonon,
				  vector<int>& psi_f, Scalar_t del){
  return 0.0;
}

//void 
//SlaterKosterTBKPT::setParticleSet(Particle_t* ptclin) {
//  //here goes all the allocation
//  myPtcl = ptclin;
//
//  if(NeverInitialized) {
//    //Set the QuantumHamiltonian::Parameters
//    setESParameters(myPtcl);
//
//    //Create storage
//    vector<SparseMatrix_t*> dMats;
//    dMats.push_back(&RhoR);
//    dMats.push_back(&RhoI);
//    if(!Identity){ 
//      dMats.push_back(&HdR);
//      dMats.push_back(&HdI);
//    }
//    manageStorage(dMats);
//
//    //create a solver
//    thisSolver = createHerSolver(States);
//
//    //prevent doing the same thing
//    NeverInitialized=false;
//  }
//
//  if(Onsite) Onsite->setParticleSet(myPtcl);
//}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: SlaterKosterTBKPT.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/
