//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Potentials/ShortRangePotential.h"
#include "SlaterKosterTB/TBModelFactory.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"
#include "Utilities/OhmmsInfo.h"
#define ENABLE_INVPOWERTB
#define ENABLE_LENOSKYTB
#define ENABLE_NRLTB
#define ENABLE_SPLINETB
#define ENABLE_GSPTB
#define ENABLE_ZBSURFACE

#if defined(ENABLE_INVPOWERTB)
#include "SlaterKosterTB/Models/InvPowerTB.h"
#endif
#if defined(ENABLE_LENOSKYTB)
#include "SlaterKosterTB/Models/LenoskyTB.h"
#endif
#if defined(ENABLE_NRLTB)
#include "SlaterKosterTB/Models/NRLTB.h"
#endif
#if defined(ENABLE_SPLINETB)
#include "SlaterKosterTB/Models/SplineTB.h"
#endif
#if defined(ENABLE_GSPTB)
#include "SlaterKosterTB/Models/GSPTB.h"
#endif
#if defined(ENABLE_ZBSURFACE)
#include "SlaterKosterTB/Models/ZincBlendSurface.h"
#endif
#include "SlaterKosterTB/SKParameters.h"

namespace OHMMS {

  vector<TBModelBase*> TBModelFactory::TB;

  bool
  TBModelFactory::init(const string& fname,
		       SKParameters<Scalar_t>& skp,
		       TBOnsiteBase* Onsite,
		       vector<ShortRangePotential*>& pot_s) 
  {

    ifstream fin(fname.c_str());
    if(!fin) {
      ERRORMSG("TB parameters are missing. File " << fname << " does not exist.")
      return false;
    }
    
    // make private references to Species
    SpeciesBase* mySpecies = SpeciesCollection::getSpecies();

    // make private references to Species
    int icharge = SpeciesCollection::addAttrib("q");
    int ins     = SpeciesCollection::addAttrib("ns");
    int imass   = SpeciesCollection::addAttrib("mass");
    int irmax   = SpeciesCollection::addAttrib("rmax");
    
    int Lmax = 0;
    vector<string> slist, vlist;
    bool tofindsk = true;
    int nspecies = 0;
    string onsitetype("none");
    ///////////////////////////////////////
    // first, get the species information
    // Until, XML parser works on many platforms, use SimpleParser::getXwords
    ///////////////////////////////////////
    do {
      int nw=getXwords(slist,fin);
      if(nw == 0) continue;
      if(slist[0] == "species") {
	getXwords(vlist, fin, "/species");
	int iargc=0, newid;
	int ns = 0;
	while(iargc<vlist.size()) {
	  if(vlist[iargc] == "name") {
	    newid =SpeciesCollection::addSpecies(vlist[++iargc].c_str());
	    iargc++;
	  } else if(vlist[iargc] == "mass") {
	    (*mySpecies)(imass,newid)=atof(vlist[++iargc].c_str()); // -m
	    iargc++;
	  } else if(vlist[iargc] == "valence") {
	    (*mySpecies)(icharge,newid)=atof(vlist[++iargc].c_str());//-q
	    iargc++;
	  } else if(vlist[iargc] == "orbital") {
	    int l = atoi(vlist[++iargc].c_str());
	    Lmax = max(l,Lmax);
	    ns += (2*l+1);
	    iargc++;
	  }
	  iargc++;
	}
	(*mySpecies)(ins,newid)=ns;
	vlist.erase(vlist.begin(), vlist.end());
	nspecies++;
      } else if(slist[0] == "onsite") {
	onsitetype = slist[2];
	///////////////////////////////////////////
	// create onsite potential
	///////////////////////////////////////////
	Onsite = TBOnsiteFactory::init(onsitetype,fin);
      } 
      else if(slist[0] == "TBPARAM") {
	tofindsk = false;
	skp.SKModel = slist[2];
	//tbmodelname = slist[2];
      }
    } while(tofindsk && !fin.eof());
    
    /////////////////////////////////
    // bond information is reset, whenver new species are added
    /////////////////////////////////
    skp.Bond.resize(nspecies, nspecies);
    skp.Bond = -1; // no bond is connected
    
    ///////////////////////////////////////////////////
    // model-specific initialization of Slater-Koster blocks 
    // for the Hamiltonian and Overlap Matrix
    // Returns a repulsive potential if a model has one
    ///////////////////////////////////////////////////
    //     ShortRangePotential* rep =
    //       TBModelFactory::init(tbmodelname, 
    // 			   skp.SK_h, skp.SK_s,skp.Bond, 
    // 			   fin);
    //if(rep) pot_s.push_back(rep);

    bool success = TBModelFactory::init(tbmodelname,skp, pot_s, fin);

    if(sucess) {
      for(int ia=0; ia<nspecies; ia++) {
	for(int ja=0; ja<nspecies; ja++) {
	  skp.reset(ia,ja,
		    static_cast<int>((*mySpecies)(ins,ia)),
		    static_cast<int>((*mySpecies)(ins,ja)),
		    Lmax);
	}
      }
    }
    return success;
  }


  bool TBModelFactory::init(SKParameters<Scalar_t>& skp,
			    std::vector<ShortRangePotental*>& pot_s,
			    istream& fin) {

    enum {harrison,lenosky,nrl,gsp,spline};
    map<string,int> model_map;
    model_map["lenosky"] = lenosky;
    model_map["nrl"] = nrl;
    model_map["gsp"] = gsp;
    model_map["spline"] = spline;
    model_map["invpower"] = harrison;

    map<string,int>::iterator it = model_map.find(skp.SKModel);
    if(it == model_map.end()) {
      ERRORMSG(skp.SKModel << " not available currently. Choose [lenosky|nrl|gsp|spline|invpower]")
      return false;
    }

    switch((*it).second) {
#ifdef ENABLE_INVPOWERTB
    case(harrison):
      LOGMSG("Using (r_0/r)^n model.")
      tbmodel = new InvPowerTB;
      break;
#endif
#ifdef ENABLE_LENOSKYTB
    case(lenosky):
      LOGMSG("Using Lenksy model: cubic spline functions.")
      tbmodel = new LenoskyTB;
      break;
#endif
#ifdef ENABLE_NRLTB
    case(nrl):
      LOGMSG("Using NRL model: consider using spline functions.")
      tbmodel = new NRLTB;
      break;
#endif
#ifdef ENABLE_SPLINETB
    case(spline):
      LOGMSG("Using generic model: cubic spline functions")
      tbmodel = new SplineTB;
      break;
#endif
#ifdef ENABLE_GSPTB
    case(gsp):
      LOGMSG("Using GSP model: (r_0/r)^n - (r_0/r_c)^n")
      tbmodel = new GSPTB;
    break;
#endif
    }

    if(tbmodel) {
      tbmodel->init(skp,pot_s,fin);
      skp->curSK = tbmodel;
      return true;
    } else {
      ERRORMSG(skp.SKModel << " is not enabled. Check src/SlaterKosterTB/CMakeLists.txt or src/SlaterKosterTB/object.pl")
      return false;
    }
  }

  ShortRangePotential*
  TBModelFactory::init(const string& mname, 
		       SKArray_t& SK_h,  SKArray_t& SK_s, 
		       Matrix<int>& Bond, istream& fin) {
    
    enum {harrison,lenosky,nrl,gsp,spline};
    int tbmode = -1;
    int id = TB.size();
    
    // connect mname with enumeration
    if(mname == "lenosky") {
      tbmode = lenosky;
    } else if(mname == "nrl") {
      tbmode = nrl;
    } else if(mname == "gsp") {
    tbmode = gsp;
    } else if(mname == "spline") {
      tbmode = spline;
    } else if(mname == "invpower") {
      tbmode = harrison;
    }
    
  // creat a model
  switch(tbmode) {
#ifdef ENABLE_INVPOWERTB
  case(harrison):
    LOGMSG("Using (r_0/r)^n model.")
    TB.push_back(new InvPowerTB);
    break;
#endif
#ifdef ENABLE_LENOSKYTB
  case(lenosky):
    LOGMSG("Using Lenksy model: cubic spline functions.")
    TB.push_back(new LenoskyTB);
    break;
#endif
#ifdef ENABLE_NRLTB
  case(nrl):
    LOGMSG("Using NRL model: consider using spline functions.")
    TB.push_back(new NRLTB);
    break;
#endif
#ifdef ENABLE_SPLINETB
  case(spline):
    LOGMSG("Using generic model: cubic spline functions")
    TB.push_back(new SplineTB);
    break;
#endif
#ifdef ENABLE_GSPTB
  case(gsp):
    LOGMSG("Using GSP model: (r_0/r)^n - (r_0/r_c)^n")
    TB.push_back(new GSPTB);
    break;
#endif
  //defaults:
  //  break;
  }

  if(id == TB.size()) {
    //No TB is added. Return NULL
    ERRORMSG(mname << " is not available. Do nothing.")
    return NULL;
  } else {
    return TB[id]->init(SK_h, SK_s, Bond, fin);
  }
}

/////////////////////////////////////////////////////////
// returns an onsite term
/////////////////////////////////////////////////////////
TBOnsiteBase* TBOnsiteFactory::init(const string& oname, istream& fin) {

  TBOnsiteBase* onsite = NULL;
#if defined(ENABLE_ZBSURFACE)
  if(oname == "zbsurface") {
    onsite = new ZincBlendSurface;
  } 
#endif  
#if defined(ENABLE_NRLTB)
  if(oname == "nrl") {
    onsite = new NRLTBOnsite;
  }
#endif  
  if(onsite)  onsite->init(fin); 

  return onsite;
}

}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: TBModelFactory.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
