//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseMatrix.h"
#include "SparseMatrix/Matrix2SparseMatrixProduct.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
//BlockMatrixRef is not used due to bad performance
#include "SparseMatrix/BlockMatrixRef.h"
#include "Utilities/Clock.h"

void 
OHMMS::Dense2SparseMatrix::product(real_type alpha, 
	                           const RealMatrix_t& bra,  
                                   const RealMatrix_t& ket, 
	                           SparseMatrix_t& mat) {

#pragma omp parallel
   {
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = mat.getIndex();

    int iG=myIndex->first_diag(myID);
    std::size_t             stride(mat.getBlockSize());
    cout << "***** Size of stride = " << stride << endl;
    
    SparseMatrix_t::pointer dPtr=mat.diag(myID,0);
    SparseMatrix_t::pointer oPtr=mat.offdiag(myID,0);
    for(int iL=0; iL < myIndex->num_diag(myID); iL++,iG++){    

      int ioff = myIndex->col_loc(iG);
      int nbi = myIndex->col_block(iG);

      //Product<real_type,TB_BLOCK>::make(mat.diag(myID,iL),nbi,nbi, bra,ioff,ket,ioff,alpha);
      Product<real_type,TB_BLOCK>::make(dPtr,nbi,nbi, bra,ioff,ket,ioff,alpha);
      dPtr+=stride; 

      //overhead seems to be too great, use template functions
      //BlockMatrixRef<Scalar_t> diag_block(mat.diag(myID,iL),nbi,nbi);
      //diag_block.multadd(alpha,bra,ioff,ket,ioff);

      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++, oPtr+=stride) {
        int jG = myIndex->iadj(iL, nni, myID);
        if(iG == jG) continue;
	int nbj = myIndex->col_block(jG);
	//Product<real_type,TB_BLOCK>::make(mat.offdiag(myID,nn),nbi,nbj,bra,ioff,ket,myIndex->col_loc(jG),alpha);
	Product<real_type,TB_BLOCK>::make(oPtr,nbi,nbj,bra,ioff,ket,myIndex->col_loc(jG),alpha);
	//overhead seems to be too great, use template functions
	//BlockMatrixRef<Scalar_t> offdiag_block(mat.offdiag(myID,nn),nbi,nbj);
	//offdiag_block.multadd(alpha,bra,ioff,ket,myIndex->col_loc(jG));
      }// pair-list
    }
  }// closing pramga

}

void 
OHMMS::Dense2SparseMatrix::product(const RealMatrix_t& bra, 
                                   const RealMatrix_t& ket, 
		                   SparseMatrix_t& mat) {
  product(1.0,bra,ket,mat);
}

void 
OHMMS::Dense2SparseMatrix::product(const RealMatrix_t& bra, 
                                   const RealMatrix_t& ket, 
		                   const vector<real_type>& w, 
                                   SparseMatrix_t& mat) {

#pragma omp parallel
   {
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = mat.getIndex();
    int ncol = bra.cols();

    int iG=myIndex->first_diag(myID);
    int stride(mat.getBlockSize());

    SparseMatrix_t::pointer restrict dPtr=mat.diag(myID,0);
    SparseMatrix_t::pointer restrict oPtr=mat.offdiag(myID,0);

    for(int iL=0; iL < myIndex->num_diag(myID); iL++,iG++){    

      int nbi(myIndex->col_block(iG));
      int ioff(myIndex->col_loc(iG));

      Product<real_type,TB_BLOCK>::make(dPtr,nbi,nbi,bra,ioff,ket,ioff,w);
      dPtr+=stride;

      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++, oPtr+=stride) {
        int jG(myIndex->iadj(iL, nni, myID));
        if(iG == jG) continue;
	int nbj(myIndex->col_block(jG));
	int joff(myIndex->col_loc(jG));
	Product<real_type,TB_BLOCK>::make(oPtr,nbi,nbj, bra,ioff,ket, joff,w);
      }// pair-list
    }
  }// closing pramga

}

void 
OHMMS::Dense2SparseMatrix::product(const RealMatrix_t& bra, 
                                   const RealMatrix_t& ket, 
		                   const vector<real_type>& w, 
		                   SparseMatrix_t& rho_r, 
                                   SparseMatrix_t& rho_i) {

  product(bra,ket,w,rho_r);
}


void 
OHMMS::Dense2SparseMatrix::product(const ComplexMatrix_t& bra, 
                                   const ComplexMatrix_t& ket, 
		                   const vector<real_type>& w, 
                                   SparseMatrix_t& rho) {

}

void 
OHMMS::Dense2SparseMatrix::product(const ComplexMatrix_t& bra, 
                                   const ComplexMatrix_t& ket, 
		                   const vector<real_type>& w, 
		                   SparseMatrix_t& mat_r, 
                                   SparseMatrix_t& mat_i) {

#pragma omp parallel
  {
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = mat_r.getIndex();

    int iG=myIndex->first_diag(myID);
    int nn = 0;
    for(int iL=0; iL < myIndex->num_diag(myID); iL++,iG++){    

      int ioff = myIndex->col_loc(iG);
      int nbi = myIndex->col_block(iG);

      Product<real_type,TB_BLOCK>::make(mat_r.diag(myID,iL),mat_i.diag(myID,iL),
				        nbi,nbi, bra,ioff,ket,ioff,w);
//        BlockMatrixRef<complex<Scalar_t> > 
//  	diag_block(mat_r.diag(myID,iL), mat_i.diag(myID,iL), nbi,nbi);
//        diag_block.multadd(bra,ioff,ket,ioff,w);

      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
        if(iG == jG) continue;
	int nbj = myIndex->col_block(jG);
	Product<real_type,TB_BLOCK>
	  ::make(mat_r.offdiag(myID,nn),mat_i.offdiag(myID,nn),
		 nbi,nbj, bra,ioff,ket,myIndex->col_loc(jG),w);

//  	BlockMatrixRef<complex<Scalar_t> > 
//  	  offdiag_block(mat_r.offdiag(myID,nn), mat_i.offdiag(myID,nn), nbi,nbj);
//  	offdiag_block.multadd(bra,ioff,ket,myIndex->col_loc(jG),w);
      }// pairs
    }
  }// omp
}

void 
OHMMS::Dense2SparseMatrix::product(const SparseMatrix_t& h, 
                                   const RealMatrix_t& psi,
		                   RealMatrix_t& hpsi) {

#ifdef OHMMS_MTL
  set_value(hpsi,0.0);
#else
  //hpsi.clear();
  hpsi = 0.0;
#endif

#pragma omp parallel
  {
    omp_int_t myID = omp_get_thread_num();
//  #pragma omp parallel for
//    for(int myID=0; myID<h.getIndex()->NumThreads; myID++) {
    const SparseIndex_t* myIndex = h.getIndex();

    int iG = myIndex->first_diag(myID);
    int first = myIndex->col_loc(iG);
    int last = myIndex->col_loc(myIndex->last_diag(myID));

    int nn = 0;
    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    

      int nbi = myIndex->col_block(iG);
      //////////////////////////////////////////
      // \todo check if the performance is a problem
      //////////////////////////////////////////
      ConstBlockMatrixRef<real_type> diag_block(h.diag(myID,iL), nbi,nbi);
      diag_block.multadd(psi,first,hpsi,first);

      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
	int nbj = myIndex->col_block(jG);

	ConstBlockMatrixRef<real_type> 
	  offdiag_block(h.offdiag(myID,nn), nbi,nbj);
	offdiag_block.multadd(psi,myIndex->col_loc(jG), hpsi, first);
      }
      first +=nbi;
    }
  }// end-of-omp
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: Matrix2SparseMatrixProduct.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
