//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseMatrix.h"
#include "OhmmsNumeric/Matrix.h"
#include "SparseMatrix/SparseMatrix2MatrixCopy.h"
#include "Utilities/Clock.h"
using namespace OHMMS;

#ifndef OHMMS_MTL
template<class Mat_t, class T>
inline void set_value(Mat_t& M, T a) {
  std::fill(M.data(), M.data()+M.size(), a);
}
#endif

///////////////////////////
// special copy function for sparse matrix -> symmetric/hermitan matrix 
// note that only upper triangle is set up
///////////////////////////
template<class SMat_t, class Mat_t>
void copy_upper(const SMat_t& h, Mat_t& mout) {

  set_value(mout,0.0);

#pragma omp parallel
  {
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = h.getIndex();
    typedef typename SMat_t::const_pointer const_pointer;

    int iG(myIndex->first_diag(myID));
    int stride(h.getBlockSize());
    int nloc(myIndex->num_diag(myID));
    const_pointer dPtr=h.diag(myID,0);     
    const_pointer oPtr=h.offdiag(myID,0);     
    for(int iL=0; iL<nloc; iL++,iG++){    
      int nbi(myIndex->row_block(iG));
      int ioff(myIndex->row_loc(iG));
      const_pointer dptr=dPtr;
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
        for(int kp=0,jj=ioff; kp<nbi; kp++, jj++) {
          if(kp>=k) { mout(ii,jj)+=*dptr;}
          ++dptr;
        }
      }
      dPtr+=stride;

      int num_nn(myIndex->nadj(iL,myID));
      for(int nni=0; nni < num_nn; nni++,oPtr+=stride ) {
	int jG = myIndex->iadj(iL, nni, myID);
	if(jG < iG) continue;
	int nbj(myIndex->col_block(jG));
        int joff(myIndex->col_loc(jG));
        const_pointer optr=oPtr;
        for(int k=0,ii=ioff; k<nbi; k++,ii++) {
          for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
            if(jj>=ii) mout(ii,jj)+=*optr;
            ++optr;
          }//kp
        }//k
      }//nni

    }// iL
  }
  //myclock.stop();
}

template<class SMat_t, class Mat_t>
void copy_all(const SMat_t& h, Mat_t& mout) {

  set_value(mout,0.0);
#pragma omp parallel
  {
    typedef typename SMat_t::const_pointer const_pointer;

    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = h.getIndex();

    int iG(myIndex->first_diag(myID));
    int stride(h.getBlockSize());
    int nloc(myIndex->num_diag(myID));
    const_pointer dPtr=h.diag(myID,0);     
    const_pointer oPtr=h.offdiag(myID,0);     

    for(int iL=0; iL<nloc; iL++,iG++){    
      int nbi = myIndex->row_block(iG);
      int ioff = myIndex->row_loc(iG);

      const_pointer dptr=dPtr;
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
        for(int kp=0,jj=ioff; kp<nbi; kp++, jj++) {
          mout(ii,jj)+=*dptr++;
        }
      }
      dPtr+=stride;

      int num_nn(myIndex->nadj(iL,myID));
      for(int nni=0; nni < num_nn; nni++,oPtr+=stride ) {
	int jG = myIndex->iadj(iL, nni, myID);
	int nbj(myIndex->col_block(jG));
        int joff(myIndex->col_loc(jG));
        const_pointer optr=oPtr;
        for(int k=0,ii=ioff; k<nbi; k++,ii++) {
          for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
            mout(ii,jj)+=*optr++;
          }//kp
        }//k
      }//nni

    }// iL
  }
}

//////////////////////////////////////////////////////
// sparse matrix -> real matrix
//////////////////////////////////////////////////////
/*!\fn void OHMMS::copy(const SparseMatrix_t& h, PackedSymMatrix_t>&  mout)
 *\brief sparse matrix -> packed symmetric matrix
 */
void 
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                PackedSymMatrix_t&  mout) {
  copy_upper(h,mout);
}


/*!\fn void OHMMS::copy(const SparseMatrix_t& h, PackedSymMatrix_t>&  mout)
 *\brief sparse matrix -> packed symmetric matrix
 */
void 
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                const SparseMatrix_t& s,
	                        real_type mu, RealMatrix_t& mout) {

  set_value(mout,0.0);

#pragma omp parallel
  {
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = h.getIndex();

    int iG = myIndex->first_diag(myID);
    int nn = 0;

    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    

      int nbi = myIndex->row_block(iG);
      int ioff = myIndex->row_loc(iG);
      int ii = ioff;
      int kk=0;
      for(int k=0; k<nbi; k++) {
	int jj= ioff;
	for(int kp=0; kp<nbi; kp++, jj++, kk++) {
	  if(jj>= ii) 
	    mout(ii,jj) = h.diag(myID,iL)[kk]+mu*s.diag(myID,iL)[kk];
	}
	ii++;
      }

      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
	int jG = myIndex->iadj(iL, nni, myID);
	if(jG<iG) continue;
	int nbj = myIndex->row_block(jG);
	ii = ioff;
	kk=0;
	for(int k=0; k<nbi; k++) {
	  int jj = myIndex->col_loc(jG);
	  for(int kp=0; kp<nbj; kp++,jj++,kk++)  {
	    if(jj>= ii) 
	      mout(ii,jj)+= h.offdiag(myID,nn)[kk]+mu*s.offdiag(myID,nn)[kk];
	  }
	  ii++;
	}
      
      } // nni: nadj(myID, iL)
    }// iL
  }
  //myclock.stop();
}

/*!\fn void OHMMS::copy(const SparseMatrix_t& h, SymMatrix_t>&  mout)
 *\brief sparse matrix -> symmetric matrix
 */
//void OHMMS::copy(const SparseMatrix_t& h, SymMatrix_t& mout) {
//
//  copy_upper(h,mout);
//}

/*!\fn void OHMMS::copy(const SparseMatrix_t& h, RealMatrix_t>&  mout)
 *\brief sparse matrix -> real matrix
 */
void 
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                               RealMatrix_t& mout, bool up) {

  if(up)
    copy_upper(h,mout);
  else
    copy_all(h,mout);
}

//////////////////////////////////////////////////////
// sparse matrix -> "hermitian" matrix (skip general complex matrix)
//////////////////////////////////////////////////////
/*!\fn void OHMMS::copy(const SparseMatrix_t& h, PackedHerMatrix_t>&  mout)
 *\brief sparse matrix -> packed hermitian matrix
 */
void 
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                PackedHerMatrix_t&  mout) {
  copy_upper(h,mout);
}

/*!\fn void OHMMS::copy(const SparseMatrix_t& h, HerMatrix_t>&  mout)
 *\brief sparse matrix -> hermitian matrix
 */
//void OHMMS::copy(const SparseMatrix_t& h, HerMatrix_t& mout) {
//  copy_upper(h,mout);
//}

/*!\fn void OHMMS::copy(const SparseMatrix_t& h, RealMatrix_t>&  mout)
 *\brief sparse matrix -> real matrix
 */
void 
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                ComplexMatrix_t& mout, bool up) {

  if(up)
    copy_upper(h,mout);
  else
    copy_all(h,mout);
}

//////////////////////////////////////////////////////
// sparse matrix*phase factor -> "hermitian" matrix (skip general complex matrix)
//////////////////////////////////////////////////////
/*!\fn void OHMMS::copy(const SparseMatrix_t& h, PackedHerMatrix_t>&  mout)
 *\brief sparse matrix -> packed hermitian matrix
 */
void
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                const complex_type* phi, 
                                PackedSymMatrix_t& mout) {
  // ignore phase
  copy_upper(h,mout);
}

//void
//OHMMS::copy(const SparseMatrix_t& h, const complex<Scalar_t>* phi, SymMatrix_t& mout) {
//  // ignore phase
//  copy_upper(h,mout);
//}

void
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                const complex_type* phi, 
                                RealMatrix_t& mout, bool up) {
  // ignore phase
  if(up)
    copy_upper(h,mout);
  else
    copy_all(h,mout);
}

///////////////////////////
// special copy function for sparse matrix*phi -> hermitan or general matrix 
// note that only upper triangle is set up
///////////////////////////
template<class SMat_t, class T, class Mat_t>
void copy_upper(const SMat_t& h, const complex<T>* phi, Mat_t& mout) {

  set_value(mout,0.0);

#pragma omp parallel
  {
    typedef typename SMat_t::const_pointer const_pointer;
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = h.getIndex();

    int iG(myIndex->first_diag(myID));
    //int nn = 0;
    int nntot = myIndex->first_offdiag(myID);
    const_pointer dPtr=h.diag(myID,0);     
    const_pointer oPtr=h.offdiag(myID,0);     
    int stride(h.getBlockSize());

    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
      int nbi  = myIndex->row_block(iG);
      int ioff = myIndex->row_loc(iG);

      const_pointer dptr=dPtr;
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
        for(int kp=0,jj=ioff; kp<nbi; kp++, jj++) {
          if(kp>=k) { mout(ii,jj) += *dptr; }
          ++dptr;
        }
      }
      dPtr+=stride;
      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++,nntot++,oPtr+=stride ) {
	int jG = myIndex->iadj(iL, nni, myID);
	if(jG < iG) continue;
	int nbj(myIndex->col_block(jG));
        int joff(myIndex->col_loc(jG));
        complex<T> p(phi[nntot]);
        const_pointer optr=oPtr;
        for(int k=0,ii=ioff; k<nbi; k++,ii++) {
          for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
#ifdef USE_NAG
            if(jj > ii) mout(ii,jj) += p*(*optr);
            if(jj== ii) mout(ii,jj) += p.real()*(*optr);
#else
            if(jj>=ii) mout(ii,jj)+=p*(*optr);
#endif
            ++optr;
          }//kp
        }//k
      } // nni
    }//iL
  }// end-of-omp
}

template<class SMat_t, class T, class Mat_t>
void copy_all(const SMat_t& h, const complex<T>* phi, Mat_t& mout) {

  set_value(mout,0.0);

#pragma omp parallel
  {
    typedef typename SMat_t::const_pointer const_pointer;
    omp_int_t myID = omp_get_thread_num();
    const SparseIndex_t* myIndex = h.getIndex();

    int iG(myIndex->first_diag(myID));
    //int nn = 0;
    int nntot = myIndex->first_offdiag(myID);
    const_pointer dPtr=h.diag(myID,0);     
    const_pointer oPtr=h.offdiag(myID,0);     
    int stride(h.getBlockSize());

    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
      int nbi  = myIndex->row_block(iG);
      int ioff = myIndex->row_loc(iG);

      const_pointer dptr=dPtr;
      for(int k=0,ii=ioff; k<nbi; k++,ii++) {
        for(int kp=0,jj=ioff; kp<nbi; kp++, jj++) {
          mout(ii,jj) += *dptr++; 
        }
      }
      dPtr+=stride;

      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++,nntot++,oPtr+=stride ) {
	int jG = myIndex->iadj(iL, nni, myID);
	int nbj(myIndex->col_block(jG));
        int joff(myIndex->col_loc(jG));
        complex<T> p(phi[nntot]);
        const_pointer optr=oPtr;
        for(int k=0,ii=ioff; k<nbi; k++,ii++) {
          for(int kp=0,jj=joff; kp<nbj; kp++, jj++) {
            mout(ii,jj)+=p*(*optr++);
          }//kp
        }//k
      }// nni

    }// iL

  }// end-of-omp
}

void
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h, 
                                const complex_type* phi, 
	                        PackedHerMatrix_t& mout) {
  copy_upper(h,phi,mout);
}

//void
//OHMMS::copy(const SparseMatrix_t& h,  const complex<Scalar_t>* phi, HerMatrix_t& mout) {
//
//  copy_upper(h,phi,mout);
//}

void
OHMMS::Sparse2DenseMatrix::copy(const SparseMatrix_t& h,  
     	                        const complex_type* phi, 
	                        ComplexMatrix_t& mout, bool up) {

  if(up) 
    copy_upper(h,phi,mout);
  else
    copy_all(h,phi,mout);

}

bool 
OHMMS::Sparse2DenseMatrix::matrix_equal(const SparseMatrix_t& h, 
                                        RealMatrix_t& mout, bool up) {


  for(int myID=0; myID<h.getIndex()->NumThreads; myID++) {

    const SparseIndex_t* myIndex = h.getIndex();

    int iG = myIndex->first_diag(myID);
    int nn = 0;

    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
      
      int nbi = myIndex->row_block(iG);
      int nbj = myIndex->col_block(iG);

      int ioff = myIndex->row_loc(iG);
      int joff = myIndex->col_loc(iG);
      int ii=ioff;
      int kk=0;
      for(int k=0; k<nbi; k++) {
	int jj= joff;
	for(int kp=0; kp<nbj; kp++, jj++,kk++) {
	  if(fabs(mout(ii,jj)-h.diag(myID,iL)[kk]) > 1e-12) {
            cout << ii << " " << jj << " " << iG << endl;
	    cout << mout(ii,jj) << " " << h.diag(myID,iL)[kk]<<endl;
	    return false;
	  }
	}
	ii++;
      }

      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++, nn++) {

	int jG = myIndex->iadj(iL, nni, myID);
	nbj = myIndex->col_block(jG);

	ii = ioff;
	kk=0;
	for(int k=0; k<nbi; k++) {
	  int jj = myIndex->col_loc(jG);
	  for(int kp=0; kp<nbj; kp++,jj++,kk++)  {
	    if(fabs(mout(ii,jj)-h.offdiag(myID,nn)[kk])<1e-12) {
              cout << ii << " " << jj << " " << iG << " " << jG << endl;
	      cout << mout(ii,jj) << " " << h.offdiag(myID,nn)[kk]<<endl;
	      return false;
	    }
	  }
	  ii++;
	}
      } // nni: nadj(myID, iL)
    }// iL
  }
  return true;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: SparseMatrix2MatrixCopy.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
