//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim, Seungwon Lee and Lars Jonsson
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// Choosing ENABLE_OPENMP results in
//   openMP parallelization
//   use map<int,int> to mask sparse indecies locally instead of Work arrary
//////////////////////////////////////////////////////////////

// -*- C++ -*-

#include "Message/OpenMP.h"
#ifdef HAVE_CONFIG_H
#include "ohmms-config.h"
#endif

//#define USE_WORKSPACE
#include <iostream>
#include <iomanip>
#include <map>
using namespace std;

#ifdef USE_MMV
extern "C" {
  void matvec(const double* restrict a, const double* restrict x, 
              double* restrict y, int& n, int& m);
}
#endif
#include "SparseMatrix/SparseMatrixOMP.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "Utilities/RandomGenerator.h"
#include "Utilities/OhmmsInfo.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
using namespace OHMMS;

SparseMatrixOMP::SparseMatrixOMP()
  : Num(0), WS_size(0),Dmax(0), NumThreads(0), myIndex(0)
{
  DEBUGMSG("Calling SparseMatrixOMP::SparseMatrixOMP()")
}

SparseMatrixOMP::SparseMatrixOMP(Index_t* ind)
  : Num(0), WS_size(-1),Dmax(0), NumThreads(0), myIndex(ind)
{
  DEBUGMSG("Calling SparseMatrixOMP::SparseMatrixOMP(Index_t*)")
}

SparseMatrixOMP::~SparseMatrixOMP() { 
  //JEONGNIM: 11/22/02 something is wrong; memory violation
  DEBUGMSG("Calling SparseMatrixOMP::~SparseMatrixOMP()")

  if(d_diag.size()) {
    for(int i=0; i<NumThreads; i++) delete d_diag[i]; 
  }
  if(d_offdiag.size()) {
    for(int i=0; i<NumThreads; i++)  delete d_offdiag[i];
  }
}

void SparseMatrixOMP::setIndex(Index_t* sInd) {
  myIndex=sInd;
}

void 
SparseMatrixOMP::create() {

  NumThreads = omp_get_max_threads();
  MaxNumPairs.resize(NumThreads+1);

  Dmax = myIndex->MaxBlockSize;
  Num = myIndex->getTotNum();

  DEBUGMSG("Create blocks of SparseMatrix " << myIndex->objName())
  DEBUGMSG("max block elements " << Dmax)
  DEBUGMSG("max number of blocks " << Num)
  DEBUGMSG("total number of off-diag blocks " << myIndex->getTotNadj())

  for(int i=0; i<NumThreads; i++) {
    MaxNumPairs[i] = myIndex->maxpairs(i);
  }  

  if(d_diag.empty()) {
    for(int i=0; i<NumThreads; i++) d_diag.push_back(0);
  }

  if(d_offdiag.empty()) {
    for(int i=0; i<NumThreads; i++) d_offdiag.push_back(0);
  }
  //MaxNumPairs[NumThreads] = numpairs;
  //WS_size = 2*Dmax*(2*numatoms_per_thread+numpairs_per_thread);
  ////std::cout << "Sparse matrix " << myName << std::endl;
  ////std::cout << "The number of pairs, atoms " << numatoms_per_thread << " " 
  ////	      << numpairs_per_thread << std::endl;
  //if(!myWorkSpace) myWorkSpace = WorkSpace::create(WS_size);

#pragma omp parallel
  {
#pragma omp for nowait
    for(int myID=0; myID<NumThreads; myID++) {
      int myNumAtoms(myIndex->num_diag(myID));
      int myNumPairs(myIndex->num_offdiag(myID));
      DEBUGMSG("number of pair blocks initialized " << myNumAtoms << " " << myNumPairs)
      if(d_diag[myID]==0) {
        d_diag[myID]=new Container_t;
      }
      d_diag[myID]->resize(myNumAtoms*Dmax);
      if(d_offdiag[myID]==0) {
        d_offdiag[myID]=new Container_t;
      }
      d_offdiag[myID]->resize(myNumPairs*Dmax);
    }
  }
}

void
SparseMatrixOMP::clearDiagonal(int myID) {
  std::fill(d_diag[myID]->begin(), d_diag[myID]->end(), 0.0e0);
}

void
SparseMatrixOMP::clear(int assignmode, int myID) {

  //////////////////////////////////
  // To add re-mapping for Dynamics Simulations
  //////////////////////////////////
  if(assignmode>= 0) 
    std::fill(d_diag[myID]->begin(),d_diag[myID]->end(), 0.0);
  std::fill(d_offdiag[myID]->begin(),d_offdiag[myID]->end(), 0.0);

  /* Not used now
  int myNumAtoms(myIndex->num_diag(myID));
  int myNumPairs(myIndex->num_offdiag(myID));
  DEBUGMSG("number of pair blocks assigned " << myNumPairs)
  if(assignmode == 1) { // diagonal block
    int iG =myIndex->first_diag(myID);
    iterator dit(d_diag[myID]->begin());
    for(int iL=0; iL<myNumAtoms; iL++){    
      for(int ibi=0; ibi<myIndex->row_block(iG); ibi++) {
        int cols(myIndex->col_block(iG));
        T sum(0.0);
        iterator cit=dit+cols*ibi;
        for(int ibj=0; ibj<cols; ibj++) {
          T x(2.0*Random()-1);
          *cit++= x;
          sum += x*x;
        }
        sum = 1.0/sqrt(sum);
        cit=dit+cols*ibi;
        for(int ibj=0; ibj<cols; ibj++) {
          *cit++ *= sum;
        }
        dit += Dmax;
      }
    }
  } else if(assignmode == 2) {
    iterator dit(d_diag[myID]->begin()), dit_end(d_diag[myID]->end());
    while(dit != dit_end) {
      *dit=2.0e0*Random()-1.0e0;
      ++dit;
    }
  } else if(assignmode == 3) {
    iterator dit(d_diag[myID]->begin()), dit_end(d_diag[myID]->end());
    while(dit != dit_end) {
      *dit=2.0e0*Random()-1.0e0;
      ++dit;
    }
    dit=d_offdiag[myID]->begin();
    dit_end=d_offdiag[myID]->end();
    while(dit != dit_end) {
      *dit=2.0e0*Random()-1.0e0;
      ++dit;
    }
  } else if(assignmode == 4) {
    int iG =myIndex->first_diag(myID);
    iterator dit(d_diag[myID]->begin());
    for(int iL=0; iL<myNumAtoms; iL++,iG++){    
      int rows(myIndex->row_block(iG));
      for(int ib=0; ib<rows; ib++) {
        *(dit+ib*rows+ib)=1.0;
      }
      dit+=Dmax;
    }
  }
  */
}

/** add a constant to the diagonal
 * @param lambda value to be added
 * @param myid thread id
 *
 * @warning potential problem with row_block when the block size depends on the components.
 */
void SparseMatrixOMP::add2Diag(T lambda, int myid) {
  int myNumAtoms = myIndex->num_diag(myid);
  int iG  = myIndex->first_diag(myid);
  T* dPtr=d_diag[myid]->data();
  for(int iL=0; iL<myNumAtoms; iL++, iG++){    
    AddDiag<T,TB_BLOCK>::apply(dPtr,lambda,myIndex->row_block(iG));
    dPtr += Dmax;
  }
}
/*
SparseMatrixOMP& SparseMatrixOMP::operator +=(T lambda) 
{
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int myNumAtoms = myIndex->num_diag(myID);
    int iG  = myIndex->first_diag(myID);
    for(int iL=0; iL<myNumAtoms; iL++, iG++){    
      AddDiag<T,TB_BLOCK>::apply(d_diag[myID][iL],lambda,myIndex->row_block(iG));
    }
  }
  return *this;
}
*/

/** (*this) += mu*a
 * @param a sparse matrix with the same sparsity
 * @param mu real value to be multiplied to a
 */
void SparseMatrixOMP::axpy(const SparseMatrixOMP& a, T mu, int myID) {
  OhmmsBlas::axpy(Dmax*myIndex->num_diag(myID),mu,
      a.d_diag[myID]->data(),1,d_diag[myID]->data(),1);
  OhmmsBlas::axpy(Dmax*myIndex->num_offdiag(myID),mu,
      a.d_offdiag[myID]->data(),1,d_offdiag[myID]->data(),1);
  //OhmmsBlas::axpy(Dmax*myIndex->num_diag(myID),mu,a.d_diag[myID][0],1,d_diag[myID][0],1);
  //OhmmsBlas::axpy(Dmax*myIndex->num_offdiag(myID),mu,a.d_offdiag[myID][0],1,d_offdiag[myID][0],1);
}

/**\fn void apply(const T* restrict win, T* restrct wout, bool mpitarget)
 *\param win the address of the input vector
 *\param wout the address of the output vector
 *\param mpitarget true, if the target vector is distributed
 */
void 
SparseMatrixOMP::apply(const T* restrict win, T* restrict wout, bool mpitarget) const {

//assign(wout,myIndex->nrows(), 0.0);
#pragma omp parallel
  {
    int myID = omp_get_thread_num();

    int iG = myIndex->first_diag(myID);
    int first = myIndex->col_loc(iG);
    int last = myIndex->col_loc(myIndex->last_diag(myID));
    int locnum = last-first;
    //assign(wout+first,locnum,0.0);
    int nn = 0;
    int offset = 0;
    if(mpitarget) {
      offset = first;
      std::fill(wout, wout+locnum, 0.0);
    } else{
      std::fill(wout+first, wout+last, 0.0);
    }

    T* dPtr=d_diag[myID]->data();
    T* oPtr=d_offdiag[myID]->data();
    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    

      T* target = wout + myIndex->row_loc(iG) - offset;
      int nbi = myIndex->row_block(iG);

#ifdef USE_MMV
      matvec(dPtr, iG*TB_BLOCK, target, TB_BLOCK, TB_BLOCK);
#else
      MultAdd<T,TB_BLOCK>::apply(dPtr, win+myIndex->col_loc(iG), target, nbi, nbi);
#endif
      dPtr+=Dmax; 

      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
#ifdef USE_MMV
	matvec(oPtr, win+jG*TB_BLOCK,target, TB_BLOCK, TB_BLOCK);
#else
        MultAdd<T,TB_BLOCK>
        ::apply(oPtr, win+myIndex->row_loc(jG), target, myIndex->row_block(jG), nbi);
#endif
        oPtr+=Dmax;
      }
    }
  }// end-of-omp
}

void 
SparseMatrixOMP::apply(T** restrict win, int i,  
		       T** restrict wout, int j, 
		       int len, int factor) const {

//    assign(wout,myIndex->nrows(), 0.0);
#pragma omp parallel
  {
    omp_int_t myID = omp_get_thread_num();
    //////////////// factor == 2, apply2
    int iG = myIndex->first_diag(myID);
    int first = myIndex->col_loc(iG);
    int locnum = myIndex->col_loc(myIndex->last_diag(myID))-first;
    int ioffset = len*i*factor;
    int joffset = len*j*factor;

    // all the values to zero
    assign(&(wout[myID][joffset]),factor*len,0.0);

    T* dPtr=d_diag[myID]->data();
    T* oPtr=d_offdiag[myID]->data();

    int nn(0), ndiag(myIndex->num_diag(myID));
    for(int iL=0; iL<ndiag; iL++,iG++){    
      int jshift = myIndex->col_loc(iG)-first;
      T* target = &(wout[myID][joffset+factor*jshift]);
      // need a special apply for complex
      int nbi = myIndex->row_block(iG);
      MultAdd<T,TB_BLOCK>::apply2(dPtr, &(win[myID][ioffset+factor*jshift]),
				  target, nbi, nbi);
      dPtr+=Dmax;

      int num_nn(myIndex->nadj(iL,myID));
      for(int nni=0; nni<num_nn; nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
        int ynode = myIndex->where(jG);
        int ishift = myIndex->row_loc(jG) 
     	           - myIndex->row_loc(myIndex->first_diag(ynode));

        MultAdd<T,TB_BLOCK>::apply2(oPtr, &win[ynode][ioffset+factor*ishift], 
            target, myIndex->row_block(jG), nbi);
        oPtr+=Dmax;
      }
    }
  }// end-of-omp
}

void 
SparseMatrixOMP::apply(const complex<T>* restrict win, 
		       complex<T>* restrict wout) const {

    //assign(wout,myIndex->nrows(), complex<T>(0.0));

#pragma omp parallel 
  {
    int myID = omp_get_thread_num();

    int iG = myIndex->first_diag(myID);
    int first = myIndex->col_loc(iG);
    int locnum = myIndex->col_loc(myIndex->last_diag(myID))-first;
    assign(wout+first,locnum, complex<T>(0.0));

    T* dPtr=d_diag[myID]->data();
    T* oPtr=d_offdiag[myID]->data();
    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
      complex<T>* target = wout + myIndex->row_loc(iG);
      int nbi = myIndex->row_block(iG);

      MultAdd<T,TB_BLOCK>::apply(dPtr, win+myIndex->col_loc(iG),
				 target, nbi, nbi);
      dPtr += Dmax;
      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++) {
        int jG = myIndex->iadj(iL, nni, myID);
        MultAdd<T,TB_BLOCK>
        ::apply(oPtr, win+myIndex->row_loc(jG), target, 
                myIndex->row_block(jG), nbi);
        oPtr+=Dmax;
      }
    }
  }// end-of-omp
}

bool
SparseMatrixOMP::get(ostream& os) const {

  os << "#block size " << myIndex->row_block(0) 
     << " " << myIndex->col_block(0)  << endl;
  os << "#number of diagonal blocks " << myIndex->getTotNum() << endl;
  os << "#number of off-diagonal blocks " << myIndex->getTotNadj() << endl;

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#pragma omp critical
  {
    int i=myIndex->first_diag(myID);
    T* dPtr=d_diag[myID]->data();
    for(int iat=0; iat < myIndex->num_diag(myID); i++,iat++){    
      os << "Atom = " << i << " " << i << " "
	 << myIndex->row_loc(i) << " " << myIndex->col_loc(i) << endl;
      int kk=0;
      for(int ib=0; ib<myIndex->row_block(i); ib++) {
        for(int jb=0;jb<myIndex->col_block(i); jb++)
          os << setw(18) << setprecision(8) << dPtr[kk++];
        os << endl;
      }
      dPtr+=Dmax;
    }
  }

#pragma omp critical
  {
    int iG=myIndex->first_diag(myID);
    int nn = 0;
    T* oPtr=d_offdiag[myID]->data();
    for(int iL=0; iL < myIndex->num_diag(myID); iL++,iG++){    
      for(int nni=0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
        os << "Pair = " << iG << " " << jG << " "
	 << myIndex->row_loc(iG) << " " << myIndex->col_loc(jG) << endl;
        int kk = 0;
        for(int ib=0; ib<myIndex->row_block(iG); ib++) {
          for(int jb=0;jb<myIndex->col_block(jG); jb++)
            os << setw(18) << setprecision(8) << oPtr[kk++];
          os << endl;
	}
        oPtr+=Dmax;
      }
    }
  }
  }    // end-of-omp
  return true;
}

bool SparseMatrixOMP::put(xmlNodePtr cur) {
  return true;
}

void SparseMatrixOMP::reset(){

}

void SparseMatrixOMP::resize(int myID) {
  int dItems(myIndex->num_diag(myID)*Dmax);
  if(dItems>d_diag[myID]->size()) {
    d_offdiag[myID]->resize(dItems);
  }
  int oItems(myIndex->num_offdiag(myID)*Dmax);
  if(oItems>d_offdiag[myID]->size()) {
    d_offdiag[myID]->resize(oItems);
  }
//    int numpairs = 0;
//    int nchanged = 0;
//    for(int i=0; i<NumThreads; i++) {
//      int nnmax = myIndex->maxpairs(i);
//      if(MaxNumPairs[i] < nnmax) {
//        nchanged++;
//        nnmax += nnmax/4;
//        MaxNumPairs[i] = nnmax;
//        delete [] d_offdiag[i];
//        d_offdiag[i] = new T*[nnmax];
//      }
//      numpairs +=   MaxNumPairs[i];
//    }
//    if(nchanged == 0 && (MaxNumPairs[NumThreads] > numpairs) ) return;
//    //Overwrite the maximum pairs for checking later
//    MaxNumPairs[NumThreads] = numpairs;
//    cout << "Changes in the number of pairs " << numpairs << endl;
//  #pragma omp parallel
//      {
//        int myID = omp_get_thread_num();
//        int myNumAtoms = myIndex->num_diag(myID);
//        int myNumPairs = MaxNumPairs[myID];
//        int numElements=(myNumAtoms+myNumPairs)*Dmax;
//        //resize if necessary and reset the pointer
//        myWorkSpace->resize(myID,numElements);
//        T* dptr = myWorkSpace->New(myID,numElements);
//        for(int i=0; i<myNumAtoms; i++, dptr+= Dmax){ 
//          d_diag[myID][i] = dptr;
//        }
//        for(int j=0; j<myNumPairs; j++, dptr+= Dmax){
//          d_offdiag[myID][j] = dptr;
//        }
//     }//omp-parallel closed
}

/*
//Obsolate: experimental version for order N
//Taken out to SparseSparseOps
//#define SPARSEMATRIXDEBUG
//  #ifdef ADD_
//  #define dgemm dgemm_
//  #endif
//  extern "C" {
//    void dgemm(const char& ta, const char& tb, 
//               const int& m, const int& n, const int& k,
//               const double& al, const double* a, const int& lda, 
//               const double* b, const int& ldb,
//               const double& be, double* c, const int& ldc);
//  }

inline 
void 
TestProduct(double* targetRef, const double* xref, const double* yref, 
            int na, int ma, int mb) {
  int ii=0;
  for(int i=0; i<na; i++)
    for(int j=0; j<mb; j++) {
      double tmp = 0;
      for(int k=0; k<ma; k++)
        tmp += xref[i*ma+k]*yref[k*mb+j];
      targetRef[ii++] += tmp;
    }
}


void 
SparseMatrixOMP::product(const SparseMatrixOMP& A, const SparseMatrixOMP& B) {

    int myID = omp_get_thread_num();

#ifdef SPARSEMATRIXDEBUG
//////////////////////////////// DEBUG: start
    int NS = myIndex->row_block(0);
    int BS = myIndex->col_block(0);  

    int nrow = Num*NS;
    int ncol = Num*BS;
    int namax = A.myIndex->row_block(0)*Num;    
    int mamax = A.myIndex->col_block(0)*Num;    
    int nbmax = B.myIndex->row_block(0)*Num;
    int mbmax = B.myIndex->col_block(0)*Num;

    FullMatrix_t aref(namax,mamax);
    FullMatrix_t bref(nbmax,mbmax);
    FullMatrix_t cref(nrow,ncol);
    aref = 0.0;
    bref = 0.0;
    A.copy(aref);
    B.copy(bref);

    for(int ii=0; ii<nrow; ii++) {
      for(int jj=0; jj<ncol; jj++) {
	T tmp = 0;
        for(int kk=0; kk<mamax; kk++) {
          tmp += aref(ii*mamax+kk)*bref(kk*mbmax+jj);
	}
        cref(ii*ncol+jj) = tmp;
      }
    }
#endif
//////////////////////////////// DEBUG: end
    DEBUGMSG("number of pair blocks assigned " << myIndex->numpairs(myID))
    assign(d_diag[myID][0],Dmax*myIndex->numatoms(myID),0.0);
    assign(d_offdiag[myID][0],Dmax*myIndex->numpairs(myID),0.0);

    //int nA, mA, mB;
    int nA = A.myIndex->row_block(0);
    int mA = A.myIndex->col_block(0);
    int mB = B.myIndex->col_block(0);
    DEBUGMSG("Amat = " << nA << " " << mA)
    DEBUGMSG("Bmat = " << mA << " " << mB)

    int ynn_loc = 0;  
    int ann_loc = 0;                    

#ifndef USE_MAP_MASK
    vector<int> Work(Num,-1);
#endif

    int iG = myIndex->first_diag(myID);

    // diagonal matrix elements
    for(int iL=0; iL<myIndex->num_diag(myID); iL++, iG++){    

#ifdef USE_MAP_MASK
      map<int,int> Work; // mapping global to local index
#endif
      for(int nni=0; nni<myIndex->nadj(iL,myID); nni++, ynn_loc++) {
        Work[myIndex->iadj(iL,nni,myID)] = ynn_loc;
      }
      Work[iG] = iL;

      T* xref = A.d_diag[myID][iL];
      T* targetRef = d_diag[myID][iL];

//        nA = A.myIndex->row_block(iG);
//        mA = A.myIndex->col_block(iG);
//        mB = B.myIndex->col_block(iG);

      TestProduct(targetRef, xref, B.d_diag[myID][iL], nA, mA, mB); 
      int bnn_loc = B.myIndex->begin_offdiag(myID, iL);

      for(int hnn=0; hnn< B.myIndex->nadj(iL,myID); hnn++, bnn_loc++) {

	int hjG = B.myIndex->iadj(iL, hnn,myID);

#ifdef USE_MAP_MASK
        if(Work.find(hjG) == Work.end()) continue;         
        int target = Work[hjG];
#else
        int target = Work[hjG];
	if(target < 0) continue;
#endif
	targetRef = d_offdiag[myID][target];
//          mB = B.myIndex->col_block(hjG);

	TestProduct(targetRef, xref, B.d_offdiag[myID][bnn_loc], nA, mA, mB);
      } // non-zero block of yi


      for(int nni=0; nni<A.myIndex->nadj(iL,myID); nni++, ann_loc++) {

	int hiG = A.myIndex->iadj(iL, nni, myID);  // col index of x == row of y

	int Ynode = B.myIndex->where(hiG);
        int Ynnloc = hiG - B.myIndex->myAtoms[Ynode][0];

#ifdef USE_MAP_MASK
	int target = -1;
        if(Work.find(hiG) != Work.end()) target = Work[hiG];
#else
	int target = Work[hiG];
#endif
	xref = A.d_offdiag[myID][ann_loc];
//  	nA = A.myIndex->row_block(hiG);
//  	mA = A.myIndex->col_block(hiG);

	if(target >= 0) { // yi = xi 
	  targetRef = d_offdiag[myID][target];
//  	  mB = B.myIndex->col_block(hiG);
	  TestProduct(targetRef, xref, B.d_diag[Ynode][Ynnloc],nA, mA, mB);
	}

        int hiL = hiG - B.myIndex->myAtoms[Ynode][0];
        Ynnloc = B.myIndex->begin_offdiag(Ynode,hiL);

	for(int hnn=0; hnn< B.myIndex->nadj(hiL, Ynode); hnn++, Ynnloc++) {

	  int hjG = B.myIndex->iadj(hiL, hnn, Ynode);
#ifdef USE_MAP_MASK
          if(Work.find(hjG) == Work.end()) continue;         
          target = Work[hjG];
#else
          target = Work[hjG];
	  if(target < 0) continue;
#endif
//  	  mB = B.myIndex->col_block(hjG);
	  if(iG == hjG) targetRef = d_diag[myID][target];
	  else          targetRef = d_offdiag[myID][target];
	  TestProduct(targetRef, xref, B.d_offdiag[Ynode][Ynnloc],nA, mA, mB);
	} // non-zero block of yi
      }

#ifndef USE_MAP_MASK
      for(int nni=0; nni<myIndex->nadj(iL,myID); nni++) {
        Work[myIndex->iadj(iL,nni,myID)] = -1;
      }
      Work[iG] = -1;
#endif
    } // local rows

#ifdef SPARSEMATRIXDEBUG
    iG=myIndex->first_diag(myID);
    for(int iL=0; iL< myIndex->num_diag(myID); iL++,iG++){    
        int kk=0;
        int ii = iG*myIndex->row_block(iG);
        int jj = iG*myIndex->col_block(iG);
        for(int ib=0; ib<myIndex->row_block(iG); ib++) {
          for(int jb=0;jb<myIndex->col_block(iG); jb++) {
	   if(fabs(d_diag[myID][iL][kk++] - cref((ii+ib)*ncol + jj+jb)) > 1e-12)
	   cout << "Incorrect - apply - diag!\n";
	  }
	}

    }

    iG=myIndex->first_diag(myID);
    int nn = 0;
    for(int iL=0; iL< myIndex->num_diag(myID); iL++,iG++){    
      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
        int ii = iG*myIndex->row_block(iG);
        int jj = jG*myIndex->col_block(jG);
        int kk = 0;
        for(int ib=0; ib<myIndex->row_block(iG); ib++) {
          for(int jb=0;jb<myIndex->col_block(jG); jb++) {
            if(fabs(d_offdiag[myID][nn][kk++]-cref((ii+ib)*ncol + jj+jb)) > 1e-12)
              cout << "Incorrect-apply-offdiag!\n";
	  }
	}
      }
    }

//  //////////////////////////////// DEBUG: end
//      for(int i=0; i<nrow; i++) {
//        for(int j=0; j<ncol; j++){
//          cout << setw(8) << setprecision(3) << cref(i*ncol+j);
//        }
//  	cout << endl;
//      }
//      cout << "==============" << endl;
//      y.print(cout);
#endif

    //  }
}

inline 
void TestProduct2(double* targetRef, const double* xref, const double* yref, int NS, int BS) {
  for(int i=0; i<NS; i++)
    for(int j=0; j<NS; j++) {
      double tmp = 0;
      for(int k=0; k<BS; k++) 
        tmp += xref[i*BS+k]*yref[j*BS+k];
      targetRef[i*NS+j] += tmp;
    }
//     const char transa = 'N';
//     const char transb = 'N';
//     dgemm(transa,transb, BS, BS, BS, 1.0, yref, BS, xref, BS, 1.0,targetRef, BS);
}

inline void Assign(const double* a, double *target, int n) {
  //for(int i=0; i<n; i++) target[i] += a[i];  
  for(int i=0; i<n; i++) target[i] = a[i];  
}

void 
SparseMatrixOMP::dot(const SparseMatrixOMP& x, const SparseMatrixOMP& y) {

//     int myID = omp_get_thread_num();
// //////////////////////////////// DEBUG: start
//     int NS = x.myIndex->row_block(0);  
//     int BS = x.myIndex->col_block(0);

// #ifdef SPARSEMATRIXDEBUG
//     int nrow = Num*NS;
//     int ncol = Num*BS;

//     RealMatrix_t aref(nrow,ncol);
//     RealMatrix_t bref(nrow,ncol);
//     RealMatrix_t cref(nrow,nrow);
//     aref = 0.0;
//     bref = 0.0;
//     x.copy(aref);    
//     y.copy(bref);    

//     for(int ii=0; ii<nrow; ii++)
//       for(int jj=0; jj<nrow; jj++) {
// 	double tmp = 0.0;
//         for(int ll=0; ll<ncol; ll++) tmp += aref(ii*ncol+ll)*bref(jj*ncol+ll);
//         cref(ii*nrow+jj) = tmp;
//       }

// #endif
// //////////////////////////////// DEBUG: end
//     DEBUGMSG("number of pair blocks assigned " << myIndex->numpairs(myID))
// #ifndef USE_MAP_MASK
//     vector<int> Work(Num,-1);
// #endif
//     RealMatrix_t tblock(NS,NS);

//     int nnloc = 0;// local locator for connectivity for this
//     int xnnloc = 0;// local locator for connectivity for x

//     int iG =  x.myIndex->first_diag(myID);
//     for(int iL=0; iL<x.myIndex->num_diag(myID); iG++,iL++){    

// #ifdef USE_MAP_MASK
//       map<int,int> Work;
// #endif
//       for(int nni=0; nni<x.myIndex->nadj(iL, myID); nni++, xnnloc++) {
//         Work[x.myIndex->iadj(iL, nni, myID)] = xnnloc;
//       }     
//       Work[iG] =iL;

//       int xi = iG;
//       int Ynode = y.myIndex->where(xi);
//       int Ydiagloc = xi - y.myIndex->myAtoms[Ynode][0];
//       int Yoffloc = y.myIndex->begin_offdiag(Ynode,Ydiagloc);
//       int Ynnmax = y.myIndex->nadj(Ydiagloc, Ynode);

//       int xcol = 0;
//       int Mynnmax = myIndex->nadj(iL, myID);
//       T *xref, *yref;
//       do {
// 	int myrow =xi;
// 	int icol = 0;
// 	tblock = 0.0e0;
// 	do {
// #ifdef USE_MAP_MASK
// 	  if(Work.find(myrow) != Work.end()) {
// 	    int tcol = Work[myrow];
// #else
//           int tcol = Work[myrow];
// 	  if(tcol > -1) {
// #endif
// 	    if(myrow == iG)  xref = x.d_diag[myID][tcol];          
// 	    else             xref = x.d_offdiag[myID][tcol];
// 	    if(icol == 0) {
// 	      yref = y.d_diag[Ynode][Ydiagloc];
// 	    } else  {
// 	      yref = y.d_offdiag[Ynode][Yoffloc];
// 	    }
// 	    TestProduct2(tblock.begin(), xref, yref, NS, BS);
// 	  }
// 	  if(icol > 0) Yoffloc++;
// 	  myrow = y.myIndex->iadj(Ydiagloc, icol, Ynode);
// 	  icol++;
//         } while(icol <= Ynnmax);

// 	if(xcol == 0) {
//           Assign(tblock.begin(), d_diag[myID][iL], Dmax);
// 	} else {
//           Assign( tblock.begin(), d_offdiag[myID][nnloc], Dmax);
// 	  nnloc++;
// 	}

//         xi = myIndex->iadj(iL, xcol, myID);
//         Ynode = y.myIndex->where(xi);
//         Ydiagloc = xi - y.myIndex->myAtoms[Ynode][0];
//         Yoffloc = y.myIndex->begin_offdiag(Ynode,Ydiagloc);
//         Ynnmax = y.myIndex->nadj(Ydiagloc, Ynode);
//         xcol++;

//       } while(xcol <= Mynnmax); 
// #ifndef USE_MAP_MASK
//       for(int nni=0; nni<x.myIndex->nadj(iL, myID); nni++) {
//         Work[x.myIndex->iadj(iL, nni, myID)] = -1;
//       }     
//       Work[iG] =-1;
// #endif
//     }// local rows
}

void 
SparseMatrixOMP::outerproduct(const SparseMatrixOMP& x, const SparseMatrixOMP& y) {

}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: SparseMatrixOMP.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

