//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <vector>
#include <set>
using namespace std;
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseIndexComposite.h"

SparseIndexComposite::SparseIndexComposite(const char* name): 
SparseIndexOMP(name), NumShell(0),  baseA(NULL), baseB(NULL) 
{
  cout <<"Calling SparseIndexComposite::SparseIndexComposite() with " << name << endl;
}

SparseIndexComposite::~SparseIndexComposite(){
  cout << "Calling SparseIndexComposite::~SparseIndexComposite() " 
       << objName() << endl;
  for(int i=0; i<NumThreads; i++) delete [] J[i];
  delete [] J;
  for(int i=0; i<NumThreads; i++) delete [] M[i];
  delete [] M;
} 

void SparseIndexComposite::init(const SparseIndexOMP* a, int nshell, 
				const NS_t& ns, const NS_t& nb) {
  baseA = a;
  NumShell = nshell;

  TotalNum = a->getTotNum();
  set_offsets(ns,nb);
  create();

  OwnIndices = false;
  NodeID = a->NodeID;
  IonID = a->IonID;
  myList = NULL;

  update();
}

void SparseIndexComposite::init(const SparseIndexOMP* a, const SparseIndexOMP* b,
				const NS_t& ns, const NS_t& nb) {

  baseA = a;
  baseB = b;
  NumShell = 0;
  myList = NULL;
  TotalNum = a->getTotNum();
  set_offsets(ns,nb);
  create();

  OwnIndices = false;
  NodeID = a->NodeID;
  IonID = a->IonID;
  update();
}

void SparseIndexComposite::update() {


#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    update(myID, myAtoms[myID], myPairs[myID]);
  }// end-of-omp

  int nn=myPairs[0][2];

  myPairs[0][0] = 0;
  myPairs[0][1] = nn;
  int inode = 1;
  while(inode < NumThreads) {
    myPairs[inode][0] = nn;
    nn+= myPairs[inode][2];
    myPairs[inode][1] = nn;
    inode++;
  }

}


void SparseIndexComposite::update(int myID, int *myatoms, int *mypairs) {

    myatoms[0] = baseA->myAtoms[myID][0];
    myatoms[1] = baseA->myAtoms[myID][1];
    myatoms[2] = baseA->myAtoms[myID][2];

    vector<int> Work(TotalNum);

    if(NumShell > 0) {

      int nnmax0 = baseA->numpairs(myID)/baseA->numatoms(myID)+1;
      int nnmax = nnmax0;
      for(int ishell=1; ishell<NumShell; ishell++) {
        int nt = ishell*nnmax0;
        nnmax += nt*nt*nt/2;
      } 
      nnmax *= baseA->numatoms(myID);

      if(nnmax > mypairs[3]) {
	nnmax += nnmax/10;
	if(J[myID]) delete [] J[myID];     
	J[myID] = new int[nnmax];
	myPairs[myID][3] = nnmax;
      }

      if(!M[myID]) { M[myID] = new int[baseA->numatoms(myID)+1];}

      int iG = baseA->myAtoms[myID][0];

      int nnloc = 0;
      M[myID][0] = 0;
      for(int iL=0; iL< baseA->myAtoms[myID][2]; iL++, iG++) {

        set<int> newneig;
	int nn=0;
	for(int i=0; i<TotalNum; i++) Work[i] = -1;

	Work[iG] = iG;

	for(int inn=0; inn< baseA->nadj(myID, iL); inn++) {
	  int jG = baseA->iadj(myID, iL, inn);
	  J[myID][nnloc] = jG;
	  nnloc++; nn++;
	  newneig.insert(jG);
	  Work[jG]=jG;
	}
      
	int ishell = 1;

	while(ishell<NumShell) {

	  set<int> newshell;
	  set<int>::iterator nneig = newneig.begin();
	  while(nneig != newneig.end()) {
	    int jG = *nneig;
	    int nodeid = where(jG);
	    int jL = jG - myAtoms[nodeid][0];
	    for(int jnn=0; jnn<baseA->nadj(nodeid,jL); jnn++) {
	      int kG = baseA->iadj(nodeid, jL,jnn);
	      if(Work[kG] > -1) continue;
	      newshell.insert(kG);
	      Work[kG] = kG;
	    }
	    nneig++;
	  }

	  // add a new shell and copy the new neighbors to newneig
	  // note that previous newneig is erased to avoid any redundant cals
	  newneig.erase(newneig.begin(), newneig.end());
	  set<int>::iterator it = newshell.begin();
	  for(; it != newshell.end(); ++it) {
	    J[myID][nnloc] = *it;
	    nnloc++;
	    nn++;
	    newneig.insert(*it);
	  }
	  ishell++;
	}
	M[myID][iL+1] = M[myID][iL] + nn;
      }
      mypairs[2] = nnloc;

    } else {

      int nna = baseA->numpairs(myID)/baseA->numatoms(myID);
      int nnb = baseB->numpairs(myID)/baseB->numatoms(myID);
      int nnmax =  (nna*nnb)*baseA->numatoms(myID);

      if(nnmax > mypairs[3]) {
	nnmax += nna*nnb;
	if(J[myID]) delete [] J[myID];     
	J[myID] = new int[nnmax];
	mypairs[3] = nnmax;
      }
      if(!M[myID]) { M[myID] = new int[baseA->numatoms(myID)+1];}

      int iG = baseA->myAtoms[myID][0];
      int nnloc = 0; // node sum
      M[myID][0] = 0;
      for(int iL=0; iL<baseA->myAtoms[myID][2]; iL++, iG++) {

	int nn=0; // per atom
	for(int i=0; i<TotalNum; i++) Work[i] = -1;

	Work[iG] = iG;
	for(int jnn=0; jnn< baseA->nadj(myID, iL); jnn++) {
	  int jG = baseA->iadj(myID, iL, jnn);
	  J[myID][nnloc] = jG;
	  nnloc++; nn++;
	  Work[jG]=jG;
	}

	for(int jnn=0; jnn<baseA->nadj(myID,iL); jnn++) {
	  int jG = baseA->iadj(myID, iL, jnn);
	  int nodeid = baseB->where(jG);
	  int jL = jG - baseB->myAtoms[nodeid][0];

	  for(int knn=0; knn< baseB->nadj(nodeid, jL); knn++) {
	    int kG = baseB->iadj(nodeid, jL, knn);
	    if(Work[kG] > -1) continue;
	    J[myID][nnloc] = kG;          
	    nnloc++; nn++;
	    Work[kG]=kG;
	  }
 	}
	M[myID][iL+1] = M[myID][iL] + nn;
      }
      mypairs[2] = nnloc;
    }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * OHMMS_VERSION: $Id: SparseIndexComposite.cpp 151 2004-11-02 23:51:14Z jnkim $
 ***************************************************************************/
