//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <iostream>
#include <iomanip>
using namespace std;

#include "SparseMatrix/SparseMatrixSerial.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "OhmmsPETE/TinyMatrixRef.h"

SparseMatrixSerial::SparseMatrixSerial()
:myIndex(NULL), d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0) 
{
}

SparseMatrixSerial::SparseMatrixSerial(Index_t* ind)
:d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0)
{
  init(ind);
}

SparseMatrixSerial::~SparseMatrixSerial() { 

  if(d_diag) { 
    //for(int i=0; i<Num; i++) delete [] d_diag[i];
    //Total memory is allocated to d_diag[0], d_diag[i] for i>0 only points
    //to the addresses
    delete [] d_diag[0];
    delete d_diag;
  }
  if(d_offdiag) { 
    //for(int i=0; i<NumPairs; i++) delete [] d_offdiag[i];
    //Total memory is allocated to d_diag[0], d_diag[i] for i>0 only points
    //to the addresses
    delete [] d_offdiag[0];
    delete d_offdiag;
  }
}

void
SparseMatrixSerial::init(Index_t* newindex) {

  if(newindex) { myIndex = newindex; }

  myIndex->update();

  if(Num<myIndex->numatoms()) { // first time, allocate pointer holders

    Num = myIndex->numatoms();
    NumPairs = myIndex->numpairs();
    Dmax = myIndex->MaxBlockSize;

    if(!d_diag) { //
      d_diag = new T*[Num];
      d_diag[0] = new T[Dmax*Num];
      for(int i=1; i<Num; i++) d_diag[i] = d_diag[i-1]+Dmax;
      //for(int i=0; i<Num; i++) d_diag[i] = new T[Dmax*Dmax];
    }

    if(!d_offdiag) { //
      WS_size = (NumPairs + NumPairs/10);// add 10%
      d_offdiag = new T*[WS_size];
      d_offdiag[0] = new T[Dmax*WS_size];
      for(int i=1; i<WS_size; i++) d_offdiag[i] = d_offdiag[i-1]+ Dmax;
      //for(int i=0; i<WS_size; i++) d_offdiag[i] = new T[Dmax*Dmax];
    }
  }

}

void
SparseMatrixSerial::reset(int assignmode) {

  myIndex->update();
  NumPairs = myIndex->numpairs();

  if(WS_size < NumPairs) {// checking the size

    delete [] d_offdiag[0];
    delete d_offdiag;
    
    WS_size = (NumPairs + NumPairs/10);// add 10%
    d_offdiag = new T*[WS_size];
    d_offdiag[0] = new T[Dmax*WS_size];
    for(int i=1; i<WS_size; i++) d_offdiag[i] = d_offdiag[i-1]+ Dmax;
    //for(int i=0; i<WS_size; i++) d_offdiag[i] = new T[Dmax*Dmax];
  }

  if(assignmode == 0) {
    assign(d_diag[0], Dmax*Num, 0.0);    
  }
  assign(d_offdiag[0], Dmax*NumPairs, 0.0);
}

SparseMatrixSerial& SparseMatrixSerial::operator +=(T lambda) 
{
  for(int i=0;i<Num;i++){
#ifdef TB_BLOCK
    for(int ib=0;ib<TB_BLOCK;ib++){
      d_diag[i][ib*TB_BLOCK+ib] += lambda;
    }
#else
    int nb = myIndex->row_block(i);
    for(int ib=0;ib<nb;ib++){
      d_diag[i][ib*nb+ib] += lambda;
    }
#endif
  }
  return *this;
}

void SparseMatrixSerial::apply(const FullMatrix_t& win, FullMatrix_t& wout) {

  wout = 0.0;
  int nrow = win.nrow();

  int nn = 0; // accumulation of nearest neigbor
  for(int i=0; i < myIndex->numatoms(); i++){
    int col0 = myIndex->col_loc(i);
    int row0 = myIndex->row_loc(i);
    int nbi = myIndex->row_block(i);
    for(int irow = 0; irow<nrow; irow++) {
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>::apply(d_diag[i],win[irow]+row0, wout[irow]+col0);
#else
      MultAdd<T,0>::apply(d_diag[i], win[irow]+row0,wout[irow]+col0, nbi, nbi);
#endif
    }

    for(int nni = 0; nni<myIndex->nadj(i); nni++, nn++) {
      int j = myIndex->iadj(i,nni);
      col0 = myIndex->col_loc(j);
      int nbj = myIndex->row_block(j);
      for(int irow = 0; irow<nrow; irow++) {
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>::apply(d_offdiag[nn],win[irow]+col0,wout[irow]+row0);
#else
      MultAdd<T,0>::apply(d_offdiag[nn],win[irow]+col0,wout[irow]+row0,nbj,nbi);
#endif
      }
    }
  }
}

void SparseMatrixSerial::apply(const T* win, T* wout) const {

  assign(wout,myIndex->nrows(), 0.0);

  int nn = 0; // accumulation of nearest neigbor
  for(int i=0; i < myIndex->numatoms(); i++){
    T* wtarget = wout + myIndex->row_loc(i);
#ifdef TB_BLOCK
    MultAdd<T,TB_BLOCK>::apply(d_diag[i], win+myIndex->col_loc(i), wtarget);
#else
    int nbi = myIndex->row_block(i);
    MultAdd<T,0>::apply(d_diag[i], win+myIndex->col_loc(i), wtarget, nbi, nbi);
#endif

    for(int nni = 0; nni<myIndex->nadj(i); nni++, nn++) {
      int j = myIndex->iadj(i,nni);
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>
	::apply(d_offdiag[nn],win+myIndex->col_loc(j),wtarget);
#else
      MultAdd<T,0>
	::apply(d_offdiag[nn],win+myIndex->col_loc(j),wtarget,
		myIndex->col_block(j), nbi);
#endif
    }
  }
}

void SparseMatrixSerial::apply(const complex<T>* win, complex<T>* wout) const {

  assign(wout,myIndex->nrows(),complex<T>(0.0));
  //  for(int i=0; i<myIndex->Offset[Num]; i++) wout[i] = 0.0;

  int nn = 0; // accumulation of nearest neigbor
  for(int i=0; i < myIndex->numatoms(); i++){
    complex<T>* wtarget = wout + myIndex->row_loc(i);
#ifdef TB_BLOCK
    MultAdd<T,TB_BLOCK>::apply(d_diag[i], win+myIndex->col_loc(i), wtarget);
#else
    int nbi = myIndex->row_block(i);
    MultAdd<T,0>::apply(d_diag[i], win+myIndex->col_loc(i), wtarget, nbi, nbi);
#endif

    for(int nni = 0; nni<myIndex->nadj(i); nni++, nn++) {
      int j = myIndex->iadj(i,nni);
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>
	::apply(d_offdiag[nn],win+myIndex->col_loc(j),wtarget);
#else
      MultAdd<T,0>
	::apply(d_offdiag[nn],win+myIndex->col_loc(j),wtarget,
		myIndex->col_block(j), nbi);
#endif
    }
  }
}

void SparseMatrixSerial::copy(FullMatrix_t& h, int packing) const {

  if(packing == 1) {
    int ncol = h.ncol(); 

#ifdef TB_BLOCK 
    const int D1 = TB_BLOCK; 
    const int D2 = TB_BLOCK; 
#endif

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
#ifndef TB_BLOCK 
      int D1 = myIndex->row_block(i);
#endif
      int ii0 = myIndex->row_loc(i);
      int ii = ii0;
      int kk=0;
      for(int k=0; k<D1; k++) {
        int ishift = (ncol*2-ii-1)*ii/2;
        int jj= ii0;
        for(int kp=0; kp<D1; kp++, jj++, kk++) {
          if(jj>= ii) h(ishift+jj) = d_diag[i][kk];
        }
        ii++;
      }

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
	int j=myIndex->iadj(i,nni);
	if(j<i) continue;
#ifndef TB_BLOCK
	int D2 = myIndex->col_block(j);
#endif
	ii = ii0;
	kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj = myIndex->col_loc(j);
	  for(int kp=0; kp<D2; kp++,jj++,kk++)  {
	    if(jj >= ii) h(ishift+jj)+= d_offdiag[nn][kk];
	  }
	  ii++;
	}
      }
    }
  } else {

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
      int ii0 = myIndex->row_loc(i);
#ifdef TB_BLOCK
      h.add(d_diag[i],TB_BLOCK,TB_BLOCK,ii0, ii0);
#else
      int nbi = myIndex->row_block(i);
      h.add(d_diag[i], nbi, myIndex->col_block(i), ii0, myIndex->col_loc(i));
#endif

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
        int j=myIndex->iadj(i,nni);
        if(j<i) continue;
#ifdef TB_BLOCK
        h.add(d_offdiag[nn],TB_BLOCK, TB_BLOCK, ii0, myIndex->col_loc(j));
#else
        h.add(d_offdiag[nn],nbi,myIndex->col_block(j),ii0,myIndex->col_loc(j));
#endif
      }//non-zero block of i
    }// i=number of particles

  }
}

void
SparseMatrixSerial::copy(FullCMatrix_t& h, const complex<T>* phi, int packing)
const {

  if(packing == 1) {  //Symmetric, packed
    int ncol = h.ncol(); 

#ifdef TB_BLOCK 
    const int D1 = TB_BLOCK; 
    const int D2 = TB_BLOCK; 
#endif

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
#ifndef TB_BLOCK 
      int D1 = myIndex->row_block(i);
#endif
      int ii0 = myIndex->row_loc(i);
      int ii = ii0;
      int kk=0;
      for(int k=0; k<D1; k++) {
        int ishift = (ncol*2-ii-1)*ii/2;
        int jj= ii0;
        for(int kp=0; kp<D1; kp++, jj++, kk++) {
          if(jj>= ii) h(ishift+jj) = d_diag[i][kk];
        }
        ii++;
      }

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
	int j=myIndex->iadj(i,nni);
	if(j<i) continue;
#ifndef TB_BLOCK 
	int D2 = myIndex->col_block(j);
#endif
	ii = ii0;
	kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj = myIndex->col_loc(j);
	  for(int kp=0; kp<D2; kp++,jj++,kk++)  {
	    if(jj >= ii) h(ishift+jj)+= d_offdiag[nn][kk]*phi[nn];
	  }
	  ii++;
	}
      }// nni
    }//i

  } else {

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
      int ii0 = myIndex->row_loc(i);
#ifdef TB_BLOCK
      h.add(d_diag[i],TB_BLOCK,TB_BLOCK,ii0,ii0);
#else
      int nbi = myIndex->row_block(i);
      h.add(d_diag[i], nbi, myIndex->col_block(i), ii0, myIndex->col_loc(i));
#endif

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
	int j=myIndex->iadj(i,nni);
        if(j<i) continue;
#ifdef TB_BLOCK
        h.add(d_offdiag[nn],TB_BLOCK,TB_BLOCK,ii0,myIndex->col_loc(j), phi[nn]);
#else
        h.add(d_offdiag[nn], nbi, myIndex->col_block(j),
              ii0,myIndex->col_loc(j), phi[nn]);
#endif
      }// nni
    }// i
  }// packing == 0
}

void
SparseMatrixSerial::copy(FullCMatrix_t& h, int packing) const {


  if(packing == 1) {  //Symmetric, packed
    int ncol = h.ncol(); 

#ifdef TB_BLOCK 
    const int D1 = TB_BLOCK; 
    const int D2 = TB_BLOCK; 
#endif

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
#ifndef TB_BLOCK 
      int D1 = myIndex->row_block(i);
#endif
      int ii0 = myIndex->row_loc(i);
      int ii = ii0;
      int kk=0;
      for(int k=0; k<D1; k++) {
        int ishift = (ncol*2-ii-1)*ii/2;
        int jj= ii0;
        for(int kp=0; kp<D1; kp++, jj++, kk++) {
          if(jj>= ii) h(ishift+jj) = d_diag[i][kk];
        }
        ii++;
      }

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
	int j=myIndex->iadj(i,nni);
	if(j<i) continue;
#ifndef TB_BLOCK 
	int D2 = myIndex->col_block(j);
#endif
	ii = ii0;
	kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj = myIndex->col_loc(j);
	  for(int kp=0; kp<D2; kp++,jj++,kk++)  {
	    if(jj >= ii) h(ishift+jj)+= d_offdiag[nn][kk];
	  }
	  ii++;
	}
      }// nni
    }//i

  } else {

    int nn = 0;
    for(int i=0; i < myIndex->numatoms(); i++){
      int ii0 = myIndex->row_loc(i);
#ifdef TB_BLOCK
      h.add(d_diag[i],TB_BLOCK,TB_BLOCK,ii0,ii0);
#else
      int nbi = myIndex->row_block(i);
      h.add(d_diag[i], nbi, myIndex->col_block(i), ii0, myIndex->col_loc(i));
#endif

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
	int j=myIndex->iadj(i,nni);
        if(j<i) continue;
#ifdef TB_BLOCK
        h.add(d_offdiag[nn],TB_BLOCK,TB_BLOCK,ii0,myIndex->col_loc(j));
#else
        h.add(d_offdiag[nn], nbi, myIndex->col_block(j),ii0,myIndex->col_loc(j));
#endif
      }// nni
    }// i
  }// packing == 0

}

void 
SparseMatrixSerial::product(const FullMatrix_t& a, const FullMatrix_t& b,
                            const vector<T>& w){
  int nn = 0;
  for(int i=0; i < myIndex->numatoms(); i++){

#ifdef TB_BLOCK
    Product<T,TB_BLOCK>
    ::make(d_diag[i],a,myIndex->col_loc(i),b,myIndex->row_loc(i),w);
#else
    Product<T,0>::make(d_diag[i], myIndex->row_block(i), myIndex->col_block(i),
                       a,myIndex->row_loc(i),b,myIndex->col_loc(i),w);
#endif

    for(int nni=0; nni<myIndex->nadj(i); nni++, nn++) {
      int j=myIndex->iadj(i,nni); // col index in a full matrix
      if(i == j) continue;
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>
      ::make(d_offdiag[nn],a,myIndex->row_loc(i),b,myIndex->col_loc(j),w);
#else
      Product<T,0>::make(d_offdiag[nn], 
                         myIndex->row_block(i), myIndex->col_block(j),
                         a,myIndex->row_loc(i),b,myIndex->col_loc(j),w);
#endif
    }// off-diagonal blocks
  }  // i
}

void 
SparseMatrixSerial::product(const T* a, const T* b) {
  
  int nn = 0;
  for(int i=0; i < myIndex->numatoms(); i++){

#ifdef TB_BLOCK
      Product<T,TB_BLOCK>
      ::make(d_diag[i], a+myIndex->row_loc(i),b+myIndex->col_loc(i));     
#else
      Product<T,0>
	::make(d_diag[i],myIndex->row_block(i), myIndex->col_block(i),
	       a+myIndex->row_loc(i),b+myIndex->col_loc(i));
#endif

    for(int nni=0; nni<myIndex->nadj(i); nni++, nn++) {
      int j=myIndex->iadj(i,nni); // col index in a full matrix
      if(i == j) continue;
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>
      ::make(d_offdiag[nn],a+myIndex->row_loc(i),b+myIndex->col_loc(j));     
#else
      Product<T,0>
	::make(d_offdiag[nn],myIndex->row_block(i), myIndex->col_block(j),
	       a+myIndex->row_loc(i),b+myIndex->col_loc(j));
#endif
    }// off-diagonal blocks
  }  // i
}

void 
SparseMatrixSerial::product(SparseMatrixSerial& Imat, 
                            const FullCMatrix_t& a, const FullCMatrix_t& b,
                            const vector<T>& w){

  int nn = 0;
  for(int i=0; i < myIndex->numatoms(); i++){
#ifdef TB_BLOCK
    Product<T,TB_BLOCK>::make(d_diag[i], Imat.d_diag[i],
			      a,myIndex->row_loc(i),b,myIndex->col_loc(i),w);
#else
    Product<T,0>::make(d_diag[i], Imat.d_diag[i], 
		       myIndex->row_block(i), myIndex->col_block(i),
		       a,myIndex->row_loc(i),b,myIndex->col_loc(i),w);
#endif

    for(int nni = 0; nni<myIndex->nadj(i); nni++, nn++) {  
      int j=myIndex->iadj(i,nni); // col index in a full matrix
      if(i == j) continue;
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>::make(d_offdiag[nn],Imat.d_offdiag[nn],
				a,myIndex->row_loc(i),b,myIndex->col_loc(j),w);
#else
      Product<T,0>::make(d_offdiag[nn], Imat.d_offdiag[nn], 
			 myIndex->row_block(i), myIndex->col_block(j),
			 a,myIndex->row_loc(i),b,myIndex->col_loc(j),w);
#endif
    }  // off-diagonal blocks
  }
}

void 
SparseMatrixSerial::print(ostream& os) {

  os << "diagonal blocks\n";
  for(int i=0; i < myIndex->numatoms(); i++){
    os << "Atom = " << i << endl;
    int kk=0;
    for(int ib=0; ib<myIndex->row_block(i); ib++) {
      for(int jb=0;jb<myIndex->col_block(i); jb++)
        os << setw(12) << setprecision(3) << d_diag[i][kk++];
      os << endl;
    }
  }

  os << "off-diagonal blocks\n";
  // off-diagonal blocks
  int nn = 0;
  for(int i=0; i<Num; i++) {
    for(int nni=0; nni<myIndex->nadj(i); nni++, nn++) {
      int j=myIndex->iadj(i,nni); // col index in a full matrix
      int kk = 0;
      os << "Pair = " << i << " " << j << endl;
      for(int ib=0; ib<myIndex->row_block(i); ib++) {
	for(int jb=0;jb<myIndex->col_block(j); jb++)
	  os << setw(12) << setprecision(3) << d_offdiag[nn][kk++];
	os << endl;
      }
    }
  }
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseMatrixSerial.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

