//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "SparseMatrix/SparseMatrixSpinOrbit.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "OhmmsPETE/TinyMatrixRef.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"

//////////////////////////////////////////////////////////
SparseMatrixSerial::SparseMatrixSerial(Index_t* indexin)
  :d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0)
{
  read_so_parameters();
  init(indexin);
}

SparseMatrixSerial::SparseMatrixSerial()
  :myIndex(NULL), d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0)
{
  read_so_parameters();
}

SparseMatrixSerial::~SparseMatrixSerial() { 

  if(d_diag) { 
    //for(int i=0; i<Num; i++) delete [] d_diag[i];
    //Total memory is allocated to d_diag[0], d_diag[i] for i>0 only points
    //to the addresses
    delete [] d_diag[0];
    delete d_diag;
  }
  if(d_offdiag) { 
    //for(int i=0; i<NumPairs; i++) delete [] d_offdiag[i];
    //Total memory is allocated to d_diag[0], d_diag[i] for i>0 only points
    //to the addresses
    delete [] d_offdiag[0];
    delete d_offdiag;
  }
}

SparseMatrixSerial& SparseMatrixSerial::operator +=(T lambda) 
{
  for(int i=0;i<Num;i++){
    int nb = myIndex->row_block(i);
    for(int ib=0;ib<nb;ib++){
      d_diag[i][ib*nb+ib] += lambda;
    }
  }
  return *this;
}

void
SparseMatrixSerial::init(Index_t* newindex){

  if(newindex) { myIndex = newindex; }

  myIndex->update();

  if(Num<myIndex->numatoms()) { // first time, allocate pointer holders

    NumPairs = myIndex->numpairs();
    Num = myIndex->numatoms();
    Dmax = myIndex->MaxBlockSize; // maximum block size

    if(!d_diag) { //
      d_diag = new T*[Num];
      d_diag[0] = new T[Dmax*Num];
      for(int i=1; i<Num; i++) d_diag[i] = d_diag[i-1]+Dmax; 
    }

    if(!d_offdiag) { //
      WS_size = (NumPairs + NumPairs/10);// add 10%
      d_offdiag = new T*[WS_size];
      d_offdiag[0] = new T[Dmax*WS_size];
      for(int i=1; i<WS_size; i++) d_offdiag[i] = d_offdiag[i-1]+ Dmax;
    }
  }

}

void
SparseMatrixSerial::reset(int assignmode) {

  NumPairs = myIndex->numpairs();
  if(WS_size < NumPairs) {// checking the size

    delete [] d_offdiag[0];
    delete d_offdiag;
    
    WS_size = (NumPairs + NumPairs/10);// add 10%
    d_offdiag = new T*[WS_size];
    d_offdiag[0] = new T[Dmax*WS_size];
    for(int i=1; i<WS_size; i++) d_offdiag[i] = d_offdiag[i-1]+ Dmax;
  }

  if(assignmode == 0) assign(d_diag[0], Dmax*Num, 0.0);    
  assign(d_offdiag[0],Dmax*NumPairs, 0.0);
}

void SparseMatrixSerial::apply(const complex<T>* win, complex<T>* wout) const {

  assign(wout,2*myIndex->ncols(),complex<T>(0.0));

  int* ionid = myIndex->IonID;

  int nn = 0;//accumulating neighbors
  for(int i=0; i < Num; i++){

    complex<T>* wtarget = wout + 2*myIndex->row_loc(i);
    SO[ionid[i]]->apply(win + 2*myIndex->col_loc(i), wtarget);

#ifdef TB_BLOCK
    MultAdd2<T,TB_BLOCK>::apply(d_diag[i],win+2*myIndex->col_loc(i), wtarget);
#else
    int nbi = myIndex->col_block(i);
    MultAdd2<T,0>::apply(d_diag[i],win+2*myIndex->col_loc(i),wtarget,nbi,nbi);
#endif

    for(int nni=0; nni<myIndex->nadj(i); nni++, nn++) {
      int j = myIndex->iadj(i, nni);
#ifdef TB_BLOCK
      MultAdd2<T,TB_BLOCK>
      ::apply(d_offdiag[nn],win+2*myIndex->col_loc(j),wtarget);
#else
      MultAdd2<T,0>::apply(d_offdiag[nn],win+2*myIndex->col_loc(j),wtarget,
                           myIndex->col_block(j),nbi);
#endif

    }
  }
}

void SparseMatrixSerial::copy(FullMatrix_t& mout, int packing) const {


  int* ionid = myIndex->IonID;
  if(packing == 0) {
    int nn = 0;
    for(int i=0; i < Num; i++){
      int nbi = myIndex->row_block(i);
      int upi=  2*myIndex->row_loc(i);
      int downi=upi+nbi;

      SO[ionid[i]]->copy(mout.begin(), upi, upi, mout.ncol()); 
      mout.add(TinyMatrixRef<T>(d_diag[i],nbi,nbi), upi, upi);
      mout.add(TinyMatrixRef<T>(d_diag[i],nbi,nbi), downi, downi);

      for(int nni=0; nni < myIndex->nadj(i); nni++, nn++) {
        int j=myIndex->iadj(i,nni);
        if(j<i) continue;
        int upj=2*myIndex->col_loc(j);
        int nbj = myIndex->col_block(j);
        int downj=upj+nbj;
        mout.add(TinyMatrixRef<T>(d_offdiag[nn],nbi,nbj), upi, upj);
        mout.add(TinyMatrixRef<T>(d_offdiag[nn],nbi,nbj), downi, downj);
      }
    }
  }
}

void
SparseMatrixSerial::copy(FullMatrix_t& h, const complex<T>* phi, int packing) 
const {

  if(packing == 0) {

    int* ionid = myIndex->IonID;
    int nn = 0;
    vector<complex<T> > tb(Dmax);
    for(int i=0; i < Num; i++){
      int nbi = myIndex->row_block(i);
      int upi=2*myIndex->row_loc(i);
      int downi=upi+nbi;
      SO[ionid[i]]->copy(h.begin(), upi, upi, h.ncol()); 
      h.add(TinyMatrixRef<T>(d_diag[i],nbi,nbi), upi, upi);
      h.add(TinyMatrixRef<T>(d_diag[i],nbi,nbi), downi, downi);

      static complex<T>* tb =NULL;
      if(!tb) tb  = new complex<T>(Dmax);

      for(int nni=0; nni < myIndex->nadj(i); nni++,nn++) {
        int j=myIndex->iadj(i,nni); // col index in a full matrix
        if(j<i) continue;
        int nbj = myIndex->col_block(j);
        for(int k=0; k<nbi*nbj; k++) {
           tb[k] = phi[nn]*d_offdiag[nn][k]; 
        }
        int upj = 2*myIndex->col_loc(j);
        h.add(tb, nbi, nbj, upi, upj);
        h.add(tb, nbi, nbj, downi, upj+nbj);
      }//nni
    }//i 
  }
}

void SparseMatrixSerial::read_so_parameters(){

  ifstream fin("TBPARAM");
  if(fin) {
    vector<string> slist;
    int nw=0;
    do {
      nw = getwords(slist,fin);
    } while(nw == 0 || slist[0] != "<SO>" || slist[0] == "end"); 
    if(slist[0] != "end") {
      SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
      int numspecies =mySpecies->getTotalNum(); 
      int ins = SpeciesCollection::addAttrib("ns");
      SO.reserve(numspecies);
      for(int i=0; i<numspecies; i++) {
	SO.push_back(new SpinOrbitComp<T>);
      }
      getwords(slist,fin);
      do {
	int newid =SpeciesCollection::addSpecies(slist[0].c_str());
	int i0 = atoi(slist[2].c_str());
	double lambda = atof(slist[3].c_str());
	int ns = static_cast<int>((*mySpecies)(ins,newid));
	SO[newid]->add(new SpinOrbitPP<T>(lambda, i0, i0));
	SO[newid]->set(ns,ns);
	getwords(slist,fin);
      } while(slist[0] != "</SO>");
    }
  }  
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseMatrixSpinOrbit.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

