//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim, Seungwon Lee, and Lars Jonsson
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Message/OpenMP.h"
#include <iostream>
#include <iomanip>
using namespace std;
#include "SparseMatrix/SparseMatrixSpinOrbitMP.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "OhmmsPETE/TinyMatrixRef.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"

#ifdef USE_WORKSPACE
#include "WorkSpace/WorkSpace.h"
#endif
SparseMatrixMP::SparseMatrixMP()
:myIndex(NULL), d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0) 
{
  read_so_parameters();
}

SparseMatrixMP::SparseMatrixMP(Index_t* ind)
:d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0)
{
  read_so_parameters();
  init(ind);
}

SparseMatrixMP::~SparseMatrixMP() { 
#ifdef USE_WORKSPACE
  //  WorkSpace::release();
#endif
}

void
SparseMatrixMP::init(Index_t* newindex) {

  if(newindex) { myIndex = newindex; }
  myIndex->update();

  if(Num<myIndex->nrows()) { // first time, allocate pointer holders

    Dmax = myIndex->maxBasis;
    NumPairs = myIndex->numpairs();
    Num = myIndex->nrows();
    NumThreads = omp_get_max_threads();
    d_diag = new T**[NumThreads];
    d_offdiag = new T**[NumThreads];
    myAtoms = new int*[NumThreads];
    myPairs = new int*[NumThreads];

    // create a workspace
    int nshift = Dmax*Dmax; // block-stride
#ifdef USE_WORKSPACE
    int ws_size = (Dmax+1)*(Dmax+1)*5*(Num+NumThreads); 
    WorkSpace::create(ws_size);
#else
    int ws_size = Dmax*Dmax*(Num+NumPairs);
    T* dump = new T[ws_size];
#endif

#pragma omp parallel
    {
      int myID = omp_get_thread_num();
      myAtoms[myID] =  new int[3];
      myPairs[myID] =  new int[3];
      make_mp_map(myID, myAtoms[myID], myPairs[myID]);
#pragma omp critical 
      {
      cout << "first atom= " << myAtoms[myID][0] << endl;
      cout << "last atom= " << myAtoms[myID][1] << endl;
      cout << "first pair= " << myPairs[myID][0] << endl;
      cout << "last pair= " << myPairs[myID][1] << endl;
      }
      int myNumAtoms = myAtoms[myID][2];
      int myNumPairs = myPairs[myID][2];

      d_diag[myID] = new T*[myNumAtoms];
      d_offdiag[myID] = new T*[myNumPairs];

#ifdef USE_WORKSPACE
      T* dptr = WorkSpace::New(myID, myNumAtoms*nshift); 
      T* optr = WorkSpace::New(myID, myNumPairs*nshift);
#else
      T *dptr = dump + myAtoms[myID][0]*nshift;
      T *optr = dump + Num*nshift + myPairs[myID][0]*nshift;
#endif

      for(int i=0; i<myNumAtoms; i++, dptr+=nshift){ 
        d_diag[myID][i] = dptr;
      }
      for(int j=0; j<myNumPairs; j++, optr+=nshift){
        d_offdiag[myID][j] = optr;
      }

    }//omp-parallel closed
  }// if(Num<myIndex->nrows()) // first time, allocate pointer holders
}

void SparseMatrixMP::make_mp_map(int myID, int *myatoms, int *mypairs) {

  int bat=Num/NumThreads;
  int residue = Num%NumThreads;

  if(myID<residue){
    myatoms[0] = myID*(bat+1);
    myatoms[1] = myatoms[0] + bat + 1;
    myatoms[2] = myatoms[1] - myatoms[0];
  }
  else {
    myatoms[0] = residue*(bat+1) + (myID - residue)*bat;    
    myatoms[1] = myatoms[0] + bat;
    myatoms[2] = myatoms[1] - myatoms[0];
  }

  int istart = myatoms[0];
  int nn=0;
  while(nn<NumPairs && myIndex->irow(nn) != istart) {
    nn++;
  }

  mypairs[0] = nn;
  if(myatoms[1]== Num) mypairs[1]=NumPairs;
  else {
    int nf=mypairs[0];
    while(myIndex->irow(nf) != myatoms[1]) {
      nf++;
    }
    mypairs[1] = nf;
  }
  mypairs[2] = mypairs[1] - mypairs[0];
}

void
SparseMatrixMP::reset(bool changediag) {
#pragma omp parallel
  {
      int myID = omp_get_thread_num();
      int myNumAtoms = myAtoms[myID][2];
      int myNumPairs = myPairs[myID][2];
      int nbtot = Dmax*Dmax;
      if(changediag) {
        assign(d_diag[myID][0],nbtot*myNumAtoms,0.0);
      }
      assign(d_offdiag[myID][0],nbtot*myNumPairs,0.0);
  }//omp-parallel closed
}

SparseMatrixMP& SparseMatrixMP::operator +=(T lambda) 
{
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int myNumAtoms = myAtoms[myID][2];
    int istart = myAtoms[myID][0];
    for(int i=0; i<myNumAtoms; i++,istart++){    
      int nb = myIndex->Basis[istart]; 
      for(int ib=0;ib<nb;ib++){
        d_diag[myID][i][ib*nb+ib] += lambda;
      }
    }
  }
  return *this;
}

void 
SparseMatrixMP::apply(const dcomplex_t* win, dcomplex_t* wout) const {


  int* ionid = myIndex->IonID;
  for(int i=0; i<2*myIndex->Offset[Num]; i++) wout[i] = 0.0;

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iat = 0;
    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
      int shift = 2*myIndex->Offset[i];
      SO[ionid[i]]->apply(win+shift, wout+shift);
#ifdef TB_BLOCK
      MultAdd2<T,TB_BLOCK>::apply(d_diag[myID][iat], win+shift, wout+shift);
#else
      MultAdd2<T,0>::apply(d_diag[myID][iat], win+shift, wout+shift,
                           myIndex->Basis[i],myIndex->Basis[i]);
#endif
    }
    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);
#ifdef TB_BLOCK
      MultAdd2<T,TB_BLOCK>
      ::apply(d_offdiag[myID][iat],win+2*myIndex->Offset[j],
              wout+2*myIndex->Offset[i]);
#else
      MultAdd2<T,0>
      ::apply(d_offdiag[myID][iat],
              win+2*myIndex->Offset[j],wout+2*myIndex->Offset[i],
              myIndex->Basis[j], myIndex->Basis[i]);
#endif
    }
  }
}

void SparseMatrixMP::copy(FullMatrix_t& mout, int packing) const {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();

    unsigned int ncoltot = mout.ncol();
    int iat = 0;
    int* ionid = myIndex->IonID;

    if(packing == 0) {

      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
	const int nb = TB_BLOCK;
#else
	int nb = myIndex->Basis[i];
#endif
	int up=2*myIndex->Offset[i];
	int down=up+nb;
	SO[ionid[i]]->copy(mout.begin(), up, up, ncoltot);
	mout.add(d_diag[myID][iat], nb, nb, up, up);
	mout.add(d_diag[myID][iat], nb, nb, down, down);
      }
      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j>=i) {
#ifdef TB_BLOCK
	  const int nbi = TB_BLOCK;
	  const int nbj = TB_BLOCK;
#else
	  int nbi = myIndex->Basis[i];
	  int nbj = myIndex->Basis[j];
#endif
	  int upi=2*myIndex->Offset[i];
	  int downi=upi+nbi;
	  int upj=2*myIndex->Offset[j];
	  int downj=upj+nbj;
	  mout.add(d_offdiag[myID][iat], nbi, nbj, upi, upj);
	  mout.add(d_offdiag[myID][iat], nbi, nbj, downi, downj);
	}
      }
    } // full matrix
    else {

      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
	const int nb = TB_BLOCK;
#else
	int nb = myIndex->Basis[i];
#endif
	int up=2*myIndex->Offset[i];
	int down=up+nb;

	SO[ionid[i]]->copy(mout.begin(), up, ncoltot);
        int iuu=up, idd = down;
        int kk=0;
        for(int k=0; k<nb; k++) {
          int ushift = (ncoltot*2-iuu-1)*iuu/2;
          int dshift = (ncoltot*2-idd-1)*idd/2;
          int juu= up, jdd = down;
          for(int kp=0; kp<nb; kp++, juu++, jdd++, kk++) {
            if(juu>= iuu) mout(ushift+juu) += d_diag[myID][iat][kk];
            if(jdd>= idd) mout(dshift+jdd) += d_diag[myID][iat][kk];
	  }
          iuu++;
          idd++;
	}
      }

      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {

	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j>=i) {
#ifdef TB_BLOCK
	  const int nbi = TB_BLOCK;
	  const int nbj = TB_BLOCK;
#else
	  int nbi = myIndex->Basis[i];
	  int nbj = myIndex->Basis[j];
#endif
	  int upi=2*myIndex->Offset[i];
	  int downi=upi+nbi;
	  int upj=2*myIndex->Offset[j];
	  int downj=upj+nbj;

	  int kk=0;
	  for(int k=0; k<nbi; k++) {
	    int ushift = (ncoltot*2-upi-1)*upi/2;
	    int dshift = (ncoltot*2-downi-1)*downi/2;
	    int juu= upj, jdd = downj;
	    for(int kp=0; kp<nbj; kp++, juu++, jdd++, kk++) {
	      if(juu>= upi)   mout(ushift+juu) += d_offdiag[myID][iat][kk];
	      if(jdd>= downi) mout(dshift+jdd) += d_offdiag[myID][iat][kk];
	    }
	    upi++;
	    downi++;
	  }

	//mout.add(d_offdiag[myID][iat], nbi, nbj, upi, upj);
	//mout.add(d_offdiag[myID][iat], nbi, nbj, downi, downj);
	}
      }

    }// low-packing
  }// end-of-omp
}

void
SparseMatrixMP::copy(FullMatrix_t& h, const complex<T>* phi, int packing) const {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int ncoltot = h.ncol();
    int* ionid = myIndex->IonID;

    int iat = 0;
    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    

#ifdef TB_BLOCK
      const int nb = TB_BLOCK;
#else
      int nb = myIndex->Basis[i];
#endif
      int up=2*myIndex->Offset[i];
      int down=up+nb;
      SO[ionid[i]]->copy(h.begin(), up, up, ncoltot);
      h.add(d_diag[myID][iat], nb, nb, up, up);
      h.add(d_diag[myID][iat], nb, nb, down, down);
    }

    Matrix<dcomplex_t> hoff(Dmax);

    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);
      if(j>=i) {
#ifdef TB_BLOCK
        const int nbi = TB_BLOCK;
        const int nbj = TB_BLOCK;
#else
        int nbi = myIndex->Basis[i];
        int nbj = myIndex->Basis[j];
#endif
        int upi=2*myIndex->Offset[i];
        int downi=upi+nbi;
        int upj=2*myIndex->Offset[j];
        int downj=upj+nbj;
        cmult(hoff.begin(), phi[nn], d_offdiag[myID][iat], nbi*nbj);
        h.add(hoff.begin(), nbi, nbj, upi, upj);
	h.add(hoff.begin(), nbi, nbj, downi, downj);
      }
    }
  }
}

void 
SparseMatrixMP::product(const FullMatrix_t& a, const FullMatrix_t& b,
                      const vector<T>& w){
}

void 
SparseMatrixMP::product(SparseMatrixMP&,
                        const FullMatrix_t& a, const FullMatrix_t& b,
                        const vector<T>& w){
}

void 
SparseMatrixMP::print(ostream& os) {
}

void SparseMatrixMP::read_so_parameters(){

  ifstream fin("TBPARAM");
  if(fin) {
    vector<string> slist;
    int nw=0;
    do {
      nw = getwords(slist,fin);
    } while(nw == 0 || slist[0] != "<SO>" || slist[0] == "end"); 
    if(slist[0] != "end") {
      SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
      int numspecies =mySpecies->getTotalNum(); 
      int ins = SpeciesCollection::addAttrib("ns");
      SO.reserve(numspecies);
      for(int i=0; i<numspecies; i++) {
	SO.push_back(new SpinOrbitComp<T>);
      }
      getwords(slist,fin);
      do {
	int newid =SpeciesCollection::addSpecies(slist[0].c_str());
	int i0 = atof(slist[2].c_str());
	double lambda = atof(slist[3].c_str());
	int ns = (*mySpecies)(ins,newid);
	SO[newid]->add(new SpinOrbitPP<T>(lambda, i0, i0));
	SO[newid]->set(ns,ns);
	getwords(slist,fin);
      } while(slist[0] != "</SO>");
    }
  }  
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseMatrixSpinOrbitMP.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

