//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_SPARSEVECTORINDEX_H
#define OHMMS_SPARSEVECTORINDEX_H

#include <vector>
using std::vector;
#include "Utilities/OhmmsObject.h"

/*! \class VectorIndex
 * A specialized sparse index class for TB Hamiltonian matrices
 * Has a reference to PairList and creates extra indices.
 */
template<class PairList_t>
struct VectorIndex : public OhmmsObject {

  typedef PairList_t ThisPairList_t;
  int nnmax, maxBasis;
  vector<int> I;  //!< paired with myList.J
  vector<int> Loc;//!< multi -> one mapping of pair(i,j)
  vector<int> Basis;
  vector<int> Offset;
  int* IonID;
  PairList_t* myList;

  //!<constructor
  VectorIndex(const char* objn = "vectorindex"):
  OhmmsObject(objn), myList(NULL), nnmax(0) { }

  VectorIndex(): myList(NULL),nnmax(0){ }

  ~VectorIndex() {  } // myList has to be provided externally

  template<class NS_t, class IonID_t>
  void init(PairList_t* ptclpairs, const NS_t& ns, IonID_t& ionid) {

    myList = ptclpairs;
    Basis = vector<int>(ionid.size());
    Offset = vector<int>(ionid.size()+1);
    IonID = ionid.begin();
    Offset[0] = 0;
    maxBasis = 0;
    for(int i=0; i<ionid.size(); i++) {
      Basis[i] = static_cast<int>(ns[ionid[i]]);      
      maxBasis = max(Basis[i], maxBasis);
      Offset[i+1] = Offset[i] + Basis[i];
    }
    // record the total block size
    maxBasis = maxBasis*maxBasis;
  }

  inline int nrows() const { return myList->getLocalNum();}
  inline int numpairs() const { return myList->getTotNadj();}
  inline int totalDimension() const { return Offset[nrows()];}

  inline int nadj(int myid, int i) const { return myList->nadj(i);}
  inline int iadj(int myid, int i, int jL) const { return myList->iadj(i,jL);}
 
  //!< The row-index for the nn-th pair
  inline int irow(int nn) const { return I[nn];}

  //!< The column-index for the nn-th pair
  inline int icol(int nn) const { return myList->J[nn];}

  inline int first_diag(int myid=0) const { return 0;}
  inline int last_diag(int myid=0) const { return nrows();}
  inline int num_diag(int myid=0) const { return nrows();}
  inline int first_offdiag(int myid=0) const { return 0;}
  inline int last_offdiag(int myid=0) const { return  numpairs();}
  inline int num_offdiag(int myid=0) const { return  numpairs();}


  ////////////////////////////////////////////////////////
  // WARNING: Assuming each row and col is identicial
  //////////////////////////////////////////////////////
  inline int row_loc(int i) const { return Offset[i];}
  inline int col_loc(int i) const { return Offset[i];}

  inline int row_block(int ig) const { return Basis[ig];}
  inline int col_block(int ig) const { return Basis[ig];}

  //!< location of pair(i,j), note that both i and j are indecies
  inline int loc(int i, int j) const { // better search can be used 
    // returns the first occurence
    if(i == j) return -1;
    int jL = 0;
    while(jL<myList->nadj(i)) {
      if(j == myList->iadj(i,jL)) {return myList->loc(i,jL);}
      jL++;
    }
    return -2; // should not reach here
  }
  void update();
  bool write_data(ostream&);
  bool read_data(istream&);
};

/*!\brief Update I
 */
template<class PairList_t>
void 
VectorIndex<PairList_t>::update() {
  if(myList) {

    if(nnmax < myList->getTotNadj()) {
      nnmax = myList->getTotNadj();
      nnmax += nnmax/10; // add 10%
      I = vector<int>(nnmax);
      Loc = vector<int>(nnmax);
    }

    int nn = 0;
    for(int i=0; i<myList->getLocalNum(); i++) {
      for(int jL=0; jL<myList->nadj(i); jL++) {

        // will assign the first location for pair(i,j)
        Loc[nn] = loc(i,myList->iadj(i,jL));
	I[nn++] = i;
      }
    }
  }
}

template<class PairList_t>
bool
VectorIndex<PairList_t>::write_data(ostream&)  {

  return true;
}

template<class PairList_t>
bool
VectorIndex<PairList_t>::read_data(istream&)  {

  return true;
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * OHMMS_VERSION: $Id: SparseVectorIndex.h 151 2004-11-02 23:51:14Z jnkim $
 ***************************************************************************/
