//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

// -*- C++ -*-
#ifndef OHMMS_SPECIESBASE_H
#define OHMMS_SPECIESBASE_H

#include <string>
#include <vector>
#include <map>
#include "OhmmsData/OhmmsElementBase.h"

/*! \class SpeciesBase
 *  \brief A class containing species attributes.
*/
class SpeciesBase: public OhmmsElementBase {
public:

  typedef double                           Scalar_t;
  typedef std::vector<Scalar_t>            SpeciesAttrib_t;
  typedef std::vector<SpeciesAttrib_t* >   AttribList_t;
    
  ///the number of species
  unsigned        TotalNum;

  /// Species Name list
  std::vector<std::string>  Name;   

  /// List of species attributes
  AttribList_t    d_attrib;

  ///map for the name of an attribute to the attribute index
  std::map<std::string,int> attrib_ind;

  /// Constructor
  SpeciesBase();

  /// Destructor
  ~SpeciesBase();

  ///return the number of species
  unsigned getTotalNum() const { return TotalNum; }

  ///set the number of species
  void setTotalNum(const unsigned n) { TotalNum = n; }

  ///return the number of attributes in our list
  int numAttributes() const { return d_attrib.size();}

  /// add a new attribute to the list
  int addAttrib(const char* aname);

  /// return the value of i-th attribute for the j-th species
  inline double operator()(int i, int j) const { 
    return d_attrib[i]->operator[](j); 
  }

  ///assign the value of i-th attribute for the j-th species
  inline double& operator()(int i, int j) { 
    return d_attrib[i]->operator[](j); 
  }


  /// add m species to the list by adding m elements to each attribute.
  void create(unsigned m) {
    if(m > 0) {
      Name.insert(Name.end(), m, std::string("none"));
      AttribList_t::iterator dit = d_attrib.begin();      
      for(; dit != d_attrib.end(); ++dit) (*dit)->insert((*dit)->end(), m, 0);
      TotalNum += m;
    }
  }

  /// return an ID for the species with name. 
  unsigned find(const std::string& name) const{
    unsigned i=0;
    for(; i< TotalNum; i++) {
      if(Name[i] == name) return i; 
    }
    return i;//Not found. Returns TotalNum.
  }

  /// return the name of the ith species
  const std::string& getName(unsigned int i) const { return Name[i]; }

  /// return an ID for the species with name. Add a species if not found
  unsigned getSpeciesID(const std::string& name) { 
    unsigned i = find(name); // check if the name is registered
    if(i == TotalNum) { // if not found, add a new species
      create(1);
      Name[i] = name;
    }
    return i; // return an index for a species
  }


  bool get(std::ostream& os) const;
  bool put(std::istream& is);
  void reset();
  bool put(xmlNodePtr cur);

};
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 200 $   $Date: 2004-12-22 10:33:44 -0600 (Wed, 22 Dec 2004) $
 * $Id: OhmmsSpecies.h 200 2004-12-22 16:33:44Z jnkim $ 
 ***************************************************************************/
